static char rcsId[] = "$Header: stgrind.c,v 1.9 84/09/07 13:32:36 pkl Exp $";

/*
 *	Stgrind -- converts Smalltalk ``fileOut'' files to troff format
 *		Default macros file is /usr/local/bs/lib/stmacs.
 *	
 *	Usage: stgrind [-m macros] file.st | ditroff -ms
 */

# include	<stdio.h>
# include	<ctype.h>

# define	FALSE		0
# define	TRUE		(~FALSE)




# define	MACROS		"/usr/local/bs/lib/stmacs"
char	macros[BUFSIZ];
char	filename[BUFSIZ];	/* current input file name */
FILE	*infp;

typedef int	boolean;

# define	EXPRLEN		10000
char	buf[EXPRLEN];		/* stores fileOut expressions */
int	bp;

boolean	isTimeStamp(), isComment(), isDefinition(), isClassDefinition(),
	isClassComment(), isMethodsFor(), isDoIt(),
	getNextExpr(), checkWord();

int	bufIndex();

void	scanTimeStamp(), scanComment(), scanDefinition(), scanClassComment(),
	scanClassDefinition(), scanMethodsFor(), scanMethod(), scanDoIt(),
	extractWord(), fatal(), getargs();

char	*date();



main(argc, argv)
int argc;
char **argv;
{
	getargs(argc, argv);

	printf(".so %s\n", macros);
	printf(".ds CH\n");		/* clears top of page */
	printf(".ds CF Listing from \\fI%s\\fP at \\fI%s\\fP\n",
	    filename, date());

	while (getNextExpr(infp)) {
		if (isTimeStamp())
			scanTimeStamp();
		else if (isComment())
			scanComment();
		else if (isClassComment())
			scanClassComment();
		else if (isClassDefinition())
			scanClassDefinition();
		else if (isDefinition())
			scanDefinition();
		else if (isMethodsFor())
			scanMethodsFor();
		else if (isDoIt())
			scanDoIt();
		else 
			scanMethod();
	}
}


/*
 *	Finite state machine to read an expression at a time.
 *	An expression is
 *		!string!
 *		string!
 *		string! !
 *	where string may contain carriage returns.  Could have used
 *	lex, but the carriage returns will be nasty.
 *	Returns TRUE if an expression has been read successfully.
 *	Returns FALSE on eof.
 *	Expression read is stored in the global variable "buf".
 *	Doesn't check for buffer overflow.
 */

boolean
getNextExpr(fp)
FILE *fp;
{
	register char	ch;
	register int	state;

	/*
	 *	State names for FSM scanner
	 */

#	define	ST_START	0
#	define	ST_1		1
#	define	ST_2		2
#	define	ST_3		3
#	define	ST_4		4
#	define	ST_STOP		99999

#	define	getNextChar(ch, fp)\
		if ((ch = getc(fp)) == EOF)\
			return(FALSE);\
		else

			
	while ((ch = getc(fp)) != EOF)		/* skip all blanks */
		if (ch == EOF)
			return(FALSE);
		else if (!isspace(ch))
			break;
		
	bp = 0;
	state = ST_START;
	
	while (state != ST_STOP) {
		switch (state) {
		case ST_START:
			buf[bp++] = ch;
			state =  ch == '!'  ?  ST_1  :  ST_2;
			getNextChar(ch, fp);
			break;
		case ST_1:
			buf[bp++] = ch;
			if (ch == '!')
				state = ST_STOP;
			else
				getNextChar(ch, fp);
			break;
		case ST_2:
			buf[bp++] = ch;
			if (ch == '!')
				state = ST_3;
			getNextChar(ch, fp);
			break;
		case ST_3:
			if (ch == '!') {
				state = ST_2;
				getNextChar(ch, fp);
			}
			else if (ch == ' ') {
				state = ST_4;
				getNextChar(ch, fp);
			}
			else {
				state = ST_STOP;
				ungetc(ch, fp);
			}
			break;
		case ST_4:
			if (ch == '!') {
				buf[bp++] = ' ';
				buf[bp++] = '!';
			}
			else
				ungetc(ch, fp);
			state = ST_STOP;
			break;
		default:
			fatal("getNextExpr: unknown state: %d", state);
			break;
		}
	}
	buf[bp] = '\0';
	return(TRUE);
}


/*
 *	Reports fatal error and die.
 */

void
fatal(fmt, s1, s2, s3, s4)
char *fmt;
int s1, s2, s3, s4;
{
	char	strbuf[BUFSIZ];

	sprintf(strbuf, fmt, s1, s2, s3, s4);
	fprintf(stderr, "%s\n", strbuf);
	exit(1);
}



/*
 *	The following messages classifies a fileOut expression using
 *	various heuristics.  Returns TRUE if the expression in "buf"
 *	is of a given type.
 */

/*
 *	Detects doIt's (e.g. Class initialize!)
 */

boolean
isDoIt()
{
	return(isupper(*buf)  &&  buf[bp - 2] != ' '
	    &&  buf[bp - 1] == '!');
}


/*
 *	Detects the time stamp, which looks like
 *		'date and version'!
 *	from Smalltalk.  Not always correct.
 */

boolean
isTimeStamp()
{
	return(*buf == '\'');
}


/*
 *	Detects a comment.  ("....."!)
 */

boolean
isComment()
{
	return(*buf == '"');
}


/*
 *	Detects the definition of a new class.
 */

boolean
isDefinition()
{
	return(checkWord(2, "subclass:")
	    ||  checkWord(2, "variableSubclass:")
	    ||  checkWord(2, "variableByteSubclass:")
	    ||  checkWord(2, "variableWordSubclass:"));
}


/*
 *	Detects this,
 *		Date comment: 'This is a comment'!
 */

boolean
isClassComment()
{
	return(checkWord(2, "comment:"));
}


/*
 *	!Class methodsFor: 'controlling'!
 */

boolean
isMethodsFor()
{
	return(*buf == '!');
}


/*
 *	Definition of a class.
 *		Class class instanceVariables: ''....!
 */

boolean
isClassDefinition()
{
	return(checkWord(2, "class")
	    &&  checkWord(3, "instanceVariableNames:"));
}



/*
 *	The following messages looks at the contents of "buf",
 *	decodes it, and outputs the contents.
 */

void
scanDoIt()
{
	register int	i;

	printf(".DI \"");
	for (i = 0;  buf[i] != '!';  i++)
		putchar(buf[i]);
	printf("\"\n");
}


void
scanClassDefinition()
{
	int	i;

	printf(".DS\n");
	for (i = 0;  buf[i] != '!';  i++)
		putchar(buf[i]);
	printf("\n.DE\n");
}


void
scanTimeStamp()
{
	register int	i;

	printf(".TI \" ");
	i = 1;
	while (buf[i] != '\'')
		putchar(buf[i++]);
	printf("\"\n");
}


void
scanComment()
{
	register int	i;

	printf(".CO\n");
	for (i = 0;  buf[i] != '!';  i++)
		putchar(buf[i]);
	printf("\n.OC\n");
}


void
scanDefinition()
{
	char	word[100];
	int	wordIndex;
#	define	W_SUPERCLASS		1
#	define	W_CLASS			3
#	define	W_INSTVAR		5
	
	extractWord(W_CLASS, word);
	printf(".CL \"%s\"\n", word);

	extractWord(W_SUPERCLASS, word);
	printf(".SC \"%s\"\n", word);

	wordIndex = W_INSTVAR;
	if (checkWord(W_INSTVAR, "classVariableNames:"))
		printf(".IV \\fInone\\fP\n");
	else {
		extractWord(W_INSTVAR, word);
		printf(".IV \"%s\"\n", word);
		wordIndex++;
		while (!checkWord(wordIndex, "classVariableNames:")) {
			extractWord(wordIndex++, word);
			printf(".IC \"%s\"\n", word);
		}
	}
	
	wordIndex++;
	if (checkWord(wordIndex, "poolDictionaries:"))
		printf(".CV \\fInone\\fP\n");
	else {
		extractWord(wordIndex++, word);
		printf(".CV \"%s\"\n", word);
		while (!checkWord(wordIndex, "poolDictionaries:")) {
			extractWord(wordIndex++, word);
			printf(".CC \"%s\"\n", word);
		}
	}
	
	wordIndex++;
	if (checkWord(wordIndex, "category:"))
		printf(".PD \\fInone\\fP\n");
	else {
		extractWord(wordIndex++, word);
		printf(".PD \"%s\"\n", word);
		while (!checkWord(wordIndex, "category:")) {
			extractWord(wordIndex++, word);
			printf(".PC \"%s\"\n", word);
		}
	}
	
	extractWord(++wordIndex, word);
	printf(".CA \"%s\"\n", word);
}


/*
 *	returns the index of the first character in buf[] of the index'th
 *	word.
 */

int
bufIndex(index)
int index;
{
	int	i = 0;

	while (--index) {
		while (!isspace(buf[i]))
			i++;
		while (isspace(buf[i]) ||  buf[i] == '\''  ||  buf[i] == '#')
			i++;
	}
	return(i);
}


/*
 *	returns TRUE if index'th word in buf[] matches aString.
 */

boolean
checkWord(index, aString)
int index;
char aString[];
{
	char	word[100];
	
	extractWord(index, word);
	return(!strcmp(word, aString)  ?  TRUE  :  FALSE);
}


/*
 *	Extracts the index'th word from buf[] and stores it as a
 *	null terminated string in aString.
 */

void
extractWord(index, aString)
int index;
char aString[];
{
	int	i, j;
	
	i = bufIndex(index);
	j = 0;
	while (!isspace(buf[i])  &&  buf[i] != '\'')
		aString[j++] = buf[i++];
	aString[j] = '\0';
}


/*
 *	Copies a method definition expression to the output.
 *	Some characters must be handled differently.
 */

void
scanMethod()
{
	register int		i, column;
	register boolean	inComment = FALSE;
	
	i = 0;
	while (isspace(buf[i]))
		i++;
	printf(".MH \"");	/* method start */
	while (buf[i] != '\n')
		putchar(buf[i++]);
	printf("\"\n");

	column = 0;
	while (buf[i] != '\0') {
		switch (buf[i]) {
		case '.':
		case '\'':
			if (i == 0  ||  buf[i - 1] == '\n')
				printf("\\&%c", buf[i]);
			else
				putchar(buf[i]);
			break;
		case '\n':
			putchar(buf[i]);
			column = 0;
			break;
		case '^':
			printf("\\(ua");
			break;
		case '_':
			printf("\\(<-");
			break;
		case '\\':
			printf("\\e");
			break;
		case '~':
			printf("\\(ap");
			break;
		case '!':
			if (buf[i + 1] == '!') {
				putchar('!');
				i++;
			}
			break;
		case '"':
			if (inComment)
				printf("\"\\fP");
			else
				printf("\\fI\"");
			inComment = !inComment;
			break;
		default:
			putchar(buf[i]);
			break;
		}
		i++;
		column++;
	}
	printf("\n.ME\n");
}


/*
 *	!<class> class methodsFor: 'class protocol name'!
 *	!<class> methodsFor: 'protocol name'!
 */

void
scanMethodsFor()
{
	register int	i;
#	define	M_CLASS_KW	2
#	define	M_PROTOCOL	3

	if (checkWord(M_CLASS_KW, "class")) {
		printf(".PC \"");
		i = bufIndex(M_PROTOCOL + 1);
		while (buf[i] != '\'')
			putchar(buf[i++]);
	}
	else {
		printf(".PI \"");
		i = bufIndex(M_PROTOCOL);
		while(buf[i] != '\'')
			putchar(buf[i++]);
	}
	printf("\"\n");
}


/*
 *	<class> comment: 'This class has no comment.'!
 */

void
scanClassComment()
{
	register int	i;
#	define	COMMENT_START	3

	printf(".CS\n");		/* class comment start */
	i = bufIndex(COMMENT_START);
	for (;;) {
		if (buf[i] == '\'') {
			if (buf[i + 1] == '\'')		/* ... '' ... */
				putchar('\'');
			else if (buf[i + 1] == '!')	/* ... .'! */
				break;
			else				/* ... ' ... */
				putchar(buf[i]);
		}
		else if (buf[i] == '!')			/* ... .! */
			break;
		else if (buf[i] == '\\')
			printf("\\e");
		else if (buf[i] == '\t') {
			if (i != 0  &&  buf[i - 1] == '\n')
				printf(".br\n\t");
			else
				putchar('\t');
		}
		else
			putchar(buf[i]);
		i++;
	}
	printf("\n.CE\n");		/* class comment end */
}


void
getargs(argc, argv)
int argc;
char **argv;
{
	register char	*ch;

	strcpy(macros, MACROS);
	while (--argc > 0 && (*++argv)[0] == '-')
		for (ch=argv[0]+1; *ch!='\0'; ch++) 
			switch (*ch) {
			case 'm':
				if (*(ch + 1) == '\0') {
					strcpy(macros, *++argv);
					--argc;
					*ch = '\0';
				}
				else {
					strcpy(macros, ch + 1);
					*(ch + 1) = '\0';
				}
				break;

			default:
				fprintf(stderr, 
				    "Usage: stgrind [-m macros] file\n");
				exit(-1);
				break;
			}

	/* 
	 *	If there are any more arguments left, open that file and
	 *	set input to its fp.
	 */

	if (argc == 0) {
		infp = stdin;
		strcpy(filename, "stdin");
	}
	else if (argc != 1) {
		fprintf(stderr, "Usage: stgrind [-m macros] file\n");
		exit(-1);
	}
	else if ((infp = fopen(*argv, "r")) == NULL) {
		fprintf(stderr, "stgrind: cannot open file: %s\n", *argv);
		exit(1);
	}
	else
		strcpy(filename, *argv);
}


/*
 *	Returns the date string
 */

char *
date()
{
	long		now;
	char		*string;
	extern char	*ctime();

	now = time(0);
	string = ctime(&now);
	string[24] = '\0';
	return(string);
}
