
Transcript show: '
Alternative browsers
	W.R. Bush
	U.C. Berkeley
	May 1984
'!

Object subclass: #Explorer
	instanceVariableNames: 'branchCount classCount depthCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!Explorer methodsFor: 'probing'!
getName: selection
	"Display list elements are strings.
	 When one is selected it must be converted to its associated class."
	selection == nil ifTrue: [ ^nil ].
	(selection at: 1) = $- ifTrue: [ ^nil ].
	^ (Smalltalk at: (selection asSymbol))
! !

!Explorer methodsFor: 'probing'!
lookupSelectorFrom: class
	| selector scanner result |
	FillInTheBlank
		request: 'Method to look up:'
		displayAt: Sensor cursorPoint
		centered: true
		action: [ :answer | selector _ answer asSymbol ]
		initialAnswer: 'lookup chain'.
	scanner _ class.
	result _ WriteStream on: (String new: 1000).
	result print: selector; cr.
	[ scanner ~~ nil ] whileTrue:
		[ result tab; print: scanner; space.
	  	  (scanner includesSelector: selector)
			ifTrue: [ result nextPut: $+ ]
			ifFalse: [ result nextPut: $- ].
		  result cr.
		  scanner _ scanner superclass ].
	Transcript cr; show: (result contents); cr.
! !

!Explorer methodsFor: 'counting'!
count: aClass
	| result |
	branchCount _ Array new: 11 withAll: 0.
	classCount _ 1.
	depthCount _ 1.
	self count: aClass at: 1.
	result _ WriteStream on: (String new: 1000).
	result print: aClass; cr.
	result tab; print: classCount - 1; space; print: #subclasses; cr.
	result tab; print: depthCount - 1;
		space; print: #subclass; space; print: #levels; cr.
	(branchCount at: 1) = 0 ifFalse:
		[ result tab; print: (branchCount at: 1); space.
		  (branchCount at: 1) = 1
			ifTrue: [ result print: #leaf ]
			ifFalse: [ result print: #leaves ].
		  result cr ].
	1 to: 9 do: [ :i |
			(branchCount at: i + 1) = 0 ifFalse:
				[ result tab; print: (branchCount at: i + 1);
					space; print: i; nextPut: $-;
					print: #way; space.
				  (branchCount at: i + 1) = 1
					ifTrue: [ result print: #branching ]
					ifFalse: [ result print: #branchings ].
				  result cr ]].
	((aClass == Object) and: [(branchCount at: 11) = 1]) ifTrue:
		[ result tab; print: ((Object subclasses) size);
			space; print: #Object; space; print: #subclasses; cr.
		  branchCount at: 11 put: 0 ].
	(branchCount at: 11) = 0 ifFalse:
		[ result tab; print: (branchCount at: 11);
			space; print: #greater; space; print: #branchings; cr.
		  aClass == Object ifTrue:
			[ result tab; tab; print: ((Object subclasses) size);
				space; print: #Object; space;
				print: #subclasses; cr ]].
	Transcript cr; show: (result contents); cr.
! !

!Explorer methodsFor: 'counting'!
count: class at: level
	| subclassSet subclassSetSize |
	subclassSet _ class subclasses.
	class == Class ifTrue: [ subclassSet _ Set new ].
	subclassSetSize _ subclassSet size.
	classCount _ classCount + subclassSetSize.
	subclassSetSize > 9 ifTrue: [ subclassSetSize _ 10 ].
	branchCount at: (subclassSetSize + 1)
		put: ((branchCount at: subclassSetSize + 1) + 1).
	subclassSetSize = 0 ifTrue: [ ^self ].
	level = depthCount ifTrue: [ depthCount _ depthCount + 1 ].
	subclassSet do: [ :e | self count: e at: level + 1 ].
! !

!Explorer methodsFor: 'accessing'!
classCount
	^ classCount
! !

!Explorer methodsFor: 'accessing'!
depthCount
	^ depthCount
! !


Explorer subclass: #Navigator
	instanceVariableNames: 'root branches branch branchMenu leaves leaf leafMenu'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!Navigator methodsFor: 'initialization'!
on: aClass
	| subclassSet branchSize leafSize branchIndex leafIndex |
	root _ aClass.
	subclassSet _ aClass subclasses.
	branchSize _ 0.
	leafSize _ 0.
	subclassSet do:
		[ :e |
			((e subclasses) size) = 0
				ifTrue: [ leafSize _ leafSize + 1 ]
				ifFalse: [ branchSize _ branchSize + 1 ]].
	branches _ Array new: branchSize.
	leaves _ Array new: leafSize.
	branchIndex _ 1.
	leafIndex _ 1.
	subclassSet do:
		[ :e |
			((e subclasses) size) = 0
				ifTrue: [ leaves at: leafIndex
						put: (e name) asString.
					  leafIndex _ leafIndex + 1 ]
				ifFalse: [ branches at: branchIndex
						put: (e name) asString.
					   branchIndex _ branchIndex + 1 ]].
! !

!Navigator methodsFor: 'branch list'!
branchList
	^ branches
! !

!Navigator methodsFor: 'branch list'!
branchMenu
	branchMenu == nil ifTrue:
	    [ branchMenu _ ActionMenu
		labels: 'navigate branch\navigate root\browse branch\browse root\instance lookup\class lookup' withCRs
		lines: #(2 4)
		selectors: #(navigateBranch navigateRoot browseBranch browseRoot lookupBranchInstance lookupBranchClass) ].
	^ branchMenu
! !

!Navigator methodsFor: 'branch list'!
branchName
	^ branch
! !

!Navigator methodsFor: 'branch list'!
branchName: selection
	branch _ self getName: selection.
! !

!Navigator methodsFor: 'leaf list'!
leafList
	^ leaves
! !

!Navigator methodsFor: 'leaf list'!
leafMenu
	leafMenu == nil ifTrue:
	    [ leafMenu _ ActionMenu
		labels: 'browse leaf\browse root\navigate root\instance lookup\class lookup' withCRs
		lines: #(2 3)
		selectors: #(browseLeaf browseRoot navigateRoot lookupLeafInstance lookupLeafClass) ].
	^ leafMenu
! !

!Navigator methodsFor: 'leaf list'!
leafName
	^ leaf
! !

!Navigator methodsFor: 'leaf list'!
leafName: selection
	leaf _ self getName: selection.
! !

!Navigator methodsFor: 'branch functions'!
browseBranch
	branch == nil ifTrue: [ ^self ].
	Browser newOnClass: branch.
! !

!Navigator methodsFor: 'branch functions'!
browseRoot
	Browser newOnClass: root.
! !

!Navigator methodsFor: 'branch functions'!
navigateBranch
	branch == nil ifTrue: [ ^self ].
	NavigatorView openOn: branch.
! !

!Navigator methodsFor: 'branch functions'!
navigateRoot
	root == Object ifTrue: [ ^self ].
	NavigatorView openOn: (root superclass).
! !

!Navigator methodsFor: 'branch functions'!
lookupBranchInstance
	branch == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: branch.
! !

!Navigator methodsFor: 'branch functions'!
lookupBranchClass
	branch == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: (branch class).
! !

!Navigator methodsFor: 'leaf functions'!
browseLeaf
	leaf == nil ifTrue: [ ^self ].
	Browser newOnClass: leaf.
! !

!Navigator methodsFor: 'leaf functions'!
lookupLeafInstance
	leaf == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: leaf.
! !

!Navigator methodsFor: 'leaf functions'!
lookupLeafClass
	leaf == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: (leaf class).
! !


StandardSystemView subclass: #NavigatorView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!NavigatorView class methodsFor: 'instance creation'!
openOn: aClass
	| aNavigator rootClass minimumSize topView |
	((aClass subclasses) size) = 0
		ifTrue: [ rootClass _ aClass superclass ]
		ifFalse: [ rootClass _ aClass ].
	aNavigator _ Navigator new on: rootClass.
	rootClass == Object
		ifTrue: [ minimumSize _ 250@500 ]
		ifFalse: [ minimumSize _ 200@100 ].
	(topView _ self
			model: aNavigator
			label: ((rootClass name) asString), ' Navigator'
			minimumSize: minimumSize)
		addBranchView: (0@0 extent: 0.5@1.0) on: aNavigator;
		addLeafView: (0.5@0 extent: 0.5@1.0) on: aNavigator.
	topView controller open
! !

!NavigatorView methodsFor: 'subview creation'!
addBranchView: area on: aNavigator
	self addSubView:
		(SelectionInListView on: aNavigator
			printItems: false oneItem: false
			aspect: #branchName change: #branchName:
			list: #branchList menu: #branchMenu
			initialSelection: #branchName)
		in: area borderWidth: 1
! !

!NavigatorView methodsFor: 'subview creation'!
addLeafView: area on: aNavigator
	self addSubView:
		(SelectionInListView on: aNavigator
			printItems: false oneItem: false
			aspect: #leafName change: #leafName:
			list: #leafList menu: #leafMenu
			initialSelection: #leafName)
		in: area borderWidth: 1
! !


Explorer subclass: #Examiner
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!Examiner methodsFor: 'classes'!
addSubclasses: aClass to: aList
	| subclassList oldSize newSize newList |
	aClass == Class ifTrue: [ ^aList ].
	subclassList _ aClass subclasses.
	(subclassList size) = 0 ifTrue: [ ^aList ].
	aList == nil
		ifTrue: [ oldSize _ 0 ]
		ifFalse: [ oldSize _ (aList size) ].
	newSize _ oldSize + (subclassList size) + 1.
	newList _ Array new: newSize.
	1 to: oldSize do: [ :i | newList at: i put: (aList at: i) ].
	newList at: (oldSize _ oldSize + 1) put:
			('--', ((aClass name) asString), '--subclasses--').
	subclassList do:
		[ :e | newList at: (oldSize _ oldSize + 1)
				put: ((e name) asString) ].
	^ newList
! !


Examiner subclass: #SubExaminer
	instanceVariableNames: 'classList selectedClass menu parent child'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!SubExaminer methodsFor: 'view functions'!
classList
	^ classList
! !

!SubExaminer methodsFor: 'view functions'!
classList: aList
	classList _ aList.
! !

!SubExaminer methodsFor: 'view functions'!
examinerMenu
	menu == nil ifTrue:
	  [ menu _ ActionMenu
		  labels: 'examine\clear\browse\instance lookup\class lookup' withCRs
		  lines: #(3 4)
		  selectors: #(examineClass clearClasses browseClass lookupClassInstance lookupClassClass) ].
	^ menu
! !

!SubExaminer methodsFor: 'view functions'!
selectedClass
	^ selectedClass
! !

!SubExaminer methodsFor: 'view functions'!
selectClass: selection
	selectedClass _ self getName: selection.
! !

!SubExaminer methodsFor: 'menu functions'!
browseClass
	selectedClass == nil ifTrue: [ ^self ].
	Browser newOnClass: selectedClass.
! !

!SubExaminer methodsFor: 'menu functions'!
examineClass
	child == nil ifTrue: [ ^self ].
	(selectedClass == nil) ifTrue: [ self examineClasses. ^self ].
	((selectedClass subclasses) size) = 0 ifTrue: [ ^self ].
	child examine: (Array with: ((selectedClass name) asString)).
	child change.
! !

!SubExaminer methodsFor: 'menu functions'!
examineClasses
	child == nil ifTrue: [ ^self ].
	classList == nil ifTrue: [ ^self ].
	child clearClasses.
	child examine: classList.
	child change.
! !

!SubExaminer methodsFor: 'menu functions'!
clearClasses
	child ~~ nil ifTrue: [ child clearClasses ].
	parent ~~ nil ifTrue: [ classList _ nil ].
	self changed: #selectedClass.
! !

!SubExaminer methodsFor: 'menu functions'!
lookupInstance
	selectedClass == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: selectedClass.
! !

!SubExaminer methodsFor: 'menu functions'!
lookupClass
	selectedClass == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: (selectedClass class).
! !
!SubExaminer methodsFor: 'menu functions'!
change
	child ~~ nil ifTrue: [ child change ].
	self changed: #selectedClass.
! !

!SubExaminer methodsFor: 'accessing'!
parent: aSubExaminer
	parent _ aSubExaminer.
! !

!SubExaminer methodsFor: 'accessing'!
child: aSubExaminer
	child _ aSubExaminer.
! !

!SubExaminer methodsFor: 'accessing'!
child
	^ child
! !

!SubExaminer methodsFor: 'private'!
examine: superList
	| addList addClass newList |
    (Cursor wait) showWhile:
      [ addList _ nil.
	1 to: (superList size) do:
		[ :i | addClass _ self getName: (superList at: i).
		  addClass ~~ nil ifTrue:
			[ addList _ self addSubclasses: addClass to: addList ]].
	child ~~ nil ifTrue: [ child examine: addList ].
	((classList ~~ nil) and: [addList ~~ nil])
		ifTrue: [ newList _ Array new:
				(classList size) + (addList size).
			  1 to: (classList size) do:
				[ :i | newList at: i put: (classList at: i) ].
			  1 to: (addList size) do:
				[ :i | newList at: i + (classList size) put:
					(addList at: i) ].
			  classList _ newList ].
	classList == nil
		ifTrue: [ classList _ addList ] ].
! !


StandardSystemView subclass: #ExaminerView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!ExaminerView class methodsFor: 'instance creation'!
openOn: aClass
	| rootClass numberOfWindows anExaminer topView singleWidth totalWidth
		totalLength firstSubExaminer nextSubExaminer lastSubExaminer |
	anExaminer _ (Examiner new).
	((aClass subclasses) size) = 0
		ifTrue: [ rootClass _ aClass superclass ]
		ifFalse: [ rootClass _ aClass ].
	(Cursor wait) showWhile:
		[ anExaminer count: rootClass.
		  numberOfWindows _ (anExaminer depthCount) - 1 ].
	singleWidth _ 1.0 / numberOfWindows.
	numberOfWindows < 6
		ifTrue: [ totalWidth _ 150 * numberOfWindows ]
		ifFalse: [ totalWidth _ 900 ].
	rootClass == Object
		ifTrue: [ totalLength _ 600 ]
		ifFalse: [ totalLength _ 200 ].
	topView _ self
		model: anExaminer
		label: ((rootClass name) asString), ' Examiner'
		minimumSize: (totalWidth@totalLength).
	firstSubExaminer _ (SubExaminer new).
	lastSubExaminer _ firstSubExaminer.
	2 to: numberOfWindows do:
		[ :i | nextSubExaminer _ (SubExaminer new).
		  nextSubExaminer parent: lastSubExaminer.
		  lastSubExaminer child: nextSubExaminer.
		  lastSubExaminer _ nextSubExaminer ].
	(Cursor wait) showWhile:
		[ rootClass == Object
			ifTrue: [ firstSubExaminer classList:
				(anExaminer addSubclasses: rootClass to: nil) ]
			ifFalse: [ firstSubExaminer examine:
				(Array with: ((rootClass name) asString)) ]].
	lastSubExaminer _ firstSubExaminer.
	1 to: numberOfWindows do:
		[ :i | topView
				addWindow: (singleWidth*(i-1) @ 0
						extent: singleWidth@1.0)
				on: lastSubExaminer.
			lastSubExaminer _ lastSubExaminer child ].
	topView controller open.
! !

!ExaminerView methodsFor: 'subview creation'!
addWindow: area on: aSubExaminer
	self addSubView:
		(SelectionInListView on: aSubExaminer
			printItems: false oneItem: false
			aspect: #selectedClass change: #selectClass:
			list: #classList menu: #examinerMenu
			initialSelection: #selectedClass)
		in: area borderWidth: 1.
! !


Explorer subclass: #Peruser
	instanceVariableNames: 'classList selectedClass peruserMenu'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!Peruser methodsFor: 'initialization'!
initialize
	(Cursor wait) showWhile:
		[ self count: Object.
		  classList _ Array new: (self classCount).
		  classList at: 1 put: ((Object name) asString).
		  self classEnumerate: Object at: 2 ].
! !

!Peruser methodsFor: 'private'!
classEnumerate: aClass at: index
	| subclassSet subclassIndex |
	aClass == Class ifTrue: [ ^index ].
	subclassSet _ aClass subclasses.
	(subclassSet size) = 0 ifTrue: [ ^index ].
	subclassIndex _ index.
	subclassSet do:
		[ :e | classList at: subclassIndex put: ((e name) asString).
			subclassIndex _ subclassIndex + 1.
			subclassIndex _
				(self classEnumerate: e at: subclassIndex) ].
	^ subclassIndex
! !

!Peruser methodsFor: 'list functions'!
peruserMenu
	peruserMenu == nil ifTrue:
	    [ peruserMenu _ ActionMenu
		labels: 'navigate\examine\browse\instance lookup\class lookup' withCRs
		lines: #(1 2 3)
		selectors: #(navigate examine browse lookupInstance lookupClass) ].
	^ peruserMenu
! !

!Peruser methodsFor: 'list functions'!
classList
	^ classList
! !

!Peruser methodsFor: 'list functions'!
selectedClass
	^ selectedClass
! !

!Peruser methodsFor: 'list functions'!
selectClass: selection
	selection == nil
		ifTrue: [ selectedClass _ nil ]
		ifFalse: [ selectedClass _ Smalltalk at: (selection asSymbol) ].
! !

!Peruser methodsFor: 'menu functions'!
navigate
	selectedClass == nil ifTrue: [ ^self ].
	NavigatorView openOn: selectedClass.
! !

!Peruser methodsFor: 'menu functions'!
examine
	selectedClass == nil ifTrue: [ ^self ].
	ExaminerView openOn: selectedClass.
! !

!Peruser methodsFor: 'menu functions'!
browse
	selectedClass == nil ifTrue: [ ^self ].
	Browser newOnClass: selectedClass.
! !

!Peruser methodsFor: 'menu functions'!
lookupInstance
	selectedClass == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: selectedClass.
! !

!Peruser methodsFor: 'menu functions'!
lookupClass
	selectedClass == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: (selectedClass class).
! !

!Peruser methodsFor: 'private'!
lookupSelectorFrom: class
	| selector scanner result |
	FillInTheBlank
		request: 'Method to look up:'
		displayAt: Sensor cursorPoint
		centered: true
		action: [ :answer | selector _ answer asSymbol ]
		initialAnswer: 'lookup chain'.
	scanner _ class.
	result _ WriteStream on: (String new: 1000).
	result print: selector; cr.
	[ scanner ~~ nil ] whileTrue:
		[ result tab; print: scanner; space.
	  	  (scanner includesSelector: selector)
			ifTrue: [ result nextPut: $+ ]
			ifFalse: [ result nextPut: $- ].
		  result cr.
		  scanner _ scanner superclass ].
	Transcript cr; show: (result contents); cr.
! !


StandardSystemView subclass: #PeruserView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!PeruserView class methodsFor: 'instance creation'!
open
	| aPeruser topView |
	aPeruser _ (Peruser new) initialize.
	(topView _ self
			model: aPeruser
			label: 'Peruser'
			minimumSize: 200@600)
	    addSubView:
		(SelectionInListView on: aPeruser
			printItems: false oneItem: false
			aspect: #selectedClass change: #selectClass:
			list: #classList menu: #peruserMenu
			initialSelection: #selectedClass)
		in: (0@0 extent: 1.0@1.0) borderWidth: 1.
	topView controller open.
! !


!ScreenController class methodsFor: 'class initialization'!
initialize
	ScreenYellowButtonMenu _
		PopUpMenu
			labels: 
'restore display
exit project
open project
file list
browse
explore
open workspace
system transcript
system workspace
save
suspend
quit'
			lines: #(2 9).
	ScreenYellowButtonMessages _
			#(restoreDisplay exitProject openProject 
			openFileList openBrowser explore openWorkspace openTranscript openSystemWorkspace
			save suspend quit).
	ScreenController allInstancesDo: [:c | c initializeYellowButtonMenu].
! !

!ScreenController methodsFor: 'menu messages'!
explore
	| menu index aClass |
	menu _ PopUpMenu
		labels: 'navigate\navigate Object\examine\examine Object\peruse' withCRs
		lines: #(2 4).
	index _ menu startUp.
	index = 1 ifTrue: [ FillInTheBlank
				request: 'Class to navigate:'
				displayAt: Sensor cursorPoint
				centered: true
				action: [ :answer | aClass _
					Smalltalk at: (answer asSymbol) ]
				initialAnswer: 'Object'.
			    NavigatorView openOn: aClass ].
	index = 2 ifTrue: [ NavigatorView openOn: Object ].
	index = 3 ifTrue: [ FillInTheBlank
				request: 'Class to examine:'
				displayAt: Sensor cursorPoint
				centered: true
				action: [ :answer | aClass _
					Smalltalk at: (answer asSymbol) ]
				initialAnswer: 'Object'.
			    ExaminerView openOn: aClass ].
	index = 4 ifTrue: [ ExaminerView openOn: Object ].
	index = 5 ifTrue: [ PeruserView open ].
! !

ScreenController initialize!
