
Object subclass: #Navigator
	instanceVariableNames: 'root branches branch branchMenu leaves leaf leafMenu'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!Navigator methodsFor: 'initialization'!
on: aClass
	| subclassSet branchSize leafSize branchIndex leafIndex |
	root _ aClass.
	subclassSet _ aClass subclasses.
	branchSize _ 0.
	leafSize _ 0.
	subclassSet do:
		[ :e |
			((e subclasses) size) = 0
				ifTrue: [ leafSize _ leafSize + 1 ]
				ifFalse: [ branchSize _ branchSize + 1 ]].
	branches _ Array new: branchSize.
	leaves _ Array new: leafSize.
	branchIndex _ 1.
	leafIndex _ 1.
	subclassSet do:
		[ :e |
			((e subclasses) size) = 0
				ifTrue: [ leaves at: leafIndex
						put: (e name) asString.
					  leafIndex _ leafIndex + 1 ]
				ifFalse: [ branches at: branchIndex
						put: (e name) asString.
					   branchIndex _ branchIndex + 1 ]].
! !

!Navigator methodsFor: 'branch list'!
branchList
	^ branches
! !

!Navigator methodsFor: 'branch list'!
branchMenu
	branchMenu == nil ifTrue:
	    [ branchMenu _ ActionMenu
		labels: 'navigate branch\navigate root\browse branch\browse root\instance lookup\class lookup' withCRs
		lines: #(2 4)
		selectors: #(navigateBranch navigateRoot browseBranch browseRoot lookupBranchInstance lookupBranchClass) ].
	^ branchMenu
! !

!Navigator methodsFor: 'branch list'!
branchName
	^ branch
! !

!Navigator methodsFor: 'branch list'!
branchName: selection
	selection == nil
		ifTrue: [ branch _ nil ]
		ifFalse: [ branch _ Smalltalk at: (selection asSymbol) ].
! !

!Navigator methodsFor: 'leaf list'!
leafList
	^ leaves
! !

!Navigator methodsFor: 'leaf list'!
leafMenu
	leafMenu == nil ifTrue:
	    [ leafMenu _ ActionMenu
		labels: 'browse leaf\browse root\navigate root\instance lookup\class lookup' withCRs
		lines: #(2 3)
		selectors: #(browseLeaf browseRoot navigateRoot lookupLeafInstance lookupLeafClass) ].
	^ leafMenu

! !

!Navigator methodsFor: 'leaf list'!
leafName
	^ leaf
! !

!Navigator methodsFor: 'leaf list'!
leafName: selection
	selection == nil
		ifTrue: [ leaf _ nil ]
		ifFalse: [ leaf _ Smalltalk at: (selection asSymbol) ].
! !

!Navigator methodsFor: 'branch functions'!
browseBranch
	branch == nil ifTrue: [ ^self ].
	Browser newOnClass: branch.
! !

!Navigator methodsFor: 'branch functions'!
browseRoot
	Browser newOnClass: root.
! !

!Navigator methodsFor: 'branch functions'!
navigateBranch
	branch == nil ifTrue: [ ^self ].
	NavigatorView openOn: branch.
! !

!Navigator methodsFor: 'branch functions'!
navigateRoot
	root == Object ifTrue: [ ^self ].
	NavigatorView openOn: (root superclass).
! !

!Navigator methodsFor: 'branch functions'!
lookupBranchInstance
	branch == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: branch.
! !

!Navigator methodsFor: 'branch functions'!
lookupBranchClass
	branch == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: (branch class).
! !

!Navigator methodsFor: 'leaf functions'!
browseLeaf
	leaf == nil ifTrue: [ ^self ].
	Browser newOnClass: leaf.
! !

!Navigator methodsFor: 'leaf functions'!
lookupLeafInstance
	leaf == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: leaf.
! !

!Navigator methodsFor: 'leaf functions'!
lookupLeafClass
	leaf == nil ifTrue: [ ^self ].
	self lookupSelectorFrom: (leaf class).
! !

!Navigator methodsFor: 'private'!
lookupSelectorFrom: class
	| selector scanner result |
	FillInTheBlank
		request: 'Method to look up:'
		displayAt: Sensor cursorPoint
		centered: true
		action: [ :answer | selector _ answer asSymbol ]
		initialAnswer: 'lookup chain'.
	scanner _ class.
	result _ WriteStream on: (String new: 1000).
	result print: selector; cr.
	[ scanner ~~ nil ] whileTrue:
		[ result tab; print: scanner; space.
	  	  (scanner includesSelector: selector)
			ifTrue: [ result nextPut: $+ ]
			ifFalse: [ result nextPut: $- ].
		  result cr.
		  scanner _ scanner superclass ].
	Transcript cr; show: (result contents); cr.
! !



StandardSystemView subclass: #NavigatorView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Browser'!

!NavigatorView class methodsFor: 'instance creation'!
openOn: aClass
	| aNavigator minimumSize topView |
	aNavigator _ Navigator new on: aClass.
	aClass == Object
		ifTrue: [ minimumSize _ 200@400 ]
		ifFalse: [ minimumSize _ 200@100 ].
	(topView _ self
			model: aNavigator
			label: ((aClass name) asString), ' Navigator'
			minimumSize: minimumSize)
		addBranchView: (0@0 extent: 0.5@1.0) on: aNavigator;
		addLeafView: (0.5@0 extent: 0.5@1.0) on: aNavigator.
	topView controller open
! !

!NavigatorView methodsFor: 'subview creation'!
addBranchView: area on: aNavigator
	self addSubView:
		(SelectionInListView on: aNavigator
			printItems: false oneItem: false
			aspect: #branchName change: #branchName:
			list: #branchList menu: #branchMenu
			initialSelection: #branchName)
		in: area borderWidth: 1
! !


!NavigatorView methodsFor: 'subview creation'!
addLeafView: area on: aNavigator
	self addSubView:
		(SelectionInListView on: aNavigator
			printItems: false oneItem: false
			aspect: #leafName change: #leafName:
			list: #leafList menu: #leafMenu
			initialSelection: #leafName)
		in: area borderWidth: 1
! !



!ScreenController methodsFor: 'menu messages'!
openNavigator
	| aClass |
	FillInTheBlank
		request: 'Class to navigate:'
		displayAt: Sensor cursorPoint
		centered: true
		action: [ :answer | aClass _
				Smalltalk at: (answer asSymbol) ]
		initialAnswer: 'Object'.
	((aClass subclasses) size) = 0 ifTrue: [ aClass _ aClass superclass ].
	NavigatorView openOn: aClass.
! !

!ScreenController class methodsFor: 'class initialization'!
initialize
	ScreenYellowButtonMenu _
		PopUpMenu
			labels: 
'restore display
exit project
open project
file list
browse
navigate
open workspace
system transcript
system workspace
save
quit'
			lines: #(2 9 ).
	ScreenYellowButtonMessages _
			#(restoreDisplay exitProject openProject 
			openFileList openBrowser openNavigator openWorkspace openTranscript openSystemWorkspace
			save quit).
	ScreenController allInstancesDo: [:c | c initializeYellowButtonMenu].
! !

ScreenController initialize!
