'From Smalltalk-80, version 2, of April 1, 1983 on 2 August 1983 at 12:23:37 am'!

!DisplayScreen class methodsFor: 'current display'!

displayExtent: t1 
	| t2 t3 |
	t2 _ t1 x truncateTo: 32.
	t3 _ t1 y truncateTo: 2.
	DisplayScreen currentDisplay: (DisplayScreen new
			extent: t2 @ t3
			offset: 0 @ 0
			bits: (DisplayBitmap new: 0)).
	ScheduledControllers restore! !

Object subclass: #Boot
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Berkeley Additions'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Boot class
	instanceVariableNames: ''!


!Boot class methodsFor: 'filing in'!

fileIn
	^(ReadWriteStream with: self primitiveBoot) reset fileIn! !

!Boot class methodsFor: 'primitives'!

primitiveBoot
	<primitive: 206>
	^self primitiveFailed! !


!SystemDictionary methodsFor: 'system backup/out'!

snapshotAs: t1 thenQuit: t2 
	| t3 |
	Cursor execute show.
	self logChange: '''----SNAPSHOT----'''.
	Disk == nil ifFalse: [(FileStream fileNamed: t1 , '.im') beSnapshotFile; readWrite; close].
	self releaseExternalViews.
	Transcript cr; show: 'Snapshot at: ' , Time dateAndTimeNow printString.
	Delay preSnapshot.
	Cursor wait show.
	t3 _ self snapshotPrimitive isNil.
	t3 & t2
		ifTrue: [self quitPrimitive]
		ifFalse: 
			[Delay postSnapshot.
			self install.
			ScheduledControllers restore].
	Cursor normal show.
	^t3! !

!SystemDictionary methodsFor: 'system compression'!

reorganize
	"perform offline reorganization: global mark and sweep with depth first reordering for Old space."
	<primitive: 214>
	^self primitiveFailed
	"Smalltalk reorganize"! !

!SystemDictionary methodsFor: 'private'!

lowSpaceNotificationLoop
	[true]
		whileTrue: 
			[LowSpaceSemaphore wait.
			self oldLeft <= self coreLeftLimit
				ifTrue: 
					[ScheduledControllers interruptName: 'Space is Low'.
					(BinaryChoice message: 'Do you wish to reclaim old space (takes 5 min.)?')
						ifTrue: 
							[Transcript show: self oldLeft printString , ' words left in old space before reorganization.'; cr.
							self reorganize.
							Transcript show: self oldLeft printString , ' words left in old space after reorganization.'; cr.
							ScheduledControllers restore]
						ifFalse: [self halt: 'Proceed to restart space notification, or close after a snapshot.']]
				ifFalse: 
					[ScheduledControllers interruptName: 'Infinite recursion'.
					self halt: 'Proceed to restart space notification, or close after a snapshot.'].
			self resetLowSpaceSignal]! !

File subclass: #UnixFile
	instanceVariableNames: 'fd '
	classVariableNames: 'NewLine '
	poolDictionaries: ''
	category: 'Files-Berkeley Unix'!


!UnixFile methodsFor: 'initialization'!

fd: t1 
	fd _ t1! !

!UnixFile methodsFor: 'testing'!

characterSize
	| t1 t2 |
	t2 _ UnixFile
				primitiveSeekFd: fd
				offset: 0
				pointer: 1.
	t2 < 0 ifTrue: [^self error: 'seek error'].
	t1 _ UnixFile
				primitiveSeekFd: fd
				offset: 0
				pointer: 2.
	UnixFile
		primitiveSeekFd: fd
		offset: t2
		pointer: 0.
	^t1!

size
	| t1 |
	t1 _ UnixFilePage new pageSize.
	^self characterSize + t1 - 1 quo: t1! !

!UnixFile methodsFor: 'pages'!

pageClass
	^UnixFilePage! !

!UnixFile methodsFor: 'nonhomogeneous accessing'!

readOrAdd: t1 
	^self read: t1!

seek: t1 
	| t2 |
	t2 _ UnixFile
				primitiveSeekFd: fd
				offset: t1 address
				pointer: 0.
	t2 < 0 ifTrue: [^self error: 'seek error']! !

!UnixFile methodsFor: 'file status'!

close
	^self release! !

!UnixFile methodsFor: 'file accessing'!

endFile: t1 
	| t2 |
	t1 isNil
		ifTrue: 
			[self release.
			self directory newFile: self].
	t2 _ self write: t1.
	UnixFile primitiveTruncateFd: fd length: (UnixFile
			primitiveSeekFd: fd
			offset: 0
			pointer: 1).
	^t2! !

!UnixFile methodsFor: 'snapshots'!

beSnapshotFile
	UnixFile primitiveBeSnapshotFile: fileName! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

UnixFile class
	instanceVariableNames: ''!


!UnixFile class methodsFor: 'primitives'!

primitiveBeSnapshotFile: fileName 
	<primitive: 209>
	^self primitiveFailed!

primitiveCloseFd: t1 
	<primitive: 204>
	^self primitiveFailed!

primitiveReadFd: t1 buffer: t2 length: t3 
	<primitive: 202>
	^self primitiveFailed!

primitiveSeekFd: t1 offset: t2 pointer: t3 
	<primitive: 201>
	^self primitiveFailed!


primitiveTruncateFd: fd length:len
	<primitive: 210>
	^self primitiveFailed!

primitiveWriteFd: t1 buffer: t2 length: t3 
	<primitive: 203>
	^self primitiveFailed! !

FileDirectory subclass: #UnixFileDirectory
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Berkeley Unix'!


!UnixFileDirectory methodsFor: 'file access'!

fileClass
	^UnixFile! !

!UnixFileDirectory methodsFor: 'private'!

find: t1 ifAbsentDo: t2 
	| t3 |
	t3 _ UnixFileDirectory primitiveOpenName: t1 fileName mode: 0.
	t3 < 0 ifTrue: [^t2 value: t1].
	^t1 fd: t3! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

UnixFileDirectory class
	instanceVariableNames: ''!


!UnixFileDirectory class methodsFor: 'instance creation'!

directory: t1 directoryName: t2 
	^super new setDirectory: t1 directoryName: t2!

new
	^self directory: nil directoryName: '.'! !

!UnixFileDirectory class methodsFor: 'primitives'!

primitiveCreatName: t1 mode: t2 
	<primitive: 205>
	self error: 'creat primitive failed'.
	^-1!

primitiveOpenName: t1 mode: t2 
	<primitive: 200>
	self error: 'open failed'.
	^-1! !

FilePage subclass: #UnixFilePage
	instanceVariableNames: 'size pageNumber '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Berkeley Unix'!


!UnixFilePage methodsFor: 'testing'!

size
	^size!

size: t1 
	size _ t1! !

!UnixFilePage methodsFor: 'page accessing'!

address
	^self pageNumber - 1 * self dataSize!

pageNumber
	^pageNumber!

pageNumber: t1 
	pageNumber _ t1!

serialNumber: t1 
	^self! !

!SystemOrganization!
('Numeric-Magnitudes' Date Magnitude Time)
('Numeric-Numbers' Float Fraction Integer LargeNegativeInteger LargePositiveInteger Number Random SmallInteger)
('Collections-Abstract' ArrayedCollection Collection SequenceableCollection)
('Collections-Unordered' Bag Dictionary IdentityDictionary IdentitySet Set)
('Collections-Sequenceable' Interval LinkedList MappedCollection OrderedCollection SortedCollection)
('Collections-Text' Character String Symbol Text)
('Collections-Arrayed' Array ByteArray RunArray)
('Collections-Streams' PositionableStream ReadStream ReadWriteStream Stream WriteStream)
('Collections-Support' Association Link LookupKey)
('Graphics-Primitives' Pen Point Quadrangle Rectangle)
('Graphics-Display Objects' Cursor DisplayMedium DisplayObject DisplayScreen DisplayText Form InfiniteForm OpaqueForm Paragraph Path)
('Graphics-Paths' Arc Circle Curve Line LinearFit Spline)
('Graphics-Views' DisplayTextView FormHolderView FormView)
('Graphics-Editors' BitEditor FormButtonCache FormEditor FormMenuController FormMenuView)
('Graphics-Support' BitBlt CharacterBlock CharacterBlockScanner CharacterScanner CompositionScanner DisplayBitmap DisplayScanner StrikeFont TextLineInterval TextStyle WordArray)
('Kernel-Objects' Boolean False Object True UndefinedObject)
('Kernel-Classes' Behavior Class ClassDescription Metaclass MetaclassForMultipleInheritance)
('Kernel-Methods' BlockContext CompiledMethod ContextPart InstructionPrinter InstructionStream Message MethodContext MethodDescription)
('Kernel-Processes' Delay Process ProcessorScheduler Semaphore SharedQueue)
('Kernel-Support' ClassCategoryReader ClassOrganizer MethodDictionary RemoteString SystemOrganizer)
('Interface-Framework' Controller ControlManager NoController View WindowingTransformation)
('Interface-Support' Explainer MouseMenuController ScreenController ScrollController StandardSystemController StandardSystemView)
('Interface-Lists' ListController ListView LockedListController SelectionInListController SelectionInListView TextList)
('Interface-Text' AlwaysAcceptCodeController CodeController CodeView OnlyWhenSelectedCodeController OnlyWhenSelectedCodeView ParagraphEditor StringHolder StringHolderController StringHolderView TextController TextView)
('Interface-Menus' ActionMenu BinaryChoiceView BooleanView Button IndicatorOnSwitchController LockedSwitchController OneOnSwitch PopUpMenu Switch SwitchController SwitchView)
('Interface-Prompt/Confirm' BinaryChoice BinaryChoiceController CRFillInTheBlankController FillInTheBlank FillInTheBlankController FillInTheBlankView)
('Interface-Browser' Browser BrowserView MethodListBrowser)
('Interface-Inspector' ContextInspector DictionaryInspector Inspector InspectorView)
('Interface-Debugger' Debugger NotifierController NotifierView ProcessHandle SyntaxError)
('Interface-File Model' FileList FileModel)
('Interface-Transcript' TextCollector TextCollectorController TextCollectorView)
('Interface-Projects' Project ProjectController ProjectView)
('Interface-Changes' ChangeController ChangeList ChangeListController ChangeListView)
('Berkeley Additions' Boot)
('System-Support' Benchmark InputSensor InputState KeyboardEvent MessageTally SystemDictionary)
('System-Changes' Change ChangeScanner ChangeSet ClassChange ClassCommentChange ClassDefinitionChange ClassOtherChange ClassRelatedChange MethodChange MethodDefinitionChange MethodOtherChange OtherChange)
('System-Compiler' AssignmentNode BlockNode CascadeNode Compiler Decompiler DecompilerConstructor Encoder LeafNode LiteralDictionary LiteralNode MessageNode MethodNode ParseNode Parser ParseStack ReturnNode Scanner SelectorNode VariableNode)
('System-Releasing' Checker SystemTracer)
('Files-Streams' ExternalStream FileStream)
('Files-Abstract' File FileDirectory FilePage)
('Files-Xerox Alto' AltoFile AltoFileAddressArray AltoFileDirectory AltoFilePage)
('Files-Berkeley Unix' UnixFile UnixFileDirectory UnixFilePage)
!
'From Smalltalk-80, version 2, of April 1, 1983 on 7 August 1983 at 5:42:50 am'!



!UnixFileDirectory class methodsFor: 'primitives'!

primitiveLinkName: oldName to: newName
	<primitive: 212>
	self error: 'link primitive failed'.
	^-1!

primitiveUnlinkName: t1
	<primitive: 211>
	self error: 'unlink primitive failed'.
	^-1! !


!FileDirectory methodsFor: 'removing'!

remove: aFileName ifAbsentDo: aBlock 
	"Remove aFile if found in the receiver.  Answer the result of evaluating aBlock if the file is not found.  Modified 8/7/83 D. Ungar to fix bug."

	^self removeOld: (self find: aFileName ifAbsentDo: [:ff | ^aBlock value: ff])! !

!UnixFileDirectory reorganize!
('file accessing' fileClass rename:newName:)
('private' find:ifAbsentDo:)
('adding' addNew:)
('removing' removeOld:)
!



!UnixFileDirectory methodsFor: 'file accessing'!

rename: oldFile newName: newName
	-1 == (UnixFileDirectory primitiveLinkName: oldFile fileName to:newName) ifFalse:[
		self removeOld: oldFile.
		oldFile fileName: newName]! !

!UnixFileDirectory methodsFor: 'removing'!

removeOld: aFile
	UnixFileDirectory primitiveUnlinkName: aFile fileName! !

'From Smalltalk-80, version 2, of April 1, 1983 on 9 August 1983 at 12:11:08 am'!



!UnixFile class methodsFor: 'class initialization'!

initialize
	NewLine _ 10 asCharacter.
	super initialize
	"UnixFile initialize"! !
UnixFile initialize!


!SequenceableCollection methodsFor: 'accessing'!

replaceElement: oldElement withElement: newElement
	"replaces all occurences of oldElement with newElement"

	| i size |
	i _ 1.
	size _ self size.
	[i <= size] whileTrue: [(self at: i) == oldElement ifTrue:[self at:i put: newElement].
i _ i + 1]! !

'From Smalltalk-80, version 2, of April 1, 1983 on 9 August 1983 at 1:40:25 am'!



!UnixFile methodsFor: 'nonhomogeneous accessing'!

read: t1 
	| t2  |
	self seek: t1.
	t2 _ UnixFile
				primitiveReadFd: fd
				buffer: t1 page
				length: t1 dataSize.
	-1 = t2 ifTrue: [^self error: 'read error'].
	self isBinary ifFalse:[
		t1 page replaceElement: NewLine asInteger withElement: Character cr asInteger].
	t1 size: t2.
	^t1!

write: t1 
	self isBinary ifFalse:[
		t1 page replaceElement: Character cr asInteger withElement: NewLine asInteger].
	self seek: t1.
	(UnixFile
		primitiveWriteFd: fd
		buffer: t1 page
		length: t1 size)
		= t1 size
		ifFalse: 
			[self close.
			fd _ UnixFileDirectory primitiveOpenName: fileName mode: 2.
			self seek: t1.
			(UnixFile
				primitiveWriteFd: fd
				buffer: t1 page
				length: t1 size)
				= t1 size ifFalse: [^self error: 'write error']].
	^t1! !

'From Smalltalk-80, version 2, of April 1, 1983 on 9 August 1983 at 3:46:46 pm'!


!UnixFile methodsFor: 'file status'!

release
	fd notNil ifTrue: [UnixFile primitiveCloseFd: fd]! !


!UnixFileDirectory methodsFor: 'adding'!

addNew: t1 
	| t2 |
	t1 release.
	t2 _ UnixFileDirectory primitiveCreatName: t1 fileName mode: 432.
	t2 < 0 ifTrue: [^self error: 'could not create ' , t1 fileName].
	(t1 fd: t2) close.
	^t1 reopen! !


'From Smalltalk-80, version 2, of April 1, 1983 on 15 August 1983 at 10:50:32 pm'!

!Collection methodsFor: 'private'!

growSize
	"Answer an amount by which the receiver should grow to make room for more elements (in response to the message 'grow').
	Changed by D. Ungar 8/15/83 to avoid becomes."

	self basicSize >= self maxSize ifTrue: [self error: 'unable to grow this collection'].
	^(self basicSize max: 2) * 3   min:  self maxSize - self basicSize! !
