StandardSystemController subclass: #PerformanceMonitorController
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!
PerformanceMonitorController comment:
'I make sure the daemon stops when necessary.'!


Process subclass: #PeriodicDaemon
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!
PeriodicDaemon comment:
'This class repeatedly executes a block
after a delay.'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PeriodicDaemon class
	instanceVariableNames: ''!


!PeriodicDaemon class methodsFor: 'instance termination'!

terminateAll
	"emergency stop"
	self allInstancesDo: [:d | d terminate]
	"PeriodicDaemon terminateAll"! !

!PeriodicDaemon class methodsFor: 'instance creation'!

forContext: aBlock delay: delay
	"create daemon that will execute aBlock, every delay"
	^(self forContext: [[aBlock value. delay wait] forever] priority: Processor userInterruptPriority) 
	"PeriodicDaemon forContext: [Display reverse; reverse] delay: (Delay forSeconds: 2)"! !

Object subclass: #PerformanceMonitor
	instanceVariableNames: ''
	classVariableNames: 'CurrentURU Daemon DeltaURU '
	poolDictionaries: ''
	category: 'Performance Monitoring'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PerformanceMonitor class
	instanceVariableNames: ''!


!PerformanceMonitor class methodsFor: 'inquiry'!

deltaOf: aSelector
	^DeltaURU perform: aSelector!

previousOf: aSelector
	^(self totalOf: aSelector) - (self deltaOf: aSelector)!

rateOf: aSelector
	^(DeltaURU perform: aSelector) / DeltaURU realTime!

totalOf: aSelector
	^CurrentURU perform: aSelector! !

!PerformanceMonitor class methodsFor: 'updating'!

update
	| aURU |
	aURU _ UnixResourceUsage new.
	DeltaURU _ aURU - CurrentURU.
	CurrentURU _ aURU.
	self changed! !

!PerformanceMonitor class methodsFor: 'initialization'!

initialize
	CurrentURU _ UnixResourceUsage new.
	(Delay forSeconds: 1) wait.
	Daemon == nil ifFalse: [Daemon terminate].
	Daemon _ PeriodicDaemon forContext: [self update] delay: (Delay forSeconds: 5).
	Daemon resume.
	"PerformanceMonitor initialize"! !

!PerformanceMonitor class methodsFor: 'instance creation'!

open
	| topView |
	topView _ PerformanceMonitorView
				model: self
				label: 'Performance Monitor'
				minimumSize:  16 @ 16.
	topView
		addSubView: (CPUView new model: self)
		in: (0 @ 0 extent: 1 / 3 @ 1)
		borderWidth: 1.
	topView
		addSubView: (PageFaultView new model: self)
		in: (1 / 3 @ 0 extent: 1 / 3 @ 1)
		borderWidth: 1.
	topView
		addSubView: (TenureView new model: self)
		in: (2 / 3 @ 0 extent: 1 / 3 @ 1)
		borderWidth: 1.
	topView controller open
	"PerformanceMonitor open "! !

!PerformanceMonitor class methodsFor: 'termination'!

terminateDaemon
	^Daemon terminate
	"PerformanceMonitor terminateDaemon"! !

PerformanceMonitor initialize!


Object subclass: #UnixResourceUsage
	instanceVariableNames: 'userTime systemTime maxRSS ixRSS idRSS isRSS minorFaults majorFaults swaps inBlocks outBlocks messagesSent messagesReceived signals voluntaryContextSwitches involuntaryContextSwitches realTime oldLength '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!
UnixResourceUsage comment:
'Interface to Unix rusage() system call.
All times in seconds.'!


!UnixResourceUsage methodsFor: 'primitives'!

getrusagePrimitive
	<primitive: 218>
	^self primitiveFailed! !

!UnixResourceUsage methodsFor: 'access'!

idRSS
	^idRSS / self userTime * 1024!

majorFaults
	^majorFaults!

maxRSS
	^maxRSS * 1024!

minorFaults
	^minorFaults!

oldLength
	^oldLength!

realTime
	^realTime!

systemTime
	^systemTime!

userTime
	^userTime! !

!UnixResourceUsage methodsFor: 'arithmetic'!

- aURU
	| result |
	result _ self class basicNew.
	1 to: self class instSize do: [:i | result instVarAt: i put: (self instVarAt: i) - (aURU instVarAt: i)].
	^result! !

!UnixResourceUsage methodsFor: 'initialization'!

initialize
	self getrusagePrimitive.
	realTime _ Time millisecondClockValue asFloat / 1000.0.
	oldLength _ Smalltalk oldUsed! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

UnixResourceUsage class
	instanceVariableNames: ''!


!UnixResourceUsage class methodsFor: 'instance creation'!

new
	^super new initialize! !

View subclass: #PerformanceMonitorSubview
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!


!PerformanceMonitorSubview methodsFor: 'initialization'!

initialize
	insideColor _ Form white.
	super initialize! !

!PerformanceMonitorSubview methodsFor: 'updating'!

update: ignored
	(ScheduledControllers scheduledControllers includes: self topView controller) ifTrue: [self displayView]! !

PerformanceMonitorSubview subclass: #TwoPartBarGraphView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!


!TwoPartBarGraphView methodsFor: 'defaults'!

bottom
	^0!

middle
	^0!

top
	^1! !

!TwoPartBarGraphView methodsFor: 'displaying'!

displayView
	| box b m |
	box _ self insetDisplayBox.
	b _ box bottom - (self bottom / self top * box height) rounded max: box top.
	m _ b - (self middle / self top * box height) rounded max: box top.
	
	
	Display fill: (box origin corner: box right @ m) mask: Form white.
	Display fill: (box left @ m  corner: box right @ b) mask: Form lightGray.
	Display fill: (box left @ b  corner: box corner) mask: Form darkGray! !

TwoPartBarGraphView subclass: #CPUView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!


!CPUView methodsFor: 'values'!

bottom
	^model rateOf: #userTime!

middle
	^model rateOf: #systemTime! !

TwoPartBarGraphView subclass: #PageFaultView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!


!PageFaultView methodsFor: 'values'!

bottom
	^model rateOf: #majorFaults!

middle
	^model rateOf: #minorFaults!

top
	^4! !

TwoPartBarGraphView subclass: #TenureView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!


!TenureView methodsFor: 'values'!

bottom
	^(PerformanceMonitor previousOf: #oldLength) - ((PerformanceMonitor previousOf: #maxRSS) / 4)!

middle
	^model deltaOf: #oldLength!

top
	^1024 * 128! !

StandardSystemView subclass: #PerformanceMonitorView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Performance Monitoring'!
PerformanceMonitorView comment:
'A StandardSystemView using PerformanceMonitorCOntroller.'!


!PerformanceMonitorView methodsFor: 'controller access'!

defaultControllerClass
	^PerformanceMonitorController! !