'From Smalltalk-80, version 2, of April 1, 1983 on 7 July 1984 at 5:27:02 pm'!

!ClassDescription methodsFor: 'fileIn/Out'!

printOutCategory: aString 
	"Create a readable version of the message category aString, and send to a printer."
	| fileName fileStream |
	fileName _ Disk checkName: self name , '-' , aString, '.bs' fixErrors: true.
	fileStream _ Disk file: fileName.
	self printOutStartUp: fileStream.
	self printOutTimeStamp: fileStream.
	self printOutCategory: aString on: fileStream.
	fileStream shorten; close!

printOutCategory: aString on: aFileStream 
	"File a description of the receiver's category, aString, onto 
	aFileStream. "

	self printOutMacro: 'PN' with: aString on: aFileStream.
	(self organization listAtCategoryNamed: aString)
		do: [:sel | self printOutMessage: sel on: aFileStream]!

printOutCommentOn: fileStream
	| comm prev | 
	comm _ self comment.
	comm size = 0 ifTrue: [^self].
	self printOutMacro: 'CS' with: nil on: fileStream.
	prev _ Character space.
	comm do: [:this |
		((this = Character tab) and: [prev = Character cr])
			ifTrue:
				[self printOutMacro: 'br' with: nil on: fileStream].
		fileStream nextPut: this.
		prev _ this].
	fileStream cr.
	self printOutMacro: 'CE' with: nil on: fileStream!

printOutDefinitionOn: fileStream
	| names first sc | 
	self printOutMacro: 'CL' with: self name on: fileStream.
	sc _ self superclass.
	sc isNil
		ifTrue: [self printOutMacro: 'SC' with: 'nil' on: fileStream]
		ifFalse: [self printOutMacro: 'SC' with: sc name on: fileStream].

	names _ self instVarNames.
	names size = 0
		ifTrue: [self printOutMacro: 'IV' with: '\fInone\fP' on: fileStream].
	names size > 0
		ifTrue: [self printOutMacro: 'IV' with: (names at: 1) on: fileStream].
	names size > 1
		ifTrue: [2 to: names size
				do: [:index | self printOutMacro: 'IC' with: (names at: index) on: fileStream]].

	names _ self classPool keys asSortedCollection.
	names size = 0
		ifTrue: [self printOutMacro: 'CV' with: '\fInone\fP' on: fileStream]
		ifFalse:
			[first _ true.
			names do: [:each |
				first
					ifTrue:
						[self printOutMacro: 'CV' with: each on: fileStream.
						first _ false]
				ifFalse: [self printOutMacro: 'CC' with: each on: fileStream]]].

	names _ self sharedPools.
	names size = 0
		ifTrue: [self printOutMacro: 'PD' with: '\fInone\fP' on: fileStream]
		ifFalse:
			[first _ true.
			names do: [:each |
				first
					ifTrue:
						[self printOutMacro: 'PD'
							with: (Smalltalk keyAtValue: each) on: fileStream.
						first _ false]
					ifFalse:
						[self printOutMacro: 'PC'
							with: (Smalltalk keyAtValue: each) on: fileStream]]].

	self category ~= nil
		ifTrue: [self printOutMacro: 'CA' with: self category on: fileStream]!

printOutMacro: macroName with: argString on: aFileStream
	aFileStream nextPut: $.; nextPutAll: macroName.
	argString ~= nil
		ifTrue: [aFileStream space; nextPut: $"; nextPutAll: argString asString; nextPut: $"].
	aFileStream cr!

printOutMessage: aString 
	"Create a readable version of the message with selector aString, and send to a printer.
	Defaults to fileOut."
	self printOutMessage: aString
		fileName: (Disk checkName: self name , '-' , aString , '.bs' fixErrors: true)!

printOutMessage: aString fileName: fileName 
	| fileStream |
	fileStream _ Disk file: fileName.
	self printOutStartUp: fileStream.
	self printOutTimeStamp: fileStream.
	self printOutMessage: aString on: fileStream.
	fileStream close!

printOutMessage: aString on: aFileStream 
	"Create troff ready file for the method aString"

	| sourceCode bodyStart inComment |
	sourceCode _ self sourceCodeAt: aString asSymbol.
	bodyStart _ sourceCode findString: '\' withCRs startingAt: 1.
	self printOutMacro: 'MH'
		with: (sourceCode copyFrom: 1 to: (bodyStart - 1))
		on: aFileStream.
	bodyStart = 0
		ifTrue: 
			[aFileStream tab; nextPutAll: '\(uaself'; cr; nextPutAll: '.ME'; cr.
			^self].
	inComment _ false.
	(sourceCode copyFrom: bodyStart + 1 to: sourceCode size)
		do: [:char |
			char = $^
				ifTrue: [aFileStream nextPutAll: '\(ua']
				ifFalse: [char = $\
					ifTrue: [aFileStream nextPutAll: '\e']
					ifFalse: [char = $~
						ifTrue: [aFileStream nextPutAll: '\(ap']
						ifFalse: [char = $_
							ifTrue: [aFileStream nextPutAll: '\(<-']
							ifFalse: [(char = $') | (char = $.)
								ifTrue: [aFileStream nextPutAll: '\&'; nextPut: char]
								ifFalse: [char = $"
									ifTrue: [aFileStream nextPutAll:
										(inComment ifTrue: ['"\fP'] ifFalse: ['\fI"']).
										inComment _ inComment not]
									ifFalse: [aFileStream nextPut: char]]]]]]].
	(sourceCode at: sourceCode size) ~= Character cr
		ifTrue: [aFileStream cr].
	aFileStream nextPutAll: '.ME'; cr!

printOutOn: aFileStream
	"print me out on aFileStream"
	self printOutDefinitionOn: aFileStream.
	self printOutCommentOn: aFileStream.
	self organization categories
		do: [:heading | self printOutCategory: heading on: aFileStream]!

printOutStartUp: fileStream
	fileStream nextPutAll: '.so /usr/local/bs/lib/tmac.st'; cr!

printOutTimeStamp: aStream
	aStream nextPutAll: '.TI "'.
	Smalltalk timeStamp: aStream.
	aStream nextPut: $"; cr! !


!Class methodsFor: 'fileIn/Out'!

printOut
	"Create a readable version (in *roff format) of my definition."
	| fileStream |
	Transcript refresh; cr; cr; show: 'Printing out class:'.
	fileStream _ Disk file: self name , '.bs'.
	self printOutStartUp: fileStream.
	self printOutTimeStamp: fileStream.
	self printOutOn: fileStream.
	fileStream shorten; close.
	self removeFromChanges!

printOutOn: aFileStream
	"File a description of the receiver on aFileStream."

	Transcript cr; show: name.
	super printOutOn: aFileStream.
	self class nonTrivial ifTrue: [self class printOutOn: aFileStream]! !

!SystemOrganizer methodsFor: 'fileIn/Out'!

printOutCategory: category
	| aFileStream |
	Transcript refresh; cr; cr; show: ('Printing out category: ', category).
	aFileStream _ Disk file: (category, '.bs') asFileName.
	Object printOutStartUp: aFileStream.
	Object printOutTimeStamp: aFileStream.
	self printOutCategory: category on: aFileStream.
	aFileStream shorten; close!

printOutCategory: category on: aFileStream
	| class first |
	first _ true.
	(self superclassOrder: category) do: 
		[:class | 
		first
			ifTrue: [first _ false]
			ifFalse: [class printOutMacro: 'BP' with: nil on: aFileStream].
		class printOutOn: aFileStream.
		class removeFromChanges]! !
