'From Smalltalk-80, version 2, of April 1, 1983 on 11 April 1984 at 1:51:51 pm'!

Object subclass: #ProjectBrowser
	instanceVariableNames: 'projects currentProject '
	classVariableNames: 'ProjectMenu TextMenu '
	poolDictionaries: ''
	category: 'Interface-Projects'!
ProjectBrowser comment:
'This class represents a browser fpr createing and accessing projects.
Create one by evaluating 

		ProjectBrowser open'!


!ProjectBrowser methodsFor: 'initialize-release'!

initialize
	"Set the instance variables."

	projects _ Project allInstances! !

!ProjectBrowser methodsFor: 'list access'!

currentProject
	"Answer the currently selected project."

	^currentProject!

currentProject: aProject 
	"Set the currently selected project to be aProject"

	currentProject _ aProject.
	self changed: #text!

projectList
	"Answer the collection of items to be displayed in the menu."

	^projects!

projectMenu
	"Answer a menu fpr accessing or changing projects."

	currentProject isNil
		ifTrue: [ProjectMenu _ ActionMenu
						labels: 'add project'
						lines: #()
						selectors: #(addProject )]
		ifFalse: [ProjectMenu _ ActionMenu
						labels: 'update\add project\rename\remove\enter' withCRs
						lines: #(1 4)
						selectors: #(updateProject addProject renameProject removeProject enterProject)].
	^ProjectMenu! !

!ProjectBrowser methodsFor: 'list function'!

addProject
	"Add another project to the menu of projects."

	| aProject aString |
	aString _ FillInTheBlank request: 'Title for Project?' initialAnswer: 'a Project'.
	aString isEmpty ifTrue: [^self].
	aProject _ Project new title: aString.
	projects add: aProject.
	currentProject _ aProject.
	self changed: #currentProject!

enterProject
	"Make the selected project the current one."

	currentProject notNil ifTrue: [currentProject enter]!

removeProject
	"Remove the currently selected project from the menu."

	currentProject isNil ifTrue: [^self].
	currentProject == Project current
		ifTrue: [^self error: 'You cannot remove the project in which you are currently running.']
		ifFalse: [(BinaryChoice message: 'Are you sure you want to
remove everything in this project?')
				ifTrue: 
					[currentProject release.
					projects remove: currentProject ifAbsent: [].
					currentProject _ nil.
					self changed: #currentProject]
				ifFalse: [^self]]!

renameProject

	| aString |
	aString _ FillInTheBlank request: 'New Title for Project?' initialAnswer: (currentProject title).
	aString isEmpty ifTrue: [^self].
	currentProject title: aString.
	self changed: #currentProject!

updateProject
	"update the menu of projects."

	self changed: #currentProject! !

!ProjectBrowser methodsFor: 'text access'!

acceptText: aText from:whoCares "Store the text to be displayed in the code view."currentProject contents: aText string. ^true!

text
	"Answer the text to be displayed in the code view."

	currentProject isNil
		ifTrue: [^'' asText]
		ifFalse: [^currentProject contents asText]!

textMenu
	"Answer a menu for editing the text in the code view."

	currentProject isNil ifTrue: [^nil].
	TextMenu isNil ifTrue: [TextMenu _ ActionMenu
					labels: 'again\undo\copy\cut\paste\accept\cancel' withCRs
					lines: #(2 5 )
					selectors: #(again undo copySelection cut paste accept cancel )].
	^TextMenu! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ProjectBrowser class
	instanceVariableNames: ''!


!ProjectBrowser class methodsFor: 'instance creation'!

open
	"Create and schedule a browser for all the projects in the system."

	| topView aBrowser |
	aBrowser _ super new initialize.
	topView _ StandardSystemView
				model: aBrowser
				label: 'Project Browser'
				minimumSize: 200 @ 150.
	topView
		addSubView: (SelectionInListView
				on: aBrowser
				printItems: true
				oneItem: false
				aspect: #currentProject
				change: #currentProject:
				list: #projectList
				menu: #projectMenu
				initialSelection: nil)
		in: (0 @ 0.0 extent: 1 @ 0.4)
		borderWidth: 1.
	topView
		addSubView: (CodeView
				on: aBrowser
				aspect: #text
				change: #acceptText:from:
				menu: #textMenu
				initialSelection: nil)
		in: (0 @ 0.4 extent: 1 @ 0.6)
		borderWidth: 1.
	topView controller open! !

!ScreenController methodsFor: 'menu messages'!

openProjectBrowser
	ProjectBrowser open! !

!ScreenController class methodsFor: 'class initialization'!

initialize
	ScreenYellowButtonMenu _
		PopUpMenu
			labels: 
'restore display
project browser
file list
browse
navigate
open workspace
system transcript
system workspace
save
suspend
quit'
			lines: #(2 8).
	ScreenYellowButtonMessages _
			#(restoreDisplay openProjectBrowser 
			openFileList openBrowser openNavigator openWorkspace openTranscript openSystemWorkspace
			save suspend quit).
	ScreenController allInstancesDo: [:c | c initializeYellowButtonMenu].

	"ScreenController initialize."! !

ScreenController initialize!
