Object subclass: #ProtocolBrowser
	instanceVariableNames: 'list classDictionary selectedClass selectedSelector '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Protocol'!
ProtocolBrowser comment:
'ProtocolBrowser openForClass: className

See Goldberg book, page 300.'!


!ProtocolBrowser methodsFor: 'private'!

on: aClass 
	"Create the protocol browser the the class, aClass."

	| defClass |
	list _ OrderedCollection new.
	classDictionary _ Dictionary new.
	aClass allSelectors asSortedCollection do: 
		[:selector | 
		defClass _ aClass whichClassIncludesSelector: selector.
		list add: selector printString , '	(' , defClass name , ')'.
		classDictionary add: (Association key: selector value: defClass)]! !

!ProtocolBrowser methodsFor: 'list access'!

selector: aString 
	"Answer the message selector for changing the information that 
	shows in the list menu."

	aString == nil
		ifTrue: 
			[selectedSelector _ nil.
			^self].
	selectedSelector _ (aString copyUpTo: Character tab) asSymbol.
	selectedClass _ classDictionary at: selectedSelector.
	self changed: #text!

selectorList
	"Answer the list of strings that show in the list menu."

	^list!

selectorListMenu
	"Answer a menu for making queries about messages and methods."

	^selectedSelector == nil ifFalse: [ActionMenu
			labels: 'senders\implementors\messages' withCRs
			lines: #()
			selectors: #(browseSenders browseImplementors browseMessages )]! !

!ProtocolBrowser methodsFor: 'text access'!

acceptText: aText from: aController 
	"Answer the message selector for changing the displayed text."

	| newSelector |
	newSelector _ selectedClass
				compile: aText
				classified: (selectedClass organization categoryOfElement: selectedSelector)
				notifying: aController.
	newSelector == nil ifTrue: [^false].
	newSelector == selectedSelector ifFalse: [self newSelectorList: newSelector].
	^true!

text
	"Answer the Symbol identifying the text of the menu."

	^selectedSelector == nil
		ifTrue: ['' asText]
		ifFalse: [(selectedClass sourceCodeAt: selectedSelector) asText makeSelectorBoldIn: selectedClass]!

textMenu
	"Answer a menu for requesting edits of the text of a method."

	^ActionMenu
		labels: 'again\undo\copy\cut\paste\accept\cancel' withCRs
		lines: #(2 5 )
		selectors: #(again undo copySelection cut paste accept cancel )! !

!ProtocolBrowser methodsFor: 'list functions'!

browseImplementors
	"Create a message-set browser for the implementors of the
	selected message."

	Smalltalk browseAllImplementorsOf: selectedSelector!

browseMessages
	"Create a menu of messages in the selected method and then, if the user selects a menu item, create a message-set browser for the methods that implement it."

	Smalltalk showMenuThenBrowse: (selectedClass compiledMethodAt: selectedSelector) messages asSortedCollection!

browseSenders
	"Create a message-set browser for the metjods that include the 
	selected message."

	Smalltalk browseAllCallsOn: selectedSelector! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ProtocolBrowser class
	instanceVariableNames: ''!


!ProtocolBrowser class methodsFor: 'instance creation'!

openForClass: aClass 
	"Create and schedule a browser fpr the entire protocol of the class."

	| topView aPBrowser label |
	aPBrowser _ super new on: aClass.
	label _ 'Entire protocol of: ' , aClass name.
	topView _ StandardSystemView
				model: aPBrowser
				label: label
				minimumSize: 200 @ 200.
	topView
		addSubView: (SelectionInListView
				on: aPBrowser
				aspect: #selector
				change: #selector:
				list: #selectorList
				menu: #selectorListMenu
				initialSelection: nil)
		in: (0 @ 0.0 extent: 1 @ 0.3)
		borderWidth: 1.
	topView
		addSubView: (CodeView
				on: aPBrowser
				aspect: #text
				change: #acceptText:from:
				menu: #textMenu
				initialSelection: nil)
		in: (0 @ 0.3 extent: 1 @ 0.7)
		borderWidth: 1.
	topView controller open! !

ProtocolBrowser subclass: #SingleMessageProtocolBrowser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Protocol'!
