File subclass: #UnixFile
	instanceVariableNames: 'fd '
	classVariableNames: 'NewLine '
	poolDictionaries: ''
	category: 'Files-Berkeley Unix'!


!UnixFile methodsFor: 'initialization'!

fd: t1 
	fd _ t1! !

!UnixFile methodsFor: 'testing'!

characterSize
	| t1 t2 |
	t2 _ UnixFile
				primitiveSeekFd: fd
				offset: 0
				pointer: 1.
	t2 < 0 ifTrue: [^self error: 'seek error'].
	t1 _ UnixFile
				primitiveSeekFd: fd
				offset: 0
				pointer: 2.
	UnixFile
		primitiveSeekFd: fd
		offset: t2
		pointer: 0.
	^t1!

size
	| t1 |
	t1 _ UnixFilePage new pageSize.
	^self characterSize + t1 - 1 quo: t1! !

!UnixFile methodsFor: 'pages'!

pageClass
	^UnixFilePage! !

!UnixFile methodsFor: 'nonhomogeneous accessing'!

readOrAdd: t1 
	^self read: t1!

seek: t1 
	| t2 |
	t2 _ UnixFile
				primitiveSeekFd: fd
				offset: t1 address
				pointer: 0.
	t2 < 0 ifTrue: [^self error: 'seek error']! !

!UnixFile methodsFor: 'file status'!

close
	^self release! !

!UnixFile methodsFor: 'file accessing'!

endFile: t1 
	| t2 |
	t1 isNil
		ifTrue: 
			[self release.
			self directory newFile: self].
	t2 _ self write: t1.
	UnixFile primitiveTruncateFd: fd length: (UnixFile
			primitiveSeekFd: fd
			offset: 0
			pointer: 1).
	^t2! !

!UnixFile methodsFor: 'snapshots'!

beSnapshotFile
	UnixFile primitiveBeSnapshotFile: self fullName! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

UnixFile class
	instanceVariableNames: ''!


!UnixFile class methodsFor: 'primitives'!

primitiveBeSnapshotFile: fileName 
	<primitive: 209>
	^self primitiveFailed!

primitiveCloseFd: t1 
	<primitive: 204>
	^self primitiveFailed!

primitiveReadFd: t1 buffer: t2 length: t3 
	<primitive: 202>
	^self primitiveFailed!

primitiveSeekFd: t1 offset: t2 pointer: t3 
	<primitive: 201>
	^self primitiveFailed!


primitiveTruncateFd: fd length:len
	<primitive: 210>
	^self primitiveFailed!

primitiveWriteFd: t1 buffer: t2 length: t3 
	<primitive: 203>
	^self primitiveFailed! !

FileDirectory subclass: #UnixFileDirectory
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Berkeley Unix'!


!UnixFileDirectory methodsFor: 'file access'!

fileClass
	^UnixFile! !

!UnixFileDirectory methodsFor: 'private'!

find: t1 ifAbsentDo: t2 
	| t3 |
	t3 _ UnixFileDirectory primitiveOpenName: t1 fullName mode: 0.
	t3 < 0 ifTrue: [^t2 value: t1].
	^t1 fd: t3! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

UnixFileDirectory class
	instanceVariableNames: ''!


!UnixFileDirectory class methodsFor: 'instance creation'!

directory: t1 directoryName: t2 
	^super new setDirectory: t1 directoryName: t2!

new
	^self directory: nil directoryName: '.'! !

!UnixFileDirectory class methodsFor: 'primitives'!

primitiveCreatName: t1 mode: t2 
	<primitive: 205>
	self error: 'creat primitive failed'.
	^-1!

primitiveOpenName: t1 mode: t2 
	<primitive: 200>
	self error: 'open failed'.
	^-1! !

FilePage subclass: #UnixFilePage
	instanceVariableNames: 'size pageNumber '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Berkeley Unix'!


!UnixFilePage methodsFor: 'testing'!

size
	^size!

size: t1 
	size _ t1! !

!UnixFilePage methodsFor: 'page accessing'!

address
	^self pageNumber - 1 * self dataSize!

pageNumber
	^pageNumber!

pageNumber: t1 
	pageNumber _ t1!

serialNumber: t1 
	^self! !



!UnixFileDirectory class methodsFor: 'primitives'!

primitiveLinkName: oldName to: newName
	<primitive: 212>
	self error: 'link primitive failed'.
	^-1!

primitiveUnlinkName: t1
	<primitive: 211>
	self error: 'unlink primitive failed'.
	^-1! !


!FileDirectory methodsFor: 'removing'!

remove: aFileName ifAbsentDo: aBlock 
	"Remove aFile if found in the receiver.  Answer the result of evaluating aBlock if the file is not found.  Modified 8/7/83 D. Ungar to fix bug."

	^self removeOld: (self find: aFileName ifAbsentDo: [:ff | ^aBlock value: ff])! !

!UnixFileDirectory reorganize!
('file accessing' fileClass rename:newName:)
('private' find:ifAbsentDo:)
('adding' addNew:)
('removing' removeOld:)
!


!UnixFileDirectory methodsFor: 'file accessing'!

rename: oldFile newName: newName 
	-1 == (UnixFileDirectory primitiveLinkName: oldFile fullName to: newName)
		ifFalse: 
			[self removeOld: oldFile.
			UnixFileDirectory directoryFromName: newName setFileName: [:fileName | oldFile fileName: fileName]]! !

!UnixFileDirectory methodsFor: 'removing'!


removeOld: t1 
	UnixFileDirectory primitiveUnlinkName: t1 fullName! !

'From Smalltalk-80, version 2, of April 1, 1983 on 9 August 1983 at 12:11:08 am'!



!UnixFile class methodsFor: 'class initialization'!

initialize
	NewLine _ 10 asCharacter.
	super initialize
	"UnixFile initialize"! !
UnixFile initialize!


!SequenceableCollection methodsFor: 'accessing'!

replaceElement: oldElement withElement: newElement
	"replaces all occurences of oldElement with newElement"

	| i size |
	i _ 1.
	size _ self size.
	[i <= size] whileTrue: [(self at: i) == oldElement ifTrue:[self at:i put: newElement].
i _ i + 1]! !

'From Smalltalk-80, version 2, of April 1, 1983 on 9 August 1983 at 1:40:25 am'!



!UnixFile methodsFor: 'nonhomogeneous accessing'!

read: t1 
	| t2  |
	self seek: t1.
	t2 _ UnixFile
				primitiveReadFd: fd
				buffer: t1 page
				length: t1 dataSize.
	-1 = t2 ifTrue: [^self error: 'read error'].
	self isBinary ifFalse:[
		t1 page replaceElement: NewLine asInteger withElement: Character cr asInteger].
	t1 size: t2.
	^t1!


write: t1 
	self isBinary ifFalse: [t1 page replaceElement: Character cr asInteger withElement: NewLine asInteger].
	self seek: t1.
	(UnixFile
		primitiveWriteFd: fd
		buffer: t1 page
		length: t1 size)
		= t1 size
		ifFalse: 
			[self close.
			fd _ UnixFileDirectory primitiveOpenName: self fullName mode: 2.
			self seek: t1.
			(UnixFile
				primitiveWriteFd: fd
				buffer: t1 page
				length: t1 size)
				= t1 size ifFalse: [^self error: 'write error']].
	^t1! !

'From Smalltalk-80, version 2, of April 1, 1983 on 9 August 1983 at 3:46:46 pm'!


!UnixFile methodsFor: 'file status'!

release
	fd notNil ifTrue: [UnixFile primitiveCloseFd: fd]! !


!UnixFileDirectory methodsFor: 'adding'!

addNew: t1 
	| t2 |
	t1 release.
	t2 _ UnixFileDirectory primitiveCreatName: t1 fullName mode: 432.
	t2 < 0 ifTrue: [^self error: 'could not create ' , t1 fullName].
	(t1 fd: t2) close.
	^t1 reopen! !


'From Smalltalk-80, version 2, of April 1, 1983 on 19 July 1984 at 2:59:45 pm'!



!UnixFileDirectory methodsFor: 'file accessing'!

filesMatching: pattern 
	"Answer an Array of the names of files that match the string, pattern."
	| names name |
	names _ WriteStream on: (Array new: 10).
	self do: 
		[:each |
		name _ each fileName.
		(pattern match: name)
			ifTrue:
				[names nextPut: name]].
	^names contents! !


!UnixFileDirectory methodsFor: 'enumerating'!

next
	| name |
	name _ self class primitiveReadDir.
	^name isNil ifFalse: [(self fileClass new) directory: self; fileName: name]!


reset
	self class primitiveOpenDir: self fullName! !

!UnixFileDirectory class methodsFor: 'primitives'!


primitiveOpenDir
	<primitive: 221>
	self error: 'opendir primitive failed'.
	^-1!

primitiveReadDir
	<primitive: 222>
	^self error: 'readdir primitive failed'! !

!FileDirectory class methodsFor: 'instance creation'!

directoryFromName: fileDesignator setFileName: localNameBlock 
	"Answer the file directory implied from the designator, presumably at the 
	 start of the  
	designator, and evaluate the block with the rest of the designator  
	(presumably the  
	file name itself."

	| directoryName fileName lastSlash |
	lastSlash _ fileDesignator findLast: [:c | c == $/].
	lastSlash = 0
		ifTrue: 
			[directoryName _ '.'.
			fileName _ fileDesignator]
		ifFalse: 
			[directoryName _ fileDesignator copyFrom: 1 to: (lastSlash - 1 max: 1).
			fileName _ fileDesignator copyFrom: lastSlash + 1 to: fileDesignator size].
	localNameBlock value: fileName.
	^UnixFileDirectory directory: nil directoryName: directoryName! !
'From Smalltalk-80, version 2, of April 1, 1983 on 20 July 1984 at 3:33:46 pm'!



!UnixFileDirectory methodsFor: 'file accessing'!

fullName
	"Answer the name of the receiver in a form recognized when parsing file names (i.e., using [...])."
	directoryName = '' ifTrue:[^'.'].
	^directoryName! !

'From Smalltalk-80, version 2, of April 1, 1983 on 20 July 1984 at 3:23:34 pm'!



!UnixFile methodsFor: 'file accessing'!

fullName
	| dn | 
	"Answer the name of the receiver, including directory names if 
	necessary. "
	dn _ self directory fullName.
	dn = '.' ifTrue:[^self fileName].
	dn = '/' ifTrue:[^'/', self fileName].
	^dn, '/',  self fileName! !

'From Smalltalk-80, version 2, of April 1, 1983 on 20 July 1984 at 3:10:25 pm'!



!FileStream methodsFor: 'file accessing'!

name
	"Answer the name of the file for the page the receiver is streaming over."
	^page file fullName! !

'From Smalltalk-80, version 2, of April 1, 1983 on 20 July 1984 at 2:49:35 pm'!



!UnixFile methodsFor: 'error handling'!

error: e 
	"Construct an error message using the argument, e.  Subclassed for full name."

	| s errorString |
	s _ WriteStream on: (String new: 50).
	s nextPutAll: self fullName.
	s nextPutAll: ' '.
	s nextPutAll: e.
	s nextPutAll: ', '.
	s nextPutAll: error.
	errorString _ s contents.
	error _ ''.
	^super error: errorString! !

'From Smalltalk-80, version 2, of April 1, 1983 on 20 July 1984 at 2:24:39 pm'!



!UnixFile methodsFor: 'printing'!

printOn: aStream 
	self fullName printOn: aStream! !

'From Smalltalk-80, version 2, of April 1, 1983 on 20 July 1984 at 2:15:22 pm'!



!FileDirectory methodsFor: 'adding'!

add: aFile 
	"Add the file, aFile, to the receiver.   This creates a new version of the 
	file rather than generating an error if a file with the same name exists."

	| old file |
	file _ aFile.
	self versionNumbers
		ifTrue: ["ignore explicit version and directory will create a next version"
			file _ self initFileName: (file fileName copyUpTo: $!!)]
		ifFalse: [(self find: file) notNil
				ifTrue: 
					["otherwise, if file already exists,  
					  rename it to name$, deleting that file first if it exists"
					old _ self initFileName: (file fileName copyWith: $$).
					self remove: old ifAbsentDo: [:f | false].
					self rename: file newName: old fullName.
					file fileName: (file fileName copyUpTo: $$).
					(self find: file) notNil ifTrue: [file error: 'insert/rename ??']]].
	self addNew: file.
	^file! !
