'From Smalltalk-80, version 2, of April 1, 1983 on 7 June 1984 at 11:04:01 pm'!

String organization addCategory: #primitive before: nil!

!String methodsFor: 'primitive'!

system
	"Type the string into the Unix system, by DMU, 5/14/84"
	<primitive: 217>
	^self primitiveFailed
	"'echo Hi there >/dev/tty' system"! !

!String class methodsFor: 'instance creation'!

fromString: aString 
	"Answer a new String that is a copy of the argument, aString."
	
	| newString |
	newString _ self new: aString size.
	1 to: aString size do: [:i | newString at: i put: (aString at: i)].
	^newString! !

!String methodsFor: 'converting'!

asUnixCommand
	"Answer a Paragraph whose text string is the receiver."

	^UnixCommand withString: self! !

!Text methodsFor: 'converting'!

asUnixCommand
	"Converts the string to a Unix Command."

	^UnixCommand withString: string! !



Object subclass: #UnixCommand
	instanceVariableNames: 'command redirectInput redirectOutput '
	classVariableNames: 'Shell Stdout '
	poolDictionaries: ''
	category: 'Berkeley Additions'!
UnixCommand comment:
'I am an interface to the wonderful world of Unix.

by PKL@Berkeley, 5/18/84.'!


!UnixCommand methodsFor: 'command execution'!

unixItWithInputFrom: aFileName WithShell: aShell 
	"adds input and output redirection only if they have not already been 
	redirected.  Return the content of the command output as a string.  
	Throws away the terminating carriage return."

	| aCommand output |
	aCommand _ aShell , ' "' , command.
	redirectInput ifFalse: [aCommand _ aCommand , ' < ' , aFileName].
	redirectOutput ifFalse: [aCommand _ aCommand , ' >&!! ' , Stdout].
	aCommand _ aCommand , '"'.
	Cursor execute showWhile: [aCommand system].
	redirectOutput
		ifTrue: [output _ '']
		ifFalse: 
			[output _ (FileStream oldFileNamed: Stdout) contentsOfEntireFile.
			output size ~= 0 
				ifTrue: [(output at: output size) = Character cr 
					ifTrue: [output at: output size put: Character space]]].
	^output

	"(UnixCommand new: 'echo Hi There') unixItWithInputFrom: '/dev/null' WithShell: '/bin/csh -f -c'"!

unixItWithShell: aShell 
	"Prevents hanging forever if the command requires input from   
	Standard Input by always redirecting input from /dev/null."

	^self unixItWithInputFrom: '/dev/null' WithShell: Shell
	"(UnixCommand new: 'echo Hi There') unixItWithShell: '/bin/csh -f -c'"!

unixIt
	^self unixItWithShell: Shell
	"(UnixCommand new: 'echo Hi There') unixIt"! !


!UnixCommand methodsFor: 'private'!

setCommand: aString 
	"replace all CRs with spaces.  Also test if the command has > or <."

	| index |
	command _ (String fromString: aString)
				replaceElement: Character cr withElement: Character space.
	redirectInput _ redirectOutput _ false.
	index _ command indexOf: $<.
	(index ~= 0) & (index ~= 1)
		ifTrue: [(command at: index - 1) ~= $\ 
			ifTrue: [redirectInput _ true]].
	index _ command indexOf: $>.
	(index ~= 0) & (index ~= 1)
		ifTrue: [(command at: index - 1) ~= $\ 
			ifTrue: [redirectOutput _ true]]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

UnixCommand class
	instanceVariableNames: ''!


!UnixCommand class methodsFor: 'class initialization'!

defaultShell: aString 
	"set default shell"

	Shell _ String fromString: aString
	"UnixCommand defaultShell: '/bin/csh -c'"!

initialize
	self defaultShell: '/bin/csh -f -c'.
	Stdout _ '.##UnixIt.Stdout'
	"UnixCommand initialize"! !

!UnixCommand class methodsFor: 'instance creation'!

withString: aString 
	^self new setCommand: aString! !

UnixCommand initialize!


!CodeController methodsFor: 'menu messages'!

doIt
	"Evaluate the current text selection as an expression.  Left-shift doIt is a unix-doit."

	| result selectionStart oldTextSize selection |
	self controlTerminate.
	selectionStart _ startBlock stringIndex.
	oldTextSize _ self text size.
	selection _ self selection.
	Sensor leftShiftDown
		ifTrue: [result _ self selection asUnixCommand unixIt]
		ifFalse: [result _ model doItReceiver class evaluatorClass new
						evaluate: self selectionAsStream
						in: model doItContext
						to: model doItReceiver
						notifying: self
						ifFail: 
							[self controlInitialize.
							^#failedDoit]].
	self selection asString = selection asString ifFalse: [self selectFrom: selectionStart to: selectionStart + selection size - 1 + (self text size - oldTextSize)
		"Reselect doIt range after compiler interaction"].
	Smalltalk logChange: self selection string.
	model doItValue: result.
	self controlInitialize.
	^result! !


!StringHolderController methodsFor: 'menu messages'!

doIt
	"Treat the current text selection as an expression; evaluate it.  Left-Shift doit is a Unix-doit."

	| result |
	self controlTerminate.
	Sensor leftShiftDown
		ifTrue: [result _ self selection asUnixCommand unixIt]
		ifFalse: [result _ model doItReceiver class evaluatorClass new
						evaluate: self selectionAsStream
						in: model doItContext
						to: model doItReceiver
						notifying: self
						ifFail: 
							[self controlInitialize.
							^#failedDoit]].
	Smalltalk logChange: self selection string.
	self controlInitialize.
	^result! !


StringHolder subclass: #UnixWorkspace
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Text'!
UnixWorkspace comment:
'Inherits everything from StringHolder.'!


StringHolderController subclass: #UnixWorkspaceController
	instanceVariableNames: ''
	classVariableNames: 'UnixYellowButtonMenu UnixYellowButtonMessages '
	poolDictionaries: ''
	category: 'Interface-Text'!
UnixWorkspaceController comment:
'Defines a new yellow button menu, which contains a unixIt selection.'!


!UnixWorkspaceController methodsFor: 'private'!

initializeYellowButtonMenu
	self yellowButtonMenu: UnixYellowButtonMenu 
		yellowButtonMessages: UnixYellowButtonMessages! !

!UnixWorkspaceController methodsFor: 'menu messages'!

unixIt
	self afterSelectionInsertAndSelect: 
		self selection asUnixCommand unixIt printString! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

UnixWorkspaceController class
	instanceVariableNames: ''!


!UnixWorkspaceController class methodsFor: 'class initialization'!

initialize
	"Initialize the yellow button pop-up menu and corresponding messages."

	UnixYellowButtonMenu _ 
		PopUpMenu 
			labels: 'again\undo\copy\cut\paste\unix it\accept\cancel' withCRs
			lines: #(2 5 6).
	UnixYellowButtonMessages _ 
		#(again undo copySelection cut paste unixIt accept cancel)
	"UnixWorkspaceController initialize"! !


UnixWorkspaceController initialize!


StringHolderView subclass: #UnixWorkspaceView
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Interface-Text'!
UnixWorkspaceView comment:
'Defines a bigger default view size.'!


!UnixWorkspaceView methodsFor: 'controller access'!

defaultControllerClass
	^UnixWorkspaceController! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

UnixWorkspaceView class
	instanceVariableNames: ''!


!UnixWorkspaceView class methodsFor: 'instance creation'!

open
	self open: UnixWorkspace new label: 'Unix Workspace'!

open: anUnixWorkspace label: aString 
	"Create a UnixWorkspaceView of the model, anUnixWorkspace, as 
	viewed by an instance of me.  The label of the view is aString."

	| anUnixWorkspaceView topView |
	anUnixWorkspaceView _ self container: anUnixWorkspace.
	topView _ StandardSystemView new.
	topView borderWidth: 1.
	topView model: anUnixWorkspaceView model.
	topView addSubView: anUnixWorkspaceView.
	topView label: aString.
	topView minimumSize: 400 @ 150.
	topView controller open! !


!ScreenController methodsFor: 'menu messages'!

openWorkspace
	"left shift workspace gives an unix workspace."

	Sensor leftShiftDown
		ifTrue: [UnixWorkspaceView open]
		ifFalse: [StringHolderView open]! !
