static char rcsId[] = "$Header: IOPrims.c,v 44.5 84/07/26 00:26:10 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	This file contains routines for Input/Output and Process Scheduling.
 *	See picture.c for the routines that are called from here.
 *	
 *	ARGSUSED has been added to stop lint from complaining.
 */

# include	<stdio.h>
# include	<math.h>
# include	<sgtty.h>
# include	<signal.h>
# include	<ctype.h>
# include	<sunwindow/window_hs.h>
# include	<pixrect/pixrect_hs.h>
# include	<sys/timeb.h>
# include	<sys/vtimes.h>

# include	"keytables.h"
# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"prims.h"
# include	"execute.h"
# include	"cursor.h"


char		SnapName[BUFSIZ];
static boolean	timeOrTickWordsInto();
static void	OnInterrupt(), OnCont(), OnIO();

/*
 *	When it's SignalAtTickTime signal SignalAtTickSemaphore.
 */

static long	signalAtTickTime = 0;

/* 
 *	InputSensor primMousePt InputState primMousePt
 *
 *	Optional primitive for polling mouse coordinates.
 *	This is the basis of the kludgy and imperfect think time detection.
 *	The idea is that if you just sit around polling the mouse
 *	you are not doing anything useful.  The idea is false.
 *	(think time only done for VAX/AED system)
 */

/*ARGSUSED*/
rp_t
SensorMousePrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	aPoint;
	u_int		x, y;

	aPoint = instantiateClass(classPointOop, 0);
	getMouseCoordinates(&x, &y);
	xPt(aPoint) = integerValueOf(x),  yPt(aPoint) = integerValueOf(y);
	*rp = (oop_t)aPoint;
	return (ac);
}


/*
 *	InputState primCursorLocPut:
 *
 *	Moves the cursor to a new location.  Only effective
 *	when cursor not linked to mouse.  The location is given by
 *	a point.
 */

rp_t
InpStCrsPutPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register oop_t	x, y;
	register rp_t	newLoc;

	newLoc = (rp_t)rp[1];
	if (isIntegerObject(newLoc)
	    ||  classRp(newLoc) != classPointOop)
		return (NullOop);
	x = xPt(newLoc);
	y = yPt(newLoc);
	if (!isIntegerObject(x)  ||  !isIntegerObject(y))
		return (NullOop);

	convertToInteger(x);
	convertToInteger(y);

	setCursorToXY((int)x, (int)y);
	return (ac);
}


/*
 *	Cursor class cursorLink:
 *
 *	Links and unlinks the cursor to/from the mouse.
 */

rp_t
CFCCursLinkPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	aBoolean;

	aBoolean = (rp_t)rp[1];
	if (aBoolean == trueOop)
		setCursorLink(TRUE);
	else if (aBoolean == falseOop)
		setCursorLink(FALSE);
	else
		return (NullOop);
	return (ac);
}


/*
 *	Cursor class cursorChange:
 *
 *	Enables/Disables interpreter's cursors.
 */

rp_t
CFCCursChgPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	aBoolean;

	aBoolean = (rp_t)rp[1];
	if (aBoolean == trueOop)
		setCursorChange(TRUE);
	else if (aBoolean == falseOop)
		setCursorChange(FALSE);
	else
		return (NullOop);
	return (ac);
}


/*
 *	InputState primInputSemaphore:
 *
 *	Sets the Input semaphore used to notify image when input
 *	is in input queue.
 */

rp_t
InpStInSemPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	sema;

	sema = (rp_t)rp[1];
	if (isIntegerObject((oop_t)sema)  
	    ||  classRp(sema) != classSemaphoreOop)
		return (NullOop);
	InputSemaphore = sema;
	return (ac);
}


/*
 *	InputState primSampleInterval:
 *
 *	regulates frequency of mouse polling
 */

/*ARGSUSED*/
rp_t
InpStSmpIntPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	warning("InputState primitiveSampleInterval does nothing");
	return (ac);
}


/*
 *	InputState primInputWord
 *
 *	Fetches next input event word out of the input buffer.
 *	Assumes word will fit in smallInteger: only valid for
 *	32 bit OOP system.
 */

rp_t
InpStInpWdPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	int	inputWord;

	if (isEmptyQueue(&InputBuffer))
		return (NullOop);
	dequeue(&InputBuffer, inputWord);
	*rp = integerObjectOf(inputWord);	/* 32 bit OOPS only */
	return (ac);
}


/*
 *	BitBlt copyBits
 *
 *	Interface to bitblt.  See bitblt.c for more details.
 */

rp_t
BBCopyBitsPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	if (!copybits((rp_t)*rp))
		return (NullOop);
	return (ac);
}


/*
 *	BitBlt drawLoopXY
 *
 *	interface to bitblt.
 */

rp_t
BBDrawLoopXY(rp, ac)
oop_t *rp;
rp_t ac;
{
	if (!drawLoopXY((oop_t *)*rp))
		return (NullOop);
	return (ac);
}



/*
 *	SystemDictionary SnapshotPrimitive
 *
 *	Save image on disk.  Current process has to be suspended and
 *	stored into object memory so that it will be included in the
 *	snapshot.
 */

/*ARGSUSED*/
rp_t
SDSnapshotPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	actProc;

	actProc = activeProcessSc(schedulerOop());
	suspendedContextPr(actProc) = ac;
	ggcStore(actProc, ac);

	dumpToFile(SnapName[0] ? SnapName : "snapshot.im", (oop_t *)&ac);
	SnapName[0] = '\0';
	return (ac);
}


/*
 *	Time class timeWordsInto:
 *
 *	Returns time in seconds since 1970.
 *	Code is factored with tickWordsInto.
 */

rp_t
TimeCTimePrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	if (!timeOrTickWordsInto(rp, timeNotTick))
		return (NullOop);
	return (ac);
}


/*
 *	Time class tickWordsInto:
 *
 *	returns time in milliseconds.
 *	Code is factored with timeWordsInto.
 */

rp_t
TimeCTickPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	if (!timeOrTickWordsInto(rp, tickNotTime))
		return (NullOop);
	return (ac);
}


/*
 *	ProcessorScheduler signal:atTick:
 *
 *	Kick SignalAtTickSemaphore in tick milliseconds.
 *	The time is passed in a 4 byte large positive integer.
 *	It sets SignalAtTickSemaphore and signalAtTickTime
 *	which are then checked at input polling time.
 *	Of course, it will be late.
 */

rp_t
PSAtTickPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	semoop;
	register rp_t	tickoop;
	register byte_t	*p;
	register long	now;

	semoop = (rp_t)rp[1];
	tickoop = (rp_t)rp[2];

	if (isIntegerObject((oop_t)semoop)
	    ||  classRp(semoop) != classSemaphoreOop) {
		SignalAtTickSemaphore = nilOop;
		return (ac);
	}
	if (isIntegerObject((oop_t)tickoop)  ||  formatRp(tickoop) != BYTES
	    ||  lengthRp(tickoop) < 4)
		return (NullOop);

	SignalAtTickSemaphore = semoop;

	/*
	 *	Converts 4 bytes into an integer.
	 */

	p = byteFieldRp(tickoop);
	signalAtTickTime =  p[0]  <<  0 * 8
	    |   p[1]  <<  1 * 8
	    |   p[2]  <<  2 * 8
	    |   p[3]  <<  3 * 8;

	getMillisecondClock(now);
	signalAtTickTime -= now;
	return (ac);
}


/*
 *	CursorForm beCursor 
 *
 *	The 16 by 16 form becomes the new cursor.
 */

rp_t
CFBeCursPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	beCursor(BEFieldRp(bitmapFm((rp_t)*rp)));
	return (ac);
}


/*
 *	DisplayForm beDisplay
 *
 *	Subsequent Bitblts to this map will be displayed.
 *	It is promoted because it is too large to conveniently scavenge.
 */

rp_t
DFBeDispPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	DisplayForm = (rp_t)promote(*rp, (oop_t *)&ac);
	return (ac);
}


/*
 *	TextScanner scanCharactersFrom:to:in:
 *
 *	interface to scanCharacters subclass of Bitblt.
 *	Used for all scanning and text composition.
 */

rp_t
TSScanPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	if (!scanCharacters(rp))
		return (NullOop);	/* for now */
	return (ac);
}


/*
 *	inputEvent is called whenever a button is pushed, the mouse
 *	changes position, or a key is pressed.  The event is encoded into
 *	an input buffer event word which is passed as an argument.
 *	Each word is 16 bits wide.  The higher 4 bits holds a type 
 *	code and the lower 12 bits holds the time in milliseconds since 
 *	the last event is enqueued.  This may take multiple words if 
 *	the time is too large (more than 12 bits).  The Semaphore is 
 *	signaled once for EACH word enqueued.
 *	Tvp is the timestamp from input events.
 */

void
inputEvent(event, tvp, acp)
u_int event;
register struct timeval *tvp;
register rp_t *acp;
{
	static u_long	lastEventTime = 0;
	register u_long	thisEventTime;

	/*
	 *	Put event x on the queue and signal that a new event 
	 *	has occured.
	 */

#	define queueIt(x) \
		if (TRUE) {\
			enqueue(&InputBuffer, x);\
			synchronousSignal(InputSemaphore, acp);\
		}\
		else


	if (InputSemaphore == nilOop)
		return;

	thisEventTime = tvp->tv_sec * 1000L  +  tvp->tv_usec / 1000L;
	if (thisEventTime - lastEventTime  &  ~0xfff) { /* not in 12 bits */
		queueIt(ABSOLUTE_TIME);
		queueIt(thisEventTime >> 16  &  0xffff);
		queueIt(thisEventTime & 0xffff);
	}
	else
		queueIt(DELTA_TIME  |  thisEventTime - lastEventTime);

	lastEventTime = thisEventTime;
	queueIt(event);
}


/*
 *	routine executed in case of terminal interrupt.
 *	Will go into halt loop for debugging.
 */

static void
OnInterrupt()
{
	halt("Interrupt");
}


void
pollInput()
{
	int count;
	
	ioctl(BSWindowFd, FIONREAD, &count);
	if (count)
		forcePoll();
}




/*
 *	Encoded keyboard checkInput() routine.  
 *	Works with any keyboard, but does not support some of 
 *	the Smalltalk good stuff, like ^1 or shift-control-C, etc.
 *	This checkInput routine should be used when you are booting up 
 *	your image because the image we got from Xerox assumes an 
 *	ALTO/Dolphin/Dorado keyboard layout.
 *
 *	Grotty
 */


void
checkInputEncoded(acp)
rp_t *acp;
{
	static boolean			receivedControlPrefix = FALSE;
	struct inputevent		ie;
	register struct inputevent	*iep;
	register short			code;
	int				count;
	/*
	 *	We shouldn't have to keep track of these, but we can't get
	 *	sunwindows to tell us about negative events.
	 */
	static boolean			left, middle, right;

#	define	controlPrefix	'\030'		/* ctrl X */

#	define	sendInputEvent(event)	inputEvent(event, &iep->ie_time, acp)
#	define	sendOnOffEvent(code)\
			if (TRUE) {\
				sendInputEvent(BISTATE_ON | (u_int)(code));\
				sendInputEvent(BISTATE_OFF | (u_int)(code));\
			}\
			else

#	define	releaseButton(flag, button)\
		if (flag) {\
			sendInputEvent(BISTATE_OFF | (button));\
			flag = FALSE;\
		}\
		else

#	define pressButton(flag, button)\
		if (!flag) {\
			sendInputEvent(BISTATE_ON | (button));\
			flag = TRUE;\
		}\
		else

	/*
	 *	Read until no more input is left.
	 *	Have to block out SIGALRM because it hangs otherwise.
	 *	Can't figure it out w/o source for input_readevent().
	 */

	iep = &ie;

	sighold(SIGALRM);
	while(input_readevent(BSWindowFd, iep) == 0) {
		code = iep->ie_code;
		if (code == controlPrefix) {
			receivedControlPrefix = TRUE;
			sendInputEvent(BISTATE_ON | (u_int)138);
		}
		else if (receivedControlPrefix) {
			sendOnOffEvent(code);
			sendInputEvent(BISTATE_OFF | (u_int)138);
			receivedControlPrefix = FALSE;
		}
		else if (win_inputnegevent(iep)) {
			releaseButton(left, SMALLTALK_LEFT);
			releaseButton(middle, SMALLTALK_MIDDLE);
			releaseButton(right, SMALLTALK_RIGHT);
		}
		else if (code == LOC_MOVE)
			setMouseToXY(iep->ie_locx, iep->ie_locy);
		else if (code == MS_LEFT)
			pressButton(left, SMALLTALK_LEFT);
		else if (code == MS_MIDDLE)
			pressButton(middle, SMALLTALK_MIDDLE);
		else if (code == MS_RIGHT)
			pressButton(right, SMALLTALK_RIGHT);
		else if (iscntrl(code))
			switch(code) {
			case '\010':		/* backspace */
			case '\011':		/* tab */
			case '\015':		/* carriage return */
			case '\033':		/* escape */
				sendOnOffEvent(code);
				break;
			/*
			 *	Signals cannot be caught, thus I am
			 *	interpreting the control keys here.
			 */

			case '\032':	/* ^Z - to stop BS */
				OnSusp();
				break;

			case '\177':	/* ^? - intr to interpreter */
				OnInterrupt();
				return;

			default:	/* other virtual control */
				sendInputEvent(BISTATE_ON | (u_int)138);
				sendOnOffEvent(code | 0140);
				sendInputEvent(BISTATE_OFF | (u_int)138);
				break;
			}
		else
			sendOnOffEvent(code);
	}
	sigrelse(SIGALRM);
}



/*
 *	Checks input events on an unencoded keyboard.
 *	Some keys (like ^ and _) may not work until
 *	InputState class initialize has been fixed
 *	with the appropriate keymap  (VT100 keymap at Berkeley).  
 *	You should file in vt100Keymap.st.  
 *	This checkInput() does not process ^Z because it is too
 *	klugy.  To suspend BS, file in suspend.st (should be 
 *	distributed with our BS distribution tape).
 *	Processes ^? because it is a life-saver.
 */

void
checkInputUnencoded(acp)
rp_t *acp;
{
	struct inputevent		ie;
	register struct inputevent	*iep;
	register short			code;
	register short			stCode;

	/*
	 *	Queues a mouse button event.  Checks if the button is 
	 *	going up or down.
	 */

#	define	mouseEvent(button)\
		inputEvent((win_inputnegevent(iep) ? BISTATE_OFF :\
		    BISTATE_ON) | button,  &iep->ie_time,  acp);


	/*
	 *	Read until no more input is left.
	 *	Have to block out SIGALRM because it hangs otherwise.
	 *	Can't figure it out w/o source for input_readevent().
	 */

	iep = &ie;

	sighold(SIGALRM);
	while(input_readevent(BSWindowFd, iep) == 0) {
		code = iep->ie_code;
		if (devId(code) == MOUSE_DEVID)
			switch (KEYOF(code) + VKEY_FIRSTPSEUDO) {

			case MS_LEFT:	mouseEvent(SMALLTALK_LEFT);	break;
			case MS_MIDDLE:	mouseEvent(SMALLTALK_MIDDLE);	break;
			case MS_RIGHT:	mouseEvent(SMALLTALK_RIGHT);	break;

			case LOC_MOVE:
				setMouseToXY(iep->ie_locx, iep->ie_locy);
				break;
			default:
				break;
			}
		else {
			stCode = stKeyCode(code);
			if (stCode == ST_HALT)
				OnInterrupt();
			else if (stCode != NULL)
				inputEvent((keyDown(code) ? BISTATE_ON
				      : BISTATE_OFF) | stCode,  
				    &iep->ie_time,  acp);
		}
	}
	sigrelse(SIGALRM);
}


/*
 *	reset the tty modes on suspend signal iff BS is in foreground
 *	Then, send stop to really stop.
 *	Only used by checkInputEncoded().
 */

void
OnSusp()				/* no longer stolen from more */
{
	CloseWindow(TRUE);
	kill(getpid(), SIGSTOP);
	OpenWindow(NULL);		/* use previous values */
}


/*
 *	Common routine to get seconds since 1970 or milliseconds since
 *	whenever.  Must file in unixTime.st to get Date to use unix's epoch.
 *	It is passed a large positive integer into which it puts the answer.
 *	CPU_MS may be defined to return CPU ms instead of real ms
 *	for benchmarking on a timeshared system.
 */

static boolean
timeOrTickWordsInto(rp, timeOrTick)
oop_t *rp;
enum timeOrTick timeOrTick;
{

	/*
	 *	Unix returns seconds since midnight 1/1/70.
	 */

	register u_long	answer;
	register rp_t	oop;
	register byte_t	*p;

	oop = (rp_t)rp[1];
	if (isIntegerObject((oop_t)oop)  ||  formatRp(oop) != BYTES
	    ||  lengthRp(oop) < 4)
		return (FALSE);
	switch (timeOrTick) {
	default:
		fatal("timeOrTick is neither", NullOop);
	case timeNotTick:
		answer = time((time_t *)0);
		break;
	case tickNotTime:
#		ifdef CPU_MS
			{
				struct vtimes	vtimeb;

				vtimes(&vtimeb, 0);
				answer = (vtimeb.vm_utime + vtimeb.vm_stime)
				    * 1000 / HZ;
			}
#		else
			getMillisecondClock(answer);
#		endif
		break;
	}
	p = byteFieldRp(oop);

	/*
	 *	break down int into 4 bytes.
	 */

	p[0] = answer  >>  0 * 8   &   0xff;
	p[1] = answer  >>  1 * 8   &   0xff;
	p[2] = answer  >>  2 * 8   &   0xff;
	p[3] = answer  >>  3 * 8   &   0xff;
	return (TRUE);
}


/*
 *	Decrements the timer semaphore.  Used by  signal:atTick:  message.
 *	Forces polling (which wakes up processes) if time left is zero.
 */

void
pollTimeSem(elapsedMs)
int elapsedMs;
{
	if (SignalAtTickSemaphore != nilOop) {
		signalAtTickTime -= elapsedMs;
		if (signalAtTickTime <= 0)
			forcePoll();
	}
}


/*
 *	Called at poll time.
 */

void
checkSignalAtTick(acp)
rp_t *acp;
{
	if (SignalAtTickSemaphore != nilOop  &&  signalAtTickTime <= 0) {
		synchronousSignal(SignalAtTickSemaphore, acp);
		SignalAtTickSemaphore = nilOop;
	}
}
