static char rcsId[] = "$Header: SIPrims.c,v 41.2 84/03/29 20:26:41 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

# include	<stdio.h>
# include	<math.h>
# include	<sgtty.h>
# include	<signal.h>
# include	<ctype.h>

# include	<sys/types.h>
# include	<sys/vtimes.h>
# include	<sys/timeb.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"prims.h"


/*
 *	Macros used in implementation SmallInteger Primitives.
 *
 *	Unlike the other macros, which can fit anywhere
 *	an ordinary statement can, these macros form
 *	building blocks for the following routines.
 *	Possible sequences are:
 *
 *	SmallIntegerRATest
 *	[SmallIntegerRAConvert]
 *	<some random code>
 *	SmallIntegerOperationFinish[WithOverflow]  or  relationFinish
 */


/*
 *	Setup routine for smallInteger primitives.  Reg1 and reg2 are 
 *	registers allocated to the calling procedure.  Can only be called
 *	at the beginning of a block.
 *
 *	Rp points to the receiver on the stack, which is the first 
 *	operand for the arithmetic/comparison operation.  Both of them
 *	must be SmallInteger, so no check is necessary.
 */

# define SmallIntegerRATest\
	register oop_t	reg1, reg2;\
	\
	reg1 = *rp;\
	reg2 = rp[1];\
	if (!isIntegerObject(reg2))\
		return (NullOop)


/*
 *	Rest of setup, used when conversion is needed.   
 *	Converts both operands to integer.
 */

# define SmallIntegerRAConvert\
	convertToInteger(reg1);\
	convertToInteger(reg2)


/*
 *	Stores result of arithmetic operation to receiver (rp) with 
 *	overflow check.  Overflow check is performed by checking both the 
 *	overflow bit in assembly language and the leftmost bit in the 
 *	result (since SIs are only 31 bits).
 *	Overflow check is only needed for multiplication (*) 
 *	and left shift (<<) (but * is funny on the SUN!).
 *	This is the only place where we need asm() (barf, barf).
 */

# define SmallIntegerOperationFinishWithOverflow(label) \
	asm("	jvs	label");\
	if (!isIntegerValue(reg1)) {\
		asm("label:");\
		return (NullRp);\
	}\
	*rp++ = integerObjectOf(reg1);\
	return (ac)


/*
 *	Stores result of arithmetic operation into receiver without 
 *	overflow check.
 */

# define SmallIntegerOperationFinish \
	if (!isIntegerValue(reg1))\
		return (NullRp);\
	*rp = integerObjectOf(reg1);\
	return (ac)


/*
 *	Implementation of arithmetic primitives 
 */


/*
 *	SmallInteger +
 */

rp_t
SIAddPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	reg1 += reg2;
	SmallIntegerOperationFinish;
}


/*
 *	SmallInteger -
 */

rp_t
SISubPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	reg1 -= reg2;
	SmallIntegerOperationFinish;
}


/*
 *	SmallInteger <
 */

rp_t
SILtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	relationFinish(reg1 < reg2);
}


/*
 *	SmallInteger >
 */

rp_t
SIGtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	relationFinish(reg1 > reg2);
}


/*
 *	SmallInteger <=
 */

rp_t
SILePrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	relationFinish(reg1 <= reg2);
}


/*
 *	SmallInteger >=
 */

rp_t
SIGePrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	relationFinish(reg1 >= reg2);
}


/*
 *	SmallInteger =
 */

rp_t
SIEqPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	relationFinish(reg1 == reg2);
}


/*
 *	SmallInteger ~=
 */

rp_t
SINeqPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	relationFinish(reg1 != reg2);
}


/*
 *	SmallInteger *
 */

rp_t
SIMulPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	reg1 *= reg2;
	SmallIntegerOperationFinishWithOverflow(SIMulPrmOvfl);
}


/*
 *	SmallInteger /
 *	Fails on division by zero, or if result is not an integer.
 */

rp_t
SISlashPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	if (reg2 == 0  ||  reg1 % reg2  !=  0)
		return (NullOop);
	reg1 /= reg2;
	SmallIntegerOperationFinish;
}


/*
 *	SmallInteger \\ (modulus),
 *	Quotient always rounded down towards negative infinity.  
 *	Because the C % operator works differently, we have to consider
 *	the four cases separately.
 */

rp_t
SIModPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	if (reg2 == 0)			/* cannot modulo by zero */
		return (NullOop);
	if (reg1 >= 0  &&  reg2 > 0)
		reg1 %= reg2;
	else if (reg1 >= 0) {		/* negative divisor only */
		reg2 = -reg2;		/* make it positive */
		reg1 %= reg2;		/* but 3 mod -4 is 3-4=-1 */
		if (reg1)		
			reg1 -= reg2;	/* so subtract 4 more to make it -1 */
	}
	else if (reg2 > 0) {		/* negative dividend only */
		reg1 = -reg1;
		reg1 %= reg2;		/* but -3 mod 4 is -3--4  is 1 */
		if (reg1)
			reg1 = reg2 - reg1;/* subtract it from 4 to make it 3 */
	}
	else {				/* both negative */
		reg1 = -reg1;
		reg2 = -reg2;
		reg1 %= reg2;		/* -3 mod -4 is -3 - 0 is -3 */
		if (reg1)
			reg1 = -reg1;	/* so negate it to make 3 */
	}
	SmallIntegerOperationFinish;
}


/*
 *	SmallInteger // (divide),
 *	Quotient always rounded down towards negative infinity.
 */

rp_t
SISlSlPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	if (reg2 == 0)
		return (NullOop);
	if ((reg1 ^ reg2) < 0  		/* different signs */
	    &&  reg1 % reg2)		/* and do not divide evenly */
		reg1 = reg1 / reg2  -  1; /* rounds down */
	else
		reg1 /= reg2;
	SmallIntegerOperationFinish;
}


/*
 *	SmallInteger quo: 
 *	Quotient rounded towards zero.
 */

rp_t
SIQuoPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	if (reg2 == 0)
		return (NullOop);
	reg1 /= reg2;
	SmallIntegerOperationFinish;
}


/*
 *	SmallInteger bitAnd:
 *	No need to convert to integer and back.
 */

rp_t
SIAndPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	*rp = reg1 & reg2;
	return (ac);
}


/*
 *	SmallInteger bitOr:
 *	No need to convert to integer and back.
 */

rp_t
SIOrPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	*rp = reg1 | reg2;
	return (ac);
}


/*
 *	SmallInteger bitXor:
 *	No need to convert to integer and back.  But the result must 
 *	be a smallInteger, so INTEGER_BITS is ORed wit the result of
 *	XORing reg1 and reg2.
 */

rp_t
SIXorPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	*rp = INTEGER_BITS  |  reg1 ^ reg2;
	return (ac);
}


/*
 *	SmallInteger bitShift: 
 *	Arithmetic shift left.	If the argument is negative, do a right 
 *	shift instead.
 *	Unlike the (rational) VAX, the 68k does NOT do a right
 *	shift automatically, hence the explicit test.
 */

rp_t
SIAslPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	SmallIntegerRATest;
	SmallIntegerRAConvert;
	if (reg2 < 0)
		reg1 >>= -reg2;
	else
		reg1 <<= reg2;
	SmallIntegerOperationFinishWithOverflow(SIAslPrmOvfl);
}


/*
 *	Number @ 
 *	Make point.  Will create the new object newPoint and stored into
 *	the receiver.
 */

rp_t
NumbAtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	newPoint;	/* point object to be returned */
	register oop_t	newXY;		/* new x or y coordinate */

	newPoint = instantiateClass(classPointOop, 0);

	xPt(newPoint) = newXY = *rp;
	ggcStore(newPoint, newXY);

	yPt(newPoint) = newXY = rp[1];
	ggcStore(newPoint, newXY);

	*rp = (oop_t)newPoint;
	return (ac);
}
