static char rcsId[] = "$Header: bitblt.c,v 42.5 84/07/25 22:32:06 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	This file contains implementation of the class Bitblt described
 *	in the Book.  Most of the routines are direct translation of 
 *	those written in Smalltalk in the Book.  However, the more 
 *	important cases have been optimized so that it will run faster on
 *	the SUN.
 *
 *	The MSB bit of each Smalltalk 16-bit word is bit zero.
 *	Bit coordinates are zero-origin.  The top left is (0, 0).
 *
 *	Can be compiled with YAK for debugging, and SAFE for sanity
 *	checks in scanCharacters().
 */

# include	<stdio.h>
# include	"framebuf.h"

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"bitblt.h"
# include	"prims.h"

static boolean	getArgs(), clipForX(), clipForY();
void	BBSetup(), internalCopyBits();


/*
 *	Macros for bitblt.c.
 */

/*
 *	getForm() takes apart form arguments that are passed in.
 */

# define getForm(foop, fs) \
	if (TRUE) {\
		register rp_t	rp;\
		\
		getInt(widthFm(foop), fs.f_width, FALSE);\
		getInt(heightFm(foop), fs.f_height, FALSE);\
		\
		rp = bitmapFm(foop);\
		if (isIntegerObject((oop_t)rp)) {\
			warning("getForm: bad bitmap");\
			return (FALSE);\
		}\
		if (fs.f_width < 0  ||  fs.f_height < 0) {\
			warning("getForm: form has neg size");\
			fs.f_width = fs.f_height = 0;\
		}\
		if (foop != DisplayForm\
		  &&  fs.f_width * fs.f_height  >  wordsToBits(lengthRp(rp))) {\
			warning("getForm: bitmap has bad size");\
			fs.f_width = fs.f_height = 0;\
			if (DisplayForm == nilOop) {\
				warning("getForm: forcing beDisplay");\
				DisplayForm = foop;\
			}\
		}\
		fs.f_mapp = BEFieldRp(rp);\
	}\
	else


/*
 *	Combine src and dst according to combination rule.
 *	Implementation of  merge:with:  message on p361 of the book.
 */

# define combineSrcDst(src, dst, result) \
	if (TRUE) \
		switch (combinationRule) {\
		default:\
			fatal("copybits: internal optimization error",\
			    NullOop);\
		case 1:  (result) =  (src)  &  (dst); 	break;\
		case 2:  (result) =  (src)  & ~(dst);	break;\
		case 3:  (result) =  (src);		break;\
		case 4:  (result) = ~(src)  &  (dst);	break;\
		case 6:  (result) =  (src)  ^  (dst);	break;\
		case 7:  (result) =  (src)  |  (dst);	break;\
		case 8:  (result) = ~(src)  & ~(dst);	break;\
		case 9:  (result) = ~(src)  ^  (dst);	break;\
		case 11: (result) =  (src)  | ~(dst);	break;\
		case 12: (result) = ~(src);		break;\
		case 13: (result) = ~(src)  |  (dst);	break;\
		case 14: (result) = ~(src)  | ~(dst);	break;\
		}\
	else


/*
 *	Rule maps for sun
 *		sun 1 is white, Smalltalk 1 is black
 */

# define	DISPSOURCE	(~GXSOURCE & GXMASK)
# define	MEMSOURCE	(GXSOURCE & GXMASK)
# define	STDEST		~GXDEST


/*
 *	Combination Functions.
 *	The ~'s result from the fact that SUN's colors
 *	are reversed from Smalltalk's.
 */

static BE_t	sunRmSrcMem[] = {
	~0,
	~(MEMSOURCE & STDEST),
	~(MEMSOURCE & ~STDEST),
	~MEMSOURCE,

	~(~MEMSOURCE & STDEST),
	~STDEST,
	~(MEMSOURCE ^ STDEST),
	~(MEMSOURCE | STDEST),

	~(~MEMSOURCE & ~STDEST),
	~(~MEMSOURCE ^ STDEST),
	~~STDEST,
	~(MEMSOURCE | ~STDEST),

	~~MEMSOURCE,
	~(~MEMSOURCE | STDEST),
	~(~MEMSOURCE | ~STDEST),
	~~0
};


/*
 *	Combination Rules.  Similar to sunRmSrcMem[], but this time
 *	the source is the display.
 */

static BE_t sunRmSrcDisp[] = {
	~0,
	~(DISPSOURCE & STDEST),
	~(DISPSOURCE & ~STDEST),
	~DISPSOURCE,

	~(~DISPSOURCE & STDEST),
	~STDEST,
	~(DISPSOURCE ^ STDEST),
	~(DISPSOURCE | STDEST),

	~(~DISPSOURCE & ~STDEST),
	~(~DISPSOURCE ^ STDEST),
	~~STDEST,
	~(DISPSOURCE | ~STDEST),

	~~DISPSOURCE,
	~(~DISPSOURCE | STDEST),
	~(~DISPSOURCE | ~STDEST),
	~~0
};


/*
 *	From Bitblt description in Book, p350.
 */

int		combinationRule;	/* 0 thru 15 */
int		destX, destY;
int		width, height;		/* bit size of xfer */
int		sourceX, sourceY;
int		clipX, clipY, clipWidth, clipHeight;


/*
 *	Internal structure to represent a form.
 *	Defines dest form (df), source form (sf), halftone form (hf).
 */

static struct	form {
	int		f_width, f_height;
	BE_t		*f_mapp;
} df, sf, hf;


/*
 *	Instance variables in class BitBltSimulation.
 */

static short		sx, dx, w, sy, dy, h;	/* after clipping */
static short		dleft, dtop;		/* always top left */
static short		sleft, stop;		/* always top left */

static short		destRaster, sourceRaster, /* see book */
			mask1, mask2, skewMask, nwords;
static int		skew;			/* int for compiler */

static short		hDir, vDir;		/* direction to move */


/*
 *	From book.
 *	True to indicate that it is necessary to preload some source bits
 *	to be shifted prior to the inner loop of the copy.
 */

static boolean		preload;


static int		sourceIndex, destIndex;
static short		sourceDelta, destDelta;


/*
 *	The SUN hardware requires us to treat these cases differently.
 *	Used later by internalCopyBits() (modified
 *	version of copyLoop described in the Book) to decide which
 *	copy routine to call.
 */

static boolean		dstIsDisplay, srcIsDisplay, htIsNil, srcIsNil;


/*
 *	external interface to BitBlt
 */

boolean
copybits(bb)
rp_t bb;
{
	if (!getArgs(bb, TRUE, FALSE))
		return (FALSE);
	internalCopyBits(bb);
	return (TRUE);
}


/*
 *	Hook for scanCharacters and turtleDrawLoop.
 *	Follows book except for optimizations. 
 *	Dispatches to appropriate routine for special case.
 */

void
internalCopyBits(bb)
rp_t bb;
{
	if (!clipForX()  ||  !clipForY())
		return;
	BBSetup(bb);

	if (srcIsNil)
		if (htIsNil)
			if (dstIsDisplay)
				BBDstDisp();
			else
				BBDstMem();
		else
			if (dstIsDisplay)
				BBHtDstDisp();
			else
				BBHtDstMem();
	else
		if (htIsNil)
			if (dstIsDisplay)
				if (srcIsDisplay)
					BBSrcDispDstDisp();
				else
					BBSrcMemDstDisp();
			else
				if (srcIsDisplay)
					BBSrcDispDstMem();
				else
					BBSrcMemDstMem();
		else
			if (dstIsDisplay)
				if (srcIsDisplay)
					BBSrcDispHtDstDisp();
				else
					BBSrcMemHtDstDisp();
			else
				if (srcIsDisplay)
					BBSrcDispHtDstMem();
				else
					BBSrcMemHtDstMem();
}


/*
 *	Gets all the arguments.
 *	Extracts fields from BitBlt and Scanner as needed.
 */

static boolean
getArgs(bb, willCopybits, willScan)
rp_t bb;
boolean willCopybits, willScan;
{
	register rp_t	rp;

	if (willCopybits) {
		if (destFormBB(bb) == nilOop) {
			warning("bitblt failing: no destForm");
			return (FALSE);
		}
		getInt(combinationRuleBB(bb), combinationRule, FALSE);
		combinationRule &= 15;
		getInt(destYBB(bb), destY, FALSE);
		getInt(heightBB(bb), height, FALSE);
		getInt(sourceYBB(bb), sourceY, FALSE);
		getInt(clipXBB(bb), clipX, FALSE);
		getInt(clipYBB(bb), clipY, FALSE);
		getInt(clipWidthBB(bb), clipWidth, FALSE);
		getInt(clipHeightBB(bb), clipHeight, FALSE);

		getForm(destFormBB(bb), df);
		switch (combinationRule) {
		case 0:
		case 5:
		case 10:
		case 15:
			srcIsNil = htIsNil = TRUE;
			break; 
		default:
			if (!(srcIsNil  =  sourceFormBB(bb) == nilOop)) {
				getForm(sourceFormBB(bb), sf);
				srcIsDisplay =  sourceFormBB(bb) == DisplayForm;
			}
			if (!(htIsNil  =  halftoneFormBB(bb) == nilOop)) {
				register BE_t	*p;

				getForm(halftoneFormBB(bb), hf);
				p = hf.f_mapp;
				if (isOnesBitmap(p))
					htIsNil = TRUE;
				else if (isZerosBitmap(p)) {

					/*
					 *	Reduced rules when bitmap is 
					 *	all zeros.  Makes it faster.
					 */

					static short rm[] = {
						0, 0, 0, 0,
						5, 5, 5, 5,
						10, 10, 10, 10,
						15, 15, 15, 15};
					combinationRule = rm[combinationRule];
					htIsNil = srcIsNil = TRUE;
				}
			}
			break;
		}

		dstIsDisplay =  destFormBB(bb) == DisplayForm;
	}
	getInt(destXBB(bb), destX, FALSE);

	/*
	 *	If called by ScanCharacters(), must also check for
	 *	the validity of the font table.
	 */

	if (willScan) {
		register rp_t	xt;

		xt = xTableCS(bb);
		if (isIntegerObject((oop_t)xt)  
		    ||  classRp(xt)  != classArrayOop) {
			warning("scanCharacters failing: XTable not array");
			return (FALSE);
		}
		rp = stopConditionsCS(bb);
		if (isIntegerObject((oop_t)rp)
		    ||  classRp(rp) != classArrayOop) {
			warning("scanCharacters failing: stopConditions not array");
			return (FALSE);
		}
	}
	else {
		getInt(widthBB(bb), width, FALSE);
		getInt(sourceXBB(bb), sourceX, FALSE);
	}
	return (TRUE);
}


/*
 *	Basically out of the Book (p356).  Appears as method
 *	clipRange in the Book, but it has been broken down 
 *	into two separate routines here since they are logically
 *	separate.
 *	This adjusts the width of the window in the destination.
 *	Must ensure that we don't clobber memory by clipping.
 *	Must also take min of source, dest, and clip rects.
 *	Returns FALSE if clipped window shrinkds to zero or
 *	negative size.
 */

static boolean
clipForX()
{
	register int	i;

	if (width <= 0)
		return (FALSE);

	/*
	 *	Clips dest to clip rectangle so that sx lies within
	 *	the destination rectangle.
	 */

	i = clipX - destX;
	if (i <= 0) {			/* set origin */
		sx = sourceX;
		dx = destX;
		w = width;
	}
	else {
		sx = i + sourceX;
		dx = clipX;
		w = width - i;
	}
	if (dx < 0) {
		sx += -dx;
		w -= -dx;
		dx = 0;
	}

	/*
	 *	Sets the width in case part of the clipped window
	 *	falls outside the destination window.
	 */

	i = dx + w  -  (clipX + clipWidth);
	if (i > 0)
		w -= i;
	i = dx + w - df.f_width;
	if (i > 0)
		w -= i;

	/*
	 *	Now clip to source rectangle.  Adjust sx and w in case 
	 *	the source is larger than the destination.
	 */

	if (!srcIsNil) {	/* now clip to source rect */
		if (sx < 0) {
			dx += -sx;
			w -= -sx;
			sx = 0;
		}
		i = sx + w - sf.f_width;
		if (i > 0)
			w -= i;
	}
	if (w <= 0)
		return (FALSE);
	return (TRUE);
}


/*
 *	Now, do the same for the height (along vertical axis).
 */

static boolean
clipForY()
{
	register int i;

	if (height <= 0)
		return (FALSE);

	i = clipY - destY;
	if (i <= 0) {
		sy = sourceY;
		dy = destY;
		h = height;
	}
	else {
		sy = i + sourceY;
		dy = clipY;
		h = height - i;
	}
	if (dy < 0) {
		sy += -dy;
		h -= -dy;
		dy = 0;
	}

	i = dy + h  -  (clipY + clipHeight);
	if (i > 0)
		h -= i;
	i = dy + h - df.f_height;
	if (i > 0)
		h -= i;

	if (!srcIsNil) {	/* now clip to source rect */
		if (sy < 0) {
			dy += -sy;
			h -= -sy;
			sy = 0;
		}
		i = sy + h - sf.f_height;
		if (i > 0)
			h -= i;
	}
	if (h <= 0)
		return (FALSE);
	return (TRUE);
}


/*
 *	Bitblt setup: straight from the book.  
 *	Original ST code is shown in comments.  See the book for
 *	more detailed description.
 *
 *	Optimized by omitting certain initialization when either
 *	destination and/or source is the display.  Those values
 *	won't be used in such cases.
 */

void
BBSetup(bb)
rp_t bb;
{
	register int	startBits, endBits;

	/*
	 *	See book.  But C's index is zero-origin while ST's is
	 *	one-origin.  So adding one to the index is not necessary.
	 */

	static BE_t rightMasks[] = {0,
			0x1,	0x3,	0x7,	0xf,
			0x1f,	0x3f,	0x7f,	0xff,
			0x1ff,	0x3ff,	0x7ff,	0xfff,
			0x1fff,	0x3fff,	0x7fff,	0xffff};


	/*
	 *	self computeMasks.
	 */

	if (!dstIsDisplay) {
		destRaster = nonNegBitsToWords(df.f_width - 1)  +  1;
		startBits = 16 - (dx & 15);
		mask1 = rightMasks[startBits];
		endBits = 15 - ((dx + w - 1) & 15);
		mask2 = ~rightMasks[endBits];
	}
	if (!srcIsNil  &&  !srcIsDisplay) {
		sourceRaster = nonNegBitsToWords(sf.f_width - 1)  +  1;
		if (!dstIsDisplay) {
			skew = sx - dx  &  15;
			skewMask = skew == 0  ?  0  :  rightMasks[16 - skew];
		}
	}

	if (w < startBits) {
		mask1 &= mask2;
		mask2 = 0;
		nwords = 1;
	}
	else
		nwords = nonNegBitsToWords(w - startBits - 1) + 2;

	/*
	 *	The top left corner of the destination window.  For
	 *	determining if the cursor is in the window to be changed.
	 *	If so, remove cursor before bitblting.
	 */

	dleft = dx;
	dtop = dy;
	sleft = sx;
	stop = sy;

	/*
	 *	self checkOverlap.
	 *	Check for overlapping of source and destination to 
	 *	avoid destroying the source while it is being copied.
	 */

	hDir = vDir = 1;

	if (!srcIsNil  &&  sourceFormBB(bb) == destFormBB(bb)  
	    &&  dy >= sy)		/* not moving up */
		if (dy > sy) {		/* moving down */

			/*
			 *	must start copy from bottom 
			 *	and work upwards.
			 */

			vDir = -1;
			sy += h - 1;
			dy += h - 1;
		}
		else if (dx > sx) {	/* moving horizontally right */
			
			/*
			 *	must start copy from the right and 
			 *	work towards left.
			 */

			hDir = -1;
			sx += w - 1;
			dx += w - 1;
			if (!dstIsDisplay) {
				short	t;

				skewMask = ~skewMask;
				t = mask1,  mask1 = mask2,  mask2 = t;
			}
		}


	/*
	 *	self calculateOffsets.
	 */

	if (!srcIsNil  &&  !srcIsDisplay) {
		if (!dstIsDisplay) {
			preload  =  skew != 0 && skew <= (sx & 15);
			if (hDir < 0)
				preload = !preload;
		}
		sourceIndex = sy * sourceRaster  +  nonNegBitsToWords(sx);
		sourceDelta = sourceRaster * vDir  -  (nwords + preload) * hDir;
	}
	if (!dstIsDisplay) {
		destIndex = dy * destRaster  +  nonNegBitsToWords(dx);
		destDelta = destRaster * vDir  -  nwords * hDir;
	}

}


/*
 *	ScanCharacters implemented from the source code.  This 
 *	routine is translated from the Smalltalk source code for:
 *
 *	scanCharactersFrom: startIndex to: stopIndex in: sourceString
 *		rightX: rightX stopConditions: stops displaying: display 
 *		| lastIndex  nextDestX |
 *		lastIndex _ startIndex.
 *		[ lastIndex <= stopIndex ] whileTrue: [
 *			ascii _ (sourceString at: lastIndex) asciiValue.
 *			(stopConditions at: ascii + 1) ~~ nil  
 *			    ifTrue: [ ^stops at: ascii + 1 ].
 *			sourceX _ xTable at: ascii + 1.
 *			nextDestX _ destX + 
 *			    (width _ (xTable at: ascii + 2) - sourceX).
 *			nextDestX > rightX
 *			    ifTrue: [ ^stops at: CrossedX ].
 *			display ifTrue: [ self copyBits ].
 *			destX _ nextDestX.
 *			lastIndex _ lastIndex + 1
 *		].
 *		^stops at: EndOfRun
 *
 *
 *	This is the inner loop of text display.
 *	March through source string from startIndex to stopIndex.
 *	If any character is flagged with a non-nil entry in stops, 
 *	then return the corresponding value. Determine width of 
 *	each character from xTable. If dextX would exceed rightX, 
 *	then return stops at: 258. If displaying is true, then 
 *	display the character. Advance destX by the width of the 
 *	character. If stopIndex has been reached, then return
 *	stops at: 257.  Fail under the same conditions that the 
 *	Smalltalk code would cause an error. Optional.
 *
 *	This code is optimized for the SUN; as each character's position
 *	is computed, the character itself is displayed on the screen with
 *	unrolled loops.
 *
 *	The original Smalltalk code is repeated in the comments
 *	to help out.
 */

boolean
scanCharacters(rp)
oop_t *rp;
{
	register short	*dsty, sc, *srcp, sinc, dstw;
	short		*oldSrcp;
	short		*dstx;
	boolean		easyBitblt;	/* can use fast copyBits */

	register rp_t	rpreg;		/* scratch register */
	register rp_t	bb;		/* receiver bitblt */
	register oop_t	*xTablep;	/* font table */
	register rp_t	*stopConditionsp; /* argument stopCondition */
	register byte_t	*sourceStringp;	  /* argument sourceString */

	register u_int	xTableLength, stopConditionsLength;
	register oop_t	oop;

	oop_t	startIndex, stopIndex, rightX;	/* arguments */
	u_int	sourceStringSize;
	rp_t	*stopsp;
	u_int	stopsLength;
	boolean	willCopybits;		/* argument display */

	byte_t	*lastSourcep, *stopSourcep;

	u_int	ascii;
	int	nextDestX;

	u_int	ret;			/* return value from method */

	
#	ifdef SAFE
		getInt(startIndexRp(rp), startIndex, FALSE);
		if (startIndex < 1)
			startIndex = 1;
		getInt(stopIndexRp(rp), stopIndex, FALSE);
		getInt(rightXRp(rp), rightX, FALSE);
#	else
		startIndex = nonNegIntegerValueOf(startIndexRp(rp));
		stopIndex = nonNegIntegerValueOf(stopIndexRp(rp));
		rightX = nonNegIntegerValueOf(rightXRp(rp));
#	endif

	rpreg = sourceStringRp(rp);
#	ifdef SAFE
		if (isIntegerObject((oop_t)rpreg) || formatRp(rpreg) != BYTES) {
			warning("scanCharacters failing: not byte object");
			return (FALSE);
		}
#	endif
	sourceStringp = byteFieldRp(rpreg);
	sourceStringSize = lengthRp(rpreg);

	rpreg = stopsRp(rp);
	stopsp = rpFieldRp(rpreg);
	stopsLength = lengthRp(rpreg);

	rpreg = displayRp(rp);
#	ifdef SAFE
		if (rpreg == trueOop)
			willCopybits = TRUE;
		else if (rpreg == falseOop)
			willCopybits = FALSE;
		else {
			warning("scanCharacters failing: display not boolean");
			return (FALSE);
		}
#	else
		willCopybits =  rpreg == trueOop;
#	endif


	/*
	 *	Get bitblt object to display, and extract various fields.
	 */

	bb = (rp_t)*rp;
	if (!getArgs(bb, willCopybits, TRUE)){
		warning("scanCharacters failing: bad Bitblt");
		return (FALSE);
	}

	if (willCopybits)
		if (!clipForY())
			willCopybits = FALSE;
		else if (sf.f_height <= 16  &&  dstIsDisplay  &&  htIsNil
		    &&  destX >= clipX  &&  destY >= clipY) {
			easyBitblt = TRUE;

			srcIsNil = TRUE;	/* don't clip source */
			width = clipWidth;
			(void)clipForX();
			srcIsNil = FALSE;

			sinc = nonNegBitsToWords(sf.f_width + 15);
			cursorOff(dx, dy, w, h);
			GXpattern = ~0;
			GXfunction = sunRmSrcMem[combinationRule];

			dstx = (short *)(GXBase + GXselectX + GXset0) + dx;
			*dstx = 1;
			dsty = (short *)(GXBase + GXselectY + GXset0
			    + GXsource + GXupdate) + dy;
		}
		else
			easyBitblt = FALSE;

#	ifdef SAFE
		if (stopIndex > sourceStringSize)
			stopIndex = sourceStringSize;

		if (minStopsLength > stopsLength) {
			warning("scanCharacters: bad stops subscript");
			return (FALSE);
		}
#	endif


	/*
	 *	Starts Smalltalk code
	 */

	/*
	 *	lastIndex <- startIndex.
	 *
	 *	lastIndex indexes the next character to be printed, and
	 *	lastSourcep is its address.  stopIndex indexes the last
	 *	character to be printed, and stopSourcep is its address.
	 */

	lastSourcep = &sourceStringp[startIndex - 1];
	stopSourcep = &sourceStringp[stopIndex - 1];

#	ifdef SAFE
		xTableLength = lengthRp(xTableCS(bb));
#	endif

	/*
	 *	stopConditionsp is the address of an array of oops.
	 *	A character constitutes a stop condition if the element
	 *	indexed by its ascii value is not nil.
	 */

	stopConditionsp = rpFieldRp(stopConditionsCS(bb));
	stopConditionsLength = lengthRp(stopConditionsCS(bb));

	for (;;) {

		/*
		 *	[lastIndex <= stopIndex] whileTrue:  [
		 */

		if (lastSourcep > stopSourcep) {
# 			ifdef YAK
				warning("scanCharacters: stopIndex");
# 			endif
			lastSourcep = stopSourcep;
			ret = stopIndexIndex - 1;
			break;
		}

		/*
		 *	ascii <- (sourceString at: lastIndex) asciiValue.
		 */

		ascii = *lastSourcep;

		/*
		 *	(stopConditions at: ascii + 1) ~~ nil ifTrue:
		 *
		 *	The ASCII character set starts at zero, but Smalltalk
		 *	arrays starts at one.  That's why we add one to "ascii"
		 *	in the Smalltalk code.  This does not happen in our 
		 *	C code since C has zero-origin arrays.
		 */

#		ifdef SAFE
			if (ascii >= stopConditionsLength) {
				warning("scanCharacters: stopCond subscrpt");
				return (FALSE);
			}
#		endif
		if (stopConditionsp[ascii] != nilOop) {
#			ifdef YAK
				warning("scanCharacters: stopCondition");
#			endif

		 	/*
			 *	[^stops at: ascii + 1].
			 */

			ret = ascii;
			break;
		}

		/*
		 *	sourceX <- xTable at: ascii + 1.
		 *
		 *	Get bit position of character to copy.
		 */

#		ifdef SAFE
			if (ascii + 1  >=  xTableLength) {
				warning("scanChars failing: ascii >= xTblLen");
				return (FALSE);
			}
#		endif
#		ifdef YAK
			putchar(ascii);
			fflush(stdout);
#		endif
		xTablep = &oopFieldRp(xTableCS(bb))[ascii];
#		ifdef SAFE
			getInt(*xTablep++, oop, FALSE);
			sourceX = oop;
#		else
			sourceX = nonNegIntegerValueOf(*xTablep++);
#		endif

		/*
		 *	nextDestX <-
		 *	    destX + (width <- (xTable at: ascii + 2) - sourceX).
		 *
		 *	Find the width of the character to be printed, and
		 *	use that to compute where the cursor will stop at.
		 */

#		ifdef SAFE
			getInt(*xTablep, oop, FALSE);
#		else
			oop = nonNegIntegerValueOf(*xTablep);
#		endif
		nextDestX = destX + (width = oop - sourceX);

		/*
		 *	nextDestX > rightX ifTrue: [^stops at: CrossedX].
		 */

		if (nextDestX > rightX) {
# 			ifdef YAK
				warning("\r\nscanCharacters: nextDestX > rightX");
# 			endif
			ret = crossedXIndex - 1;
			break;
		}

		/*
		 *	display ifTrue: [self copyBits].
		 */

		if (willCopybits  &&  easyBitblt  
		    &&   destX + width  <=  dx + w) {

			/*
			 *	speeded up for small copyBits that 
			 *	does not need clipping the source.
			 */

			oldSrcp = srcp = (short *)&sf.f_mapp[
			    nonNegBitsToWords(sourceX)];
			sc = sourceX & 15;
			dstw = bitsPerBE - sc;
			if (dstw > width)
				dstw = width;
			GXwidth = dstw;

			/*
			 *	Copies skewed region on the left.
			 *	Unrolled loop (disguised as switch
			 *	statement) for speed.
			 */

#			define shft(x)\
			    case x:  *dsty++ = *srcp << sc;  srcp += sinc;
			switch (h) {
				shft(16); shft(15); shft(14); shft(13);
				shft(12); shft(11); shft(10); shft(09);
				shft(08); shft(07); shft(06); shft(05);
				shft(04); shft(03); shft(02); shft(01);
			}
			*(dstx += dstw) = 1;
			dsty -= h;
			width -= dstw;

			/*
			 *	Copies words of 16 bits in the middle.
			 */

			for ( ;  width >= bitsPerBE;  width -= bitsPerBE) {
				GXwidth = bitsPerBE;
				srcp = ++oldSrcp;

#				define ns(x) \
				    case x:  *dsty++ = *srcp; srcp += sinc;
				switch (h) {
					ns(16); ns(15); ns(14); ns(13);
					ns(12); ns(11); ns(10); ns(09);
					ns(08); ns(07); ns(06); ns(05);
					ns(04); ns(03); ns(02); ns(01);
				}
				*(dstx += bitsPerBE) = 1;
				dsty -= h;
			}

			if (width > 0) {

				/*
				 *	Copies skew region on right.
				 */

				GXwidth = width;
				srcp = ++oldSrcp;

				switch (h) {
					ns(16); ns(15); ns(14); ns(13);
					ns(12); ns(11); ns(10); ns(09);
					ns(08); ns(07); ns(06); ns(05);
					ns(04); ns(03); ns(02); ns(01);
				}
				*(dstx += width) = 1;
				dsty -= h;
			}
		}
		else if (willCopybits)
			internalCopyBits(bb);


		/*
		 *	destX <- nextDestX.
		 */

		destX = nextDestX;

		/*
		 *	lastIndex <- lastIndex + 1].
		 */

		++lastSourcep;
	}


	/*
	 *	^stops at: EndOfRun
	 */

	rpreg = stopsp[ret];
	resultRp(rp) = rpreg;

	/*
	 *	End Smalltalk code
	 */


	/*
	 *	lastIndex <- stopIndex.
	 *
	 *	last - first is last, but with one-origin must increment.
	 */

	putInt(lastSourcep - sourceStringp + 1,  lastIndexCS(bb));

	/*
	 *	Write all changed values back to receiver Bitblt.
	 */

	putInt(destX, destXBB(bb));
	putInt(sourceX, sourceXBB(bb));
	putInt(width, widthBB(bb));

	if (willCopybits  &&  easyBitblt)
		cursorOn();

#	ifdef YAK
		warning("scanCharacters succeded");
#	endif

	return (TRUE);
}


/*
 *	Straight from method  drawLoopX:Y:  in the book, p352.
 */

boolean
drawLoopXY(rp)
oop_t *rp;
{
	register oop_t	px, py;		/* xDelta, yDelta */
	register int	signX, signY, p, i;
	register rp_t	bb;		/* receiver bitblt */

	bb = (rp_t)*rp;
	if (!getArgs(bb, TRUE, FALSE))
		return (FALSE);
	getInt(rp[1], px, FALSE);
	getInt(rp[2], py, FALSE);

	if (px < 0) {
		px = -px;
		signX = -1;
	}
	else
		signX = 1;

	if (py < 0) {
		py = -py;
		signY = -1;
	}
	else
		signY = 1;

	internalCopyBits(bb);

	if (py > px) 			/* more horizontal */
		for (i = 1,  p = py >> 1;   i <= py;   i++) {
			destX += signX;
			if ((p -= px) < 0) {
				destY += signY;
				p += py;
			}
			internalCopyBits(bb);
		}
	else				/* more vertical */
		for (i = 1,  p = px >> 1;   i <= px;   i++) {
			destY += signY;
			if ((p -= py) < 0) {
				destX += signX;
				p += px;
			}
			internalCopyBits(bb);
		}

	return (TRUE);
}

/*
 *	The following are subroutines for internalCopyBits().
 *	Each one of them handles a special case of copyLoop.
 */


/*
 *	Bitblt with no source or halftone (i.e. halftone = ones) to memory.
 */
 
BBDstMem()
{
	register BE_t	*p, *pr, *pe;

	/*
	 *	Do the trivial cases of one or two words separately 
	 *	for speed.
	 */

	switch (combinationRule) {
	case 0:			/* zero fill */
	case 4:
	case 8:
	case 12:
		if (nwords == 1)
			for (p = &df.f_mapp[destIndex],
			      pe = p + h * destRaster;
			    p < pe;
			    p += destRaster)
				*p &= ~mask1;
		else if (nwords == 2)
			for (p = &df.f_mapp[destIndex],
			      pe = p + h * destRaster;
			    p < pe;
			    p += destRaster) {
				*p &= ~mask1;
				p[1] &= ~mask2;
			}
		else
			for (p = &df.f_mapp[destIndex],
				pe = p  +  h * destRaster;
			    p < pe;
			    p += destRaster) {
				*p &= ~mask1;
				sfill((short)0,  (short *)(p + 1),
				    (u_int)(nwords - 2));
				p[nwords - 1] &= ~mask2;
			}
		break;

	case 1:			/* do nothing: merge with dst word (itself!) */
	case 5:	
	case 9:
	case 13:
		break;

	case 2:			/* invert dest */
	case 6:
	case 10:
	case 14:
		if (nwords == 1)
			for (p = &df.f_mapp[destIndex],
			      pe = p + h * destRaster;
			    p < pe;
			    p += destRaster)
				*p ^= mask1;
		else if (nwords == 2)
			for (p = &df.f_mapp[destIndex],
			      pe = p + h * destRaster;
			    p < pe;
			    p += destRaster) {
				*p ^= mask1;
				p[1] ^= mask2;
			}
		else
			for (p = &df.f_mapp[destIndex],
				pe = p  +  h * destRaster;
			    p < pe;
			    p += destDelta) {
				*p ^= mask1;
				for (pr = p + nwords - 1,  ++p;  p < pr;  p++)
					*p = ~*p;
				*p++ ^= mask2;
			}
		break;

	case 3:			/* one fill */
	case 7:
	case 11:
	case 15:
		if (nwords == 1)
			for (p = &df.f_mapp[destIndex],
			      pe = p + h * destRaster;
			    p < pe;
			    p += destRaster)
				*p |= mask1;
		else if (nwords == 2)
			for (p = &df.f_mapp[destIndex],
			      pe = p + h * destRaster;
			    p < pe;
			    p += destRaster) {
				*p |= mask1;
				p[1] |= mask2;
			}
		else
			for (p = &df.f_mapp[destIndex],
				pe = p  +  h * destRaster;
			    p < pe;
			    p += destRaster) {
				*p |= mask1;
				sfill((short)~0, (short *)(p + 1),
				    (u_int)(nwords - 2));
				p[nwords - 1] |= mask2;
			}
		break;
	}
}


/*
 *	Non-trivial Bitblt routines.  The general algorithm is:
 *
 *		for each row in Form
 *			copy first word (or partial word), if any.
 *			copy complete word(s) in the middle.
 *			copy first word (or partial word), if any.
 */


/*
 *	Bitblt routine: no src, with halftone, dest in memory.
 */

BBHtDstMem()
{
	register BE_t	*destp, *hfp, *endhfp;

	register int	word;
	register int	mergeWord;

	register int	halftoneWord;

	int	i;

	halftoneWord = ~0;
	destp = &df.f_mapp[destIndex];
	hfp = &hf.f_mapp[dy & 15];
	endhfp = &hf.f_mapp[HALFTONE_WORDS];

	for (i = 1;  i <= h;  i++) {
		halftoneWord = *hfp++;
		if (hfp >= endhfp)
			hfp = hf.f_mapp;

		/*
		 *	do first word.
		 */

		combineSrcDst(halftoneWord, *destp, mergeWord);
		*destp = mask1 & mergeWord  |  ~mask1 & *destp;
		destp++;

		for (word = 2;  word < nwords;  ++word, ++destp)
			combineSrcDst(halftoneWord, *destp, *destp);

		/*
		 *	do last word
		 */

		if (nwords > 1) {
			combineSrcDst(halftoneWord, *destp, mergeWord);
			*destp = mask2 & mergeWord |  ~mask2 & *destp;
			destp++;
		}
		destp += destDelta;
	}
}


/*
 *	Bitblt routine: src in memory, no halftone, dest in memory.
 *	An all ones mask is no mask at all.
 *	Cheat by assigning a halftone mask and calling a more general routine.
 */

BBSrcMemDstMem()
{
	static BE_t onesHt[] = {
		~0, ~0, ~0, ~0,
		~0, ~0, ~0, ~0,
		~0, ~0, ~0, ~0,
		~0, ~0, ~0, ~0};

	hf.f_mapp = onesHt;
	BBSrcMemHtDstMem();
}


/*
 *	Bitblt routine: src in memory, with halftone, dest in memory.
 *	This is the one routine that could stand more optimization;
 *	memory to memory bitblts are too slow.
 */

BBSrcMemHtDstMem()
{
	register BE_t	*destp, *sourcep, *hfp, *endhfp;

	register int	word;
	int	skewWord;		/* compiler barfs on reg */
	register int	prevWord, thisWord;
	register int	mergeMask;
	register int	mergeWord;

	short	halftoneWord;
	int	i;

	destp = &df.f_mapp[destIndex];
	hfp = &hf.f_mapp[dy & 15];
	endhfp = &hf.f_mapp[HALFTONE_WORDS];
	sourcep = &sf.f_mapp[sourceIndex];

	for (i = 1;  i <= h;  i++) {
		halftoneWord = *hfp++;
		if (hfp >= endhfp)
			hfp = hf.f_mapp;

		if (preload) {
			prevWord = *sourcep;
			sourcep += hDir;
		}
		else
			prevWord = 0;

		mergeMask = mask1;

		/*
		 *	Inner horizontal loop.
		 *
		 *	Straight from the book.  sourceForm notNil test
		 *	is not needed.
		 */

		for (word = 1;  word <= nwords;  ++word) {
			prevWord &= skewMask;
			thisWord = *sourcep;
			skewWord = prevWord  |  thisWord & ~skewMask;
			prevWord = thisWord;
			skewWord = skewWord << skew  
			    |   (skewWord & 0xffff)  >>  16 - skew;
			combineSrcDst(skewWord & halftoneWord,  *destp,
			    mergeWord);
			*destp = mergeMask & mergeWord  |  ~mergeMask & *destp;
			sourcep += hDir;
			destp += hDir;
			mergeMask = word == nwords - 1  ?  mask2  :  ~0;
		}
		sourcep += sourceDelta;
		destp += destDelta;
	}
}



/*
 *	Bitblt routine: src in display, no halftone, dest in memory
 *	optimized scroll bar case (comb. rule 3): moves vertically.
 *	Cursor must be turned off while changing the display, otherwise
 *	it will be moved together with the background.
 */


BBSrcDispDstMem()
{
	register short	*destp, *oldDst;
	register short	*srcx, *srcy, *endx, *endy, rastReg, mergeMask;
	register short	hreg, dummy;
	static BE_t onesMap[] = {
		~0, ~0, ~0, ~0,
		~0, ~0, ~0, ~0,
		~0, ~0, ~0, ~0,
		~0, ~0, ~0, ~0};

	if (combinationRule != 3  ||  dx  &  bitsPerBE - 1) {
		hf.f_mapp = onesMap;
		BBSrcDispHtDstMem();
		return;
	}


	cursorOff(sleft, stop, w, h);

	srcx = (short *)(GXBase + GXselectX + GXset0) + sx;
	endx = srcx + w - bitsPerBE;

	srcy = (short *)(GXBase + GXselectY + GXset0 + GXsource) + sy;
	hreg = h;
	endy = srcy + hreg;

	oldDst = destp = (short *)&df.f_mapp[destIndex];
	rastReg = destRaster;


	/*
	 *	The Programmer's manual lies; the GXwidth register 
	 *	effects reads.
	 */

	GXwidth = bitsPerBE;	

	/*
	 *	The Sun programmer's manual explains the pipelining in
	 *	in the graphics board that requires us to start with a dummy
	 *	read.  The data must be assigned to dummy so the C compiler
	 *	doesn't optimize it out.
	 */

	if (w < bitsPerBE) {
		mergeMask = mask1;
		for (*srcx = 1,  dummy = *srcy++;
		    srcy < endy;
		    destp += rastReg)
			*destp =  mergeMask & ~*srcy++  |  ~mergeMask & *destp;
		*destp = mergeMask & ~*(srcy -= hreg)  |  ~mergeMask & *destp;
		srcx += bitsPerBE;
		destp = ++oldDst;
	}
	while (srcx < endx) {
		for (*srcx = 1,  dummy = *srcy++;
		    srcy < endy;
		    destp += rastReg)
			*destp = ~*srcy++;
		*destp = ~*(srcy -= hreg);

		srcx += bitsPerBE;
		destp = ++oldDst;
	}
	if (w >= bitsPerBE) {
		mergeMask = mask2;
		for (*srcx = 1,  dummy = *srcy++;
		    srcy < endy;
		    destp += rastReg)
			*destp =  mergeMask & ~*srcy++  |  ~mergeMask & *destp;
		*destp = mergeMask & ~*(srcy -= hreg)  |  ~mergeMask & *destp;
	}
	cursorOn();
}


/*
 *	Src is display, with halftone to memory: moves horizontally.
 *	Works okay with nil halftone.
 */

BBSrcDispHtDstMem()
{
	register short	*srcx, *endx;
	register BE_t	*destp;
	register BE_t	halftoneWord, mergeWord;

	register short	*srcy, *endy, *oldsx;
	register BE_t	*hfp, *endhfp;

	register short	leftWidth, leftSkip, rightWidth;

	cursorOff(sleft, stop, w, h);
	cursorOff(dleft, dtop, w, h);

	oldsx = srcx = (short *)(GXBase + GXselectX + GXset0 + GXsource) + sx;

	srcy = (short *)(GXBase + GXselectY + GXset0) + sy;
	endy = srcy + h;

	destp = &df.f_mapp[destIndex];
	hfp = &hf.f_mapp[dy & 15];
	endhfp = &hf.f_mapp[HALFTONE_WORDS];

	leftSkip = dx  &  bitsPerBE - 1;
	leftWidth = bitsPerBE - leftSkip;

	if (w <= leftWidth) {
		leftWidth = w;
		rightWidth = 0;
	}
	else
		rightWidth =  dx + w  &  bitsPerBE - 1;
	endx = srcx + w - rightWidth;

	/*
	 *	The Programmer's manual lies; width of register 
	 *	effects reads.
	 */

	GXwidth = bitsPerBE;	

	for (  ;   srcy < endy;   srcx = oldsx,  destp += destDelta) {
		halftoneWord = *srcy++;
		halftoneWord = *srcx;	/* for pipelining */
		srcx += leftWidth;

		halftoneWord = *hfp++;
		if (hfp == endhfp)
			hfp = hf.f_mapp;

		/*
		 *	do first word
		 */

		combineSrcDst(~*srcx >> leftSkip  &  halftoneWord,  *destp,
		    mergeWord);
		*destp = mask1 & mergeWord |  ~mask1 & *destp;
		++destp;

		for ( ;  srcx < endx;  ++destp) {
			srcx += bitsPerBE;
			combineSrcDst(~*srcx & halftoneWord, *destp, *destp);
		}

		if (rightWidth) {	/* if more bits left */
			combineSrcDst(~*srcx & halftoneWord,  *destp,
			    mergeWord);
			*destp = mask2 & mergeWord |  ~mask2 & *destp;
			++destp;
		}
	}
	cursorOn();
}


 /*
  *	Bitblt to display, no src no ht.
  *	There is some trickiness here to use 32-bit instructions for speed.
  */

BBDstDisp()
{
	register long	*dsty, *endy;
	register short	*dstx, *endx;
	register short	hreg, leftWidth, hHi, hLo;
	register boolean	isOdd;

	cursorOff(dleft, dtop, w, h);

	GXfunction = sunRmSrcMem[combinationRule];	/* comb. rule */
	GXsetSource = ~0;				/* no source */
	GXpattern = ~0;					/* no halftone */

	/*
	 *	dstx and dsty are the X and Y register pairs in in SUN's
	 *	graphics board.
	 *	SUN's graphics board requires that the y coordinates (dy)
	 *	be left shifted by one bit.
	 */

	dstx = (short *)(GXBase + GXselectX + GXset0) + dx;
	dsty = (long *)(GXBase + GXselectY + GXset0 + GXupdate + (dy << 1));

	/*
	 *	The algorithm is:
	 *		align Form on the right hand side,
	 *		copy the left partial words (or full 32-bit words
	 *		    if w is a factor of 16).
	 *		copy the rest in chunks of 32-bits in a loop.
	 */

	hreg = h >> 1;
	hHi = hreg & ~7;
	hLo = hreg & 7;
	isOdd = h & 1;
	endx = dstx + w;

	leftWidth = w  &  bitsPerBE - 1;
	if (leftWidth == 0)
		leftWidth = bitsPerBE;

	for (GXwidth = leftWidth,  *dstx = isOdd,  endy = dsty + hHi;
	    dsty < endy; ) {		/* copy left word */

		/*
		 *	Each assignment is an 68k instruction moving
		 *	four bytes (long ints) to update the display.
		 *
		 *	isOdd is just used here as a convenient dummy register.
		 */

		*dsty++ = isOdd;  *dsty++ = isOdd;
		*dsty++ = isOdd;  *dsty++ = isOdd;
		*dsty++ = isOdd;  *dsty++ = isOdd;
		*dsty++ = isOdd;  *dsty++ = isOdd;
	}
	for (endy += hLo;  dsty < endy; )
		*dsty++ = isOdd;
	if (isOdd)
		*(short *)dsty = isOdd;

	/*
	 *	Copy the rest
	 */

	for (GXwidth = bitsPerBE,  dstx += leftWidth;
	    dstx < endx;
	    dstx += bitsPerBE) {
		for (*dstx = isOdd,  dsty -= hreg,  endy = dsty + hHi;
		    dsty < endy; ) {
			*dsty++ = isOdd;  *dsty++ = isOdd;
			*dsty++ = isOdd;  *dsty++ = isOdd;
			*dsty++ = isOdd;  *dsty++ = isOdd;
			*dsty++ = isOdd;  *dsty++ = isOdd;
		}
		for (endy += hLo;  dsty < endy; )
			*dsty++ = isOdd;
		if (isOdd)
			*(short *)dsty = isOdd;
	}
	cursorOn();
}


 /*
  *	Bitblt to display with halftone only.
  */

BBHtDstDisp()
{
	register short	*dsty, *endy;
	register char	*dstx, *endx;
	register BE_t	*hfp, *hfendp;
	register short	wreg, leftWidth, restWidth, wHi, wLo;

	cursorOff(dleft, dtop, w, h);

	GXfunction = sunRmSrcMem[combinationRule];
	GXsetSource = ~0;

	dstx = (char *)(GXBase + GXselectX + GXset0 + GXupdate) + (dx << 1);
	dsty = (short *)(GXBase + GXselectY + GXset0) + dy;

	endy = dsty + h;

	hfp = &hf.f_mapp[dy  &  15];
	hfendp = &hf.f_mapp[HALFTONE_WORDS];

	leftWidth = w  &  15;

	wreg = w << 1;
	wHi = (wreg & ~31)  &  ~255;
	wLo = (wreg & ~31)  &  255;

	restWidth = bitsPerBE << 1;

	for ( ;  dsty < endy;  dstx -= wreg) {	/* vertical loop */
		GXpattern = *hfp++;		/* halftone mask */
		if (hfp == hfendp)
			hfp = hf.f_mapp;

		*dsty++ = wreg;			/* set up row */
		if (leftWidth) {
			GXwidth = leftWidth;
			*(short *)dstx = wreg;	/* write first word */
		}
		for (dstx += leftWidth,  dstx += leftWidth,
		      GXwidth = bitsPerBE,  endx = dstx + wHi;
		    dstx < endx; ) {		/* horizontal loop */
			*(short *)dstx = wreg,  dstx += restWidth;
			*(short *)dstx = wreg,  dstx += restWidth;
			*(short *)dstx = wreg,  dstx += restWidth;
			*(short *)dstx = wreg,  dstx += restWidth;

			*(short *)dstx = wreg,  dstx += restWidth;
			*(short *)dstx = wreg,  dstx += restWidth;
			*(short *)dstx = wreg,  dstx += restWidth;
			*(short *)dstx = wreg,  dstx += restWidth;
		}
		for (endx += wLo;  dstx < endx; )
			*(short *)dstx = wreg,  dstx += restWidth;
	}
	cursorOn();
}


/*
 *	Bitblt from memory to display, no haftone.
 */

BBSrcMemDstDisp()
{
	register BE_t	*sourcep;
	register short	*dsty, *endy;
	register short	*dstx, *endx;
	register short	hreg, leftWidth, leftSkip, rightWidth;
	register short	rastReg;
	register BE_t	*oldSrc;

	cursorOff(dleft, dtop, w, h);

	oldSrc = sourcep = &sf.f_mapp[sourceIndex];
	rastReg = sourceRaster;

	dstx = (short *)(GXBase + GXselectX + GXset0) + dx;
	dsty = (short *)(GXBase + GXselectY + GXset0 + GXsource + GXupdate)
	    + dy;

	hreg = h;
	endy = dsty + hreg;

	leftSkip = sx  &  bitsPerBE - 1;
	leftWidth = bitsPerBE - leftSkip;

	if (w <= leftWidth) {
		leftWidth = w;
		rightWidth = 0;
	}
	else
		rightWidth =  sx + w  &  bitsPerBE - 1;
	endx = dstx + w - rightWidth;

	GXfunction = sunRmSrcMem[combinationRule];
	GXpattern = ~0;			/* no halftone pattern */

	if (leftSkip == 0)
		for (GXwidth = leftWidth,  *dstx = hreg; 
		    dsty < endy;
		    sourcep += rastReg)
			*dsty++ = *sourcep;
	else
		for (GXwidth = leftWidth,  *dstx = hreg;
		    dsty < endy;
		    sourcep += rastReg)
			*dsty++ = *sourcep  <<  leftSkip;

	for (dstx += leftWidth,  GXwidth = bitsPerBE;
	    dstx < endx;
	    dstx += bitsPerBE)
		for (sourcep = ++oldSrc,  dsty -= hreg,  *dstx = hreg;
		    dsty < endy;
		    sourcep += rastReg)
			*dsty++ = *sourcep;

	if (rightWidth)
		for (GXwidth = rightWidth,  sourcep = ++oldSrc,
		      dsty -= hreg,  *dstx = hreg;
		    dsty < endy;
		    sourcep += rastReg)
			*dsty++ = *sourcep;

	cursorOn();
}


/*
 *	Bitblt from memory to display, with halftone.
 */

BBSrcMemHtDstDisp()
{
	register short	*dsty, *endy;
	register BE_t	*sourcep;
	register short	*dstx, *endx;
	register BE_t	*hfp, *endhfp;
	register short	leftWidth, leftSkip, rightWidth, wreg, srcDel;

	cursorOff(dleft, dtop, w, h);

	hfp = &hf.f_mapp[dy & 15];
	endhfp = &hf.f_mapp[HALFTONE_WORDS];

	sourcep = &sf.f_mapp[sourceIndex];
	dstx = (short *)(GXBase + GXselectX + GXset0 + GXsource + GXupdate) 
	    + dx;
	dsty = (short *)(GXBase + GXselectY + GXset0) + dy;

	endy = dsty + h;

	leftSkip = sx  &  bitsPerBE - 1;
	leftWidth = bitsPerBE - leftSkip;

	wreg = w;
	if (wreg <= leftWidth) {
		leftWidth = wreg;
		rightWidth = 0;
	}
	else
		rightWidth =  sx + wreg  &  bitsPerBE - 1;

	wreg -= rightWidth;
	endx = dstx + wreg;

	srcDel = sourceRaster  -  ((sx + w + 15  >>  4) - (sx >> 4));

	GXfunction = sunRmSrcMem[combinationRule];

	for (  ;  dsty < endy;  sourcep += srcDel,  dstx -= wreg) {
		GXpattern = *hfp++;
		if (hfp == endhfp)
			hfp = hf.f_mapp;

		GXwidth = leftWidth;
		*dsty++ = wreg;
		*dstx = *sourcep++ << leftSkip;	/* copy first word */
		dstx += leftWidth;

		for (GXwidth = bitsPerBE;  dstx < endx;  dstx += bitsPerBE)
			*dstx = *sourcep++;

		if (rightWidth) {		/* copy last word, if any */
			GXwidth = rightWidth;
			*dstx = *sourcep++;
		}
	}
	cursorOn();
}


 /*
  *	Bitblt routine src and dst are display, no halftone.
  *	Sets up the graphics board functions and mask pattern (no pattern),
  *	then call the appropriate routine depending on the direction
  *	of movement.
  */

BBSrcDispDstDisp()
{
	cursorOff(sleft, stop, w, h);
	cursorOff(dleft, dtop, w, h);

	GXfunction = sunRmSrcDisp[combinationRule];
	GXpattern = ~0;

	if (hDir == 1)
		if (vDir == 1)
			BBSrcDispDstDispRD();
		else
			BBSrcDispDstDispRU();
	else
		if (vDir == 1)
			BBSrcDispDstDispLD();
		else
			fatal("bitblt: never go left and up", NullOop);
	cursorOn();
}


/*
 *	The following three routines copies from one place to another
 *	one on the display.  They could have been integrated together
 *	with the BBSrcDispHtDisp??() routines, but treating them as 
 *	separate cases is faster since these routines will be heavily
 *	used.
 *	Each of the BBSrcDistDisp??() differs in the direction 
 *	in which the for loops are controlled.
 */

/*
 *	From display to display, no halftone, moving right and down.
 */

BBSrcDispDstDispRD()
{
	register short	*srcy, *dsty, *endsrcy;
	register short	*dstx, *srcx, *endsrcx;
	register short	hreg, leftWidth;

	dstx = (short *)(GXBase + GXselectX + GXset0) + dx;
	srcx = (short *)(GXBase + GXselectX + GXset1) + sx;
	dsty = (short *)(GXBase + GXselectY + GXset0 + GXupdate) + dy;
	srcy = (short *)(GXBase + GXselectY + GXset1 + GXsource) + sy;
	endsrcx = srcx + w;
	hreg = h;
	endsrcy = srcy + hreg;

	/*
	 *	Right adjust Form.
	 */

	leftWidth = w  &  bitsPerBE - 1;
	if (leftWidth == 0)
		leftWidth = bitsPerBE;


	/*
	 *	Copy first word (may be a partial word) for each row
	 */

	for (GXwidth = leftWidth,  *dstx = *srcx;
	    srcy < endsrcy;
	    *dsty++ = *srcy++)
		;


	/*
	 *	Copy the rest of the words.
	 */

	for (GXwidth = bitsPerBE,  srcx += leftWidth,  dstx += leftWidth;
	    srcx < endsrcx;
	    srcx += bitsPerBE,  dstx += bitsPerBE)	/* horizontal loop */
		for (srcy -= hreg,  dsty -= hreg,  *dstx = *srcx;
		    srcy < endsrcy;
		    *dsty++ = *srcy++)			/* vertical loop */
			;
}


/*
 *	From display to display, no halftone, moving right and up.
 *	Similar to BBSrcDispDstDispRD() except we are copying the form
 *	from the bottom up.
 */

BBSrcDispDstDispRU()
{
	register short	*srcy, *dsty, *endsrcy;
	register short	*dstx, *srcx, *endsrcx;
	register short	hreg, leftWidth;

	hreg = h;
	dstx = (short *)(GXBase + GXselectX + GXset0) + dx;
	srcx = (short *)(GXBase + GXselectX + GXset1) + sx;
	dsty = (short *)(GXBase + GXselectY + GXset0 + GXupdate) + dy + 1;
	srcy = (short *)(GXBase + GXselectY + GXset1 + GXsource) + sy + 1;
	endsrcx = srcx + w;
	endsrcy = srcy - hreg;

	leftWidth = w  &  bitsPerBE - 1;
	if (leftWidth == 0)
		leftWidth = bitsPerBE;

	for (GXwidth = leftWidth,  *dstx = *srcx;
	    srcy > endsrcy;
	    *--dsty = *--srcy)
		;

	for (GXwidth = bitsPerBE,  srcx += leftWidth,  dstx += leftWidth;
	    srcx < endsrcx;
	    srcx += bitsPerBE,  dstx += bitsPerBE)
		for (*dstx = *srcx,  srcy += hreg,  dsty += hreg;
		    srcy > endsrcy;
		    *--dsty = *--srcy)
			;
}


/*
 *	From display to display, no halftone, moving left and down
 *	Similar to BBSrcDispDstDispRD(), except it left adjusts the
 *	form and copies from right to left.
 */

BBSrcDispDstDispLD()
{
	register short	*srcy, *dsty, *endsrcy;
	register short	*dstx, *srcx, *endsrcx;
	register short	hreg, rightWidth, wreg;

	wreg = w;
	dstx = (short *)(GXBase + GXselectX + GXset0) + dx + 1;
	srcx = (short *)(GXBase + GXselectX + GXset1) + sx + 1;
	dsty = (short *)(GXBase + GXselectY + GXset0 + GXupdate) + dy;
	srcy = (short *)(GXBase + GXselectY + GXset1 + GXsource) + sy;
	endsrcx = srcx - wreg;
	hreg = h;
	endsrcy = srcy + hreg;

	rightWidth = wreg  &  bitsPerBE - 1;	/* left adjusted */
	if (rightWidth == 0)
		rightWidth = bitsPerBE;

	for (GXwidth = rightWidth,  srcx -= rightWidth,  dstx -= rightWidth,
	      *dstx = *srcx;
	    srcy < endsrcy;
	    *dsty++ = *srcy++)
		;

	for (GXwidth = bitsPerBE;  srcx > endsrcx;  )	/* horizontal loop */
		for (srcx -= bitsPerBE,  dstx -= bitsPerBE,
		     *dstx = *srcx,  srcy -= hreg,  dsty -= hreg;
		    srcy < endsrcy;
		    *dsty++ = *srcy++)			/* vertical loop */
			;
}

/*
 *	The following routines are for bitblting from display to display
 *	with a halftone.  BBSrcDispHtDstDisp() is similar to BBSrcDispDstDisp()
 *	except it calls different routines for doing the actual bitblting.
 */


/*
 *	Bitblt routine src, dst is display with halftone.
 */

BBSrcDispHtDstDisp()
{
	/*
	 *	Removes cursor from src or dst.
	 */

	cursorOff(sleft, stop, w, h);
	cursorOff(dleft, dtop, w, h);

	GXfunction = sunRmSrcDisp[combinationRule];

	if (hDir == 1)
		if (vDir == 1)
			BBSrcDispHtDstDispRD();
		else
			BBSrcDispHtDstDispRU();
	else
		if (vDir == 1)
			BBSrcDispHtDstDispLD();
		else
			fatal("bitblt: never go left and up", NullOop);
	cursorOn();
}


/*
 *	These routines are generalizations of BBSrcDispDstDisp??().
 *	They works fine even when the halftone is nil (i.e. the 
 *	degenerated cases above).
 *	They have the vertical loop on the outside to save
 *	calculations on which halftone pattern word to use.
 */	


/*
 *	Going to the right and down.
 */

BBSrcDispHtDstDispRD()
{
	register short	*dstx, *srcx, *endsrcx, *srcy, *dsty, *endsrcy;
	register BE_t	*hfp, *hfendp;
	register short	wreg, leftWidth;

	hfp = &hf.f_mapp[dy  &  HALFTONE_WORDS - 1];
	hfendp = &hf.f_mapp[HALFTONE_WORDS];

	dstx = (short *)(GXBase + GXselectX + GXset0 + GXupdate) + dx;
	srcx = (short *)(GXBase + GXselectX + GXset1 + GXsource) + sx;
	dsty = (short *)(GXBase + GXselectY + GXset0) + dy;
	srcy = (short *)(GXBase + GXselectY + GXset1) + sy;
	wreg = w;
	endsrcx = srcx + wreg;
	endsrcy = srcy + h;

	leftWidth = wreg  &  bitsPerBE - 1;	/* left adjust */
	if (leftWidth == 0)
		leftWidth = bitsPerBE;

	for (;  srcy < endsrcy;  dstx -= wreg,  srcx -= wreg) {	/* vert loop */
		GXpattern = *hfp++;
		if (hfp == hfendp)
			hfp = hf.f_mapp;

		*dsty++ = *srcy++;

		GXwidth = leftWidth,
		*dstx = *srcx;			/* do first word */

		for (GXwidth = bitsPerBE,
		      srcx += leftWidth,   dstx += leftWidth;
		    srcx < endsrcx;
		    srcx += bitsPerBE,  dstx += bitsPerBE) /* hort loop */
			*dstx = *srcx;		/* dummy data */
	}
}


/*
 *	Going to the right and up.
 */

BBSrcDispHtDstDispRU()
{
	register short	*dstx, *srcx, *endsrcx, *srcy, *dsty, *endsrcy;
	register BE_t	*hfp, *hfendp;
	register short	wreg, leftWidth;

	hfp = &hf.f_mapp[dy  &  HALFTONE_WORDS - 1];
	hfendp = &hf.f_mapp[0];

	dstx = (short *)(GXBase + GXselectX + GXset0 + GXupdate) + dx;
	srcx = (short *)(GXBase + GXselectX + GXset1 + GXsource) + sx;
	dsty = (short *)(GXBase + GXselectY + GXset0) + dy + 1;/* bottom */
	srcy = (short *)(GXBase + GXselectY + GXset1) + sy + 1;
	wreg = w;
	endsrcx = srcx + wreg;
	endsrcy = srcy - h;

	leftWidth = wreg  &  bitsPerBE - 1;
	if (leftWidth == 0)
		leftWidth = bitsPerBE;

	for (  ;   srcy > endsrcy;   dstx -= wreg,  srcx -= wreg) {
		if (hfp == hfendp)
			hfp = &hf.f_mapp[HALFTONE_WORDS];
		GXpattern = *--hfp;

		*--dsty = *--srcy;

		GXwidth = leftWidth;
		*dstx = *srcx;		/* dummy data */

		for (srcx += leftWidth,  dstx += leftWidth,
		      GXwidth = bitsPerBE;
		    srcx < endsrcx;
		    srcx += bitsPerBE,  dstx += bitsPerBE)
			*dstx = *srcx;	/* dummy data */
	}
}


/*
 *	Going to the left and down.
 */

BBSrcDispHtDstDispLD()
{
	register short	*dstx, *srcx, *endsrcx, *srcy, *dsty, *endsrcy;
	register BE_t	*hfp, *hfendp;
	register short	wreg, rightWidth;

	hfp = &hf.f_mapp[dy  &  HALFTONE_WORDS - 1];
	hfendp = &hf.f_mapp[HALFTONE_WORDS];

	wreg = w;
	dstx = (short *)(GXBase + GXselectX + GXset0 + GXupdate) + dx + 1;
	srcx = (short *)(GXBase + GXselectX + GXset1 + GXsource) + sx + 1;
	dsty = (short *)(GXBase + GXselectY + GXset0) + dy;
	srcy = (short *)(GXBase + GXselectY + GXset1) + sy;
	endsrcx = srcx - wreg;
	endsrcy = srcy + h;

	rightWidth = wreg  &  bitsPerBE - 1;
	if (rightWidth == 0)
		rightWidth = bitsPerBE;

	for (  ;   srcy < endsrcy;   dstx += wreg,  srcx += wreg) {
		GXpattern = *hfp++;
		if (hfp == hfendp)
			hfp = hf.f_mapp;

		*dsty++ = *srcy++;

		GXwidth = rightWidth;
		*(dstx -= rightWidth) = *(srcx -= rightWidth);

		for (GXwidth = bitsPerBE;
		    srcx > endsrcx;
		    *(dstx -= bitsPerBE) = *(srcx -= bitsPerBE))
			;
	}
}
