static char rcsId[] = "$Header: bs.c,v 43.9 84/08/10 13:38:07 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Main program for BS.
 */


# include	<stdio.h>

# include	"utils.h"
# include	"mem.h"
# include	"execute.h"
# include	"io.h"
# include	"cursor.h"

# define	usage\
	"usage: %s [-d display] [-m mouse] [-k keyboard] [-i] [-e] [-s] [image]\n"


static boolean	getImageName();

FILE	*traceFp = NULL;		/* for trace information */


/*
 *	True if environment variable ``Encoded'' or -e is set.
 *	Cause BS to generate ASCII events to Smalltalk.
 *	Turn it on if you do not have a VT100 keyboard, or you 
 *	have not filed in vt100Keymap.st yet. 
 */

boolean		EncodedKeyboard;


/*
 *	main() opens files and parses arguments,
 *	Reads in the image, initializes the SUN,
 *	and calls execute() which runs the interpreter.
 */

main(argc, argv)
int argc;
char *argv[];
{
	char		*namep;
	char		imageName[BUFSIZ];
	extern char	*getenv(), *getlogin();

	moncontrol(0);			/* turn off profiling */
	if ((namep = getenv("TRACE")) != NULL)
		traceFp = xfopen(namep, "w");
	else
		traceFp = stdout;


	/*
	 *	Cannot suspend bs if my name starts with '-' (i.e. a
	 *	login shell).
	 */
	
	setSuspend(**argv != '-');


	/*
	 *	If no image name is given to bs, it will look for the
	 *	file loginName.im.
	 */

	if (!getImageName(argc, argv, imageName)) {
		strcpy(imageName, getlogin());
		if (!imageName[0])	/* cannot find login name */
			fatal("Who are you?", NullOop);
		strcat(imageName, ".im");
	}

	if (!EncodedKeyboard)
		EncodedKeyboard = (boolean)getenv("Encoded");


	traceInitialize();
	SunInitialize();


	/*
	 *	WINDOW_PARENT tells us whether it is invoked from
	 *	within suntools or not.
	 *	Must open window first if running under suntools.  
	 *	Otherwise, my process size will grow
	 *	so big that the fork() in OpenWindow() may fail.
	 */

	if (getenv("WINDOW_PARENT")) {		/* from suntools */
		OpenWindow(argv);
		loadFromFile(imageName);
	}
	else {
		loadFromFile(imageName);
		OpenWindow(argv);
	}
	execute();
}


/*
 *	Looks at the argument line, skips over the arguments that can 
 *	will be handled by win_initscreenfromargv(), and store the 
 *	image name in name[].  Returns TRUE if there is an image name 
 *	given, FALSE otherwise.
 */

static boolean
getImageName(argc, argv, name)
int argc;
char **argv;
char name[];
{
	register int	i;
	
	i = 1;
	while (i < argc  &&  argv[i][0] == '-')
		switch (argv[i][1]) {
		/*
		 *	These are sunwindow options.
		 */
		case 'd':		/* display */
		case 'm':		/* mouse */
		case 'k':		/* keyboard */
			i += 2;		/* skip over next argument */
			break;
		case 'f':		/* color maps? */
		case 'b':
			i += 3;		/* skip over red/green/blue args */
			break;
		case 'i':		/* invert black and white */
			i++;
			break;
		/*
		 *	These are BS options.
		 */
		case 'e':		/* 'e'ncoded keyboard */
			i++;
			EncodedKeyboard = TRUE;
			break;
		case 's':		/* for 's'teady cursor */
			i++;
			setCursorChange(FALSE);
			break;
		default:
			fprintf(stderr, usage, argv[0]);
			fatal("getImageName: unknown option.\n", NullOop);
			break;
		}

	if (i < argc) {
		strcpy(name, argv[i]);
		return(TRUE);
	}
	else				/* no image name given */
		return(FALSE);
}


/*
 *	Two factors determines whether BS can be suspended.
 *	You can suspend BS only if (1) it is not your login shell
 *	and (2) it is not invoked from suntools.  Suspendable tells
 *	you whether you can suspend it.
 */

static suspendable = TRUE;		/* can suspend */

void
setSuspend(flag)
boolean flag;
{
	suspendable = flag;
}


boolean
canSuspend()
{
	return(suspendable);
}
