static char rcsId[] = "$Header: ctrlPrims.c,v 41.1 84/03/14 19:58:19 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	This file consists of the implemenation of the primitive messages
 *	for blockContext and Objects.  Most of the code are translated
 *	from the Smalltalk code in Goldberg and Robson, p. 638-41.
 *
 *	If it weren't for these, the primitive interface would be
 *	faster and simpler.
 */

# include	<stdio.h>
# include	<math.h>
# include	<sgtty.h>
# include	<signal.h>
# include	<ctype.h>

# include	<sys/types.h>
# include	<sys/vtimes.h>
# include	<sys/timeb.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"prims.h"
# include	"execute.h"
# include	"mc.h"

/*
 *	Both value primitives must decrement the sp in the ac:
 *		It is normally set to include the answer for the prim.
 *		But, these prims with return, which will push the answer on.
 */

/*
 *	BlockContext value, value:, value:value:, etc.
 *
 *	The number of arguments passed is in argumentCount.
 *	After the receiver has been initialized with the active
 *	context's stack, stack pointer and instruction pointer, 
 *	it becomes the active context.
 */

rp_t
BCValuePrm(rp, ac, argumentCount)
oop_t *rp;
rp_t ac;
register u_int argumentCount;
{
	register rp_t	newContext;

	newContext = (rp_t)*rp;

	if (integerValueOf(argCountBC(newContext)) !=  argumentCount)
		return (NullOop);

	addToSmallInteger(spC(ac), -1);
	callerBC(newContext) = ac;
	ipC(newContext) = initIPBC(newContext);
	spC(newContext) = integerObjectOf(argumentCount);
	copy_words(&rp[1], tempFrameC(newContext), argumentCount);

	return (newContext);
}


/*
 *	BlockContext valueWithArguments: anArray
 *
 *	Basically the same as BCValuePrm, except the arguments 
 *	comes in a single array instead of separately.
 *	Fails if the argument is not an array.
 */

rp_t
BCValArgsPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	newContext;
	register rp_t	args;
	register u_int	argLength;

	args = (rp_t)rp[1];
	if (isIntegerObject((oop_t)args)
	    ||  classRp(args) != classArrayOop)
		return (NullOop);
	argLength = lengthRp(args);

	newContext = (rp_t)*rp;
	if (integerValueOf(argCountBC(newContext)) != argLength
	    ||  callerBC(newContext) != nilOop)
		return (NullOop);

	addToSmallInteger(spC(ac), -1);
	callerBC(newContext) = ac;
	ipC(newContext) = initIPBC(newContext);
	spC(newContext) = integerObjectOf(argLength);
	copy_words(oopFieldRp(args), tempFrameC(newContext), argLength);

	return (newContext);
}


/*
 *	The following two routines implement the perform messages.
 */

/*
 *	Object perform:, perform:with:, perform:with:with, etc.
 *
 *	The receiver is initialized with the correct number of arguments.
 *	Must get rid of selector from the stack.  Thus, the number of
 *	arguments for the send is the number of arguments to ``perform:''
 *	minus one.
 */

rp_t
ObjPerfPrm(rp, ac, argumentCount)
register oop_t *rp;
rp_t ac;
u_int argumentCount;
{
	register struct methodCache	*cs;
	register rp_t	class;
	register rp_t	selector;

	register oop_t	rcvr;
	register u_int	newArgCount;

	rcvr = *rp;
	class = isIntegerObject(rcvr) ? classSmallIntegerOop
	    : classRp((rp_t)rcvr);
	selector = (rp_t)rp[1];

	/*
	 *	Do method lookup in order to check argument count.
	 *	If can't find method, assume correct count so that
	 *	subsequent send code can send messageNotUnderstood.
	 */

	cs = cacheSlot(class, selector);
	if (cs->mc_sel == selector  &&   cs->mc_class == class
	    ||    lookupMethodInClass(class, (oop_t)selector)) {
		/* 
		 *	found it 
		 */
		argumentCountCM(cs->mc_method, headerCM(cs->mc_method),
		    newArgCount);
		if (argumentCount - 1  !=  newArgCount)
			return (NullOop);
	}
	else
		newArgCount = argumentCount - 1;

	copy_words(&rp[2],  &rp[1],  newArgCount);
	storeSP(ac,  rp + newArgCount + 1);

	return (methodSend(ac, selector, newArgCount, FALSE));
}


/*
 *	Object perform:withArguments
 *
 *	This is the same as ObjPerfPrm(), except the arguments
 *	comes in a single array instead of as multiple arguments to 
 *	the perform: message.
 *	Do some extra work to make this look like a normal send.
 *	The work is looking up the method to verify the arg count.
 *	If good, copy arguments from the array into the current context.
 *	This can fail if there's not enough room in the current
 *	context for the arguments, but this does not seem to be
 *	a problem.
 */

rp_t
ObjPerfArgsPrm(rp, ac)
register oop_t *rp;
register rp_t ac;
{
	register rp_t	argArray;
	register struct	methodCache	*cs;

	register oop_t	class;
	register oop_t	selector;
	register u_int	newArgCount;
	register oop_t	rcvr;

	rcvr = *rp;
	selector = rp[1];
	argArray = (rp_t)rp[2];

	if (isIntegerObject((rp_t)argArray)  
	    ||  classRp(argArray) != classArrayOop)
		return (NullOop);

	class = isIntegerObject(rcvr) ? (oop_t)classSmallIntegerOop
	    : (oop_t)classRp((rp_t)rcvr);

	cs = cacheSlot(class, selector);
	if (cs->mc_sel == (rp_t)selector  &&  cs->mc_class == (rp_t)class
	    ||    lookupMethodInClass((rp_t)class, selector)) {
		argumentCountCM(cs->mc_method, headerCM(cs->mc_method),
		    newArgCount);
		if (lengthRp(argArray) !=  newArgCount)
			return (NullOop);
	}
	else
		newArgCount = lengthRp(argArray);


	/*
	 *	Check that the receiver has enough room on the stack
	 *	to hold all the arguments before copying the arguments
	 *	from the argument array to receiver's stack.
	 */

	if (&fetchSP(ac)[newArgCount - 2] >= &oopFieldRp(ac)[lengthRp(ac)])
		return (NullOop);
	copy_words(oopFieldRp(argArray), &rp[1], newArgCount);
	storeSP(ac,  rp + newArgCount + 1);

	return (methodSend(ac, (rp_t)selector, newArgCount, FALSE));
}


/*
 *	Actual blockCopy operation, creating a blockContext.  
 *
 *	Gets the old context and argument count on the stack.
 *	It allocates and initializes the blockContext,
 *	and fills in the header.
 */

rp_t
CBlockCopyPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	newcon, oldcon;
	register oop_t	count;		/* number of block arguments */

	/*
	 *	The number of bytecodes after the send bytecode.
	 *	It is 3 in the implementation of primitiveBlockCopy
	 *	in the book, but 2 is what works for us because we have
	 *	already stored the incremented pc.
	 */

#	define	lengthOfJumpAroundBlock	2

	
	oldcon = (rp_t)rp[0];
	getInt(rp[1], count, NullRp);

	/*
	 *	If the receiver is a BlockContext, its home context
	 *	becomes the new context's home context.
	 */

	if (isBlock(oldcon))
		oldcon = homeBC(oldcon);

	/*
	 *	Check to see that arguments will fit in a context
	 *	the same size as oldcon.
	 */

	if (&tempFrameC((rp_t)0)[(u_int)count]  
	    >= &oopFieldRp((rp_t)0)[lengthRp(oldcon)])
		return (NullOop);
	newcon = instantiateClass(classBlockContextOop,
	    lengthRp(oldcon) - fixedFieldsC);

	homeBC(newcon) = oldcon;
	argCountBC(newcon) = integerObjectOf(count);

	{
		register u_int	newIP;

		/*
		 *	Computes new ip and stores it in the new context.
		 */

		newIP = integerObjectOf(nonNegIntegerValueOf(ipC(ac))
		    + lengthOfJumpAroundBlock);
		initIPBC(newcon) = newIP;
		ipC(newcon) = newIP;
	}

	*rp = (oop_t)newcon;
	spC(newcon) = integerObjectOf(0);

	return (ac);
}
