static char rcsId[] = "$Header: cursor.c,v 44.6 84/07/25 22:03:47 pkl Exp $ (Berkeley)";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */


# include	<stdio.h>
# include	<sunwindow/window_hs.h>
# include	<pixrect/pixrect_hs.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"cursor.h"
# include	"bitblt.h"

/*
 *	simulates Alto unlinked cursor
 */

static boolean	altoCursorLinked = TRUE;
static u_int	altoCursorX, altoCursorY;
static void	paintUnlinkedCursor(), removeUnlinkedCursor();


/*
 *	The current cursor is cached.
 *	The cursor is initialized to some
 *	interesting pattern (the BS logo cursor).
 */

static BE_t	altoCursorBitmap[cursorBitmapLength]  = {
		0xfc00, 0xfc00, 0xc300, 0xc300,
		0xfc3c, 0xfc3c, 0xc3c0, 0xc3c0,
		0xfc3f, 0xfc3f, 0x0003, 0x0003,
		0x00c3, 0x00c3, 0x003c, 0x003c
	};

mpr_static(cursorShape, CursorWidth, CursorHeight, 1, altoCursorBitmap);

static struct cursor	altoCursor = {0, 0, PIX_SRC | PIX_DST, &cursorShape};

/*
 *	True to enable ggc, swap and snapshot cursor.
 */

static boolean	cursorChange = TRUE;


/*
 *	Initialize cursor to some interesting pattern.  This also makes
 *	sure that the cursor is the Smalltalk cursor as soon as it starts up.
 */

InitCursor()
{
	altoCursorX = 0, altoCursorY = 0;

	if (!windowOpened(BSWindowFd))
		return;
	win_setcursor(BSWindowFd, &altoCursor);
}


/*
 *	Status of cursor flag
 */

void
setCursorChange(flag)
boolean flag;
{
	cursorChange = flag;
}
	

/*
 *	The cursor will change into an empty bucket at the beginning
 *	of a snapshot.  After every transfer of data, a 2x3 halftone 
 *	block will be added to the cursor, until the cursor is filled 
 *	up or snapshot is finished.
 *	halftoneMask[] contains masks for making 2x3 halftone blocks.
 */

static BE_t	halftoneMask[] = {0x3800, 0x0700, 0x00e0, 0x001c};
# define	halftonePattern		0xaaaa

/*
 *	The lower two bits of patternIndex is an index to 
 *	halftoneMask[] to select the pattern to use.  The 
 *	higher bits translate to the row number within the cursor.
 *	Incrementing patternIndex will move it to the next
 *	pattern and next row (if necessary).
 */

static int	patternIndex;


/*
 *	Generates an empty bucket and stores it in cursor.
 *	This cursor will be displayed during a snapshot.
 */

void
initSnapshotCursor(bucket, saved)
BE_t bucket[], saved[];
{
	static BE_t	emptyBucket[cursorBitmapLength] = {
		0xc003, 0xc003, 0xc003, 0xc003,
		0xc003, 0xc003, 0xc003, 0xc003,
		0xc003, 0xc003, 0xc003, 0xc003,
		0xc003, 0xc003, 0xffff, 0xffff};
	
	if (!cursorChange)
		return;
	
	patternIndex = 0;
	copyCursor(altoCursorBitmap, saved);
	copyCursor(emptyBucket, bucket);
	beCursor(bucket);
}


/*
 *	Puts one more 2x3 halftone block in bucket cursor.  
 */

void
snapshotCursor(bucket)
BE_t bucket[];
{
	/*
	 *	Extracts and translates row and column from patternIndex.
	 *	Row is obtained from patternIndex by discarding the lower
	 *	two bits and multiplying it by two (blocks are two pixels
	 *	high).
	 */

#	define bottomRow	13	/* lowest row inside bucket */
#	define columnField	0x0003
#	define column		(patternIndex & columnField)
#	define row \
		(bottomRow   -   (patternIndex >> 2)  *  2)

	if (!cursorChange  ||  row < 0)/* cursor disables or bucket is full */
		return;

	bucket[row]     |=  halftonePattern & halftoneMask[column];
	bucket[row - 1] |= ~halftonePattern & halftoneMask[column];
	patternIndex++;
	beCursor(bucket);
}


/*
 *	The following two routines save the current cursor in the
 *	16x16 bit array (argument ``saved'') and display a special
 *	cursor.  The next routine restores the saved cursor.
 */

/*
 *	Save cursor and display first cursor for GGC (the square cursor).
 */

void
ggcCursor(saved)
BE_t *saved;
{
	static BE_t	mc[cursorBitmapLength] = {
		0xffff, 0xffff, 0xc003, 0xc61b,
		0xc61b, 0xc003, 0xffff, 0xc003,
		0xc003, 0xc003, 0xc0c3, 0xd8c3,
		0xd803, 0xc003, 0xffff, 0xffff};

	if (!cursorChange)
		return;

	copyCursor(altoCursorBitmap,  saved);
	beCursor(mc);
}


/*
 *	Called from swap pointer to objects for long becomes (the rake 
 *	cursor).
 */

void
swapCursor(saved)
BE_t *saved;
{
	static BE_t	mc[cursorBitmapLength] = {
		0x000c, 0x000c, 0x0018, 0x0018,
		0x0030, 0x0030, 0x0060, 0x0060,
		0x00c0, 0x00c0, 0x0180, 0x0180,
		0xffff, 0xffff, 0x5555, 0x5555};

	if (!cursorChange)
		return;

	copyCursor(altoCursorBitmap, saved);
	beCursor(mc);
}


/*
 *	Gives a non-zero bitmap to the cursor.
 */

void
restoreCursor(saved)
BE_t *saved;
{
	if (!cursorChange)
		return;

	beCursor(saved);
}


/*
 *	beCursor:  takes a new cursor bitmap and loads it into the cursor.
 */

void
beCursor(srcp)
BE_t *srcp;
{
	copyCursor(srcp, altoCursorBitmap);
	if (!windowOpened(BSWindowFd))
		return;

	else if (altoCursorLinked)
		win_setcursor(BSWindowFd, &altoCursor);
	else {
		cursorOff(altoCursorX, altoCursorY, CursorWidth, CursorHeight);
		removeUnlinkedCursor();
		paintUnlinkedCursor();
		cursorOn();
	}
}


/*
 *	Better kernel interface would clean this up.
 *	Used only in setCursorLink(), but mpr_static can only be
 *	declared globally (although it says "static", not all its
 *	components are static!).
 */

static BE_t	invisCursorBitmap[1];		/* implicit init */

static mpr_static(invisCursorShape, 0, 0, 1, invisCursorBitmap);

static struct cursor	invisCursor = {0, 0, PIX_DST, &invisCursorShape};


void
setCursorLink(link)
boolean link;
{
	if (link && altoCursorLinked  ||  !link && !altoCursorLinked)
		return;

	if (altoCursorLinked = link) {
		cursorOff(altoCursorX, altoCursorY, CursorWidth, CursorHeight);
		removeUnlinkedCursor();
		cursorOn();
		win_setmouseposition(BSWindowFd, altoCursorX, altoCursorY);
		setMouseToXY(altoCursorX, altoCursorY);
		win_setcursor(BSWindowFd, &altoCursor);
	}
	else {
		win_setcursor(BSWindowFd, &invisCursor);
		cursorOff(altoCursorX, altoCursorY, CursorWidth, CursorHeight);
		getMouseCoordinates(&altoCursorX, &altoCursorY);
		paintUnlinkedCursor();
		cursorOn();
	}
}


void
setCursorToXY(x,y)
register int x, y;
{
	register int	winWidth, winHeight;	/* current width and height */

	winWidth = win_getwidth(BSWindowFd);
	winHeight = win_getheight(BSWindowFd);

	if (x < 0)
		x = 0;
	else if (x > winWidth)
		x = winWidth;

	if (y < 0)
		y = 0;
	else if (y > winHeight)
		y = winHeight;

	if (altoCursorLinked) {
		setMouseToXY(x, y);
		win_setmouseposition(BSWindowFd, x, y);
		altoCursorX = x, altoCursorY = y;
	}
	else {
		cursorOff(altoCursorX, altoCursorY, CursorWidth, CursorHeight);
		removeUnlinkedCursor();
		altoCursorX = x, altoCursorY = y;
		paintUnlinkedCursor();
		cursorOn();
	}
}


/*
 *	Save area under cursor, paint it, or restore area under cursor
 *	(for unlinked cursor).
 *	``unlinkedCursorPr'' saves the 16x16 bitmap currently under the
 *	cursor.
 */

static short	unlinkedCursorBitCache[cursorBitmapLength];

static mpr_static(unlinkedCursorPr, cursorBitmapLength, cursorBitmapLength,
    PIXEL_DEPTH, unlinkedCursorBitCache);

static mpr_static(altoCursorPr, cursorBitmapLength, cursorBitmapLength,
    PIXEL_DEPTH, altoCursorBitmap);


/*
 *	Save the 16x16 bitmap currently under the cursor, and paint the
 *	cursor shape on it.
 */

static void
paintUnlinkedCursor()
{
	register struct pixrect	*curPr, *altoPr; /* for speed */
	register struct pixwin	*regPixwin;

	regPixwin = BSPixwin;
	curPr = &unlinkedCursorPr;
	altoPr = &altoCursorPr;
	pw_read(curPr, 0, 0, cursorBitmapLength, cursorBitmapLength,
	    PIX_COPY, regPixwin, altoCursorX, altoCursorY);
	pw_write(regPixwin, altoCursorX, altoCursorY, cursorBitmapLength,
	    cursorBitmapLength, PIX_PAINT, altoPr, 0, 0);
}


/*
 *	Removes cursor by repainting the saved 16x16 bitmap (in
 *	unlinkedCursorPr) over the current cursor position.
 */

static void
removeUnlinkedCursor()
{
	register struct pixrect	*curPr;
	register struct pixwin	*regPixwin;

	regPixwin = BSPixwin;
	curPr = &unlinkedCursorPr;
	pw_write(regPixwin, altoCursorX, altoCursorY, cursorBitmapLength,
	    cursorBitmapLength, PIX_COPY, curPr, 0, 0);
}


/*
 *	Locks the window so that kernel cursor track will not
 *	interfere with bitblting.
 *
 *	No need to call pw_lock if it is already locked.
 *	Will only be used by bs.  Bs.tool uses pw_rop which does
 *	locking already.
 */

static boolean	pixwinLocked = FALSE;

void
cursorOff(x, y, w, h)
int x, y, w, h;
{
	struct rect	r;
	
	/*
	 *	The entire window is locked because we are using
	 *	the low level interface to the graphics board.
	 *	If raster ops are used, we can just lock the 
	 *	area affected by bitblt.
	 */

	if (!pixwinLocked) {
		r.r_left = r.r_top = 0;
		r.r_width = BSPixwin->pw_pixrect->pr_width;
		r.r_height = BSPixwin->pw_pixrect->pr_height;
		pw_lock(BSPixwin, &r);
		pixwinLocked = TRUE;
	}
}


/*
 *	Unlocks only when window is locked.  Also resets flag so that 
 *	cursorOff() will really lock window the next time it is 
 *	called.
 *
 *	Could have allowed cursorOff() to lock all the time and use 
 *	pw_reset() to release all locks, but it may
 *	be slower because window will be locked twice.
 */

void
cursorOn()
{
	if (pixwinLocked) {
		pw_unlock(BSPixwin);
		pixwinLocked = FALSE;
	}
}
