static char executeBlock_h[] = "$Header: execBlock.h,v 41.1 84/03/14 19:59:21 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	executeBlock.h: interpreter macros for bytecodes in blocks.
 *	Block execute is not too heavily optimized (yet).
 *	Most of this is similar to executeMethod.h, see comments there.
 */


/*
 *	fetch local variables
 */

# define fetchLocals(ac)\
	if (TRUE) {\
		hc = homeBC((rp_t)ac);\
		method = methodMC(hc);\
		pc = fetchPC((rp_t)ac, method);\
		sp = fetchSP((rp_t)ac);\
		rcvr = (oop_t)receiverMC(hc);\
	}\
	else


/*
 *	store locals e.g. before a send
 */

# define storeLocals(ac)\
	if (TRUE) {\
		storePC(ac, method, pc);\
		storeSP(ac, sp);\
	}\
	else


/*
 *	since sends usually go to methods, may not be much point in
 *	doing it inline, since will have to return to execute() to
 *	call executeMethod() anyway.  On the other hand
 *	the primitives would benefit.
 */

# define doASend(sel, nargs)\
	if (TRUE) {\
		storeLocals((rp_t)ac);\
		r1 = (oop_t)methodSend((rp_t)ac, (rp_t)(sel), (u_int)(nargs), FALSE);\
		if (r1 != ac  ||  checkPoll())\
			return ((rp_t)r1);\
		fetchLocals((rp_t)ac);\
		break;\
	}\
	else

/*
 *	send to super
 */

# define doASuperSend(sel, nargs)\
	if (TRUE) {\
		storeLocals((rp_t)ac);\
		r1 = (oop_t)methodSend((rp_t)ac, (rp_t)(sel),\
		    (u_int)(nargs), TRUE);\
		if (r1 != ac  ||  checkPoll())\
			return ((rp_t)r1);\
		fetchLocals((rp_t)ac);\
		break;\
	}\
	else


/*
 *	push bytecodes.  "x" is the bytecode, "y" is the offset of the 
 *	receiver variable/temporary variable/literal constant, and "lab" 
 *	is a label for the assembler.
 */

# define pushRcvrVar(x, y, lab) \
	case x: \
		adb(lab);\
		*sp++ = oopFieldRp((rp_t)rcvr)[y]; \
		break

# define pushTempVar(x, y, lab) \
	case x:\
		adb(lab);\
		*sp++ = tempFrameC(hc)[y];\
		break

# define pushLitConst(x, y, lab) \
	case x:\
		adb(lab);\
		*sp++ = literalFrameCM(method)[y];\
		break


/*
 *	y is the index of the literal that points to the association
 *	whose value is what we want.
 */

# define pushLitVar(x, y, lab) \
	case x:\
		adb(lab);\
		*sp++ = valueAs((rp_t)literalFrameCM(method)[y]);\
		break

# define pushConstObj(x, y, lab) \
	case x:\
		adb(lab);\
		*sp++ = y;\
		break

/*
 *	stores.
 */

# define storePopRcvrVar(x, y, lab) \
	case x:\
		adb(lab);\
		r1 = *--sp;\
		oopFieldRp((rp_t)rcvr)[y] = r1;\
		ggcStore((rp_t)rcvr, r1);\
		break

/*
 *	unlike executeMethod, stores to temps in a block must
 *	be scavenge-checked in case a context is being stored.
 *	Consider the following example:
 *		bar: aBlock
 *			aBlock value: thisContext
 *
 *		foo: a  |b|  
 *			self bar: [:d | b <- d].
 *			b whatever.
 *
 *	Bar's context must be marked nonlifo when it is stored into b,
 *	because b is in foo's context and will outlive bar's context.
 *	If every thisContext were to mark the active context as nonlifo
 *	each block's home would be wastefully retained.
 */

# define storePopTempVar(x, y, lab) \
	case x:\
		adb(lab);\
		r1 = *--sp;\
		tempFrameC(hc)[y] = r1;\
		ggcStore(hc, r1); /* because this is a block */\
		break

# define storePopLitVar(x, y) \
	case x:\
		r1 = literalFrameCM(method)[y];\
		valueAs((rp_t)r1) = *--sp;\
		ggcStore((rp_t)r1, *sp);\
		break

/*
 *	jumps
 */

# define jump(x, y, lab) \
	case x: \
		adb(lab);\
		pc += y;\
		break


/*
 *	conditional jump.  Jump if TOS == yes, fall through if TOS == no.
 *	Send error message otherwise.
 */

# define popJump(x, y, yes, no, lab) \
	case x: \
		adb(lab);\
		r1 = *--sp;\
		if (r1 == (oop_t)yes)\
			pc += y;\
		else if (r1 != (oop_t)no) {\
			--pc;\
			*sp++ = r1;\
			doASend(mustBeBooleanSelectorOop, 0);\
		}\
		break


/*
 *	extended uncond jump.  y is derived from the bytecode.
 *	check polling.
 */

# define extendedJump(x, y, lab) \
	case x:\
		adb(lab);\
		pc += *pc + 1 + (y - 4  <<  8);\
		if (checkPoll()) {\
			storeLocals((rp_t)ac);\
			return ((rp_t)ac);\
		}\
		break


/*
 *	extended condtional jump.  Jump iff TOS == yes, fall through for no.
 */

# define extendedPopJump(x, y, yes, no, lab) \
	case x:\
		r1 = *--sp;\
		if (r1 == (oop_t)yes)\
			pc += y * 256  +  *pc;\
		else if (r1 != (oop_t)no) { \
			pc -= 1;\
			*sp++ = r1;\
			doASend(mustBeBooleanSelectorOop, 0);\
		}\
		++pc;\
	break

/*
 *	sends
 */

# define sendLitSel(x, sel, nargs, lab)\
	case x:\
		adb(lab);\
		doASend(literalFrameCM(method)[sel], nargs);\
		break


/*
 *	for direct primitives.  These are very similar to those 
 *	with the same name in executeMethod.h.  See those macros
 *	for more details.
 */

# define DirectRARecover(sel)	/* failure */\
	if (TRUE) {\
		sp += 2;\
		doASend(specialSelOf(sel), specialNargOf(sel));\
	}\
	else


# define DirectRAEnd(ans) 	/* success */\
	if (TRUE) {\
		*sp++ = ans;\
		break;\
	}\
	else

# define DirectRATest(sel) 	/* prolog for direct binary primis */\
	if (TRUE) {\
		r2 = *--sp;\
		r1 = *--sp;\
		if (!isIntegerObject(r1 & r2))\
			DirectRARecover(sel);\
	}\
	else

# define DirectRAConvert 	/* more prolog for smallInteger prims */\
	if (TRUE) {\
		convertToInteger(r1);\
		convertToInteger(r2);\
	}\
	else


/*
 *	epilogue for smallInt arith prims.  sel is the code needed to obtain
 *	the selector if this fails.
 */

# define DirectOperationFinish(sel)\
	if (TRUE) {\
		if (!isIntegerValue(r1)) \
			DirectRARecover(sel);\
		DirectRAEnd(integerObjectOf(r1));\
	}\
	else


# define DirectOperationFinishWithOverflow(sel, label)\
	if (TRUE) {\
		asm("	jvs	label");/* for overflow, needed for * and << */\
		if (!isIntegerValue(r1)) {\
			asm("label:");\
			DirectRARecover(sel);\
		}\
		DirectRAEnd(integerObjectOf(r1));\
	}\
	else


# define DirectRelationFinish(pred) 	/* relation primitive epilogue */\
	if (pred) \
		DirectRAEnd((oop_t)trueOop);\
	else if (TRUE)\
		DirectRAEnd((oop_t)falseOop);\
	else

/*
 *	returns
 */

/*
 *	The context reclamation scheme relies on
 *	detecting references to method contexts that might outlive
 *	the referent and marking all such contexts.
 *	One way to create such a reference is to return an OOP
 *	for a context.
 *
 *	This code checks to see if the returned value
 *	is a context.  It could be optimized by adding a flag
 *	to every object indicating whether it's a context.
 *
 *	If a BlockContext is returned, its home must also be
 *	retained.
 */

# define returnIt(res, newcon) \
	if (TRUE) {\
		r1 = (oop_t)(res);\
		method = (rp_t)r1;\
		if (!isIntegerObject(r1)  &&  flagsRp(method) & CONTEXT)\
			ggcStore(method, r1);\
		\
		method = (rp_t)ac;  /* steal method reg */\
		ac = (oop_t)newcon; /* mustn't change ac or hc before newcon */\
		assert(FutureSurvivors.sp_bottom > ac \
		    ||  ac >= FutureSurvivors.sp_top,\
		    "ac in wrong space");\
		if ((rp_t)ac == nilOop) {\
			ac = (oop_t)method;\
			method = methodMC(hc);\
			*sp++ = rcvr;\
			*sp++ = r1;\
			doASend(cannotReturnSelectorOop, 1);\
		}\
		\
		ipC(method) = (oop_t)nilOop; /* to avoid reentry */\
		callerBC(method) = nilOop;\
		\
		if (isBlock((rp_t)ac)) {\
			fetchLocals((rp_t)ac);\
			*sp++ = r1;\
		}\
		else {\
			sp = fetchSP((rp_t)ac);\
			*sp++ = r1;\
			storeSP((rp_t)ac, sp);\
			return ((rp_t)ac);\
		}\
		break;\
	}\
	else

/*
 *	faster version without GGC context check
 */

# define returnNonCon(res) \
	if (TRUE) {\
		r1 = (oop_t)(res);\
		method = (rp_t)ac;\
		ac = (oop_t)senderMC(hc);\
		assert(FutureSurvivors.sp_bottom > ac \
		    ||  ac >= FutureSurvivors.sp_top,\
		    "ac in wrong space");\
		if (ac == (oop_t)nilOop) {\
			ac = (oop_t)method;\
			method = methodMC(hc);\
			*sp++ = rcvr;\
			*sp++ = r1;\
			doASend(cannotReturnSelectorOop, 1);\
		}\
		ipC(method) = (oop_t)nilOop;\
		callerBC(method) = nilOop;\
		if (isBlock((rp_t)ac)) {\
			fetchLocals((rp_t)ac);\
			*sp++ = r1;\
		}\
		else {\
			sp = fetchSP((rp_t)ac);\
			*sp++ = r1;\
			storeSP((rp_t)ac, sp);\
			return ((rp_t)ac);\
		}\
		break;\
	}\
	else
