static char rcsId[] = "$Header: execMethod.c,v 43.1 84/04/05 23:51:35 dmu Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	This file contains the bytecode interpreter for method
 *	contexts.  This is the most important interpreter in BS,
 *	and has been optimized.  See executeMethod.h, for the macros
 *	that are at the heart of this code.
 */

# include	<stdio.h>

# include	"utils.h"
# include	"mem.h"
# include	"mc.h"
# include	"prims.h"
# include	"execute.h"
# include	"execMethod.h"
# include	"primTable.h"


rp_t
executeMethod(ac)
register rp_t ac;
{
	/*
	 *	active context cache
	 */

	register byte_t *pc;		/* addr of next bytecode */
	register oop_t	*sp;		/* push = *sp++, pop = *--sp */
	register rp_t	method;

	register oop_t	rcvr;		/* put receiver in a data register */
	register oop_t	r1;		/* scratch */
	register oop_t	r2;		/* scratch */
	register oop_t	fastSendSel;	/* selector for fast sends */
	/* plus one for ggcStore */


	fetchLocals(ac);
	for (;;) {
		assert(!isBlock(ac), "executeMethod: ac is a block");
		if (!checkBreakpoint()  ||  !checkDbTrace(pc, sp)) {
			storeLocals(ac);
			return (ac);
		}
#		ifdef CYCLE
			++cycle;
#		endif
		asm("# SWITCH");	/* for awk script to speed up switch */
		switch (*pc++) {
		/*
		 *	They following are macros defined in 
		 *	executeMethod.h.  Each of them will be
		 *	expanded to one case of the switch.  
		 *
		 *	They implement the stack Bytecodes.  
		 *	Refer to their definitions (in executeMethod.h)
		 *	for details.
		 */

		pushRcvrVar(0, 0, _MpushRcvrVar0);
		pushRcvrVar(1, 1, _MpushRcvrVar1);
		pushRcvrVar(2, 2, _MpushRcvrVar2);
		pushRcvrVar(3, 3, _MpushRcvrVar3);
		pushRcvrVar(4, 4, _MpushRcvrVar4);
		pushRcvrVar(5, 5, _MpushRcvrVar5);
		pushRcvrVar(6, 6, _MpushRcvrVar6);
		pushRcvrVar(7, 7, _MpushRcvrVar7);
		pushRcvrVar(8, 8, _MpushRcvrVar8);
		pushRcvrVar(9, 9, _MpushRcvrVar9);
		pushRcvrVar(10, 10, _MpushRcvrVar10);
		pushRcvrVar(11, 11, _MpushRcvrVar11);
		pushRcvrVar(12, 12, _MpushRcvrVar12);
		pushRcvrVar(13, 13, _MpushRcvrVar13);
		pushRcvrVar(14, 14, _MpushRcvrVar14);
		pushRcvrVar(15, 15, _MpushRcvrVar15);

		pushTempVar(16 + 0,  0,  _MpushTempVar0);
		pushTempVar(16 + 1,  1,  _MpushTempVar1);
		pushTempVar(16 + 2,  2,  _MpushTempVar2);
		pushTempVar(16 + 3,  3,  _MpushTempVar3);
		pushTempVar(16 + 4,  4,  _MpushTempVar4);
		pushTempVar(16 + 5,  5,  _MpushTempVar5);
		pushTempVar(16 + 6,  6,  _MpushTempVar6);
		pushTempVar(16 + 7,  7,  _MpushTempVar7);
		pushTempVar(16 + 8,  8,  _MpushTempVar8);
		pushTempVar(16 + 9,  9,  _MpushTempVar9);
		pushTempVar(16 + 10,  10,  _MpushTempVar10);
		pushTempVar(16 + 11,  11,  _MpushTempVar11);
		pushTempVar(16 + 12,  12,  _MpushTempVar12);
		pushTempVar(16 + 13,  13,  _MpushTempVar13);
		pushTempVar(16 + 14,  14,  _MpushTempVar14);
		pushTempVar(16 + 15,  15,  _MpushTempVar15);

		pushLitConst(32 + 0,  0,  _MpushLitConst0);
		pushLitConst(32 + 1,  1,  _MpushLitConst1);
		pushLitConst(32 + 2,  2,  _MpushLitConst2);
		pushLitConst(32 + 3,  3,  _MpushLitConst3);
		pushLitConst(32 + 4,  4,  _MpushLitConst4);
		pushLitConst(32 + 5,  5,  _MpushLitConst5);
		pushLitConst(32 + 6,  6,  _MpushLitConst6);
		pushLitConst(32 + 7,  7,  _MpushLitConst7);
		pushLitConst(32 + 8,  8,  _MpushLitConst8);
		pushLitConst(32 + 9,  9,  _MpushLitConst9);
		pushLitConst(32 + 10,  10,  _MpushLitConst10);
		pushLitConst(32 + 11,  11,  _MpushLitConst11);
		pushLitConst(32 + 12,  12,  _MpushLitConst12);
		pushLitConst(32 + 13,  13,  _MpushLitConst13);
		pushLitConst(32 + 14,  14,  _MpushLitConst14);
		pushLitConst(32 + 15,  15,  _MpushLitConst15);
		pushLitConst(32 + 16,  16,  _MpushLitConst16);
		pushLitConst(32 + 17,  17,  _MpushLitConst17);
		pushLitConst(32 + 18,  18,  _MpushLitConst18);
		pushLitConst(32 + 19,  19,  _MpushLitConst19);
		pushLitConst(32 + 20,  20,  _MpushLitConst20);
		pushLitConst(32 + 21,  21,  _MpushLitConst21);
		pushLitConst(32 + 22,  22,  _MpushLitConst22);
		pushLitConst(32 + 23,  23,  _MpushLitConst23);
		pushLitConst(32 + 24,  24,  _MpushLitConst24);
		pushLitConst(32 + 25,  25,  _MpushLitConst25);
		pushLitConst(32 + 26,  26,  _MpushLitConst26);
		pushLitConst(32 + 27,  27,  _MpushLitConst27);
		pushLitConst(32 + 28,  28,  _MpushLitConst28);
		pushLitConst(32 + 29,  29,  _MpushLitConst29);
		pushLitConst(32 + 30,  30,  _MpushLitConst30);
		pushLitConst(32 + 31,  31,  _MpushLitConst31);

		pushLitVar(0 + 64,  0, _MpushLitVar0);
		pushLitVar(1 + 64,  1, _MpushLitVar1);
		pushLitVar(2 + 64,  2, _MpushLitVar2);
		pushLitVar(3 + 64,  3, _MpushLitVar3);
		pushLitVar(4 + 64,  4, _MpushLitVar4);
		pushLitVar(5 + 64,  5, _MpushLitVar5);
		pushLitVar(6 + 64,  6, _MpushLitVar6);
		pushLitVar(7 + 64,  7, _MpushLitVar7);
		pushLitVar(8 + 64,  8, _MpushLitVar8);
		pushLitVar(9 + 64,  9, _MpushLitVar9);
		pushLitVar(10 + 64,  10, _MpushLitVar10);
		pushLitVar(11 + 64,  11, _MpushLitVar11);
		pushLitVar(12 + 64,  12, _MpushLitVar12);
		pushLitVar(13 + 64,  13, _MpushLitVar13);
		pushLitVar(14 + 64,  14, _MpushLitVar14);
		pushLitVar(15 + 64,  15, _MpushLitVar15);
		pushLitVar(16 + 64,  16, _MpushLitVar16);
		pushLitVar(17 + 64,  17, _MpushLitVar17);
		pushLitVar(18 + 64,  18, _MpushLitVar18);
		pushLitVar(19 + 64,  19, _MpushLitVar19);
		pushLitVar(20 + 64,  20, _MpushLitVar20);
		pushLitVar(21 + 64,  21, _MpushLitVar21);
		pushLitVar(22 + 64,  22, _MpushLitVar22);
		pushLitVar(23 + 64,  23, _MpushLitVar23);
		pushLitVar(24 + 64,  24, _MpushLitVar24);
		pushLitVar(25 + 64,  25, _MpushLitVar25);
		pushLitVar(26 + 64,  26, _MpushLitVar26);
		pushLitVar(27 + 64,  27, _MpushLitVar27);
		pushLitVar(28 + 64,  28, _MpushLitVar28);
		pushLitVar(29 + 64,  29, _MpushLitVar29);
		pushLitVar(30 + 64,  30, _MpushLitVar30);
		pushLitVar(31 + 64,  31, _MpushLitVar31);


		pushConstObj(113, (oop_t)trueOop, _MpushTrue);
		pushConstObj(114, (oop_t)falseOop, _MpushFalse);
		pushConstObj(115, (oop_t)nilOop, _MpushNil);
		pushConstObj(116, integerObjectOf(-1), _MpushMinOne);
		pushConstObj(117, integerObjectOf(0), _MpushZero);
		pushConstObj(118, integerObjectOf(1), _MpushOne);
		pushConstObj(119, integerObjectOf(2), _MpushTwo);

		case 112:	/* push receiver */
			adb(_MpushRcvr);
			*sp++ = (oop_t)rcvr;
			break;

		case 136: 	/* dup stack top */
			adb(_MdupStackTop);
			*sp = sp[-1];
			++sp;
			break;

		case 137:	 /* push active context */
			adb(_MpushActiveContext);
			*sp++ = (oop_t)ac;
			break;

		case 128:	/* extended push */
			adb(_MextendedPush);
			r1 = *pc++;
			switch (r1 >> 6  &  0x3) { /* extracts bits 8 and 9 */
			case 0:	/* receiver var */
				*sp++ = oopFieldRp((rp_t)rcvr)[r1 & 0x3f];
				break;
			case 1: /* temporary var */
				*sp++ = tempFrameC(ac)[r1 & 0x3f];
				break;
			case 2: /* literal const */
				*sp++ = literalFrameCM(method)[r1 & 0x3f];
				break;
			case 3:	/* literal variable */
				*sp++ = valueAs((rp_t)
				    literalFrameCM(method)[r1 & 0x3f]);
				break;
			default:
				fatal("illegal extended push", NullOop);
				break;
			}
			break;

		/*
		 *	end of pushes
		 */

		/*
		 *	pops
		 *
		 *	Stores an object pointer found on the stack 
		 *	into the object memory or the temporary frame.
		 */

		storePopRcvrVar(96 + 0, 0, _MstorePopRcvrVar0);
		storePopRcvrVar(96 + 1, 1, _MstorePopRcvrVar1);
		storePopRcvrVar(96 + 2, 2, _MstorePopRcvrVar2);
		storePopRcvrVar(96 + 3, 3, _MstorePopRcvrVar3);
		storePopRcvrVar(96 + 4, 4, _MstorePopRcvrVar4);
		storePopRcvrVar(96 + 5, 5, _MstorePopRcvrVar5);
		storePopRcvrVar(96 + 6, 6, _MstorePopRcvrVar6);
		storePopRcvrVar(96 + 7, 7, _MstorePopRcvrVar7);

		storePopTempVar(104 + 0, 0, _MstorePopTempVar0);
		storePopTempVar(104 + 1, 1, _MstorePopTempVar1);
		storePopTempVar(104 + 2, 2, _MstorePopTempVar2);
		storePopTempVar(104 + 3, 3, _MstorePopTempVar3);
		storePopTempVar(104 + 4, 4, _MstorePopTempVar4);
		storePopTempVar(104 + 5, 5, _MstorePopTempVar5);
		storePopTempVar(104 + 6, 6, _MstorePopTempVar6);
		storePopTempVar(104 + 7, 7, _MstorePopTempVar7);

		case 135:	/* pop stack top and discard */
			adb(_MpopStacktop);
			--sp;
			break;

		case 130:	/* extended store pop */
			adb(_MextendedStorePop);
			r2 = *pc++;
			switch (r2 >> 6  &  0x3) {  /* extracts bits 8 & 9 */
			case 0:	/* receiver var */
				r1 = *--sp;
				oopFieldRp((rp_t)rcvr)[r2 & 0x3f] = r1;
				ggcStore(rcvr, r1);
				break;
			case 1: /* temp var */
				tempFrameC(ac)[r2 & 0x3f] = *--sp;
				break;

			/*
			 *	case 2: is illegal (by definition)
			 */

			case 3:	/* literal var */
				r1 = literalFrameCM(method)[r2 & 0x3f];
				r2 = *--sp;
				valueAs((rp_t)r1) = r2;
				ggcStore(r1, r2);
				break;
			default:
				fatal("illegal extended store and pop", NullOop);
				break;
			}
			break;

		/*
		 *	end of pops
		 */

		/*
		 *	Extended store.
		 */

		case 129:
			adb(_MextendedStore);
			r2 = *pc++;
			switch (r2 >> 6  &  0x3) { /* extracts bits 8 & 9 */
			default:
				fatal("illegal bytecode in MC", (oop_t)ac);
			case 0:	/* receiver var */ 
				r1 = sp[-1];
				oopFieldRp((rp_t)rcvr)[r2 & 0x3f] =  r1;
				ggcStore(rcvr, r1);
				break;
			case 1: /* temp var */
				tempFrameC(ac)[r2 & 0x3f] = sp[-1];
				break;
			
			/*
			 *	case 2 is illegal (by definition).
			 */

			case 3:	/* literal var */
				r1 = literalFrameCM(method)[r2 & 0x3f];
				r2 = sp[-1];
				valueAs((rp_t)r1) = r2;
				ggcStore(r1, r2);
				break;
			}
			break;
		/*
		 *	end of stores
		 */

		/*
		 *	jumps
		 *
		 *	These are all macros defined in executeMethod.h.
		 *	Each macro call will be expanded into a case,
		 *	and executes the unconditional/condition jump
		 *	bytecodes.
		 */

		jump(144, 1, _Mjump1); jump(145, 2, _Mjump2);
		jump(146, 3, _Mjump3); jump(147, 4, _Mjump4);
		jump(148, 5, _Mjump5); jump(149, 6, _Mjump6);
		jump(150, 7, _Mjump7); jump(151, 8, _Mjump8);

		popJump(152, 1, falseOop, trueOop, _MpopJump1);
		popJump(153, 2, falseOop, trueOop, _MpopJump2);
		popJump(154, 3, falseOop, trueOop, _MpopJump3);
		popJump(155, 4, falseOop, trueOop, _MpopJump4);
		popJump(156, 5, falseOop, trueOop, _MpopJump5);
		popJump(157, 6, falseOop, trueOop, _MpopJump6);
		popJump(158, 7, falseOop, trueOop, _MpopJump7);
		popJump(159, 8, falseOop, trueOop, _MpopJump8);

		extendedJump(160, 0, _MextendedJump0);
		extendedJump(161, 1, _MextendedJump1);
		extendedJump(162, 2, _MextendedJump2);
		extendedJump(163, 3, _MextendedJump3);
		extendedJump(164, 4, _MextendedJump4);
		extendedJump(165, 5, _MextendedJump5);
		extendedJump(166, 6, _MextendedJump6);
		extendedJump(167, 7, _MextendedJump7);

		extendedPopJump(168, 0, trueOop, falseOop, _MextendedPopJumpT0);
		extendedPopJump(169, 1, trueOop, falseOop, _MextendedPopJumpT1);
		extendedPopJump(170, 2, trueOop, falseOop, _MextendedPopJumpT2);
		extendedPopJump(171, 3, trueOop, falseOop, _MextendedPopJumpT3);

		extendedPopJump(172, 0, falseOop, trueOop, _MextendedPopJumpF0);
		extendedPopJump(173, 1, falseOop, trueOop, _MextendedPopJumpF1);
		extendedPopJump(174, 2, falseOop, trueOop, _MextendedPopJumpF2);
		extendedPopJump(175, 3, falseOop, trueOop, _MextendedPopJumpF3);

		/*
		 *	end of jumps
		 */

		/*
		 *	sends
		 *
		 *	Each macro call will be expanded to one case of 
		 *	the switch statement.  Within each case, control
		 *	will be passed to the label "fastSend*Args"
		 *	after storing the selector in fastSendSel.
		 *	sendLitSel() uses the number of arguments to 
		 *	determine which label to go to.
		 */


		/*
		 *	Zero argument (bytecodes 208 to 223)
		 */

		sendLitSel(208 + 0,  0,  0); sendLitSel(208 + 1,  1,  0);
		sendLitSel(208 + 2,  2,  0); sendLitSel(208 + 3,  3,  0);
		sendLitSel(208 + 4,  4,  0); sendLitSel(208 + 5,  5,  0);
		sendLitSel(208 + 6,  6,  0); sendLitSel(208 + 7,  7,  0);
		sendLitSel(208 + 8,  8,  0); sendLitSel(208 + 9,  9,  0);
		sendLitSel(208 + 10,  10,  0); sendLitSel(208 + 11,  11,  0);
		sendLitSel(208 + 12,  12,  0); sendLitSel(208 + 13,  13,  0);
		sendLitSel(208 + 14,  14,  0); sendLitSel(208 + 15,  15,  0);

	fastSendZeroArgs:
		adb(_MfastSendZeroArgs);
		fastSend(0);


		/*
		 *	One argument (bytecodes 224 to 239).
		 */

		sendLitSel(224 + 0,  0,  1); sendLitSel(224 + 1,  1,  1);
		sendLitSel(224 + 2,  2,  1); sendLitSel(224 + 3,  3,  1);
		sendLitSel(224 + 4,  4,  1); sendLitSel(224 + 5,  5,  1);
		sendLitSel(224 + 6,  6,  1); sendLitSel(224 + 7,  7,  1);
		sendLitSel(224 + 8,  8,  1); sendLitSel(224 + 9,  9,  1);
		sendLitSel(224 + 10,  10,  1); sendLitSel(224 + 11,  11,  1);
		sendLitSel(224 + 12,  12,  1); sendLitSel(224 + 13,  13,  1);
		sendLitSel(224 + 14,  14,  1); sendLitSel(224 + 15,  15,  1);

	fastSendOneArg:
		adb(_MfastSendOneArg);
		fastSend(1);


		/*
		 *	Two arguments (bytecodes 240 to 255).
		 */

		sendLitSel(240 + 0,  0,  2); sendLitSel(240 + 1,  1,  2);
		sendLitSel(240 + 2,  2,  2); sendLitSel(240 + 3,  3,  2);
		sendLitSel(240 + 4,  4,  2); sendLitSel(240 + 5,  5,  2);
		sendLitSel(240 + 6,  6,  2); sendLitSel(240 + 7,  7,  2);
		sendLitSel(240 + 8,  8,  2); sendLitSel(240 + 9,  9,  2);
		sendLitSel(240 + 10,  10,  2); sendLitSel(240 + 11,  11,  2);
		sendLitSel(240 + 12,  12,  2); sendLitSel(240 + 13,  13,  2);
		sendLitSel(240 + 14,  14,  2); sendLitSel(240 + 15,  15,  2);

	fastSendTwoArgs:
		adb(_MfastSendTwoArgs);
		fastSend(2);


		/* 
		 *	Send lit sel #kkkkk with jjj args.  
		 *	Use fastSend for speed if the number of arguments 
		 *	(r1 >> 5) is less than 3.
		 */

		case 131:	
			adb(_MsendLitSeljjjkkkkk);
			r1 = *pc++;
			fastSendSel = literalFrameCM(method)[r1 & 0x1f];
			switch (r1 >> 5) {
			case 0:
				goto fastSendZeroArgs;
			case 1:
				goto fastSendOneArg;
			case 2:
				goto fastSendTwoArgs;
			default:
				doASend(fastSendSel, r1 >> 5);
			}
			break;


		case 133:	/* Send lit sel #kkkkk to super with jjj args */
			adb(_MsendLitSelSupjjjkkkkk);
			r1 = *pc++;
			doASuperSend(literalFrameCM(method)[r1 & 0x1f],
			    r1 >> 5);


		/* 
		 *	Send lit sel #k8 with j8 args.  If number of arguments
		 *	is less than 3, use fastSend for speed.
		 */

		case 132:		
			adb(_MsendLitSelj8k8);
			r1 = *pc++;
			fastSendSel = literalFrameCM(method)[*pc++];
			switch (r1) {
			case 0:
				goto fastSendZeroArgs;
			case 1:
				goto fastSendOneArg;
			case 2:
				goto fastSendTwoArgs;
			default:
				doASend(fastSendSel, r1);
			}
			break;


		case 134:	/* send lit sel #k8 to super with j8 args */
			adb(_MsendLitSelSupj8k8);
			r2 = *pc++;
			r1 = *pc++;
			doASuperSend(literalFrameCM(method)[r1], r2);
			break;

		/*
		 *	Here are the direct primitives.  They are also 
		 *	implemented in *Prims.c.  The code is not factored 
		 *	because there are slight differences.  This is 
		 *	speed-critical code.  Since there is no lookup 
		 *	involved, the direct primitives must check the 
		 *	class of the receiver.
		 *
		 *	The "magic" number in DirectRATest(), 
		 *	DirectOperationFinish(), ... is the offset of 
		 *	the selector in the object pointed to by
		 *	SpecialSelectorsOop.  It is calculated from
		 *	(bytecode - 176) * 2.
		 */
	
		case 176:	/* send + */
			adb(_MsendPlus);
			DirectRATest(0);
			DirectRAConvert;
			r1 += r2;
			DirectOperationFinish(0);
	
		case 177:	/* send - */
			adb(_MsendMinus);
			DirectRATest(2);
			DirectRAConvert;
			r1 -= r2;
			DirectOperationFinish(2);
	
		case 178:	/* send < */
			adb(_MsendLt);
			DirectRATest(4);
			DirectRAConvert;
			DirectRelationFinish(r1 < r2);
	
		case 179:	/* send > */
			adb(_MsendGt);
			DirectRATest(6);
			DirectRAConvert;
			DirectRelationFinish(r1 > r2);
	
		case 180:	/* send <= */
			adb(_MsendLe);
			DirectRATest(8);
			DirectRAConvert;
			DirectRelationFinish(r1 <= r2);
	
		case 181:	/* send >= */
			adb(_MsendGe);
			DirectRATest(10);
			DirectRAConvert;
			DirectRelationFinish(r1 >= r2);
	
		case 182:	/* send = */
			adb(_MsendEq);
			DirectRATest(12);
			DirectRelationFinish(r1 == r2);
	
		case 183:	/* send ~= */
			adb(_MsendNeq);
			DirectRATest(14);
			DirectRelationFinish(r1 != r2);
	
		case 184:	/* send *, -2**31 <= r1,r2 < 2**31 must hold */
			adb(_MsendTimes);
			DirectRATest(16);
			DirectRAConvert;
			if (r1 >= 0x7fff  ||  r1 < -0x7fff
			    ||  r2 >= 0x7fff  ||  r2 < -0x7fff)
				DirectRARecover(16);
			r1 = (short)r1 * (short)r2;
			DirectOperationFinish(16);
	
		case 185:	/* send /,  result must be exact */
			adb(_MsendSlash);
			DirectRATest(18);
			DirectRAConvert;
			if (r2 == 0  ||  r1 % r2  !=  0)
				DirectRARecover(18);
			r1 /= r2;
			DirectOperationFinish(18);
	
		/*
		 *	send \\ (modulus), quotient always rounded 
		 *	down towards - infinity.  See SIModPrm() in 
		 *	SIPrims.c about the funny % business.
		 */

		case 186:
			adb(_MsendMod);
			DirectRATest(20);
			DirectRAConvert;
			if (r2 == 0)
				DirectRARecover(20);
			if (r1 >= 0  &&  r2 > 0)
				r1 %= r2;
			else if (r1 >= 0) {
				r2 = -r2;
				r1 %= r2; 	/* but 3 mod -4 is 3-4 = -1 */
				if (r1)
					r1 -= r2;
			}
			else if (r2 > 0) {
				r1 = -r1;
				r1 %= r2;	/* but -3 mod 4 is -3--4 is 1 */
				if (r1)
					r1 = r2 - r1;
			}
			else {
				r1 = -r1;
				r2 = -r2;
				r1 %= r2;	/* -3 mod -4 is -3 - 0 is -3 */
				if (r1)
					r1 = -r1;
			}
			DirectOperationFinish(20);
	
	
		/*
		 *	This one turns out to be expecially critical,
		 *	that's why I don't call instantiate object,
		 *	but instead do it here.  The real pointer
		 *	"method" (only register rp_t in this block)
		 *	is used as a scratch pad for making the point 
		 *	object.  
		 */

		case 187:	/* @ (makePoint) */
			adb(_MsendAt);
			r1 = (oop_t)method;	/* swipe method */

			allocAnObj(&New, pointLength, method);
			classRp(method) = classPointOop;
			lengthRp(method) = pointLength;
			flagsRp(method) = 0;
			ageRp(method) = 0;
			formatRp(method) = OOPS;
			fixedFieldsRp(method) = pointLength;
			if (!isIntegerValue(NewHash))
				NewHash = lastRegXoop + 1;
			hashRp(method) = integerObjectOf(NewHash);

			yPt(method) = *--sp;
			xPt(method) = *--sp;
			*sp++ = (oop_t)method;	/* return point object */

			method = (rp_t)r1;	/* get it back */
			break;
	

		/*
		 *	Send bitShift:  Do a right shift if argument is 
		 *	negative.
		 */

		case 188:
			adb(_MsendLsh);
			DirectRATest(24);
			DirectRAConvert;
			if (r2 < 0)
				r1 >>= -r2;
			else
				r1 <<= r2;
			DirectOperationFinishWithOverflow(24, MAslOvfl);
	

		/*
		 *	send // (divide),
		 *	quotient always rounded down towards negative infinity
		 */

		case 189:
			adb(_MsendSlsl);
			DirectRATest(26);
			DirectRAConvert;
			if (r2 == 0)
				DirectRARecover(26);
			if ((r1 ^ r2) < 0  &&  r1 % r2)
				r1 = r1 / r2  -  1;
			else
				r1 /= r2;
			DirectOperationFinish(26);
	
		case 190:	/* bitAnd: */
			adb(_MsendAnd);
			DirectRATest(28);
			DirectRAEnd(r1 & r2);
			break;
	
		case 191:	/* bitOr: */
			adb(_MsendOr);
			DirectRATest(30);
			DirectRAEnd(r1 | r2);
			break;
	
		/*
		 *	at, atPut, and size handle Arrays here for speed.
		 *	I don't know if they should also handle strings.
		 *	Like in @ (makePoint), "method" is used as a 
		 *	scratch pad for the receiver object.
		 */
	
		case 192:	/* at: */
			adb(_MsendAtCol);
			r1 = (oop_t)method;	/* steal method */
			method = (rp_t)sp[-2];	/* get receiver object */
			r2 = sp[-1];		/* get index */
			if (!isIntegerObject((oop_t)method) 
			    &&  fixedFieldsRp(method) == 0  
			    &&  formatRp(method) == OOPS
			    &&  isIntegerObject(r2)) {
				r2 = nonNegIntegerValueOf(r2) - 1;
				if ((u_int)r2 < (u_int)lengthRp(method)) {
					sp[-2] = oopFieldRp(method)[r2];
					--sp;
					method = (rp_t)r1; /* restore method */
					break;	/* next bytecode */
				}
			}
			method = (rp_t)r1;	/* restore method */
			fastSendSel =  (oop_t)specialSelOf(32);
			goto fastSendOneArg;


		/*
		 *	Since both scratch registers (r1, r2) are used,
		 *	the stolen method has to be restored from the
		 *	active context.
		 */

		case 193:	/* at:put: */
			adb(_MsendAtPut);
			method = (rp_t)sp[-3];	/* steal method */
			if (!isIntegerObject((oop_t)method)  
			    &&  fixedFieldsRp(method) == 0
			    &&  formatRp(method) == OOPS) {
				r1 = sp[-2];	/* get index */
				if (isIntegerObject(r1)) {
					r1 = nonNegIntegerValueOf(r1) - 1;
					if ((u_int)r1
					    < (u_int)lengthRp(method)) {
						r2 = *--sp;
						oopFieldRp(method)[r1] = r2;
						sp--;
						sp[-1] = r2;
						ggcStore(method, r2);
						method = methodMC(ac);
						break;
					}
				}
			}
			method = methodMC(ac);
			fastSendSel =  (oop_t)specialSelOf(34);
			goto fastSendTwoArgs;

		/*
		 *	another cheat for speed
		 */

		case 194:	/* size */
			adb(_MsendSize);
			r1 = sp[-1];
			if (!isIntegerObject((oop_t)r1)
			    &&  fixedFieldsRp((rp_t)r1) == 0) {
				sp[-1] = integerObjectOf(lengthRp((rp_t)r1));
				break;
			}
			fastSendSel = (oop_t)specialSelOf(36);
			goto fastSendZeroArgs;


		/*
		 *	Stream primitives
		 */

		case 195:	/* next */
			adb(_MsendNext);
			fastSendSel = (oop_t)specialSelOf(38);
			goto fastSendZeroArgs;

		case 196:	/* nextPut: */
			adb(_MsendNextPut);
			fastSendSel = (oop_t)specialSelOf(40);
			goto fastSendOneArg;

		case 197:	/* atEnd */
			adb(_MsendAtEnd);
			fastSendSel = (oop_t)specialSelOf(42);
			goto fastSendZeroArgs;
	
		/*
		 *	end of stream primitives
		 */

		case 198:	/* send == */
			adb(_MsendEqual);
			r1 = *--sp;
			if (*--sp == r1)
				*sp++ = (oop_t)trueOop;
			else
				*sp++ = (oop_t)falseOop;
			break;
	
		case 199:	/* class */
			adb(_MsendClass);
			r1 = *--sp;
			if (isIntegerObject(r1))
				*sp++ = (oop_t)classSmallIntegerOop;
			else
				*sp++ = (oop_t)classRp((rp_t)r1);
			break;
	
		case 200:	/* blockCopy: */
			adb(_MsendBlockCopy);
			r1 = sp[-2];
			if (!isIntegerObject(r1)  
			    &&  flagsRp((rp_t)r1) & CONTEXT) { /* is context */
				storeLocals(ac);
				if (CBlockCopyPrm(sp - 2,  ac) != NullRp) {
					--sp;
					break;
				}
			}
			fastSendSel = (oop_t)specialSelOf(48);
			goto fastSendOneArg;
	
		case 201:	/* value with zero args */
			adb(_MsendValue0);
			r1 = *--sp;	/* steal method */
			storeLocals(ac);
			method = (rp_t)r1;
			if (isIntegerObject(r1)  
			    ||  classRp(method) != classBlockContextOop
			    ||  argCountBC(method)  !=  integerObjectOf(0)) {
				method = methodMC(ac);
				++sp;
				fastSendSel = (oop_t)specialSelOf(50);
				goto fastSendZeroArgs;
			}

			/*
			 *	all is well, now do it
			 */

			/*
			 *	maintain caller chain
			 */

			callerBC(method) = ac;
			ipC(method) = initIPBC(method);
			spC(method) = integerObjectOf(0);
			return (method);
	
		case 202:	/* value: with one argument */
			adb(_MsendValue1);
			sp -= 2;
			r1 = *sp;	/* steal method */
			storeLocals(ac);
			method = (rp_t)r1;
			if (isIntegerObject(r1)  
			    ||  classRp(method) != classBlockContextOop
			    ||  argCountBC(method)  !=  integerObjectOf(1)) {
				method = methodMC(ac);
				sp += 2;
				fastSendSel = (oop_t)specialSelOf(52);
				goto fastSendOneArg;
			}

			/*
			 *	all is well;  now do it.
			 */

			tempFrameC(method)[0] = sp[1];

			/*
			 *	maintain caller chain
			 */

			callerBC(method) = ac;
			ipC(method) = initIPBC(method);
			spC(method) = integerObjectOf(1);
			return (method);

		case 203:	/* do: */
			adb(_MsendDo);
			fastSendSel = (oop_t)specialSelOf(54);
			goto fastSendOneArg;

		case 204:	/* new */
			adb(_MsendNew);
			fastSendSel = (oop_t)specialSelOf(56);
			goto fastSendZeroArgs;

		case 205:	/* new: */
			adb(_MsendNewCol);
			fastSendSel = (oop_t)specialSelOf(58);
			goto fastSendOneArg;

		case 206:	/* x */
			adb(_Msendx);
			r1 = (oop_t)method; /* steal method */
			method = (rp_t)*--sp;
			if (isIntegerObject((oop_t)method)  
			    ||  classRp(method) != classPointOop) {
				sp++;
				method = (rp_t)r1;
				fastSendSel = (oop_t)specialSelOf(60);
				goto fastSendZeroArgs;
			}
			*sp++ = xPt(method);
			method = (rp_t)r1;
			break;

		case 207:	/* y */
			adb(_Msendy);
			r1 = (oop_t)method; /* steal method */
			method = (rp_t)*--sp;
			if (isIntegerObject((oop_t)method)  
			    ||  classRp(method) != classPointOop) {
				sp++;
				method = (rp_t)r1;
				fastSendSel = (oop_t)specialSelOf(62);
				goto fastSendZeroArgs;
			}
			*sp++ = yPt(method);
			method = (rp_t)r1;
			break;

		/*
		 *	end of sends
		 */

		/*
		 *	returns
		 */

		case 120: 	/* return self */
			adb(_MreturnSelf);
			returnIt(rcvr);

		case 121: 	/* return true */
			adb(_MreturnTrue);
			returnNonCon((oop_t)trueOop);

		case 122: 	/* return false */
			adb(_MreturnFalse);
			returnNonCon((oop_t)falseOop);

		case 123: 	/* return nil */
			adb(_MreturnNil);
			returnNonCon((oop_t)nilOop);

		case 124:	/* return stack top from message */
			adb(_MreturnTopMess);
			returnIt(*--sp);
		
		/*
		 *	This bytecode is legal only in a block context.
		 */

		case 125:	/* return top from block */
			adb(_MreturnTopBlock);
			fatal("cannot return from block in method", NullOop);

		/*
		 *	end of returns
		 */
		} /*	end of bytecode decode switch() */
	}	/*	end of fetch-execute for loop */
	/*NOTREACHED*/
	/*
	 *	lint won't shut up so we commented this out:
	 *	fatal("executeMethod: should never reach this point", NullOop);
	 */
}
