static char executeMethod_h[] = "$Header: execMethod.h,v 41.1 84/03/14 19:59:53 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Here are the macros that lie at the heart of the interpreter
 *	for method contexts.  This is the most important bytecode
 *	interpreter and has been optimized.
 */


/*
 *	The interpreter caches information from the active context in
 *	the local register variables.  These two macros fill and flush
 *	the cache.  StoreLocals() must be called before before doing
 *	a send.
 */

# define fetchLocals(ac)\
	if (TRUE) {\
		method = methodMC(ac);\
		pc = fetchPC(ac, method);\
		sp = fetchSP(ac);\
		rcvr = receiverMC(ac);\
	}\
	else

# define storeLocals(ac)\
	if (TRUE) {\
		storePC(ac, method, pc);\
		storeSP(ac, sp);\
	}\
	else


/*
 *	send selector sel with nargs arguments.
 *	Control does not return from here, it ``returns'' or ``breaks''.
 *	In many places where it is used some of the register variables
 *	(like ``method'') have already been corrupted (stolen).
 *	This is used as a last resort when there is no faster way to 
 *	send.
 */

# define doASend(sel, nargs)\
	if (TRUE) {\
		storeLocals(ac);\
		ac = methodSend(ac, (rp_t)(sel), (u_int)(nargs), FALSE);\
		assert(FutureSurvivors.sp_bottom > ac \
		    ||  ac >= FutureSurvivors.sp_top, "ac in wrong space");\
		fetchLocals(ac);\
		if (isIntegerObject(method)  ||  checkPoll())\
			return (ac);\
		break;\
	}\
	else


/*
 *	same thing but to superclass
 */

# define doASuperSend(sel, nargs)\
	if (TRUE) {\
		storeLocals(ac);\
		ac = methodSend(ac, (rp_t)(sel), (u_int)(nargs), TRUE);\
		assert(FutureSurvivors.sp_bottom > ac \
		    ||  ac >= FutureSurvivors.sp_top,  "ac in wrong space");\
		fetchLocals(ac);\
		if (isIntegerObject(method)  ||  checkPoll())\
			return (ac);\
		break;\
	}\
	else

/*
 *	The following macros implement individual bytecodes described on 
 * 	p596 on Goldberg and Robson.  See executeMethod.c for context.
 *	Having separate cases for each bytecode makes it faster.
 *	
 *	``x'' is the bytecode, ``y'' is the field index and ``lab'' is a label
 *	that will be assembled into the code (for debugging with adb).
 */


/*
 *	For bytecodes 0 to 15 (0000iiii).
 */

# define pushRcvrVar(x, y, lab) \
	case x: \
		adb(lab);\
		*sp++ = oopFieldRp((rp_t)rcvr)[y]; \
		break


/*
 *	For bytecodes 16 to 31 (0001iiii).
 */

# define pushTempVar(x, y, lab) \
	case x:\
		adb(lab);\
		*sp++ = tempFrameC(ac)[y];\
		break


/*
 *	For bytecodes 32 to 63 (001iiiii).
 */

# define pushLitConst(x, y, lab) \
	case x:\
		adb(lab);\
		*sp++ = literalFrameCM(method)[y];\
		break


/*
 *	Executes the push literal variable bytecodes (64 to 95).  
 */

# define pushLitVar(x, y, lab) \
	case x:\
		adb(lab);\
		*sp++ = valueAs((rp_t)literalFrameCM(method)[y]);\
		break


/*
 *	For bytecodes 112 to 119 (01110iii).  
 */

# define pushConstObj(x, y, lab) \
	case x:\
		adb(lab);\
		*sp++ = y;\
		break

/*
 *	stores - stores an object pointer found on the stack to the 
 *	object memory.  ggcStore() is the Generation Scavenge store check.
 */

# define storePopRcvrVar(x, y, lab) \
	case x:\
		adb(lab);\
		r1 = *--sp;\
		oopFieldRp((rp_t)rcvr)[y] = r1;\
		ggcStore(rcvr, r1);\
		break


/*
 *	For bytecodes 104 to 111.  Function is similar to storePopRcvrVar().
 *	except it stores the popped value into the temporary frame.
 */
 
# define storePopTempVar(x, y, lab) \
	case x:\
		adb(lab);\
		tempFrameC(ac)[y] = *--sp;\
		break


# define storePopLitVar(x, y, lab) \
	case x:\
		adb(lab);\
		r1 = literalFrameCM(method)[y];\
		r2 = *--sp;\
		valueAs((rp_t)r1) = r2;\
		ggcStore(r1, r2);\
		break

/*
 *	jumps
 *
 *	Executes the Jump bytecodes 144 to 151.
 */


# define jump(x, y, lab) \
	case x: \
		adb(lab);\
		pc += y;\
		break


/*
 *	Conditional jump.  Jump if TOS == yes, fall through if TOS == no.
 *	Send error message otherwise.
 */

# define popJump(x, y, yes, no, lab) \
	case x: \
		adb(lab);\
		r1 = *--sp;\
		if (r1 == (oop_t)yes)\
			pc += y;\
		else if (r1 != (oop_t)no) {\
			--pc;\
			*sp++ = r1;\
			doASend(mustBeBooleanSelectorOop, 0);\
		}\
		break


/*
 *	Extended unconditional jump.  Sends and backward branches must 
 *	check for input events and semaphore signals.
 */

# define extendedJump(x, y, lab) \
	case x:\
		adb(lab);\
		pc += *pc + 1 + (y - 4  <<  8);\
		if (checkPoll()) {\
			storeLocals(ac);\
			return (ac);\
		}\
		break


/*
 *	extended condtional jump.  Jump iff TOS == yes, fall through for no.
 */

# define extendedPopJump(x, y, yes, no, lab) \
	case x:\
		adb(lab);\
		r1 = *--sp;\
		if (r1 == (oop_t)yes)\
			pc += y * 256  +  *pc;\
		else if (r1 != (oop_t)no) { \
			pc -= 1;\
			*sp++ = r1;\
			doASend(mustBeBooleanSelectorOop, 0);\
		}\
		++pc;\
	break

/*
 *	sends
 */

/*
 *	Optimized send operation.  
 *	"sel" is the message selector to be sent and "nargs" is the
 *	number of arguments.
 *	Both nargs and tempcount are limited to commmon cases for speed.
 *	This serves as a prelude to the fastSend macro.  It finds the 
 *	message selector and stores it in fastSendSel, then goes to 
 *	the appropriate instantiation of fastSend (based on nargs).
 */

# define sendLitSel(x, sel, nargs)\
	case x:\
		fastSendSel = literalFrameCM(method)[sel];\
		if (nargs == 0)\
			goto fastSendZeroArgs;\
		else if (nargs == 1)\
			goto fastSendOneArg;\
		else if (nargs == 2)\
			goto fastSendTwoArgs;\
		else\
			fatal("fast send with too many args", (oop_t)ac);\
		break


/*
 *	define c1 as a shorthand for the pointer to 
 *	a slot in the method cache.
 */

# define c1	((struct methodCache *)method)	/* stealing method! */


/*
 *	fastSend is the macro that actually does the send.  It is
 *	instantiated once for each possible number of arguments,
 *	so that the loop can be unrolled.  sendLitSel dispatches
 *	to the right one for the send bytecodes.
 *	fastSend expects the selector in fastSendSel.
 *	If something cannot be handled here, back off with doASend.
 *
 *	For speed, lookupMethodInClass() will only be called if
 *	cacheSlot() fails.
 *
 *	There are two reasons why this is convoluted:
 *		two kinds of short sends (ret self, ret inst var)
 *		and two kinds of primitives (successes and failures)
 *		must be handled
 *
 *		common cases must be FAST (address registers are scarce)!
 */

# define fastSend(nargs)\
	if (TRUE) {\
		/* \
		 *	find method\
		 */\
		rcvr = sp[-1 - nargs];	/* swipe rcvr */\
		if (isIntegerObject(rcvr)) {\
			r1 = (oop_t)classSmallIntegerOop;\
			method = (rp_t)cacheSlot(r1, fastSendSel);\
			    /* swipe method */\
			if ((c1->mc_class != (rp_t)r1\
			      ||  c1->mc_sel != (rp_t)fastSendSel)\
			    &&  !lookupMethodInClass((rp_t)r1, fastSendSel)) {\
				method = methodMC(ac);\
				doASend(fastSendSel, nargs);\
			}\
		}\
		else /* non-integer receiver */ {\
			r2 = (oop_t)ac;\
			ac = (rp_t)rcvr;	/* swipe ac! */\
			r1 = (oop_t)classRp(ac);\
			method = (rp_t)cacheSlot(r1, fastSendSel);\
			    /* swipe method */\
			if ((c1->mc_class != (rp_t)r1\
			        ||  c1->mc_sel != (rp_t)fastSendSel)\
			      &&  !lookupMethodInClass((rp_t)r1, fastSendSel)\
			    ||   flagsRp(ac) & SMC) { /* have to nil stack */\
				ac = (rp_t)r2; /* put it back */\
				method = methodMC(ac);\
				doASend(fastSendSel, nargs);\
			}\
			ac = (rp_t)r2;\
		} /* found method */\
		if (c1->mc_large ||  c1->mc_tempCount  >  4 + nargs) {\
			method = methodMC(ac);\
			doASend(fastSendSel, nargs);\
		}\
		/*\
		 *	The if test is written this way because\
		 *	full sends are the most common case and\
		 *	the send code is reused for primitive failure.\
		 */\
		if (c1->mc_whatKind != FULL_SEND) /* do oddball first */{\
			if (c1->mc_whatKind == PRIMITIVE) {\
				sp -= nargs;		/* assume success */\
				r2 = (oop_t)method;	/* save slot */\
				method = methodMC(ac);\
				storeLocals(ac);\
				method = (rp_t)r2;	/* restore slot */\
				r1 = (oop_t)\
				    (*primitiveDispatchTable[c1->mc_index])\
				    (&sp[-1], ac, nargs);\
				if ((rp_t)r1 == ac) /* easy success */ {\
					method = methodMC(ac);\
					rcvr = receiverMC(ac);\
					break;		/* fetch next bc */\
				}\
				else if (r1 != NullOop) /* harder success */{\
					ac = (rp_t)r1;\
					assert(FutureSurvivors.sp_bottom > ac \
					    ||  ac >= FutureSurvivors.sp_top,\
					    "ac in wrong space");\
					fetchLocals(ac);\
					if (isIntegerObject(method)\
					    || checkPoll())\
						return (ac);\
					break;	/* next bc */\
				}\
				/*\
				 *	failed:\
				 *	recover, fall through to full send\
				 */\
				sp += nargs;\
			}\
			else if (c1->mc_whatKind == RET_SELF) {\
				rcvr = receiverMC(ac);\
				method = methodMC(ac);\
				break; 		/* fetch next bc */\
			}\
			else {			/* ret inst var */\
				sp[-1] = oopFieldRp((rp_t)rcvr)[c1->mc_index];\
				rcvr = receiverMC(ac);\
				method = methodMC(ac);\
				break; 		/* fetch next bc */\
			}\
		}\
		else /*	did not already try a primitive */\
			storePC(ac, methodMC(ac), pc);\
		/*\
		 *	full send code, first get a new context,\
		 *	fill it in, and set up interpreter registers\
		 */\
		pc = (byte_t *)FirstFreeContext;/* swipe pc */\
		if (isIntegerObject((oop_t)pc)) \
			pc  = (byte_t *)instantiateClass(classMethodContextOop,\
			    smallContextLength - fixedFieldsC);\
		else\
			FirstFreeContext = nextSMC((rp_t)pc);\
		senderMC((rp_t)pc) = ac;\
		receiverMC((rp_t)pc) = rcvr;\
		\
		if (nargs == 0) \
			--sp;\
		else if (nargs == 1) {\
			tempFrameC((rp_t)pc)[0] = *--sp;\
			--sp;\
		}\
		else if (nargs == 2) {\
			tempFrameC((rp_t)pc)[1] = *--sp;\
			tempFrameC((rp_t)pc)[0] = *--sp;\
			--sp;\
		}\
		else\
			fatal("send macro: unknown arg count", NullOop);\
		\
		storeSP(ac, sp);\
		ac = (rp_t)pc;\
		assert(FutureSurvivors.sp_bottom > ac \
		    ||  ac >= FutureSurvivors.sp_top,  "ac in wrong space");\
		sp = &tempFrameC(ac)[nargs];\
		pc = c1->mc_pc;\
		if (c1->mc_tempCount == (short)nargs)\
			;\
		else if (c1->mc_tempCount  ==  (short)nargs + (short)1)\
			*sp++ = (oop_t)nilOop;\
		else if (c1->mc_tempCount  ==  (short)nargs + (short)2) {\
			r1 = (oop_t)nilOop;\
			*sp++ = r1;\
			*sp++ = r1;\
		}\
		else if (c1->mc_tempCount  ==  (short)nargs + (short)3) {\
			r1 = (oop_t)nilOop;\
			*sp++ = r1;\
			*sp++ = r1;\
			*sp++ = r1;\
		}\
		else if (c1->mc_tempCount  ==  (short)nargs + (short)4) {\
			r1 = (oop_t)nilOop;\
			*sp++ = r1;\
			*sp++ = r1;\
			*sp++ = r1;\
			*sp++ = r1;\
		}\
		else\
			fatal("bad temp count", (oop_t)method);\
		methodMC(ac) = method = c1->mc_method;\
		if (checkPoll()) {\
			storeLocals(ac);\
			return (ac);\
		}\
		break;\
	}\
	else

/*
 *	These macros are for the immediate primitives.
 *
 *	This one is called when a direct primitive fails.  Operands popped
 *	from the stack are recovered, and a message will be send to the 
 *	receiver instead of executing the direct primitive.
 *	(RA means register arithmetic.)
 */

# define DirectRARecover(sel)\
	if (TRUE) {\
		sp += 2;\
		doASend(specialSelOf(sel), specialNargOf(sel));\
	}\
	else


/*
 *	Called when the direct primitive succeeds.
 */

# define DirectRAEnd(ans)\
	if (TRUE) {\
		*sp++ = ans;\
		break;\
	}\
	else


/* 
 *	Prolog for direct binary primitives.  There are "similar" versions
 *	of these marcos in SIPrims.c.
 *
 *	DirectRATest() obtains the operands from the stack and makes sure
 *	that they are SmallIntegers. sel is the index needed to obtain 
 *	the selector if this fails.
 *	
 */

# define DirectRATest(sel) \
	if (TRUE) {\
		r2 = *--sp;\
		r1 = *--sp;\
		if (!isIntegerObject(r1 & r2))\
			DirectRARecover(sel);\
	}\
	else


# define DirectRAConvert \
	if (TRUE) {\
		convertToInteger(r1);\
		convertToInteger(r2);\
	}\
	else


/*
 *	Epilogue for SmallInteger arithmetic primitives.  sel is the index
 *	needed to obtain the selector if this fails.
 */

# define DirectOperationFinish(sel)\
	if (TRUE) {\
		if (!isIntegerValue(r1)) \
			DirectRARecover(sel);\
		DirectRAEnd(integerObjectOf(r1));\
	}\
	else


# define DirectOperationFinishWithOverflow(sel, label)\
	if (TRUE) {\
		asm("	jvs	label");/* for overflow, needed for * and << */\
		if (!isIntegerValue(r1)) {\
			asm("label:");\
			DirectRARecover(sel);\
		}\
		DirectRAEnd(integerObjectOf(r1));\
	}\
	else


/* 
 *	Relation primitive epilogue.
 */

# define DirectRelationFinish(pred) \
	if (pred) \
		DirectRAEnd((oop_t)trueOop);\
	else if (TRUE)\
		DirectRAEnd((oop_t)falseOop);\
	else

/*
 *	returns
 *
 *	BS reclaims most
 *	method contexts upon a return to reduce the load on
 *	the scavenger.  The context reclaimation scheme relies on
 *	detecting the creation of references to method contexts
 *	that might outlive the referent and marking it.
 *	One way to create such a reference is to return an OOP
 *	for a context.
 *
 *	If a BlockContext is returned, its home must also be
 *	retained.
 */

# define returnIt(res) \
	if (TRUE) {\
		r1 = (res);\
		if (!isIntegerObject(r1)) {\
			method = (rp_t)r1;	/* steal method */\
			if (flagsRp(method) & CONTEXT)\
				ggcStore(method, method);\
		}\
		if (flagsRp(ac) & SMC) /* reclaim context */ {\
			nextSMC(ac) = FirstFreeContext;\
			FirstFreeContext = ac;\
			ac = senderMC(ac);\
		}\
		else /* save context, nilling prevents reentry */ {\
			method = senderMC(ac); /* swipe it */\
			senderMC(ac) = nilOop;\
			ipC(ac) = (oop_t)nilOop;\
			ac = method;\
			flagsRp(ac) &= ~SMC;\
		}\
		assert(FutureSurvivors.sp_bottom > ac \
		    ||  ac >= FutureSurvivors.sp_top,  "ac in wrong space");\
		fetchLocals(ac);\
		*sp++ = r1;\
		if (isIntegerObject((oop_t)method)) /* it's a block! */ {\
			incrementSmallInteger(spC(ac));\
			return (ac);\
		}\
		break;\
	}\
	else


# define returnNonCon(res)	/* omit ggc check */\
	if (TRUE) {\
		r1 = (res);\
		if (flagsRp(ac) & SMC) {\
			nextSMC(ac) = FirstFreeContext;\
			FirstFreeContext = ac;\
			ac = senderMC(ac);\
		}\
		else {\
			method = senderMC(ac); /* swipe it */\
			senderMC(ac) = nilOop;\
			ipC(ac) = (oop_t)nilOop;\
			ac = method;\
			flagsRp(ac) &= ~SMC;\
		}\
		assert(FutureSurvivors.sp_bottom > ac \
		    ||  ac >= FutureSurvivors.sp_top, "ac in wrong space");\
		fetchLocals(ac);\
		*sp++ = r1;\
		if (isIntegerObject((oop_t)method)) {\
			incrementSmallInteger(spC(ac));\
			return (ac);\
		}\
		break;\
	}\
	else
