static char rcsId[] = "$Header: execute.c,v 44.5 84/07/26 00:26:42 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	This routine contains the top level routines for BS.
 *	There is also provision for tracing the stack for debugging.
 */

# include	<stdio.h>
# include	<sys/time.h>
# include	<signal.h>

# include	"utils.h"
# include	"mem.h"
# include	"prims.h"
# include	"execute.h"
# include	"io.h"


static boolean	halted = FALSE;
static char	*haltReason = "type c";

static u_int	breakpoint = ~0;	/* stop when cycle equals this */


u_int		pollCount;		/* used to schedule input polling */

static	FILE	*dbtrace;		/* for debugging */
static	u_int	msPerPoll;

static boolean	haltLoop();
void		serviceTimer();

/*
 *	This is the top level dispatch routine for BS.
 *	It decides which interpreter to run, and regains control
 *	when it's time to switch interpreters, halt, or poll.
 */

void
execute()
{
	rp_t	ac;			/* active context */
	struct itimerval	itimerval;

#	define defaultMsPerPoll	100
#	define usPerPoll	(msPerPoll * 1000)

	initTimeData();
	sigset(SIGALRM, serviceTimer);
	if (!getenv("NoPoll")) {

		/*
		 *	interval is reloaded into value
		 */

		msPerPoll =  getenv("MsPerPoll") ? atoi(getenv("MsPerPoll"))
		    : defaultMsPerPoll;
		itimerval.it_interval.tv_sec = 0;
		itimerval.it_interval.tv_usec = usPerPoll;
		itimerval.it_value.tv_sec = 0;
		itimerval.it_value.tv_usec = usPerPoll;
		(void)setitimer(ITIMER_REAL, &itimerval, 
		    (struct itimerval *)0);
	}

	ac = firstContext();
	assert(FutureSurvivors.sp_bottom > (oop_t *)ac 
	    ||  (oop_t *)ac >= FutureSurvivors.sp_top,
	    "ac in wrong space");


#	ifdef HALT
		halt("hello");
#	endif

#	ifdef DBTRACE
		dbtrace = xfopen("dbtrace", "r");
#       endif

	for (;;) {
		if (isBlock(ac))
			ac = executeBlock((oop_t)ac);
		else
			ac = executeMethod(ac);

		assert(FutureSurvivors.sp_bottom > (oop_t *)ac 
		    ||  (oop_t *)ac >= FutureSurvivors.sp_top,
		    "ac in wrong space");

		/*
		 *	at this point, all state is in ac
		 */

		if (halted  &&  haltLoop(haltReason, &ac))
			return;

		/*
		 *	The really expensive checks must not be done too
		 *	often.  See execute.h for polling stuff.
		 */

		if (isTimeToPoll()) {
			if (EncodedKeyboard)
				checkInputEncoded(&ac);
			else
				checkInputUnencoded(&ac);
			checkSignalAtTick(&ac);
			checkGenerationSpace(&ac);
			checkSpaceLimits(&ac);
			resetPollCount();
		}
	}
}


/*
 *	When BS is halted, read commands and do them.
 *	Return either to continue, or exit, according to return value.
 */

static boolean
haltLoop(msg, acp)
char msg[];
rp_t *acp;
{
	printf("%s\r\n", msg);

	CloseWindow(FALSE);
	for (;;) {
		char 	ci;		/* character input */

		/*
		 *	VM command loop.  Read a command from
		 *	the text terminal and do it.
		 */

		printf("\r\n");
#		ifdef BREAKPOINT
			printf("b ");
#		endif
		printf("c g q r s t T v: ");
		fflush(stdout);

		/*
		 *	EOF means quit 
		 *	can't use getchar with my version of filbuf
		 */

		if (strcmp(msg, QuitMsg) == 0  ||  read(0, &ci, 1) == 0)
			ci = 'q';

		switch(ci) {
#		ifdef BREAKPOINT
			case 'b':
				scanf("%d", &breakpoint);
				break;
#		endif

		case 'c':			/* continue */
			halted = FALSE;
			OpenWindow(NULL);	/* use previous argv */
			return (FALSE);
		 
		case 'g':			/* scavenge generations */
			traceGGCFor("Command");
			ggc(acp);
			break;
		 
		case 'q':
			adb(_quit);
			reportTimeData();
			traceReport();
			return (TRUE);

		case 'r':
			reorganize(acp);
			break;

		case 's':
			stackTrace(*acp);
			break;

		case 't':
			reportTimeData();
			break;

		case 'T':
			initTimeData();
			break;

		case 'v':
			verifyImage();
			break;

		default:
#			ifdef BREAKPOINT
				printf("b = breakpoint, ");
#			endif
			printf("c = continue, ");
			printf("g = generation scavenge, ");
			printf("q = quit, ");
			printf("r = reorganize, ");
			printf("s = stack trace, ");
			printf("t = report time data, ");
			printf("T = initialize time data, ");
			printf("v = verify image\n");
			break;
		}
	}
}


/*
 *	Called from other modules to halt machine for reason ``s.''
 *	Does not take effect until next send or backwards jump.
 */

void
halt(s)
char s[];
{
	halted = TRUE;
	haltReason = s;
	forcePoll();
}


/*
 *	Print out Smalltalk stack trace for active process.
 *	Idea is to find sending bytecode from calling context
 *	and use it to get the selector.
 *	The sender chain is chased back until it finds a nil 
 *	active context.  For each sender, the selector which 
 *	was used is printed, together with other information.
 *	Keyword ``super'' is printed for sending to super,
 *	and ``[]'' is printed for messages to a block.
 */

void
stackTrace(ac)
register rp_t ac;
{
	register oop_t	rcvr;
	register rp_t	method;
	register boolean	inABlock;
	char	rcbuf[BUFSIZ], selbuf[BUFSIZ];	/* names of rcvr and selector */
	byte_t	*pc;
	boolean	super;

	for (;;) {
		inABlock = isBlock(ac);
		rcvr = inABlock  ?  receiverMC(homeBC(ac))  :  receiverMC(ac);
		(void)asString(rcbuf, rcvr);

		ac = senderMC(ac);
		if (ac == nilOop)
			break;
		assert(FutureSurvivors.sp_bottom > (oop_t *)ac 
		    ||  (oop_t *)ac >= FutureSurvivors.sp_top, 
		    "ac in wrong space");

		method = isBlock(ac)  ?  methodMC(homeBC(ac))  :  methodMC(ac);

		pc = fetchPC(ac, method);
		super = FALSE;
		if (pc[-1] >= 0xd0) 	/* send literal selector */
			(void)asString(selbuf,
			    (oop_t)(literalFrameCM(method)[pc[-1] & 0xf]));
		else if (176 <= pc[-1]  &&  pc[-1] <= 207) /* send spec sel */
			(void)asString(selbuf,
			    (oop_t)(specialSelOf((pc[-1] - 176) * 2)));
		else 
			switch (pc[-2]) {
			/*
			 *	Send literal selector with less than
			 *	32 selectors.
			 */

			case 133:
				super = TRUE;
				/* FALL THROUGH */
			case 131:
				(void)asString(selbuf, 
				    (oop_t)(literalFrameCM(method)[
				      pc[-1] & 0x1f]));
				break;
			default:
				/*
				 *	Send literal seleltor with 32 
				 *	to 127 arguments
				 */

				switch (pc[-3]) {
				case 134:
					super = TRUE;
					/* FALL THROUGH */
				case 132:
					(void)asString(selbuf,
					    (oop_t)(literalFrameCM(method)[
					      pc[-1]]));
					break;
				default:	/* non-sends (not nonsense) */
					(void)sprintf(selbuf, "%d", pc[-1]);
					break;
				}
				break;
			}
		printf("%s%s>>%s%s\r\n",  super ? "super " : "",  
		    rcbuf, selbuf,  inABlock ? "[]" : "");
	}
}


# ifdef DBTRACE

	/*
	 *	Called before each bytecode.  Verify TOS and bytecode.
	 */

	boolean
	checkDbTrace(pc, sp)
	register byte_t *pc;
	register oop_t *sp;
	{
		static char	buf[BUFSIZ];
		byte_t	shouldBe;
		oop_t	bs1TOS;

		if ((shouldBe = (byte_t)getc(dbtrace)) != *pc) {
			(void)sprintf(buf, "*pc is %d should be %d cycle %d\n",
				*pc, shouldBe, cycle);
			halt(buf);
			fseek(dbtrace, -sizeof(char), 1);
			return (FALSE);
		}
		bs1TOS = getw(dbtrace);
		if (isIntegerObject(bs1TOS)) {
			if (bs1TOS != sp[-1]) {
				halt(sprintf(buf, "TOS is %x should be %x\n",
				    sp[-1], bs1TOS));
				fseek(dbtrace, -(sizeof(char) + sizeof(int)), 
				    1);
				return (FALSE);
			}
		}
		else if (bs1TOS <= lastRegXoop  
		    &&  nonNegIntegerValueOf(hashRp((rp_t)sp[-1])) != bs1TOS) {
			halt(sprintf(buf, "TOS is %x should be %x\n",
			    nonNegIntegerValueOf(hashRp((rp_t)sp[-1])), 
			      bs1TOS));
			fseek(dbtrace, -(sizeof(char) + sizeof(int)), 1);
			return (FALSE);
		}
		return (TRUE);
	}
# endif


# ifdef BREAKPOINT

	/*
	 *	Called before every bytecode.  
	 */

	boolean
	checkBreakpoint()
	{
		char 	buf[BUFSIZ];

		if (cycle == breakpoint) {
			halt(sprintf(buf, "breakpoint at cycle %d", cycle));
			++breakpoint;		/* for single stepping */
			return (FALSE);
		}
		return (TRUE);
	}

# endif


/*
 *	Executed when ALRM signal is caught.
 *	Called ten times per second.
 *	Does periodic sync() in case we are running single user.
 */

void
serviceTimer()
{
	static u_int msSinceLastSync = 0;
#	define msPerSync (30 * 1000)

	if ((msSinceLastSync += msPerPoll) >= msPerSync) {
		sync();
		msSinceLastSync = 0;
	}

	pollInput();
	pollTimeSem((int)msPerPoll);
	pollGenerationSpace();
}
