static char rcsId[] = "$Header: floatPrims.c,v 41.5 84/07/31 12:01:58 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

# include	<stdio.h>
# include	<math.h>
# include	<sgtty.h>
# include	<signal.h>
# include	<ctype.h>

# include	<sys/types.h>
# include	<sys/vtimes.h>
# include	<sys/timeb.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"prims.h"


/*
 *	Smalltalk's definition for truncation is different from
 *	C's floor() function.  In Smalltalk, truncate means "returns 
 *	a SmallInteger equal to the value of the receiver without any 
 *	fractional part".   
 *	Thus -3.6 truncated == -3.  But floor(-3.6) == -4.
 */

# define	truncate(flt)\
		((flt) > 0.0  ?  floor(flt)  :  - floor(fabs(flt)))


/*
 *	We use IEEE, since that's what C supports.
 *	ST assumes 1 <= mantissa < 2.
 */

# define	exponentBias		127
# define	rawExponentOf(f)	(((int)(f) & 0x7f800000)  >>  23)
# define	normalizedExponentOf(f)\
			(rawExponentOf(f) - exponentBias)


/*
 *	The next few macros are building blocks for float primitives
 *	just as in SIPrims.c.
 *
 *	Setup for floating point arithmethic/comparison/conversion
 *	operation.  "Reg1" and "reg2" are registers allocated for 
 *	storing pointers to the floating point objects.  "FloatReg1"
 *	and "floatReg2" are floating point registers allocated for
 *	storing the actual floating point number.  First argument 
 *	(reg1) to a floating point method must be a floating point 
 *	number. The second argument (reg2) may be an integer, in which 
 *	case it will be converted to a floating point number.
 */

# define floatOperationSetup	\
	register rp_t reg1, reg2;\
	register oop_t oopr;\
	register float floatReg1, floatReg2;\
	\
	if (TRUE) {\
		reg1 = (rp_t)*rp;\
		if (!isFloat(reg1))\
			return (NullRp);\
		floatReg1 = floatValueOf(reg1);\
		\
		reg2 = (rp_t)rp[1];\
		if (isIntegerObject((oop_t)reg2)) {\
			oopr = (oop_t)reg2;\
			convertToInteger(oopr);\
			floatReg2 = (float)oopr; \
		}\
		else if (classRp(reg2) == classFloatOop)\
			floatReg2 = floatValueOf(reg2);\
		else\
			return (NullRp);\
	}\
	else


# define floatOperationFinish \
	floatObjectOf(floatReg1, reg1);\
	*rp = (oop_t)reg1;\
	return (ac)


# define floatRelationFinish(pred) \
	if (pred)\
		*rp = (oop_t)trueOop;\
	else\
		*rp = (oop_t)falseOop;\
	return (ac)


/*
 *	SmallInteger asFloat
 */

rp_t
SIAsFloatPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register oop_t sioop;		/* integer value of rp */
	register rp_t fltoop;		/* floating point object of sioop */

	getInt(*rp, sioop, NullRp);
	floatObjectOf(sioop, fltoop);
	*rp = (oop_t)fltoop;
	return (ac);
}


/*
 *	The C compiler on release 1.1 has a bug in it.
 *	It doesn't like the more elegant ``+='' operator when both operands
 *	are floating point registers.  The same is true for the operators
 *	``-='', ``*='' and ``/=''.
 */

/*
 *	Float +
 */

rp_t
FloatAddPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
#	ifdef CompilerWorks
		floatReg1 += floatReg2;
#	else
		floatReg1 = floatReg1 + floatReg2;
#	endif
	floatOperationFinish;
}


/*
 *	Float -
 */

rp_t
FloatSubPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
#	ifdef CompilerWorks
		floatReg1 -= floatReg2;
#	else
		floatReg1 = floatReg1 - floatReg2;
#	endif
	floatOperationFinish;
}


/*
 *	Float <
 */

rp_t
FloatLtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
	floatRelationFinish(floatReg1 < floatReg2);
}


/*
 *	Float >
 */

rp_t
FloatGtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
	floatRelationFinish(floatReg1 > floatReg2);
}


/*
 *	Float <=
 */

rp_t
FloatLePrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
	floatRelationFinish(floatReg1 <= floatReg2);
}


/*
 *	Float >=
 */

rp_t
FloatGePrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
	floatRelationFinish(floatReg1 >= floatReg2);
}


/*
 *	Float =
 */

rp_t
FloatEqPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
	relationFinish(floatReg1 == floatReg2);
}


/*
 *	Float ~=
 */

rp_t
FloatNeqPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
	floatRelationFinish(floatReg1 != floatReg2);
}


/*
 *	Float *
 */

rp_t
FloatMulPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
#	ifdef CompilerWorks
		floatReg1 *= floatReg2;
#	else
		floatReg1 = floatReg1 * floatReg2;
#	endif
	floatOperationFinish;
}


/*
 *	Float /
 */

rp_t
FloatSlPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	floatOperationSetup;
	if (floatReg2 == 0.0)
		return (NullOop);
#	ifdef CompilerWorks
		floatReg1 /= floatReg2;
#	else
		floatReg1 = floatReg1 / floatReg2;
#	endif
	floatOperationFinish;
}


/*
 *	Float truncate
 *
 *	Rp is the floating point object, and will be returned as 
 *	a smallInteger.  
 */

rp_t
FloatTruncPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register float	floatReg1;	/* truncated fp value of fp object */
	register rp_t	fltOop;		/* points to receiver as fp object */
	register oop_t	intOop;		/* truncated smallInteger object */

	fltOop = (rp_t)*rp;
	if (!isFloat(fltOop))
		return (NullOop);
	floatReg1 = floatValueOf(fltOop);
	floatReg1 = truncate(floatReg1);
	intOop = (int)floatReg1;	/* coerce to integer */
	if ((float)intOop != floatReg1  
	    ||  !isIntegerValue(intOop)) /* if truncation error */
		return (NullOop);
	*rp = integerObjectOf(intOop);
	return (ac);
}


/*
 *	Float fractionPart
 */
 
rp_t
FloatFractPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	fltOop;		/* points to receiver as fp object */
	register float	floatReg1;	/* floating point value of receiver */

	fltOop = (rp_t)*rp;
	if (!isFloat(fltOop))
		return (NullOop);
	floatReg1 = floatValueOf(fltOop);
#	ifdef CompilerWorks
		floatReg1 -= truncate(floatReg1);
#	else
		floatReg1 = floatReg1 - truncate(floatReg1);
#	endif
	floatObjectOf(floatReg1, fltOop);
	*rp = (oop_t)fltOop;
	return (ac);
}


/*
 *	Float exponent
 *
 *	Consider the receiver to be represented as a power of two multiplied
 *	by a mantissa (between one and two).  Answer with the SmallInteger to 
 *	whose power two is raised.
 */

rp_t
FloatExpPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	u_int exponent;

	exponent = normalizedExponentOf(oopFieldRp((rp_t)*rp)[0]);
	*rp = integerObjectOf(exponent);
	return (ac);
}


/*
 *	Float timesTwoPower: anInteger
 *
 *	Answer with the receiver mulitplied by 2.0 raised to the power
 *	of the argument.  
 */

rp_t
FloatPowPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	float floatReg1;		/* floating point value of receiver */
	register oop_t powOop;		/* power of 2 to be multiplied */
	register rp_t fltOop;		/* points to receiver */

	getInt(rp[1], powOop, NullRp);

	fltOop = (rp_t)*rp;
	floatReg1 = floatValueOf(fltOop);

	/*
	 *	calculate new exponent
	 */

	powOop += rawExponentOf(*(int *)&floatReg1);
	if ((u_int)powOop >= 256)
		return (NullOop);

	/*
	 *	Combine mantissa and sign bits with exponent.
	 *	There are lots of magic here, but I hope it's the only
	 *	place these particular numbers appear.
	 *
	 *	0x807fffff will mask off the exponent part of floatReg1. 
	 *	(powOop << 23) shifts the power to the exponent part of
	 *	the floating point number.
	 */

	*(int *)&floatReg1 = *(int *)&floatReg1 & 0x807fffff  
	    |  (powOop << 23);
	floatObjectOf(floatReg1, fltOop);
	*rp = (oop_t)fltOop;
	return (ac);
}
