static char if_h[] = "$Header: if.h,v 41.1 84/03/14 20:01:05 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Instantiation and Freeing (see also mem.h)
 *	Contains headers and definitions of macros necessary 
 *	for carrying out Generation Scavenging.
 */


/*
 *	sends are optimized by keeping a linked list of allocated but
 *	unused small method contexts.  FirstFreeContext points to the list
 *	head, or contains an integer.
 */

extern	rp_t	FirstFreeContext;


/*
 *	Generation Scavenging:
 *
 *	See my quals paper.  Every so often, BS marks and  sweeps new 
 *	objects from the old objects that point to them (the remembered 
 *	ones).
 *
 *	MethodContexts are reclaimed on a return if the SMC
 *	bit is set, which indicates the context or one of its 
 *	blockContexts has been nonlocally stored or returned 
 *	as a value.
 */

# define MAX_REMEMBERED	800
extern	rp_t		Remembered[MAX_REMEMBERED];
extern	rp_t		*NextRemembered; /* first unused slot in Remembered */


/*
 *	An area of virtual memory for a particular class 
 *	of objects is a space.  
 */

struct space {
	oop_t 	*sp_bottom;		/* start of space (lowest address) */
	oop_t	*sp_next;		/* first free piece of new space */
	oop_t	*sp_top;		/* first char after space */
	int	sp_dummy;		/* pad out to power of two */
};



/*
 *	The tests for new and old assume that new space is before old space.
 *
 *	During execution, PastSurvivors holds previously scavenged stuff,
 *	FutureSurvivors is empty, and we allocate space out of New.
 *	A GS moves reachable new objects from PastSurvivors and New
 *	to FutureSurvivors, then when done toggles the PastSurvivorIndex.
 */

# define maxAge	'\20'


extern struct space	New, Survivor[2], Old;
extern u_int		PastSurvivorIndex;


# define PastSurvivors		(Survivor[PastSurvivorIndex])
# define FutureSurvivors	(Survivor[PastSurvivorIndex ^ 1])
# define ToggleSurvivors	(PastSurvivorIndex ^= 1)


/*
 *	Random space macros.
 *	Determines whether an object pointer is new or old.
 *	Argument MUST NOT be a SmallInteger.
 */

# define newRp(rp)	((oop_t *)rp < Old.sp_bottom)	/* ints FAIL!! */
# define oldRp(rp)	(!newRp(rp))


/*
 *	Allocates an object of size ``nOops'' from space pointed to by
 *	``spacep''.  Object pointer is returned in ``res''.
 */

# define allocAnObj(spacep, nOops, res) \
	if (TRUE) {\
		res = (rp_t)(spacep)->sp_next;\
		(spacep)->sp_next = &oopFieldRp(res)[nOops];\
		if ((spacep)->sp_next > (spacep)->sp_top)\
			fatal("allocAnObj: overflow", NullOop);\
		traceAllocAnObj(spacep, nOops, res);\
	}\
	else

/*
 *	ggcStore is called on all nonlocal stores.
 *	It checks for storing a new OOP into an old object, and
 *	checks for storing a context (assumes contexts are new).
 *	Optimized based on the following stats from a 4.5 M bc run:
 *
 *		src was int	100K
 *		src > dst	170K
 *		src old		170K
 *		dst new		180K
 *		total		190K
 *		remembered	322
 */

# define ggcStore(dest, src) /* STATEMENT */\
	if (TRUE) {\
		traceGGCStore(dest, src);\
		if (newRp((rp_t)src)) /* int test would be redundant */ {\
			if (flagsRp((rp_t)src) & CONTEXT)\
				if (!isBlock((rp_t)src))\
					creatingReferenceToContext((rp_t)src);\
				else /* storing Block Context */ {\
					register rp_t	_gst;\
					\
					creatingReferenceToContext((rp_t)src);\
					creatingReferenceToContext(\
					    homeBC((rp_t)src));\
					/*\
					 *	follow up the caller chain\
					 */\
					for (_gst = callerBC((rp_t)src);\
					    _gst != nilOop  &&  isBlock(_gst);\
					    _gst = callerBC(_gst))\
						creatingReferenceToContext(\
						    homeBC(_gst));\
					if (_gst != nilOop)\
						creatingReferenceToContext(\
						    callerBC((rp_t)src));\
				}\
			/*\
			 * 	storing new object into old space\
			 */\
			if (oldRp((rp_t)dest))\
				remember((rp_t)dest);\
		}\
	}\
	else


/*
 *	This macro implements the check to see if BS is returning
 *	a context up the activation chain.
 */

# define returningFromTo(old, new) \
	if (flagsRp(old) & SMC) {\
		traceReclaimedContext(old);\
		freeContext(old);\
	}\
	else if (TRUE) {\
		traceUnreclaimedContext(old);\
		creatingReferenceToContext(new);\
	}\
	else


/*
 *	This works since only small method contexts are reused.
 *	Setting SMC flag prevents returned context from being scavenged.
 */

# define creatingReferenceToContext(oop)\
	flagsRp(oop) &= ~SMC


/*
 *	This is an idempotent macros to remember an old object
 *	that might possibly point to a new one.
 */

# define remember(rp) /* STATEMENT */\
	if (!(flagsRp(rp) & REMEMBERED)) /* not remembered already */ {\
		if (NextRemembered >= &Remembered[lengthof(Remembered)])\
			fatal("overflowed Remembered", (oop_t)rp);\
		*NextRemembered++ = rp;\
		flagsRp(rp) |= REMEMBERED;\
	}\
	else

/*
 *	When allocating an object, it must get a quasi-unique hash 
 *	code between lastRegXoop+1 and max positive smallInteger.  
 *	This hash code is generated sequentially starting from 
 *	lastRegXoop+1.  The hash code is used to index dictionaries.
 *	NewHash keeps track of the next hash numbers to be generated.  
 *	When it reaches the maximum positive smallInteger, it will 
 *	start from lastRegXoop+1 again.
 */

extern	int	NewHash;	/* when allocating, contains new hash value */

extern oop_t	promote();
extern rp_t	instantiateClass();
extern void	initMem(), ggc(), ggcRegistryForward(), ggcSpaceForward(),
		unmarkObjects();
