static char rcsId[] = "$Header: kbd.c,v 1.3 84/09/05 20:56:12 pkl Exp $";

/*
 *	Copyright (C) 1984 by the Regents of University of California
 *	All Rights Reserved.
 */

/*
 *	Program to automatically generate a keyboard map for an unknown 
 *	keyboard.  The output is in a file "new_keytables.h" and it is
 *	in the same format as that used in keytables.h.
 *	Run this program.  Read "new_keytables.h" into keytables.h to 
 *	replace the dummy definition there.
 */

# include	<stdio.h>
# include	<signal.h>
# include	<sun/fbio.h>
# include	<sunwindow/window_hs.h>
# include	<pixrect/pixrect_hs.h>
# include	"keytables.h"		/* for ST constants */

# define	numberOfKeys	128

# ifdef RELEASE1_0
	static struct screen	displayScreen
	    = { "/dev/bw0", FBTYPE_SUN1BW, TRUE };
# else
	static struct screen	displayScreen;
# endif

int	WinFd;
int	KBType;
void	fatal(), outputKeyTable(), typeCapsLock(), typeKeys(), typeOneKey();

int	station[numberOfKeys];		/* implicit initialization to NULL */


/*
 *	Names of keyboards that Sun supports (at this time).
 */

struct {
	char	*fullname, *prefix
} kbName[] = {
	"Micro Switch 103SD32", "ms",
	"Keytronics VT100", "vt",
	"Sun 2 Custom Keyboard", "s2"
};


/*
 *	Maps Smalltalk codes to strings.
 */

char	*names[] = {
/*  0 */	"NULL",	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*  8 */	"ST_BACKSPACE",
			"ST_TAB",
				"ST_LINEFEED",
					NULL,	NULL,	"ST_RETURN",
								NULL,	NULL,
/* 16 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/* 24 */	NULL,	NULL,	NULL, 	"ST_ESCAPE",
						NULL,	NULL,	NULL,	NULL,
/* 32 */	"ST_SPACE",
			"'!'",	"'\"'",	"'#'",	"'$'",	"'%'",	"'&'",	"'\\''",
/* 40 */	"'('",	"')'",	"'*'",	"'+'",	"','",	"'-'",	"'.'",	"'/'",
/* 48 */	"'0'",	"'1'",	"'2'",	"'3'",	"'4'",	"'5'",	"'6'",	"'7'",
/* 56 */	"'8'",	"'9'",	"':'",	"';'",	"'<'",	"'='",	"'>'",	"'?'",
/* 64 */	"'@'",	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/* 72 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/* 80 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/* 88 */	NULL,	NULL,	NULL,	"'['",	"'\\\\'",
							"']'",	"'^'",	"'_'",
/* 96 */	"'`'",	"'a'",	"'b'",	"'c'",	"'d'",	"'e'",	"'f'",	"'g'",
/*104 */	"'h'",	"'i'",	"'j'",	"'k'",	"'l'",	"'m'",	"'n'",	"'o'",
/*112 */	"'p'",	"'q'",	"'r'",	"'s'",	"'t'",	"'u'",	"'v'",	"'w'",
/*120 */	"'x'",	"'y'",	"'z'",	"'{'",	"'|'",	"'}'",	"'~'",	"ST_DELETE",
/*128 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*136 */	"ST_LSHIFT",
			"ST_RSHIFT",
				"ST_CONTROL",
					"ST_ALPHALOCK",
						NULL,	NULL,	NULL,	NULL,
/*144 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*152 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*160 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*168 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*176 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*184 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*192 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*200 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*208 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*216 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*224 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*232 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*240 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*248 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	"ST_HALT"
};


main(argc, argv)
short argc;
char **argv;
{
	struct inputmask	im;

	signal(SIGWINCH, SIG_IGN);
	signal(SIGCHLD, SIG_IGN);

#	ifdef RELEASE1_0
		if ((WinFd = win_getnewwindow()) == NULL)
			fatal("cannot get new window.");
		win_screennew(WinFd, &displayScreen);
#	else
		if (getenv("WINDOW_PARENT"))
			fatal("cannot run inside suntools.");
		if ((WinFd = win_screennew(&displayScreen)) == NULL)
			fatal("cannot open screen.");
#	endif
		
	input_imnull(&im);
	im.im_flags  |=  IM_UNENCODED;
	win_setinputmask(WinFd, &im, &im, WIN_NULLLINK);

#	define typeOneSpecialKey(c)	typeOneKey("c", c)


	printf("Type each key after the prompt.  The keystation number\n");
	printf("of the key just typed is printed to confirm character\n");
	printf("typed.\n");
	typeKeys('0', '9');
	typeKeys('a', 'z');
	typeOneKey("delete", ST_HALT);
	printf("If the following keys are not in the unshifted\n");
	printf("position, hit DELETE instead.\n");
	typeOneSpecialKey('-');
	typeOneSpecialKey('_');
	typeOneSpecialKey('+');
	typeOneSpecialKey('=');
	typeOneSpecialKey('`');
	typeOneSpecialKey('~');
	typeOneSpecialKey(';');
	typeOneSpecialKey(':');
	typeOneSpecialKey('|');
	typeOneSpecialKey('[');
	typeOneSpecialKey(']');
	typeOneSpecialKey('{');
	typeOneSpecialKey('}');
	typeOneSpecialKey(',');
	typeOneSpecialKey('.');
	typeOneSpecialKey('<');
	typeOneSpecialKey('>');
	typeOneSpecialKey('?');
	typeOneSpecialKey('/');
	typeOneSpecialKey('!');
	typeOneSpecialKey('@');
	typeOneSpecialKey('#');
	typeOneSpecialKey('$');
	typeOneSpecialKey('%');
	typeOneSpecialKey('^');
	typeOneSpecialKey('&');
	typeOneSpecialKey('*');
	typeOneSpecialKey('(');
	typeOneSpecialKey(')');
	typeOneKey("'\\'", '\\');
	typeOneKey("'\"'", '"');
	typeOneKey("single quote", '\'');
	typeOneKey("backspace", ST_BACKSPACE);
	typeOneKey("tab", ST_TAB);
	typeOneKey("line feed", ST_LINEFEED);
	typeOneKey("return", ST_RETURN);
	typeOneKey("escape", ST_ESCAPE);
	typeOneKey("space bar", ST_SPACE);
	typeOneKey("left shift", ST_LSHIFT);
	typeOneKey("right shift", ST_RSHIFT);
	typeOneKey("control key", ST_CONTROL);
	typeCapsLock();

	outputKeyTable();
}


void
fatal(s)
char s[];
{
	fprintf(stderr, "%s\n", s);
	exit(1);
}


/*
 *	Ignores idle keys and upward transistion.
 */

getEvent(iep)
struct inputevent *iep;
{
	while (!input_readevent(WinFd, iep))
		if (KEYOF(iep->ie_code) != IDLEKEY
		    &&  keyDown(iep->ie_code)) {
			printf("%d\n", KEYOF(iep->ie_code));
			KBType = devId(iep->ie_code);
			return;
		}
	fatal("read input event failed.\n");
}


void
typeKeys(from, to)
char from, to;
{
	register char	c;
	struct inputevent	ie;

	for (c = from;  c <= to;  c++) {
		printf("'%c':   ", c); fflush(stdout);
		getEvent(&ie);
		station[KEYOF(ie.ie_code)] = c;
	}
}


void
typeOneKey(name, code)
char name[];
int code;
{
	struct inputevent	ie;

	printf("%s: ", name); fflush(stdout);
	getEvent(&ie);
	if (station[KEYOF(ie.ie_code)] != ST_HALT)
		station[KEYOF(ie.ie_code)] = code;
}


/*
 *	Caps lock is a special case because it generates
 *	X - Y  on downward transition
 *	and simply X on upward transition (for some integer X, Y).
 */

void
typeCapsLock()
{
	struct inputevent	ie;

	printf("caps lock: "); fflush(stdout);
	getEvent(&ie);
	if (station[KEYOF(ie.ie_code)] != ST_HALT) {
		getEvent(&ie);
		if (station[KEYOF(ie.ie_code)] != ST_HALT) {
			station[KEYOF(ie.ie_code)] = ST_ALPHALOCK;
			printf("Please hit 'caps lock' again to reset it.\n");
			getEvent(&ie);
		}
	}
}


void
outputKeyTable()
{
	FILE	*outp;
	int	i, j;
#	define	charPerRow	8

	if ((outp = fopen("new_keytables.h", "w")) == NULL)
		fatal("cannot open output file.");
	fprintf(outp, "/*\n *\tKeyboard Map for %s\n */\n", 
	    kbName[KBType].fullname);
	fprintf(outp, "static struct keymap\t%sKeytab_lc = {\n",
	    kbName[KBType].prefix);
	for (i = 0;  i < numberOfKeys;  i += charPerRow) {
		fprintf(outp, "/* %2d */", i);
		for (j = i;  j  <  i + charPerRow;  j++)
			fprintf(outp, "\t%s,", names[station[j]]);
		fprintf(outp, "\n");
	}			
	fprintf(outp, "};\n");
}
