static char keytables_rcsId[] = "$Header: keytables.h,v 1.4 84/09/05 20:56:46 pkl Exp $";

/*
 *	Copyright (C) 1984 by the Regents of University of California
 *	All Rights Reserved.
 */

/*
 *	Decodes an undecoded keyboard to the Smalltalk codes.
 *	OOPS is defined in kbd.h, but it is also defined in rp.h.
 *	We have to undefine it so that cpp will not complain.
 */

# undef		OOPS
# include	<sundev/kbd.h>
# undef		OOPS


/*
 *	Smalltalk want these codes instead of ASCII codes
 */

# define	ST_BACKSPACE	8
# define	ST_TAB		9
# define	ST_LINEFEED	10
# define	ST_RETURN	13
# define	ST_ESCAPE	27
# define	ST_SPACE	32
# define	ST_DELETE	127

# define	ST_LSHIFT	136
# define	ST_RSHIFT	137
# define	ST_CONTROL	138
# define	ST_ALPHALOCK	139


/*
 *	For halting BS
 */

# define	ST_HALT		255

/*
 *	Smalltalk, on the other hand wants these codes:
 *	(Row Your Boat).
 */

# define SMALLTALK_LEFT		130	/* red */
# define SMALLTALK_MIDDLE	129	/* yellow */
# define SMALLTALK_RIGHT	128	/* blue */


/* 
 *	Keyboard table for Smalltalk 
 */

static struct keymap vtKeytab_lc = {
/*  0 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*  8 */	NULL, 	NULL, 	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/* 16 */	NULL,	NULL,	NULL,	ST_ESCAPE,
						'1',	'2',	'3',	'4',
/* 24 */	'5', 	'6', 	'7',	'8',	'9',	'0',	'-',	'=',
/* 32 */	'`',	ST_BACKSPACE,
				NULL,	NULL,	NULL,	NULL,	NULL,   ST_TAB,
/* 40 */	'q',	'w',	'e',	'r',	't',	'y',	'u',	'i',
/* 48 */	'o',	'p',	'[',	']',	ST_HALT,
							NULL,	NULL,	NULL,
/* 56 */	NULL,	ST_CONTROL,
				ST_ALPHALOCK,
					'a',	's',	'd',	'f',	'g',
/* 64 */	'h',	'j',	'k',	'l',	';',	'\'',	ST_RETURN,
									'\\',
/* 72 */	NULL,	NULL,	NULL,	NULL,	NULL,	ST_LSHIFT,
								'z',	'x',
/* 80 */	'c',	'v',	'b',	'n',	'm',	',',	'.',	'/',
/* 88 */	ST_RSHIFT,
			ST_LINEFEED,
				NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/* 96 */	NULL,	NULL,	ST_SPACE,
					NULL,	NULL,	NULL,	NULL,	NULL,
/*104 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*112 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,
/*120 */	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL
};


/*
 *	Keyboard maps for Sun2 and Micro Switch keyboard for future expansion.
 *	Dummies for the time being.
 *	If you have a keyboard other than a vt100, add your own key maps here.
 *	See kbd.c for details on how to construct your own keymaps.
 */

static struct keymap s2Keytab_lc;
static struct keymap msKeytab_lc;


/*
 *	Only unshifted keymap is necessary because the rest is done
 *	in Smalltalk.
 */

static struct keyboard	keyindex_ms = {
	&msKeytab_lc,	0,
	0,	0,	0,	0,	0,	0,	0
};

static struct keyboard	keyindex_vt = {
	&vtKeytab_lc,	0,
	0,	0,	0,	0,	0,	0,	0
};

static struct keyboard	keyindex_s2 = {
	&s2Keytab_lc,	0,
	0,	0,	0,	0,	0,	0,	0
};


/*
 *	Three different kinds of keyboards, although only vt100 is 
 *	really support.  Others are just dummies.
 */

int	nkeytables = 3;

static struct keyboard	*keytables[] = {
	&keyindex_ms,
	&keyindex_vt,
	&keyindex_s2
};


/*
 *	7-bit device ids and starts from bit 8.
 */

# define	devIdBits		0x7f
# define	devIdPosition		8


/*
 *	Returns 7-bit device identifier.
 *
 *	Mouse			0x7f
 *	Micro Switch 103SD32	0x00	(not supported)
 *	Keytronics VT100	0x01
 *	Sun-2 custon keyboard	0x02	(not supported)
 */

# define	devId(code)	((code >> devIdPosition) & devIdBits)


/*
 *	stKeyCode() returns the keycode that Smalltalk wants.
 */

# define	stKeyCode(code)\
	(keytables[devId(code)]->k_normal->keymap[KEYOF(code)])



/*
 *	Test if a key is pressed.
 */

# define	keyDown(code)	((code & UPMASK)  ==  PRESSED)
