static char rcsId[] = "$Header: lookup.c,v 41.1 84/03/14 20:01:22 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	The method dictionaries.  It is called only when the method
 *	cache misses during a send.  When it finds it, it fills in 
 *	the cache with the results.
 *
 *	Conroy and Pelegri report a 95% hit rate for this scheme,
 *	but the hash function has been changed since then.
 */

# include	<stdio.h>
# include	"utils.h"
# include	"mem.h"
# include	"mc.h"


/*
 *	method cache hash table
 */

struct methodCache	MethodCache[cacheSize];


/*
 *	This is taken from the book.
 *	It returns TRUE iff it succeeds and also fills the cache slot.
 *	It is somewhat optimized.
 */

boolean
lookupMethodInClass(class, selector)
rp_t class;
register oop_t selector;
{
	register oop_t	*curp, *endp;
	register oop_t	nextSelector;
	register oop_t	nilreg;
	rp_t	currentClass;
	register rp_t	dict;

	oop_t	*first_slotp;
	u_int	length;

	if (!class  ||  !selector)	/* catches specific bugs */
		fatal("zero class or selector in lookup", NullOop);
	else if (isIntegerObject(selector)) {	/* catches SUN kernel bug */
		warning("lookupMethodInClass: int sel; resetting high bit");
		convertToInteger(selector);
	}


	/*
	 *	Climb the superclass ladder until we find it.
	 *	Using nilreg, a register, instead of nilOop makes it faster.
	 */

	nilreg = (oop_t)nilOop;
	for (currentClass = class;  currentClass != (rp_t)nilreg;
	    currentClass = superclassCl(currentClass)) {

		/*
		 *	visit a class.
		 */

		dict = messageDictionaryCl(currentClass);
		length = lengthRp(dict);

		/*
		 *	Selector portion of dict is power of two long.
		 *	Mask length-1 with selector to hash into dict.
		 */

		first_slotp = (oop_t *)&selectorsMD(dict)[
		    nonNegIntegerValueOf(hash(selector))
		      &  (length - selectorStartMD - 1)];

		/*
		 *	Do linear search for selector from *first_slotp 
		 *	to end of dictionary or nil, whichever come first.
		 */

		for (curp = first_slotp,  endp = &oopFieldRp(dict)[length];
		    curp < endp  &&  (nextSelector = *curp) != selector
		        &&  nextSelector != nilreg;
		    curp++)
			;

		/*
		 *	If we didn't find the selector, and the dict was full
		 *	(didn't find nil) then search from the start of the
		 *	dict to the *first_slotp, where we started.
		 */

		if (nextSelector != selector  &&  nextSelector != nilreg)
			for (curp = (oop_t *)selectorsMD(dict),
			      endp = first_slotp;
			    curp < endp  &&  (nextSelector = *curp) != selector
				&&  nextSelector != nilreg;
			    curp++)
				;

		/*
		 *	Try superclass if not found
		 */

		if (nextSelector != selector)
			continue;

		/*
		 *	Fill the cache
		 */

		{
			register struct methodCache	*cs;
			register oop_t	header;

			cs = cacheSlot(class, nextSelector);
			cs->mc_sel = (rp_t)nextSelector;
			cs->mc_class = class;
			cs->mc_method = rpFieldRp(methodArrayMD(dict))
			    [wordPointerDifference(curp, selectorsMD(dict))];

			/*
			 *	Set cached method header info.
			 */

			cs->mc_whatKind = FULL_SEND;
			header = headerCM(cs->mc_method);
			switch (flagBitsCMH(header)) {
			case MC_RETURN_SELF:
				cs->mc_whatKind = RET_SELF;
				break;
			case MC_RETURN_INSTANCE_VAR:
				cs->mc_whatKind = RET_INST;
				cs->mc_index = fieldIndexCMH(header);
				break;
			case MC_HEADER_EXTENSION:
				cs->mc_index = primitiveIndexCMHE
				   (headerExtensionCM(cs->mc_method,
				      header));
				if (cs->mc_index)
					cs->mc_whatKind = PRIMITIVE;
				/* FALL THROUGH */
			default:
				cs->mc_tempCount = temporaryCountCMH(header);
				cs->mc_pc = 
				    (byte_t *)&oopFieldRp(cs->mc_method)[
				    objectPointerCountCMH(header)];
				cs->mc_large = largeContextFlagCMH(header) 
				    ? TRUE : FALSE;
				break;
			}
		}
		return (TRUE);
	}
	return (FALSE);
}
