static char rcsId[] = "$Header: mem.c,v 42.4 84/05/11 16:57:26 dmu Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Contains the heart of BS; the memory system.
 *
 *	Most of the code is concerned with the become operation.
 *	It will be speeded up a lot if we can do all the sweeping
 *	in hardware (sign).
 */

# include	<stdio.h>
# include	<sys/file.h>
# include	<sys/vadvise.h>

# include	"utils.h"
# include	"mem.h"
# include	"prims.h"
# include	"execute.h"
# include	"cursor.h"


/*
 *	Used in becomes to swap two pointers.  Rcvr and arg are the
 *	pointers to be swapped.
 */

# define testAndXChange(src, dst) 	/* for becomes */\
	if ((src) == rcvr)\
		(dst) = arg;\
	else if ((src) == arg)\
		(dst) = rcvr;\
	else if (TRUE)\
		;\
	else


int		NewHash = lastRegXoop + 1;

static rp_t	rejeuvenate();

static void	swapNewNew(), swapOldOld(), swapPointersInInterpreter(),
		swapPointersInRegistry(),
		swapPointersInRemembered(), swapPointersInSpace(),
		swapPointersOfEquals();


newQueue(InputBuffer);			/* this is a macro */

/*
 *	Before reading a snapshot or converting a Xerox image,
 *	initMem must be called to set up the data structures.
 */

void
initMem()
{
	register u_int	i;
	u_int	newLength, survivorLength, oldLength, totalLength;
	char	*p;

	/*
	 *	Space length defaults (in 1024 words):
	 *	SurvivorLength is conservatively set to defaultNewLength.
	 *	This should be optimized.
	 *
	 *	The following values were in use when I wrote my paper:
	 *
	 *		defaultNewLength	35
	 *		defaultSurvivorLength	35
	 *		defaultOldLength	265
	 */

#	ifdef SM_SPACE
#		define defaultNewLength		1
#		define defaultSurvivorLength	18
#		define defaultOldLength		225
#	else
#		define defaultNewLength		20
#		define defaultSurvivorLength	20
#		define defaultOldLength		2500
#	endif

	printf("initializing object area\n");
	FirstFreeContext = (rp_t)integerObjectOf(0);
	NextRemembered = Remembered;

	for (i = 0;  i <= lastRegXoop;  i++)
		Registry[i] = NullOop;

	newLength =  (p = getenv("NewLength")) != NULL
	    ?  atoi(p)
	    :  defaultNewLength;
	newLength *= 1024;
	
	oldLength =  (p = getenv("OldLength")) != NULL
	    ?  atoi(p)
	    :  defaultOldLength;
	oldLength *= 1024;
	
	survivorLength =  (p = getenv("SurvivorLength")) != NULL
	    ?  atoi(p)
	    :  defaultSurvivorLength;
	survivorLength *= 1024;
	
	totalLength = newLength  +  2 * survivorLength  +  oldLength;

	/*
	 *	The initial memory partition is:
	 *		New Space	(low address)
	 *		Past Survivors
	 *		Future Survivors
	 *		Old Space
	 */

	New.sp_next = New.sp_bottom = 
	    (oop_t *)xvalloc(totalLength * sizeof(oop_t));

	Survivor[0].sp_next = Survivor[0].sp_bottom = New.sp_top =
	    &New.sp_bottom[newLength];

	Survivor[1].sp_next = Survivor[1].sp_bottom = Survivor[0].sp_top =
	    &Survivor[0].sp_bottom[survivorLength];

	Old.sp_next = Old.sp_bottom = Survivor[1].sp_top =
	    &Survivor[1].sp_bottom[survivorLength];

	Old.sp_top = &Old.sp_bottom[oldLength];

	PastSurvivorIndex = 0;
}

/*
 *	becomes
 */

/*
 *	Main routine for becomes.  Assumes args not smallInts.
 *	Overall strategy: first must ensure that both are old
 *	or both are new.  If they differ, we can either promote the new
 *	one or rejeuvenate the old one.  Rejeuvenation is preferable
 *	because subsequent becomes will be faster.
 *	But, rejeuvenation is impossible if they are methods (born old),
 *	and it is unwise is they are both old (would need two rejeuvs)
 *	or they are of the same length (can just swap contents).
 */

swapPointersOf(rcvr, arg, acp)
register rp_t rcvr, arg;		/* pointers to be swapped */
rp_t *acp;				/* active context */
{
	BE_t	savedCursor[cursorBitmapLength];

	if (rcvr == arg)		/* same pointer: no swap */
		return;			/* xor blows it! */

	swapCursor(savedCursor);
	if (formatRp(rcvr) == COMPILED_METHOD
	    ||  formatRp(arg) == COMPILED_METHOD
	    ||	oldRp(arg) && oldRp(rcvr)
	    ||  lengthRp(arg) == lengthRp(rcvr) && (oldRp(arg) || oldRp(rcvr)))
		swapOldOld(rcvr, arg, acp);
	else
		swapNewNew(rcvr, arg, acp);
	restoreCursor(savedCursor);
}


/*
 *	Become two old objects.  If they are the same length, one may be
 *	new and need promotion.
 *	If both objects are of the
 *	same size, we can simply swap the contents of the two objects.
 *	Otherwise, we have to go through the entire object memory to
 *	fix pointers pointing to either the rcvr or arg.
 */

static void
swapOldOld(rcvr, arg, acp)
register rp_t rcvr, arg;
rp_t *acp;
{
	if (newRp(rcvr))		/* arg is old; promote rcvr */
		rcvr = (rp_t)promote((oop_t)rcvr, (oop_t *)acp);
	else if (newRp(arg))
		arg = (rp_t)promote((oop_t)arg, (oop_t *)acp);

	if (flagsRp(rcvr) & REMEMBERED)
		remember(arg);	
	if (flagsRp(arg) & REMEMBERED)
		remember(rcvr);
	
	if (lengthRp(rcvr) == lengthRp(arg)) /* same size: exchange contents */
		swapPointersOfEquals(rcvr, arg);
	else {
		swapPointersInRegistry(rcvr, arg);
		swapPointersInInterpreter(rcvr, arg);
		vadvise(VA_SEQL);
		swapPointersInSpace((oop_t)rcvr, (oop_t)arg, &PastSurvivors);
		swapPointersInSpace((oop_t)rcvr, (oop_t)arg, &New);
		swapPointersInSpace((oop_t)rcvr, (oop_t)arg, &Old);
		vadvise(VA_NORM);
	}
}


/*
 *	Used for swapping rcvr and arg when at least one of them is new.
 *	If one is old, rejeuvenate it.  As in 
 *	swapOldOld(), equal size objects are swapped by exchanging 
 *	contents. 
 *	If the lengths are unequal we must search new space.
 *	This search can be shortened by doing a scavenge, which
 *	eliminates dead objects.
 *	Slow become ratio is the speed difference between scavenging a word
 *	and searching it.  By comparing the number of survivors, which is
 *	how much we would have to search if we did the ggc() with the current
 *	size of new space (how much to search without the ggc()) and
 *	taking into account that ggc is slower per field than a search,
 *	we decide if the ggc() is worth doing.
 */

static void
swapNewNew(rcvr, arg, acp)
register rp_t rcvr, arg;
rp_t *acp;
{
#	define slowBecomeRatio	5 	/* trial and error */

	/*
	 *	make any old object new
	 */

	if (oldRp(rcvr))
		rcvr = rejeuvenate(rcvr);
	else if (oldRp(arg))
		arg = rejeuvenate(arg);

	if (lengthRp(rcvr) == lengthRp(arg))	/* equal size: swap contents */
		swapPointersOfEquals(rcvr, arg);
	else {
		swapPointersInRegistry(rcvr, arg);
		swapPointersInInterpreter(rcvr, arg);

		if (wordPointerDifference(PastSurvivors.sp_next,
			PastSurvivors.sp_bottom)
		      * (slowBecomeRatio - 1)
		    <  wordPointerDifference(New.sp_next,
		      New.sp_bottom)) {
			traceGGCFor("NewSweep");
			ggc(acp);
			rcvr = forwardingRp(rcvr);
			arg = forwardingRp(arg);
		}
	
		/* sweep */
		swapPointersInRemembered((oop_t)rcvr, (oop_t)arg);
		swapPointersInSpace((oop_t)rcvr, (oop_t)arg, &PastSurvivors);
		swapPointersInSpace((oop_t)rcvr, (oop_t)arg, &New);
	}
}


/*
 *	Do registry portion of become.  If ``rcvr'' or ``arg'' is
 *	a special object/class in the Registry, swap them as well.
 */

static void
swapPointersInRegistry(rcvr, arg)
rp_t rcvr, arg;
{
	register u_int	xoop;

	xoop = nonNegIntegerValueOf(hashRp(rcvr));
	if (xoop <= lastRegXoop)
		Registry[xoop] = arg;

	xoop = nonNegIntegerValueOf(hashRp(arg));
	if (xoop <= lastRegXoop)
		Registry[xoop] = rcvr;
}


/*
 *	Swap with the standard interpreter objects if rcvr or arg
 *	is one of them.
 */

static void
swapPointersInInterpreter(rcvr, arg)
register rp_t rcvr, arg;
{
	testAndXChange(InputSemaphore, InputSemaphore);
	testAndXChange(SignalAtTickSemaphore, SignalAtTickSemaphore);
	testAndXChange(LowSpaceSemaphore, LowSpaceSemaphore);
	testAndXChange(DisplayForm, DisplayForm);
}

		
/*
 *	Exchange contents and registry but do not exchange hashes.
 *	Objects to be swapped must be of the same length.
 */

static void
swapPointersOfEquals(rcvr, arg)
register rp_t rcvr;
rp_t arg;
{
	register oop_t	*endp, *rcvrp, *argp, oop;
	
	endRp(rcvr, endp, oop_t *);

	for (rcvrp = (oop_t *)rcvr, argp = (oop_t *)arg; 
	    rcvrp < endp; ) {		/* swap object word by word */
		oop = *rcvrp;
		*rcvrp++ = *argp;
		*argp++ = oop;
	}

	/*
	 *	unexchange hash value.
	 */

	oop = hashRp(rcvr);
	hashRp(rcvr) = hashRp(arg);
	hashRp(arg) = oop;
}


/*
 *	Sweep a space and swap references.
 */

static void
swapPointersInSpace(rcvr, arg, spacep)
register oop_t rcvr, arg;
register struct space *spacep;
{
	register rp_t	rp;

	for (rp = (rp_t)spacep->sp_bottom; rp < (rp_t)spacep->sp_next; )
		traverseObject(TRUE, testAndXChange);
}


/*
 *	If one of the pointers to be swapped is in Remembered,
 *	this procedure will fix it.
 */

static void
swapPointersInRemembered(rcvr, arg)
register oop_t rcvr, arg;
{
	register rp_t	rp, *rpp;

	for (rpp = Remembered; rpp < NextRemembered; ) {
		rp = *rpp++;
		traverseObject(FALSE, testAndXChange);
	}
}



/*
 *	Tenure an object by aging and scavenging.  
 */

oop_t
promote(anew, acp)
oop_t anew, *acp;
{
	register rp_t	anOld;
	register rp_t	newrp;

	newrp = (rp_t)anew;
	if (isIntegerObject(anew)  ||  oldRp(newrp))
		return (anew);
	else if (flagsRp(newrp) & CONTEXT) {
		(void)warning("promote:  cannot tenure contexts");
		ageRp(newrp) = 0;
		return (anew);
	}
	ageRp(newrp) = maxAge;
	traceGGCFor("Promote");
	ggc((rp_t *)acp);

	if (!(flagsRp(newrp) & FORWARDED))	/* not promoted */
		fatal("promote: attempt at promotion failed", NullOop);
	anOld = forwardingRp(newrp);
	if (newRp(anOld))
		fatal("promote: ggc failed to promote", NullOop);
	return ((oop_t)anOld);
}


/*
 *	Rejeuvenate makes an old object new.  Inverse of promote.
 *	Copy the old object to New space and fix any references to it.
 *
 *	I used to disremember the rejuevenee, but that produced
 *	complaints from the verify command.  In order to make it easier
 *	to find real bugs, I decided to leave it and waste a little space.
 */

static rp_t
rejeuvenate(anOld)
register rp_t anOld;
{
	rp_t	aNew;

	if (formatRp(anOld) == COMPILED_METHOD)
		fatal("cannot rejeuvenate a compiled method", NullOop);

	ageRp(anOld) = 0;

	/*
	 *	allocate an object from New Space and copy
	 *	the old object into it.
	 */

	{
		register oop_t	*endp, *srcp, *dstp;

		endRp(anOld, endp, oop_t *);
		allocAnObj(&New, wordPointerDifference(endp, oopFieldRp(anOld)),
		    aNew);
		for (srcp = (oop_t *)anOld,  dstp = (oop_t *)aNew;
		    srcp < endp;
		    *dstp++ = *srcp++)
			;
	}

	flagsRp(aNew) &= ~REMEMBERED;

	if (nonNegIntegerValueOf(hashRp(anOld)) <= lastRegXoop)
		Registry[nonNegIntegerValueOf(hashRp(anOld))] = aNew;
	swapPointersInInterpreter(anOld, aNew);


	/*
	 *	Used when traversing Old space.  All pointers pointing to the
	 *	object before rejeuvenation will be fixed to point to the 
	 *	rejeuvenated object.  Pointers must be remembered because
	 *	they are in Old space.
	 */

#	define oldCheckIt(src, dst) \
		if ((src) == (oop_t)anOld) {\
			(dst) = (oop_t)aNew;\
			remember(rp);\
		}\
		else if (TRUE)\
			;\
		else


	/*
	 *	Same as OldCheckIt, but pointers do not have to be
	 *	remembered since this macro will only be used on pointers
	 *	in the New space.
	 */

#	define newCheckIt(src, dst) \
		if ((src) == (oop_t)anOld)\
			(dst) = (oop_t)aNew;\
		else if (TRUE)\
			;\
		else


	vadvise(VA_SEQL);
	{
		register rp_t	rp;

		for (rp = (rp_t)Old.sp_bottom;  rp < (rp_t)Old.sp_next;  )
			traverseObject(TRUE, oldCheckIt);

		for (rp = (rp_t)New.sp_bottom;  rp < (rp_t)New.sp_next;  )
			traverseObject(TRUE, newCheckIt);

		for (rp = (rp_t)PastSurvivors.sp_bottom; 
		    rp < (rp_t)PastSurvivors.sp_next;  )
			traverseObject(TRUE, newCheckIt);
	}
	vadvise(VA_NORM);
	return (aNew);
}

/*
 *	Instantiate an object of class ``class''.
 *	Also done in direct @ primitive 
 *	in executeMethod.c and executeBlock.c.
 */

rp_t
instantiateClass(class, indexableFields)
register rp_t class;
register u_int indexableFields;
{
	register rp_t	rp;
	oop_t	spec;
	register u_int	format;
	register u_int	fixedFields;
	register u_int	lengthInOops;
		
#	ifdef TraceNew
	{
		char buf[BUFSIZ];
		fprintf (traceFp, "instantiateClass\t%s\t%d\n",
		    asString(buf, class), indexableFields);
	}
#	endif
	fixedFields = 0;

	/*
	 *	Determine format and length
	 *	Some of the lengthInOops calcution is not portable because 
	 *	the magic numbers are dependent upon the word size of the 
	 *	machine used.
	 */

	if (class == classCompiledMethodOop) {
		format = COMPILED_METHOD;
		lengthInOops = indexableFields + 3  >> 2; /* NOT PORTABLE */
	}
	else if (isPointersIS(spec = instanceSpecificationCl(class))) {
		fixedFields = fixedFieldsIS(spec);
		format = OOPS;
		lengthInOops = fixedFields + indexableFields;
	}
	else if (isWordsIS(spec)) {
		format = BES;
		lengthInOops = indexableFields + 1  >> 1; /* NOT PORTABLE */
	}
	else {
		format = BYTES;
		lengthInOops = indexableFields + 3  >> 2; /* NOT PORTABLE */
	}

	/*
	 *	Allocate space from either New or Old,
	 *	depending on what class of object it is instantiating.
	 */

	if (class != classCompiledMethodOop  
	    &&  class != classSymbolOop
	    &&  &oopFieldRp((rp_t)New.sp_next)[lengthInOops] <= New.sp_top)
		allocAnObj(&New, lengthInOops, rp);
	else if (class == classMethodContextOop 
	    || class == classBlockContextOop)
		fatal("instantiateClass: out of new space", (oop_t)class);
	else if (&oopFieldRp((rp_t)Old.sp_next)[lengthInOops] <= Old.sp_top)
		allocAnObj(&Old, lengthInOops, rp);
	else
		fatal("instantiateClass: out of old space", (oop_t)class);

	/*
	 *	Store in the various fields of new object.
	 */

	ageRp(rp) = 0;
	classRp(rp) = class;
	fixedFieldsRp(rp) = fixedFields;
	flagsRp(rp) =    class == classMethodContextOop
	    &&    indexableFields + fixedFields  ==  smallContextLength 
	      ?  SMC  :  0;
	if (class == classBlockContextOop  
	    ||  class == classMethodContextOop)
		flagsRp(rp) |= CONTEXT;
	formatRp(rp) = format;
	if (!isIntegerValue(NewHash))
		NewHash = lastRegXoop + 1;	/* just for luck */
	hashRp(rp) = integerObjectOf(NewHash++);
	lengthRp(rp) = indexableFields + fixedFields;

	/*
	 *	Initializes fields
	 */

	switch (format) {
	default:
		fatal("bad format", NullOop);
	case BES:
	case BYTES:
	case COMPILED_METHOD:
		wfill(0, (int *)oopFieldRp(rp), lengthInOops);
		break;
	case OOPS:
		wfill((int)nilOop, (int *)oopFieldRp(rp), lengthInOops);
		break;
	}
	traceInstantiateClass(flagsRp(rp), format, lengthInOops);
	return (rp);
}


/*
 *	Make sure all instances of class class are old.
 *	This is done for someInstance so that nextInstance can sweep
 *	through Old space.  Otherwise, ggc() would shuffle instances
 *	and nextInstance would skip or repeat objects.
 */

void
ageInstances(class, spacep)
register rp_t class;
struct space *spacep;
{
	register rp_t	next, p;

	if (class == classBlockContextOop  ||  class == classMethodContextOop) {
		(void)warning("promote instances: cannot promote context");
		return;
	}

	p = (rp_t)spacep->sp_bottom;
	next = (rp_t)spacep->sp_next;
	while (p < next) {
		if (classRp(p) == class)
			ageRp(p) = maxAge;
		endRp(p, p, rp_t);
	}
}


/*
 *	Used by vim (verify integrity of image).  
 *	Checks all object pointers making sure that they are sane, i.e.
 *	    1)	is a SmallInteger or lies in some space, and
 *	    2)	is properly remembered if it is an old object and 
 *		    points to a new one.
 */

void
verifyImage()
{
	register rp_t rp, *rpp;


	/*
	 *	return true iff oop oop is in space space
	 */

#	define isInSpace(oop, space)\
		((rp_t)(space).sp_bottom <= (rp_t)(oop)\
		    &&  (rp_t)(oop) <= (rp_t)(space).sp_next)


	/*
	 *	return true iff oop oop is in any space
	 */

#	define isInAnySpace(oop)\
		(isInSpace(oop, New)  ||  isInSpace(oop, PastSurvivors)\
		    ||  isInSpace(oop, Old))


	/*
	 *	Verify that ``src'' is sane.  
	 *	Since name of this macro will be passed to tranverseObject()
	 *	it must have 2 arguments.  ``dummyDst'' serves no purpose 
	 *	other than as a place holder.
	 */

#	define verifyOOP(src, dummyDst) \
		if (TRUE) {\
			if (!isIntegerObject(src)  &&  !isInAnySpace(src))\
				fprintf(stderr,\
				    "oop 0x%x in obj 0x%x not in any space\n",\
				    (src), rp);\
			else if (oldRp(rp)  &&  newRp(src)\
			    &&  !(flagsRp(rp) & REMEMBERED))\
				fprintf(stderr,\
				    "new oop 0x%x in UNREMEMBERED obj 0x%x\n",\
				    (src), rp);\
		}\
		else


	/*
	 *	print space boundaries for adb hackers:
	 */

	fprintf(stderr, "New = 0x%x,  PastSurvivors = 0x%x,  Old = 0x%x\n",
	    New.sp_bottom, PastSurvivors.sp_bottom, Old.sp_bottom);

	vadvise(VA_SEQL);

	for (rpp = &Registry[firstRegXoop];  rpp <= &Registry[lastRegXoop];
	    ++rpp) {
		rp = *rpp;
		if (!isInAnySpace(rp))
			fprintf(stderr,
		      	    "oop 0x%x in Registry[0t%d] not in a space\n",
			    rp,  rpp - Registry);
	}

	for (rpp = Remembered;  rpp < NextRemembered;  ++rpp) {
		rp = *rpp;
		if (!isInSpace(rp, Old))
			fprintf(stderr,
			    "oop 0x%x in Remembered[0t%d] is not in OldSpace\n",
			    rp,  rpp - Remembered);
		else if (!(flagsRp(rp) & REMEMBERED))
			fprintf(stderr,
			    "oop 0x%x in Remembered[0t%d] is not REMEMBERED\n",
			    rp,  rpp - Remembered);
	}

	for (rp = (rp_t)New.sp_bottom; rp < (rp_t)New.sp_next; )
		traverseObject(TRUE, verifyOOP);

	for (rp = (rp_t)PastSurvivors.sp_bottom;
	    rp < (rp_t)PastSurvivors.sp_next; )
		traverseObject(TRUE, verifyOOP);

	for (rp = (rp_t)Old.sp_bottom; rp < (rp_t)Old.sp_next; )
		traverseObject(TRUE, verifyOOP);
	fflush(stderr);
	vadvise(VA_NORM);
}
