static char rcsId[] = "$Header: memPrims.c,v 41.3 84/05/25 19:55:00 dmu Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	This file contains various primitives which are related to 
 *	accessing and modifying objects in the object memory.  Most
 *	procedures takes two arguments, pointer to receiver and a 
 *	pointer to the active context.   The last argument is at the top
 *	of the stack.
 *	
 *	Some of the procedures are really implemented in mem.c, these 
 *	are just interfaces.
 */

# include	<stdio.h>
# include	<math.h>
# include	<sgtty.h>
# include	<signal.h>
# include	<ctype.h>

# include	<sys/types.h>
# include	<sys/vtimes.h>
# include	<sys/timeb.h>
# include	<sys/vadvise.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"prims.h"
# include	"execute.h"


static int	coreLeft(), oldCoreLeft(), lowSpaceThreshold = 0;


/*
 *	Object at:
 *
 *	General purpose subscripting.
 *	Must handle three cases: pointers, words (16-bit for bitMaps),
 *	and bytes.  With 32-bit OOPs, the subscript is always a smallInteger.
 *	If the subscript is a negative number, it will be converted to 
 *	a 31-bit positive number, and detected as subscripts out of 
 *	bounds.  Therefore, it is not necessary to check whether the 
 *	subscript is less than 1.
 *
 *	Also must convert from Smalltalk's one-origin to C's zero-origin.
 */

rp_t
ObjAtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register oop_t	subscript;
	register rp_t	rcvr;

	rcvr = (rp_t)rp[0];
	if (isIntegerObject((oop_t)rcvr))
		return (NullOop);

	subscript = rp[1];
	if (!isIntegerObject(subscript))
		return (NullOop);
	subscript = nonNegIntegerValueOf(subscript) + fixedFieldsRp(rcvr) - 1;

	if ((u_int)subscript >= (u_int)lengthRp(rcvr))
		return (NullOop);


	switch (formatRp(rcvr)) {
	default:
		fatal("bad format", NullOop);

	case BES:		/* 16 bits */
		*rp = integerObjectOf(
		    (oop_t)(u_short)BEFieldRp(rcvr)[subscript]);
		break;

	case BYTES:		/* 8 bits */
	case COMPILED_METHOD:
		*rp = integerObjectOf((oop_t)byteFieldRp(rcvr)[subscript]);
		break;

	case OOPS:		/* 32 bits */
		*rp = oopFieldRp(rcvr)[subscript];
		break;
	}

	return (ac);
}


/*
 *	Object at: put:
 *
 *	Normal subscripting primitive for lhs.
 *	Similar to Object at:.
 * 	Returns object stored into array.
 *	In converting a 32-bit oop to a byte of 16-bit word,
 *	the tag bit is discarded with the other high-order bits.
 */

rp_t
ObjAtPutPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	rcvr, result;
	register oop_t	subscript;

	rcvr = (rp_t)rp[0];
	subscript = rp[1];
	result = (rp_t)rp[2];

	if (isIntegerObject((oop_t)rcvr)  ||  !isIntegerObject(subscript))
		return (NullOop);

	subscript = nonNegIntegerValueOf(subscript) + fixedFieldsRp(rcvr) - 1;
	if ((u_int)subscript >= (u_int)lengthRp(rcvr))
		return (NullOop);


	switch (formatRp(rcvr)) {
	default:
		fatal("bad format", NullOop);

	case BES:
		if (!isIntegerObject((oop_t)result))
			return (NullOop);
		rp[0] = (oop_t)result;
		/* convertToInteger(result); TRICKY */
		BEFieldRp(rcvr)[subscript] = (BE_t)result;
		break;

	case BYTES:
	case COMPILED_METHOD:
		if (!isIntegerObject((oop_t)result))
			return (NullOop);
		rp[0] = (oop_t)result;
		/* convertToInteger(result);  TRICKY */
		byteFieldRp(rcvr)[subscript] = (byte_t)result;
		break;

	case OOPS:
		oopFieldRp(rcvr)[subscript] = (oop_t)result;
		ggcStore(rcvr, result);
		rp[0] = (oop_t)result;
		break;
	}

	return (ac);
}


/*
 *	Object size
 *
 *	Returns number of indexable fields.
 */

rp_t
ObjSizePrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	rcvr;

	rcvr = (rp_t)*rp;
	if (isIntegerObject((oop_t)rcvr))
		*rp = integerObjectOf(0);
	else
		*rp = integerObjectOf(lengthRp(rcvr) - fixedFieldsRp(rcvr));
	return (ac);
}


/*
 *	String at:
 *
 *	A String is actually a byte-array with the first character
 *	indexed by 1.  To obtain the offset from the base address
 *	of the receiver, only subtracting 1 is necessary.
 *	Implementation is like Object at:, but takes byte, and 
 *	returns corresponding OOP in character table.
 */

rp_t
StringAtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	string;
	register oop_t	subscript;

	subscript = rp[1];
	if (!isIntegerObject(subscript))
		return (NullOop);
	subscript = nonNegIntegerValueOf(subscript) - 1;

	string = (rp_t)rp[0];
	if ((u_int)subscript >= (u_int)lengthRp(string))
		return (NullOop);

	*rp = oopFieldRp(characterTableOop)[byteFieldRp(string)[subscript]];
	return (ac);
}


/*
 *	String at:put:
 *
 *	Like Object at:put:, but gets value of character and puts
 *	that byte in the string.  In addition to checking the bounds of
 *	the subscript, it will also make sure that the argument to put
 *	belongs to the class Character.
 */

rp_t
StringAtPutPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	string, newOop;
	register oop_t	subscript;

	newOop = (rp_t)rp[2];		/* character to put: */
	subscript = rp[1];
	string = (rp_t)rp[0];

	if (!isIntegerObject(subscript))
		return (NullOop);
	subscript = nonNegIntegerValueOf(subscript) - 1;
	if ((u_int)subscript >= (u_int)lengthRp(string))
		return (NullOop);

	if (isIntegerObject((oop_t)newOop)
	    ||  classRp(newOop) != classCharacterOop)
		return (NullOop);

	byteFieldRp(string)[subscript] =
	    (byte_t)nonNegIntegerValueOf(valueCh(newOop));

	rp[0] = (oop_t)newOop;
	return (ac);
}


/*
 *	ReadStream next
 *
 *	A stream has an Array or String, an index, and some limits.
 *	Check the limits, subscript the collection, and bumps the index.
 *
 *	The actual subscript is logically one more than
 *	the stored index, but the one/zero origin conversion cancels that out.
 */

rp_t
RSNextPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	stream, class;
	register oop_t	index;

	stream = (rp_t)*rp;

	index = indexSt(stream);
	if (!isIntegerObject(index))
		return (NullOop);
	index = nonNegIntegerValueOf(index);
	{
		register oop_t	limit;

		/*
		 *	checks for end of stream
		 */

		getInt(readLimitSt(stream), limit, NullRp);
		if ((u_int)index >= (u_int)limit)
			return (NullOop);
	}
	{
		register rp_t	array;

		/*
		 *	checks bounds and subscripts collection
		 */

		array = arraySt(stream);
		if (isIntegerObject((oop_t)array)  ||  index >= lengthRp(array))
			return (NullOop);
		if ((class = classRp(array)) == classArrayOop)
			*rp = oopFieldRp(array)[index];
		else if (class == classStringOop)
			*rp = oopFieldRp(characterTableOop)[
			    byteFieldRp(array)[index]];
		else
			return (NullOop);
	}
	indexSt(stream) = integerObjectOf(index + 1);	/* bumps index */
	return (ac);
}


/*
 *	WriteStream nextPut:
 *	Similar to ReadStream next, except the receiver gets the 
 *	puttee.
 *	If the stream is a string, the puttee will be converted
 *	to a character first.
 */

rp_t
WSNextPutPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	stream, nextOne, array;
	register oop_t	index, class;

	stream = (rp_t)rp[0];

	index = indexSt(stream);
	if (!isIntegerObject(index))
		return (NullOop);
	index = nonNegIntegerValueOf(index);

	{
		register oop_t limit;

		/*
		 *	checks end of stream
		 */

		getInt(writeLimitSt(stream), limit, NullRp);
		if ((u_int)index >= (u_int)limit)
			return (NullOop);
	}

	array = arraySt(stream);
	nextOne = (rp_t)rp[1];

	/*
	 *	checks for out of bounds and subscripts collection.
	 */

	if (isIntegerObject(array)  ||  index >= lengthRp(array))
		return (NullOop);
	if ((rp_t)(class = (oop_t)classRp(array)) == classArrayOop) {
		rpFieldRp(array)[index] = nextOne;
		ggcStore(array, nextOne);
	}
	else if ((rp_t)class == classStringOop) {
		if (isIntegerObject((oop_t)nextOne)  
		    ||  classRp(nextOne) != classCharacterOop)
			return (NullOop);
		byteFieldRp(array)[index] = 
		    nonNegIntegerValueOf(valueCh(nextOne));
	}
	else
		return (NullOop);
	indexSt(stream) = integerObjectOf(index + 1);
	rp[0] = (oop_t)nextOne;
	return (ac);
}


/*
 *	PositionableStream atEnd
 *
 *	Return true iff the index >= limit.  
 *	The tricky test tests whether both limit and index are
 *	integers in one comparison.
 */

rp_t
PSAtEndPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	stream;
	register oop_t	index, limit;

	stream = (rp_t)*rp;
	index = indexSt(stream);

	limit = readLimitSt(stream);
	if (!isIntegerObject(limit & index)) /* TRICKY */
		return (NullOop);

	*rp = index >= limit  ?  (oop_t)trueOop  :  (oop_t)falseOop;
	return (ac);
}


/*
 *	CompiledMethod objectAt:
 *
 *	special primitive for crazy hybrid object.  Returns the object
 *	pointer fields of the receiver (a compiled method).
 */

rp_t
CMObjAtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register oop_t	index;

	index = rp[1];
	if (!isIntegerObject(index))
		return (NullOop);
	index = nonNegIntegerValueOf(index) - 1;

	if ((u_int)index >= (u_int)objectPointerCountCMH(headerCM((rp_t)rp[0])))
		return (NullOop);
	*rp = oopFieldRp((rp_t)rp[0])[index];
	return (ac);
}


/*
 *	CompiledMethod objectAt:put:
 *
 *	Another special purpose primitive.
 */

rp_t
CMObjAtPutPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	method;
	register oop_t	index;

	index = rp[1];
	if (!isIntegerObject(index))
		return (NullOop);
	index = nonNegIntegerValueOf(index) - 1;

	method = (rp_t)rp[0];
	if ((u_int)index  >=  (u_int)objectPointerCountCMH(headerCM(method)))
		return (NullOop);
	{
		register rp_t	newOne;		/* object to put: */

		newOne = (rp_t)rp[2];
		rpFieldRp(method)[index] = newOne;
		ggcStore(method, newOne);
		rp[0] = (oop_t)newOne;
	}
	return (ac);
}


/*
 *	Behavior new
 *
 *	Instantiate an object with no indexable fields.
 */

rp_t
BehavNewPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	*rp = (oop_t)instantiateClass((rp_t)rp[0], 0);
	return (ac);
}


/*
 *	Behavior new:
 *
 *	Instantiate an object with the passed number of indexable fields.
 *	Only pointer objects have fixed fields.  
 */

rp_t
BehavNewColPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register oop_t	len;

	getInt(rp[1], len, NullRp);
	if (len < 0) {
		warning ("new: failing with negative length");
		return (NullOop);
	}
	*rp = (oop_t)instantiateClass((rp_t)*rp, (u_int)len);
	return (ac);
}


/*
 *	Object become:
 *
 *	Swap identities of receiver and argument, which means that
 *	all objects that used to point to rcvr now points to arg
 *	and vice versa.  swapPointersOf() is implemented in mem.c.
 */

rp_t
ObjBecomePrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register oop_t	rcvr, arg;

	rcvr = rp[0];
	arg = rp[1];
	if (isIntegerObject(rcvr) || isIntegerObject(arg))
		return (NullOop);

	swapPointersOf((rp_t)rcvr, (rp_t)arg, &ac);
	return (ac);
}


/*
 *	Object instanceVariableAt:
 *
 *	A version of the at: primitive to access fixed fields.
 *	In calculating the offset from base address of object,
 *	no need to add the number of fixed fields as in Object at:.
 */

rp_t
ObjIVAtPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register oop_t	index;
	register rp_t	rcvr;

	index = rp[1];
	if (!isIntegerObject(index))
		return (NullOop);
	index = nonNegIntegerValueOf(index) - 1;

	rcvr = (rp_t)*rp;
	if (isIntegerObject((oop_t)rcvr)  ||  formatRp(rcvr) != OOPS  
	    ||  (u_int)index >= (u_int)lengthRp(rcvr))
		return (NullOop);

	*rp = oopFieldRp(rcvr)[index];
	return (ac);
}


/*
 *	Object instanceVariableAt:put:
 *
 *	A version of at:put: for fixed fields.  Similar to instVarAt:.
 */

rp_t
ObjIVAtPutPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register oop_t	index, newOop;
	register rp_t	rcvr;

	newOop = rp[2];			/* object to be put: */
	index = rp[1];
	if (!isIntegerObject(index))
		return (NullOop);
	index = nonNegIntegerValueOf(index) - 1;

	rcvr = (rp_t)*rp;
	if (isIntegerObject((oop_t)rcvr)  ||  formatRp(rcvr) != OOPS 
	    ||  (u_int)index >= (u_int)lengthRp(rcvr))
		return (NullOop);

	oopFieldRp(rcvr)[index] = newOop;
	ggcStore(rcvr, newOop);
	*rp = newOop;
	return (ac);
}


/*
 *	Object asOop
 *
 *	returns the object's OOP as an integer.  Used for hashing into
 *	dictionaries.
 */

rp_t
ObjAsOopPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t rcvr;

	rcvr = (rp_t)*rp;
	if (isIntegerObject((oop_t)rcvr))
		return (NullOop);
	*rp = hash(rcvr);
	return (ac);
}


/*
 *	SmallInteger asObject
 *
 *	Turns integer into OOP for corresponding object which
 *	must exist.  What is this used for?
 */

/*ARGSUSED*/
rp_t
SIAsObjPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	return (NullOop);
}


/*
 *	Behavior someInstance
 *
 *	Returns first instance of receiver class.  The ordering used
 *	is the address of the object pointer.  The object with the
 *	smallest object pointer will be returned.
 *	Used to enumerate instances of a given class.
 *	Must age all members to avoid double, or missing enumeration.
 */

rp_t
BehavSomePrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	p;
	register rp_t	next;
	register rp_t	class;

	class = (rp_t)*rp;
	if (class == classSmallIntegerOop)
		return (NullOop);

	vadvise(VA_SEQL);
	ageInstances(class, &PastSurvivors);
	ageInstances(class, &New);
	vadvise(VA_NORM);
	traceGGCFor("promoteInstances");
	ggc(&ac);
	if (flagsRp(class) & FORWARDED)
		class = forwardingRp(class);

	/*
	 *	examine all objects in Old space until the first object of 
	 *	class ``class'' is found or exhausted all objects in Old.
	 */

	p = (rp_t)Old.sp_bottom,  next = (rp_t)Old.sp_next;
	while (p < next) {
		if (classRp(p) == class)
			break;
		endRp(p, p, rp_t);		/* for() loop re-init */
	}
	fetchSP(ac)[-1] = p >= next  ?  (oop_t)nilOop  :  (oop_t)p;
	return (ac);
}


/*
 *	Object nextInstance
 *
 *	Returns next instance of receiver's class.  Similar to someInstance.
 *	Used to enumerate instances of a given class.
 *	Assumes instances are old.
 */

rp_t
ObjNextInstPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	p;		/* previous object */
	register rp_t	next, class;

	p = (rp_t)*rp;
	if (isIntegerObject((oop_t)p))
		return (NullOop);

	if (newRp(p)) {
		warning("nextInstance failing with new arg");
		return (NullOop);
	}

	class = classRp(p);
	next = (rp_t)Old.sp_next;

	vadvise(VA_SEQL);
	do
		endRp(p, p, rp_t);
	while (p < next  &&  classRp(p) != class);
	vadvise(VA_NORM);

	*rp =  p >= next  ?  (oop_t)nilOop  :  (oop_t)p;
	return (ac);
}


/*
 *	CompiledMethod class newMethod: header:
 *
 *	This is the special instantiation routine for crazy hybrid methods.
 *	It gets the size of the byte part and the new method header.
 *	First the total size of the new method is computed.
 *	The method is then allocated as a byte object, and the header 
 *	field is filled in.
 *	The rest of the OOP part (literal frame and bytecodes) is 
 *	then nilled out.
 */

rp_t
CMCNewMethPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	newMethod;
	register oop_t	header, methodSize;
	register oop_t	*p, *endp, nilr;

	header = rp[2];
	methodSize = rp[1];
	if (!isIntegerObject(header)  ||  !isIntegerObject(methodSize))
		return (NullOop);

	methodSize = nonNegIntegerValueOf(methodSize);
	methodSize +=  objectPointerCountCMH(header) * sizeof(oop_t);

	newMethod = instantiateClass((rp_t)rp[0] , (u_int)methodSize);
	headerCM(newMethod) = header;

	for (p = literalFrameCM(newMethod),
	      endp = &oopFieldRp(newMethod)[objectPointerCountCMH(header)],
	      nilr = (oop_t)nilOop;
	    p < endp;
	    *p++ = nilr)
		;
	rp[0] = (oop_t)newMethod;
	return (ac);
}


/*
 *	Object ==
 *
 *	Reimplemented in executeBlock.c and executeMethod.c for speed.
 */

rp_t
ObjEqPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	*rp   =   rp[0] == rp[1]  ?  (oop_t)trueOop  :  (oop_t)falseOop;
	return (ac);
}


/*
 *	Object class
 *
 *	Reimplemented somewhere else for speed.
 */

rp_t
ObjClassPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register oop_t	rcvr;

	rcvr = rp[0];
	*rp = isIntegerObject(rcvr) ? (oop_t)classSmallIntegerOop 
	    : (oop_t)classRp((rp_t)rcvr);
	return (ac);
}


/*
 *	SystemDictionary coreleft
 */

rp_t
SDCorePrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	*rp = integerObjectOf(coreLeft());
	return (ac);
}


/*
 *	SystemDictionary oopsLeft
 *
 *	Returns some dummy value.  It will never be used anyway.
 */

rp_t
SDOopsPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	*rp = integerObjectOf(60000);
	return (ac);
}


/*
 *	SystemDictionary signal:atOopsLeft:wordsLeft:
 *
 *	Set semaphore and threshold so that it can be signalled.
 *	This intercepts infinite recursion.  
 *	Ignores oop limit.
 */

rp_t
SDSigAtOopPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	semoop;
	register oop_t	thresh;

	semoop = (rp_t)rp[1];

	if (isIntegerObject((oop_t)semoop)
	    ||  classRp(semoop) != classSemaphoreOop)
		return (NullOop);
	getInt(rp[3], thresh, NullRp);

	LowSpaceSemaphore = semoop;
	lowSpaceThreshold = thresh;
	return (ac);
}


/*
 *	Signal semaphore if space is low.
 */

checkSpaceLimits(acp)
rp_t *acp;
{
	if (LowSpaceSemaphore != nilOop  &&  coreLeft() <= lowSpaceThreshold) {
		synchronousSignal(LowSpaceSemaphore, acp);
		LowSpaceSemaphore = nilOop;
	}
}


/*
 *	Return Min words of space left.
 *	Must subtract size of used PastSurvivror space from unused Old
 *	space because objects in the former may be promoted to Old space
 *	by ggc.
 *	Check PastSurvivors to stop infinite recursion.
 *	Can be negative!
 */

static int
coreLeft()
{
	register int	old, new;

	new = wordPointerDifference(PastSurvivors.sp_top,
	    PastSurvivors.sp_next);
	old = oldCoreLeft();
	return (new < old  ?  new  :  old);
}


/*
 *	answers true iff Old space is low and it is time to reorganize
 */

rp_t
SDOldLeftPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	*rp = integerObjectOf(oldCoreLeft());
	return (ac);
}


static int
oldCoreLeft()
{
	return (wordPointerDifference(Old.sp_top, Old.sp_next)
	    -  wordPointerDifference(PastSurvivors.sp_next,
	      PastSurvivors.sp_bottom)); /* could promote */
}


rp_t
SDOldUsedPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	*rp = integerObjectOf(Old.sp_next - Old.sp_bottom);
	return (ac);
}
