static char rcsId[] = "$Header: names.c,v 41.1 84/03/14 20:03:08 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */


# include	<stdio.h>
# include	<math.h>

# include	"utils.h"
# include	"mem.h"

static void	crtonl();


/*
 *	Convert oop to a string in buf.  Sensibly prints the object 
 *	pointed to by oop.
 *
 *	Used for debugging messages.  Returns buf ala sprintf().
 *	Avoid using macros to bypass sanity checks; 
 *	want this to work on free objects as well.
 *	Cheats a little on what the interpreter is allowed to know.
 */

char *
asString(buf, oop)
char buf[];
register oop_t oop;
{
	register rp_t className, class;
	register u_int size;
	register rp_t rp;

	rp = (rp_t)oop;
	if (isIntegerObject(oop))
		class = classSmallIntegerOop;
	else if ((class = classRp(rp)) == 0)
		return (sprintf(buf, "????"));	/* for null class */

	/*
	 *	return smallInteger values in ascii
	 */

	if (class == classSmallIntegerOop) {
		adb(_asStringSI);
		(void)sprintf(buf, "%d", integerValueOf(oop));
	} 

	/*
	 *	return text prefixed by ``#'', truncated to fit into
	 *	buffer.  Must change carriage-returns to newlines.
	 */

#	define SLOP			10	/* allow for # and elipses */
	else if (class == classSymbolOop) {
		adb(_asStringSymbol);
		size = lengthRp(rp);
		if (size >= BUFSIZ - SLOP)
			(void)sprintf(buf, "#%.*s...",  BUFSIZ - SLOP,
			    byteFieldRp(rp));
		else
			(void)sprintf(buf, "#%.*s", size, byteFieldRp(rp));
		crtonl(buf);
	}
	else if (class == classCharacterOop) {
		adb(_asStringChar);
		(void)sprintf(buf, "$%c", valueCh(rp));
	}

	/*
	 *	just like symbols with apostrophes instead of the sharp.
	 */

	else if (class ==  classStringOop) {
		adb(_asStringString);
		size = lengthRp(rp);
		if (size >= BUFSIZ - SLOP)
			(void)sprintf(buf, "'%.*s...'",  BUFSIZ - SLOP,
			    byteFieldRp(rp));
		else
			(void)sprintf(buf, "#%.*s", size, byteFieldRp(rp));
		crtonl(buf);
	}
	else if (classNameCl(class) == rp)
		(void)sprintf(buf, "%.*s",
		    lengthRp(classNameCl(rp)), byteFieldRp(classNameCl(rp)));

	/*
	 *	Metaclasses are tricky.
	 *	Apparently one must check its class's class.
	 */

	else {
		adb(_asStringDefaultClass);
		if (class == classMetaclassOop) {
			adb(_asStringMetaClass);
			(void)sprintf(buf, "class %.*s",
			    lengthRp(classNameCl(rp)),
			    byteFieldRp(classNameCl(rp)));
		}

		/*
		 *	Just a plain object.  Choose ``a'' or ``an''.
		 *	Remember the class name is NOT null-terminated.
		 */

		else {
			className = classNameCl(class);
			switch (byteFieldRp(className)[0]) {
			case 'a':
			case 'e':
			case 'i':
			case 'o':
			case 'u':
			case 'A':
			case 'E':
			case 'I':
			case 'O':
			case 'U':
				(void)sprintf(buf, "an%.*s",
				    lengthRp(className),
					byteFieldRp(className));
				break;

			default:
				(void)sprintf(buf, "a%.*s",
				    lengthRp(className),
					byteFieldRp(className));
				break;
			}
		}
	}
	return (buf);
}


/*
 *	CR to NL.  Needed to print Smalltalk strings.  
 */

static void
crtonl(cp)
register char *cp;
{
	register char c;

	for (; c = *cp; ++cp)
		if (c == '\r')
			*cp = '\n';
}
