static char reg_h[] = "$Header: reg.h,v 41.3 84/05/16 01:22:48 dmu Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	The interpreter must know the identity and format of
 *	certain objects and classes.  This file contains that information.
 *	Identity is based on Xerox OOP.  An  array, called the registry,
 *	maps from Xerox OOP to BS address for registered objects.
 */
/* $Log:	reg.h,v $
 * Revision 41.3  84/05/16  01:22:48  dmu
 * moved float macros into reg.h for resource usage stuff
 * 
 * Revision 41.2  84/05/15  23:34:44  hilfingr
 * Added definition of b2s_..._defined to aid in conditional compilations.
 *  */

#define b2s_reg_defined

/*
 *	known Xerox OOPS
 */

# define firstRegXoop				1

# define nilXoop				1
# define falseXoop				2
# define trueXoop				3
# define schedulerAssociationXoop		4
# define classSmallIntegerXoop			6
# define classStringXoop			7
# define classArrayXoop				8
# define SmalltalkSysDictAssocXoop		9
# define classFloatXoop				10
# define classMethodContextXoop			11
# define classBlockContextXoop			12
# define classPointXoop				13
# define classLargePositiveIntegerXoop		14
# define classMessageXoop			16
# define classCompiledMethodXoop		17
# define classSemaphoreXoop			19
# define classCharacterXoop			20
# define doesNotUnderstandSelectorXoop		21
# define cannotReturnSelectorXoop		22
# define SpecialSelectorsXoop			24
# define characterTableXoop			25
# define mustBeBooleanSelectorXoop		26

/*
 *	unofficial OOPS
 */

# define classSymbolXoop			28
# define classMetaclassXoop			30
# define classClassXoop				35

# define lastRegXoop				35

extern	rp_t	Registry[];


/*
 *	convenient abbreviations.  These are actually rp_t's.
 */

# define nilOop				Registry[nilXoop]
# define falseOop			Registry[falseXoop]
# define trueOop			Registry[trueXoop]
# define schedulerAssociationOop 	Registry[schedulerAssociationXoop]
# define classSmallIntegerOop		Registry[classSmallIntegerXoop]
# define classStringOop			Registry[classStringXoop]
# define classArrayOop			Registry[classArrayXoop]
# define SmalltalkSysDictAssocOop	Registry[SmalltalkSysDictAssocXoop]
# define classFloatOop			Registry[classFloatXoop]
# define classMethodContextOop		Registry[classMethodContextXoop]
# define classBlockContextOop		Registry[classBlockContextXoop]
# define classPointOop			Registry[classPointXoop]
# define classLargePositiveIntegerOop 	Registry[classLargePositiveIntegerXoop]
# define classMessageOop		Registry[classMessageXoop]
# define classCompiledMethodOop		Registry[classCompiledMethodXoop]
# define classSemaphoreOop		Registry[classSemaphoreXoop]
# define classCharacterOop		Registry[classCharacterXoop]
# define doesNotUnderstandSelectorOop	Registry[doesNotUnderstandSelectorXoop]
# define cannotReturnSelectorOop	Registry[cannotReturnSelectorXoop]
# define SpecialSelectorsOop		Registry[SpecialSelectorsXoop]
# define characterTableOop		Registry[characterTableXoop]
# define mustBeBooleanSelectorOop	Registry[mustBeBooleanSelectorXoop]
# define classSymbolOop			Registry[classSymbolXoop]
# define classMetaclassOop 		Registry[classMetaclassXoop]
# define classClassOop			Registry[classClassXoop]

/*
 *	known lengths
 */

# define pointLength			2
# define floatLength			2
# define smallContextLength		18
# define largeContextLength		38

/*
 *	The following are accessing macros for various objects.
 */

/*
 *	contexts
 */

# define senderMC(mc)		(rpFieldRp(mc)[0])
# define callerBC(bc)		(rpFieldRp(bc)[0])
# define ipC(c)			(oopFieldRp(c)[1])
# define spC(c)			(oopFieldRp(c)[2])
# define methodMC(mc)		(rpFieldRp(mc)[3])
# define argCountBC(bc)		(oopFieldRp(bc)[3])
# define unusedMC(mc)		(oopFieldRp(mc)[4])
# define initIPBC(bc)		(oopFieldRp(bc)[4])
# define receiverMC(mc)		(oopFieldRp(mc)[5])
/*
 *	a linked list of free small method contexts speeds sends.
 *	The receiver field is stolen to hold the link.
 */
# define nextSMC(mc)		(rpFieldRp(mc)[5])
# define homeBC(bc)		(rpFieldRp(bc)[5])
# define fixedFieldsC		6
# define tempFrameC(c)		(oopFieldRp(c) + fixedFieldsC)

# define isBlock(c)		isIntegerObject(argCountBC(c))

/*
 *	The next four macros store and fetch the PC and SP
 *	from contexts to registers.  This happens on every send and return.
 *	The performance of these is so crucial that dirty
 *	and more efficient versions realized a (10%?) performance gain.
 *
 *	Formats:
 *
 *	In the context, the SP is a smallInteger containing the number
 *	of entries on the stack.  The PC is a smallInt with the byte
 *	offset of the next instruction (i.e., 1 = the first byte of the
 *	(method header).
 *	In registers, the PC is the address of the next instruction, and the
 *	SP points one past the TOS.
 */

# ifdef SLOW_AND_CLEAN

#	define fetchSP(c)\
	    (&tempFrameC(c)[nonNegIntegerValueOf(spC(c))])

#	define storeSP(c, sp)\
	    (spC(c) = integerObjectOf(wordPointerDifference(sp, tempFrameC(c))))

#	define fetchPC(c, cm)\
	    (&byteFieldRp(cm)[nonNegIntegerValueOf(ipC(c)) - 1])

#	define storePC(c, cm, pc)\
	    (ipC(c) = integerObjectOf(pc - byteFieldRp(cm) + 1))

# else

#	define fetchSP(c)\
	    ((oop_t *)((int)(c)  +  16  +  24  +  ((int)spC(c) << 2)))

#	define storeSP(c, sp)\
	    (spC(c) = ((int)(sp) - (int)(c)  >>  2) - 4 - 6 + 0x80000000)

#	define fetchPC(c, cm)\
	    ((byte_t *)(cm) + (int)ipC(c) + 0x7fffffff + 16)

#	define storePC(c, cm, pc)\
	    (ipC(c) = (int)(pc) - (int)(cm) - 16 + 1 + 0x80000000)

# endif

/*
 *	compiled methods
 */

# define headerIndexCM		0
# define headerCM(cm)		(oopFieldRp(cm)[headerIndexCM])
# define literalStartCM		1
# define literalFrameCM(cm)	(oopFieldRp(cm) + literalStartCM)

# define temporaryCountCMH(hdr)		((hdr) >> 7  &  0x1f)
# define fieldIndexCMH(hdr)		((hdr) >> 7  &  0x1f)
# define flagBitsCMH(hdr)		((hdr)  & 0x7000)
# define largeContextFlagCMH(hdr)	((hdr)  &  0x40)


/*
 *	values for flag bits
 */

# define MC_RETURN_SELF		0x5000
# define MC_RETURN_INSTANCE_VAR	0x6000
# define MC_HEADER_EXTENSION	0x7000

# define headerExtensionCM(cm, hdr) /* next to last literal */\
	(literalFrameCM(cm)[literalCountCMH(hdr) - 2])

# define primitiveIndexCMHE(cmhe)	(cmhe & 0xff) 

# define primitiveIndexCM(cm, hdr)	\
    (flagBitsCMH(hdr)  ==  MC_HEADER_EXTENSION\
      ?  primitiveIndexCMHE(headerExtensionCM(cm, hdr))  \
      :  0)


# define literalCountCMH(header)\
	(flagBitsCMH(header) == MC_RETURN_SELF \
	      ||  flagBitsCMH(header) == MC_RETURN_INSTANCE_VAR \
	    ?   0  \
	    :   (header) & 0x3f)


/*
 *	objectPointerCount... includes the header itself
 */

# define objectPointerCountCMH(hdr) \
	(literalCountCMH(hdr) + literalStartCM)

# define initialInstructionPointerCMH(hdr) \
	(objectPointerCountCMH(hdr) * sizeof(oop_t)  +  1)

# define argumentCountCM(cm, hdr, res) \
	if (TRUE) {\
		res = flagBitsCMH(hdr);\
		switch (res) {\
		default:\
			res >>= 12;\
			break;\
		case MC_RETURN_SELF:\
		case MC_RETURN_INSTANCE_VAR:\
			res = 0;\
			break;\
		case MC_HEADER_EXTENSION:\
			res = headerExtensionCM(cm, hdr) >> 8  &  0x1f;\
			break;\
		}\
	}\
	else

/*
 *	class class
 *	IS = instace spec
 */

# define superclassCl(cl)		(rpFieldRp(cl)[0])
# define messageDictionaryCl(cl)	(rpFieldRp(cl)[1])
# define instanceSpecIndexCl		2
# define instanceSpecificationCl(cl)	(oopFieldRp(cl)[instanceSpecIndexCl])

# define classNameCl(cl)		(rpFieldRp(cl)[6]) /* unofficial */

# define isPointersIS(is)		((is) & 0x4000)
# define isWordsIS(is)			((is) & 0x2000)
# define isIndexableIS(is)		((is) & 0x1000)
# define fixedFieldsIS(is)		((is) & 0x07ff)


/*
 *	fields of a message dictionary
 */

# define methodArrayMD(md)	(rpFieldRp(md)[1])
# define selectorStartMD	2
# define selectorsMD(md)	(rpFieldRp(md) + selectorStartMD)

# define hash(oop)	(isIntegerObject((oop_t)oop) ? (oop_t)oop \
			: hashRp((rp_t)(oop)))


/*
 *	Association
 */

# define nameAs(as)	(rpFieldRp(as)[0])
# define valueAs(as)	(oopFieldRp(as)[1])


/*
 *	Scheduler
 */

# define processListsSc(sc)	(rpFieldRp(sc)[0])
# define activeProcessSc(sc)	(rpFieldRp(sc)[1])


/*
 *	LinkedLists and semaphores
 */

# define firstLinkLL(linklist)	(rpFieldRp(linklist)[0])
# define lastLinkLL(linklist)	(rpFieldRp(linklist)[1])
# define excessSignalsSe(se)	(oopFieldRp(se)[2])


/*
 *	Link and Process
 */

# define nextLinkLi(li)			(rpFieldRp(li)[0])
# define suspendedContextPr(proc)	(rpFieldRp(proc)[1])
# define priorityPr(proc)		(oopFieldRp(proc)[2])
# define myListPr(proc)			(rpFieldRp(proc)[3])

# define schedulerOop()			((rp_t)valueAs(schedulerAssociationOop))

# define firstContext() \
	suspendedContextPr(activeProcessSc(schedulerOop()))


/*
 *	Messages
 */

# define selectorMe(me)		(rpFieldRp(me)[0])
# define argumentsMe(me)	(rpFieldRp(me)[1])
# define lengthMe(me)		(oopFieldRp(me)[2])


/*
 *	Streams
 */

# define arraySt(st)		(rpFieldRp(st)[0])
# define indexSt(st)		(oopFieldRp(st)[1])
# define readLimitSt(st)	(oopFieldRp(st)[2])
# define writeLimitSt(st)	(oopFieldRp(st)[3])


/*
 *	Characters
 */

# define valueCh(ch)	(oopFieldRp(ch)[0])


/*
 *	Points
 */

# define xPt(pt)		(oopFieldRp(pt)[0])
# define yPt(pt)		(oopFieldRp(pt)[1])


/*
 *	Forms
 */

# define bitmapFm(fm)	(rpFieldRp(fm)[0])
# define widthFm(fm)	(oopFieldRp(fm)[1])
# define heightFm(fm)	(oopFieldRp(fm)[2])


/*
 *	BitBlt and CharacterScanner
 */

# define destFormBB(bb)		(rpFieldRp(bb)[0])
# define sourceFormBB(bb)	(rpFieldRp(bb)[1])
# define halftoneFormBB(bb)	(rpFieldRp(bb)[2])
# define combinationRuleBB(bb)	(oopFieldRp(bb)[3])
# define destXBB(bb)		(oopFieldRp(bb)[4])
# define destYBB(bb)		(oopFieldRp(bb)[5])
# define widthBB(bb)		(oopFieldRp(bb)[6])
# define heightBB(bb)		(oopFieldRp(bb)[7])
# define sourceXBB(bb)		(oopFieldRp(bb)[8])
# define sourceYBB(bb)		(oopFieldRp(bb)[9])
# define clipXBB(bb)		(oopFieldRp(bb)[10])
# define clipYBB(bb)		(oopFieldRp(bb)[11])
# define clipWidthBB(bb)	(oopFieldRp(bb)[12])
# define clipHeightBB(bb)	(oopFieldRp(bb)[13])

# define lastIndexCS(cs)	(oopFieldRp(cs)[14])
# define xTableCS(cs)		(rpFieldRp(cs)[15])
# define stopConditionsCS(cs)	(rpFieldRp(cs)[16])

/*
 *	class Float
 *
 *	A floating point number is a byte object
 *	of length 4 of class Float.
 */

# define isFloat(oop)	\
    (!isIntegerObject((oop_t)oop)  &&  classRp((rp_t)oop) ==  classFloatOop)

# define floatValueOf(oop)	(*(float *)byteFieldRp(oop))

/*
 *	Instantiate a float, stuff val into it, and put its OOP into rp.
 *	Through the magic of macros, val can be either a float or an int.
 */

# define floatObjectOf(val, rp) \
	if (TRUE) {\
		rp = instantiateClass(classFloatOop,\
		    sizeof(float) / sizeof(BE_t));\
		*(float *)byteFieldRp(rp) = (float)val;\
	}\
	else


/*
 *	special selectors, used for direct primitives
 */

# define specialSelOf(sel)\
	(rpFieldRp(SpecialSelectorsOop)[sel])

# define specialNargOf(sel)\
	nonNegIntegerValueOf(oopFieldRp(SpecialSelectorsOop)[sel + 1])


extern char	*asString();
extern rp_t	DisplayForm, InputSemaphore, LowSpaceSemaphore,
		    SignalAtTickSemaphore;
