static char rcsId[] = "$Header: reorg.c,v 43.3 84/07/19 17:27:21 dmu Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Routines for BS reorganizer.
 *	reorganize() is now called within BS to do reorganization.
 *	It can also be called by command ``r'' in VM command loop.
 */

# include	<stdio.h>
# include	<sys/file.h>

# include	"utils.h"
# include	"mem.h"
# include	"prims.h"
# include	"execute.h"
# include	"mc.h"
# include	"cursor.h"


static void	registryForward(),
		reorgCopy(), reorgCopyToFile(), reorgForwardFile(),
		startTraversal(), spaceForward();

static oop_t	forwardingOop = NullOop;
static boolean	recursiveTraversal;	/* allows location of 
					   nil true false to be preserved */


/*
 *	Reorganization routine for sun:
 *	Uses a file in order to fit in 2Mb address space.  The file 
 *	(with wfp and rfp) is used like an array with wfp following rfp.
 *	Depth-first reorganization and GC of old space.
 *	Recursively traverse the image copying reachable old objects to 
 *	myOld space leaving behind forwarding blocks.  Then sweep spaces 
 *	and short-circuit forwarding blocks.
 */

void
reorganize(acp)
rp_t *acp;
{
	char	tempFile[BUFSIZ], *rd;
	FILE	*rfp, *wfp;
	u_int	oldOldLength, newOldLength;

	extern char	*getenv();
	extern long	lseek(), xfseek();


	/*
	 *	ReorgDir environment variable
	 *	set the directory in which the reorganization file
	 *	should be stored.  Default is /tmp.
	 *	This option is good for systems with a shortage
	 *	of space in /tmp.
	 *	
	 *	The reorg file takes about 2-16 Meg of disk space.
	 */
	 
	(void)sprintf(tempFile, "%s/reorg%d",
	    (rd = getenv("ReorgDir"))  ?  rd  :  "/tmp",
	    getpid());

	wfp = xfopen(tempFile, "w");
	rfp = xfopen(tempFile, "r");
	(void)unlink(tempFile);

	oldOldLength = Old.sp_next - Old.sp_bottom;
	if (!windowOpened(BSWindowFd))		/* if you're halted */
		printf("%d words of old space\r\n", oldOldLength);
	smashCache();			/* gonna move methods!! */
	ggc(acp);			/* to empty new space */

	/*
	 *	setup for reorgCopyToFile
	 */

	unmarkObjects(&Old);
	unmarkObjects(&PastSurvivors);
	forwardingOop = (oop_t)Old.sp_bottom;

	if (!windowOpened(BSWindowFd))
		printf("depth first copy to file. . .\r\n");
	startTraversal(*acp, wfp);

	if (!windowOpened(BSWindowFd))
		printf("forwarding. . .\r\n");
	
	{
		register rp_t	*src, *dst, rp;

		/*
		 *	Fixes pointers in Remembered so that they points 
		 *	to the real object (not a forwarding pointer).
		 */

		for (src = dst = Remembered;  src < NextRemembered;  ) {
			rp = *src++;
			if (flagsRp(rp) & FORWARDED)
				*dst++ = forwardingRp(rp);
		}
		NextRemembered = dst;
	}

	spaceForward(&PastSurvivors);
	registryForward();


	/*
	 *	Fix the pointer rp so that it points to the object.
	 */

#	define mustForwardIt(rp)\
		if (isIntegerObject((oop_t)rp)  ||  newRp(rp))\
			;\
		else if (flagsRp((rp)) & FORWARDED)\
			(rp) = forwardingRp((rp));\
		else if (TRUE)\
			fatal("mustForwardIt (reorg): cannot forward",\
			    NullOop);\
		else

	mustForwardIt(InputSemaphore);
	mustForwardIt(SignalAtTickSemaphore);
	mustForwardIt(LowSpaceSemaphore);
	mustForwardIt(DisplayForm);

	(void)xfseek(wfp, 0L, 0);
	reorgForwardFile(rfp, wfp);

	(void)fflush(wfp);
	if (!windowOpened(BSWindowFd))
		printf("reading reorganized old objects. . .\r\n");
	(void)fseek(rfp, 0L, 2);
	newOldLength = ftell(rfp) / sizeof(oop_t);
	(void)fseek(rfp, 0L, 0);
	xfread((char *)Old.sp_bottom, sizeof(oop_t), newOldLength, rfp);
	Old.sp_next = Old.sp_bottom  + newOldLength;
	(void)fclose(rfp);
	(void)fclose(wfp);
	
	if (!windowOpened(BSWindowFd)) {
		printf("%d words of old space\r\n",  newOldLength);
		printf("reclaimed %d words\r\n",  
		    oldOldLength - newOldLength);
	}
}


/*
 *	Recursive depth-first search with mark bit
 */

static void
reorgCopyToFile(rp, fp)
register rp_t rp;
FILE *fp;
{
	oop_t	myForwardingOop;

	if (isIntegerObject(rp)  ||  flagsRp(rp) & MARKED)
		return;

	if (oldRp(rp)) {
		register oop_t	*endp;

		endRp(rp, endp, oop_t *);
		xfwrite((char *)rp, (u_int)sizeof(oop_t),
		    (u_int)wordPointerDifference(endp, rp), fp);

		flagsRp(rp) |= FORWARDED;
		myForwardingOop = forwardingOop;
		forwardingOop += (char *)endp - (char *)rp;
	}

	flagsRp(rp) |= MARKED;

	/*
	 *	Call reorgCopyToFile recursively for each 
	 *	object pointer in rp.
	 */

#	define reorgCopyToFileOperation(src, dummyDst)\
		reorgCopyToFile((rp_t)src, fp)

	if (recursiveTraversal)
		traverseObject(FALSE, reorgCopyToFileOperation);
	else if (newRp(rp))
		fatal(rp,
		    "reorgCopyToFile: nonrecursive traversal of new object");
	else if (formatRp(rp) == OOPS  &&  lengthRp(rp))
		fatal(rp,
		    "reorgCopyToFile: nonrecursive traversal of nonempty obj");

	if (oldRp(rp))
		forwardingRp(rp) = (rp_t)myForwardingOop;
}


/*
 *	Looks at the fields of each object, forwarding oops as necessary.
 */

static void
reorgForwardFile(rfp, wfp)
register FILE *rfp, *wfp;
{
	struct rp	rp;
	register rp_t	class;
	register u_int	oopCount, skipCount;
	rp_t	oop;
	oop_t	header;

	while (fread((char *)&rp, (int)oopFieldRp((rp_t)0), 1, rfp) == 1) {
		class = classRp(&rp);
		if (flagsRp(class) & FORWARDED)
			classRp(&rp) = forwardingRp(class);
		xfwrite((char *)&rp, (u_int)oopFieldRp((rp_t)0), 1, wfp);
		switch (formatRp(&rp)) {
		default:
			fatal("reorgForwardFile bad format", NullOop);
			break;

		case BES:
			oopCount = 0;
			skipCount = roundUp(lengthRp(&rp),
			    sizeof(oop_t) / sizeof(BE_t))
			      /  (sizeof(oop_t) / sizeof(BE_t));
			break;

		case BYTES:
			oopCount = 0;
			skipCount = roundUp(lengthRp(&rp), sizeof(oop_t)) 
			    / sizeof(oop_t);
			break;

		case COMPILED_METHOD:

			/*
			 *	This assumes the header is the FIRST oop
			 *	and that it is a small int
			 */

			xfread((char *)&header, sizeof(header), 1, rfp);
			xfwrite((char *)&header, (u_int)sizeof(header), 1, wfp);
			oopCount = objectPointerCountCMH(header);
			skipCount = roundUp(lengthRp(&rp), sizeof(oop_t)) 
			    / sizeof(oop_t)  -  oopCount;
			--oopCount;	/* just did first oop */
			break;

		case OOPS:
			oopCount = lengthRp(&rp);
			skipCount = 0;
			break;
		}
		while (oopCount--) {
			xfread((char *)&oop, sizeof(oop), 1, rfp);
			if (!isIntegerObject(oop) 
			    &&  flagsRp(oop) & FORWARDED)
				oop = forwardingRp(oop);
			xfwrite((char *)&oop, (u_int)sizeof(oop), 1, wfp);
		}
		while (skipCount--) {
			xfread((char *)&oop, sizeof(oop), 1, rfp);
			xfwrite((char *)&oop, (u_int)sizeof(oop), 1, wfp);
		}
	}
}


/*
 *	Fixes pointers in the Registry so that they points to object,
 *	not a forwarding pointer.  
 */

static void
registryForward()
{
	register rp_t	*rpp, *endp, rp;

	for (rpp = &Registry[firstRegXoop],  endp = &Registry[lastRegXoop];
	    rpp <= endp;  ++rpp) {
		rp = *rpp;
		if (isIntegerObject(rp)  ||  newRp(rp))
			;
		else if (flagsRp(rp) & FORWARDED)
			*rpp = forwardingRp(rp);
		else
			fatal("registryForward: whoops!", NullOop);
	}
}


/*
 *	Copy interpreter objects, active context, objects in dictionary
 *	and objects in Registry to file.
 *	Takes special pains to maintain nil, true, false as first three
 *	objects.
 */

static void
startTraversal(ac, wfp)
rp_t ac;
FILE *wfp;
{
	register rp_t	*rpp, *endp;

	recursiveTraversal = FALSE;
		reorgCopyToFile((rp_t)nilOop, wfp);
		reorgCopyToFile((rp_t)falseOop, wfp);
		reorgCopyToFile((rp_t)trueOop, wfp);
	recursiveTraversal = TRUE;

	reorgCopyToFile((rp_t)SmalltalkSysDictAssocOop, wfp);
	reorgCopyToFile(ac, wfp);
	reorgCopyToFile(InputSemaphore, wfp);
	reorgCopyToFile(SignalAtTickSemaphore, wfp);
	reorgCopyToFile(LowSpaceSemaphore, wfp);
	reorgCopyToFile(DisplayForm, wfp);

	for (rpp = &Registry[firstRegXoop],  endp = &Registry[lastRegXoop];
	    rpp <= endp; )
		reorgCopyToFile(*rpp++, wfp);
}


/*
 *	Traverse space and fix all pointers to forwarded objects.
 */

static void
spaceForward(sp)
struct space *sp;
{
	register oop_t	*next;
	register rp_t	rp;

#	define forwardIt(src, dst)\
		if (isIntegerObject(src)  ||  newRp((rp_t)src))\
			;\
		else if (flagsRp((rp_t)(src)) & FORWARDED) {\
			if (forwardingRp((rp_t)(src)) >= (rp_t)Old.sp_next)\
				fatal("spaceForward: bad fwd pointer",\
				    NullOop);\
			(dst) = (oop_t)forwardingRp((rp_t)(src));\
		}\
		else if (sp == &Old)\
			fatal("forwardIt (reorg): cannot forward", NullOop);\
		else if (TRUE)\
			(dst) = (oop_t)nilOop;\
		else 

	rp = (rp_t)sp->sp_bottom;
	next = sp->sp_next;
	while ((oop_t *)rp < next)
		traverseObject(TRUE, forwardIt);
}
