static char rcsId[] = "$Header: rot.c,v 41.4 84/07/25 20:21:04 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Rot is a separate program that converts a xerox image 
 *	to a Berkeley image.
 *	Hassles with byte order, 16 to 32 bit conversion,
 *	and floats.
 */

# include	<stdio.h>
# include	<math.h>

# include	"utils.h"
# include	"mem.h"


# define signExtend(x)	((int)((short)(x)))


/*
 *	Xerox object table format as it appears to a VAX or SUN
 */

struct	XeroxOTE {
	unsigned char	xrc;			/* ref count */
#		define	XOVERFLOW	0x80
	unsigned char	xopfmseg;		/* flags + segment no.  */
#		define	xo(p)		((p)->xopfmseg & 0x80)	/* odd len */
#		define	xp(p)		((p)->xopfmseg & 0x40)	/* pointers */
#		define	xf(p)		((p)->xopfmseg & 0x20)	/* free entry */
#		define	xm(p)		((p)->xopfmseg & 0x10)	/* monitor */
#		define	xseg(p)		((p)->xopfmseg & 0x0f)	/* segment */
	unsigned char	xloc1, xloc0;		/* low and mid bytes of addr */
#		define	xoffset(p)\
			(xseg(p) << 16  |  (p)->xloc1 << 8  |  (p)->xloc0)
}; 


/*
 *	xoffset gives offset of start of object in object area.
 *	Object starts with length, then class, then data.
 */

# define XeroxLengthOffset	0
# define XeroxClassOffset	1
# define XeroxFieldsOffset	2


typedef	short	XeroxWord_t;

long	xlseek();
void	scopy();

rp_t	*FullRegistry;	/* maps Xoops to our oops */
int	Xfd;


/*
 *	local declarations
 */

void		skip_padding(), convertIM(), fixOops(), fixContexts(),
		convertBytes(), convertPointers(), readXSizes(), fixPC(),
		zapInitialPCMethod(), zapLPI();

char		*readXStuff();
u_int		xeroxSize();
oop_t		xeroxClass();
oop_t		xeroxSpec();


/*
 *	globals for the Xerox image
 */

u_int		XeroxObjLength, XeroxOTLength;
XeroxWord_t	*XeroxObjs;


/*
 *	if SM_SPACE is defined, this holds the offset into the image file
 *	of the object area.  Otherwise this is a real pointer.
 */

struct XeroxOTE *XeroxOT;

/*
 *	Read, convert, and write the image.
 */

main(argc, argv)
int argc;
char *argv[];
{
	oop_t	convertOop();
	extern long	lseek();

	if (argc != 3) {
		fprintf(stderr,
		    "usage: %s xerox-image-file berkeley-image-file\n",
		    argv[0]);
		fatal("wrong number of args", NullOop);
	}

	/*
	 *	read the image
	 */

	Xfd = xopen(argv[1], 0);

	readXSizes();

#	ifdef SM_SPACE
		XeroxObjs = (XeroxWord_t *)lseek(Xfd, 0L, 1);
		(void)xlseek(Xfd,
		    (long)XeroxObjLength * sizeof(XeroxWord_t),  1);
		skip_padding();
#	else
		XeroxObjs = (XeroxWord_t *)readXStuff("object space",
		    XeroxObjLength * sizeof(XeroxWord_t));
#	endif

	XeroxOT = (struct XeroxOTE *)readXStuff("object table",
	    XeroxOTLength * sizeof(struct XeroxOTE));

	must_be_at_end();


	/*
	 *	Convert it.
	 *
	 *	First convert objects to intermediate format
	 *	where the OOPs are FullRegistry indices.
	 *	Initialize the flags.
	 *	SMC is not set since all original contexts are unreclaimable.
	 *
	 *	Then traverse all OOPs, changing them to pointers.
	 *	Eliminate LargePositiveIntegers at this time,
	 *	and remember all old objects that point to new ones.
	 *
	 *	Next, fixup stored PCs in contexts.
	 *
	 *	Finally, zap CompiledMethod initialPC
	 *	to call a primitive instead because the one in the image assumes
	 *	16 bit words.
	 */

	initMem();
	FullRegistry = (rp_t *)xvalloc(XeroxOTLength * sizeof(rp_t));

	printf("converting to intermediate form\n");
	convertIM();
	printf("removing the OT\n");
	fixOops();
	printf("fixing contexts\n");
	fixContexts();

	/*
	 *	used to:
	 *		printf("zapping InitialPCMethod\n");
	 *		zapInitialPCMethod();
	 *	but don't have to do this with new image.
	 */

	LowSpaceSemaphore = SignalAtTickSemaphore = InputSemaphore
	    = DisplayForm = nilOop;

	printf("writing out Berkeley image files\n");
	dumpToFile(argv[2], (oop_t *)&nilOop);
	exit(0);
}


/*
 *	Read the number of 16-bit words objects and the number of OTEs.  
 *	Rotate the bytes as needed, and fast forward over the padding.
 */

void
readXSizes()
{
	struct	{
		unsigned char	objlen3, objlen2, objlen1, objlen0;
		unsigned char	otlen3, otlen2, otlen1, otlen0;
	} XeroxHeader;

	xread(Xfd, (char *)&XeroxHeader, (u_int)sizeof(XeroxHeader));
	skip_padding();

	XeroxObjLength = XeroxHeader.objlen3 << 24  |  XeroxHeader.objlen2 << 16
	    |  XeroxHeader.objlen1 << 8  |  XeroxHeader.objlen0;

	XeroxOTLength = XeroxHeader.otlen3 << 24  |  XeroxHeader.otlen2 << 16 
	    |  XeroxHeader.otlen1 << 8  |  XeroxHeader.otlen0;
	XeroxOTLength /= sizeof(struct XeroxOTE) / sizeof(XeroxWord_t);

	printf("OT length = %u,  obj space length = %u\n",
	    XeroxOTLength, XeroxObjLength);
}


/*
 *	Reads a bunch of bytes from the Xerox format file.
 */

char *
readXStuff(whatItIs, size)
char *whatItIs;
u_int size;
{
	register char	*bufp;

	printf("reading %s\n", whatItIs);
	bufp = xvalloc(size);
	xread(Xfd, bufp, size);
	skip_padding();
	return (bufp);
}


/*
 *	generator for all Xerox OOPs
 */

# define FOR_ALL_XOOPS(oop)	for (oop = 1;  oop < XeroxOTLength;  oop++)

/*
 *	address of first field of xerox object
 */

# define xFieldAddr(xotep)	(&XeroxObjs[xoffset(xotep) + XeroxFieldsOffset])


/*
 *	First convert objects to intermediate format where the OOPs 
 *	are FullRegistry indices.
 *	Converted oops are first stored in FullRegistry.  After all 
 *	oops in FullRegistry have been converted, they are copied 
 *	to Registry.
 *	Initialize the flags to 0, ensure registered ones are old.
 *	SMC is not set since all original contexts are unreclaimable.
 */

void
convertIM()
{
	register int			xoop;
	register XeroxWord_t		*xop;		/* xerox obj pointer */
	register struct XeroxOTE	*xotep;		/* xerox OT pointer */
	register oop_t			xcls;
	register u_int			xsize, xspec;

	register rp_t			rp;
	register u_int			bsFormat;
	register u_int			bsLength;
	register u_int			bsFlags;
	register u_int			OOPCount;

	FOR_ALL_XOOPS(xoop) {
		xotep = &XeroxOT[xoop];
		if (!xotep->xrc  ||  xf(xotep))		/* not free OTE */
			FullRegistry[xoop] = NullOop;	/* mark free entry */
		else {					/* not free OTE */
			xsize = xeroxSize(xoop);
			xop = xFieldAddr(xotep);
			xcls = xeroxClass(xoop);
			xspec = xeroxSpec(xcls);

			bsFlags = 0;
			if (xcls == classCompiledMethodXoop) {
				oop_t	header;

				bsFormat = COMPILED_METHOD;
				convertPointers(&xop[headerIndexCM],
				    &header, 1);
				OOPCount = objectPointerCountCMH(header);
				bsLength = xeroxSize(xoop)  +   OOPCount
				    * (sizeof(oop_t) - sizeof(XeroxWord_t));
				allocAnObj(&Old, roundUp(bsLength,
				    sizeof(oop_t)) / sizeof(oop_t), rp);
			}
			else if (xp(xotep)) {
				bsFormat = OOPS;
				bsLength = xsize / sizeof(XeroxWord_t);
				switch (xcls) {
				default:
					allocAnObj(&Old, bsLength, rp);
					break;
				case classMethodContextXoop:
				case classBlockContextXoop:
					allocAnObj(&PastSurvivors, bsLength,
					    rp);
					bsFlags |= CONTEXT;
					break;
				}
			}
			else if (isWordsIS(xspec)) {
				bsFormat = BES;
				bsLength = xsize / sizeof(XeroxWord_t);
				allocAnObj(&Old, roundUp(bsLength, 
				    sizeof(oop_t) / sizeof(BE_t)) 
				      / (sizeof(oop_t) / sizeof(BE_t)),
				    rp);
			}
			else /* bytes */ {
				bsFormat = BYTES;
				bsLength = xsize;
				allocAnObj(&Old, roundUp(bsLength,
				    sizeof(oop_t)) / sizeof(oop_t), rp);
			}
			FullRegistry[xoop] = rp;

			ageRp(rp) = 0;
			classRp(rp) = (rp_t)xcls;
			switch (bsFormat) {
			default:
				fatal("rxi: bad format", NullOop);
			case BES:
				convertBytes((char *)xop,
				    (char *)byteFieldRp(rp), xsize);
				break;
			case BYTES:
				convertBytes((char *)xop,
				    (char *)byteFieldRp(rp), xsize);
				break;
			case COMPILED_METHOD:
				convertPointers(xop, oopFieldRp(rp), OOPCount);
				convertBytes((char *)(xop + OOPCount),
				    (char *)&oopFieldRp(rp)[OOPCount],
				      xsize  -  OOPCount * sizeof(XeroxWord_t));
				break;
			case OOPS:
				convertPointers(xop, oopFieldRp(rp), bsLength);
				break;
			}
			fixedFieldsRp(rp) = fixedFieldsIS(xspec);
			flagsRp(rp) = bsFlags;
			formatRp(rp) = bsFormat;
			hashRp(rp) = integerObjectOf(xoop);
			lengthRp(rp) = bsLength;
		}
	}
	for (xoop = firstRegXoop;  xoop <= lastRegXoop;  xoop++)
		Registry[xoop] = rp = FullRegistry[xoop];
}


/*
 *	Traverse all OOPs, changing them to pointers.
 *	Eliminate LargePositiveIntegers at this time,
 *	and remember all old objects that point to new ones.
 */

void
fixOops()
{
	register int	xoop;
	register rp_t	rp;

	/*
	 *	convert largePositiveIntegers to smallInts, 
	 *	and remember rp if it is old.
	 */

#	define fixOopsOperation(src, dst)\
		if (isIntegerObject(src))\
			;\
		else if (TRUE) {\
			(dst) = (src) = (oop_t)FullRegistry[(int)(src)];\
			/*\
			 *	If either Berkeley LPI or Xerox LPI.\
			 */\
			if ((src) <= (oop_t)rp\
			      &&  classRp((rp_t)(src))\
				== classLargePositiveIntegerOop\
			  ||  (int)classRp((rp_t)(src))\
			    ==  classLargePositiveIntegerXoop) {\
				zapLPI((rp_t)(src), &(dst));\
				(src) = (dst);\
			}\
			if (newRp((rp_t)src) && oldRp(rp))\
				remember(rp);\
		}\
		else


	FOR_ALL_XOOPS(xoop) {
		rp = FullRegistry[xoop];
		if (rp != NullOop)
			traverseObject(FALSE, fixOopsOperation);
	}
}


/*
 *	Fix stored PCs in contexts.
 *	Both PC and initialPC have to be fixed if xoop is a block context.
 */

void
fixContexts()
{
	register int	xoop;
	register rp_t	rp;
	register rp_t	class;
	register rp_t 	method;

	FOR_ALL_XOOPS(xoop) {
		rp = FullRegistry[xoop];
		if (rp == NullRp)
			continue;
		if ((class = classRp(rp))  ==  classMethodContextOop)
			fixPC(&ipC(rp), methodMC(rp));
		else if (class == classBlockContextOop) {
			method = methodMC(homeBC(rp));
			fixPC(&ipC(rp), method);
			fixPC(&initIPBC(rp), method);
		}
	}
}


/*
 *	Copy and convert pointers from Xerox to BS format.
 *	The buffer makes it idempotent.
 */

void
convertPointers(srcp, dstp, n)
register XeroxWord_t *srcp;
register oop_t *dstp;
register u_int n;
{
	static XeroxWord_t	*buffp = NULL;
	static u_int	buffLen = 0;

	if (n > buffLen) {
		if (buffp != NULL)
			free((char *)buffp);
		buffp = (XeroxWord_t *)xmalloc(n * sizeof(XeroxWord_t));
		buffLen = n;
	}
#	ifdef SM_SPACE
		(void)xlseek(Xfd, (long)srcp, 0);
		xread(Xfd, (char *)buffp, (u_int)n * sizeof(XeroxWord_t));
#	else
		scopy((short *)srcp, (short *)buffp, n);
#	endif
	srcp = buffp;
	while (n--)
		*dstp++ = convertOop(*srcp++);
}


/*
 *	Bytes don't really need to be converted, just copy it.
 */

void
convertBytes(srcp, dstp, n)
register char *srcp;
register char *dstp;
register u_int n;
{
#	ifdef SM_SPACE
		(void)xlseek(Xfd, (long)srcp, 0);
		xread(Xfd, dstp, n); 
#	else
		bcopy(srcp, dstp, n);
#	endif
}


/*
 *	Fix PC to reflect longer headers.
 *	PCs are relative to beginning of methods.
 *	This happens late enough (pass2) that all compiled methods
 *	are already converted.
 */

void
fixPC(p, method)
oop_t *p;
rp_t method;
{
	register oop_t	oop;

	oop = *p;
	if (isIntegerObject(oop)) {
		convertToInteger(oop);
		oop += objectPointerCountCMH(headerCM(method))
		    *  (sizeof(oop_t) - sizeof(XeroxWord_t));
		*p = integerValueOf(oop);
	}
}


/*
 *	zaps CompiliedMethod initialPC with prototype
 *	that invokes a primitive.  The method in the image is
 *	wordsize dependant.
 *
 *	The new image (jun/83), version 2 doesn't need this zap,
 *	so it is not used anymore.
 */

void
zapInitialPCMethod()
{
	register rp_t	initPCMrp;
	rp_t	findMethod();

	static oop_t prototypeMethod[] = {
		integerObjectOf(0x7002),/* has primitive, 2 literals */
	     	integerObjectOf(207),	/* initialPC Primitive Index */
		integerObjectOf(0),	/* last literal, for super pseudo */
		0, 0};			/* extra bytes of source file address */

	initPCMrp = findMethod(classCompiledMethodOop, "initialPC");
	if (lengthRp(initPCMrp) < lengthof(prototypeMethod))
		fatal("zapInitialPCMethod: method too short", (oop_t)initPCMrp);
	bcopy((char *)prototypeMethod,  (char *)byteFieldRp(initPCMrp), 
	    (u_int)sizeof(prototypeMethod));
}


/*
 *	Find CompiledMethod initialPC so we can zap it.
 *	Actually is more general: finds any method where
 *	class is class and string is selector.
 *	Code was stolen from lookup().
 *	The difference is that, since string is not a symbol,
 *	must do value equivalence instead of pointer equivalence
 *	when searching dictionaries.
 */

rp_t
findMethod(class, string)
rp_t class;
register char *string;
{
	char	buf[BUFSIZ];
	register rp_t	currentClass;
	register rp_t	*curp, *endp, cur;
	register rp_t	dict;

	for (currentClass = class;  currentClass != nilOop;
	    currentClass = superclassCl(currentClass)) {

		dict = messageDictionaryCl(currentClass);

		for (curp = selectorsMD(dict), 
		      endp = &rpFieldRp(dict)[lengthRp(dict)];
		    curp < endp;  ) {
			cur = *curp++;
			if (cur != nilOop  
			    &&  strcmp(string, (char *)byteFieldRp(cur)) == 0)
				return (rpFieldRp(methodArrayMD(dict))
				    [curp - 1 - selectorsMD(dict)]);
		}
	}
	fatal(sprintf(buf, "findMethod could not find ``%s''",
	    string), (oop_t)class);
	/*NOTREACHED*/
	/* fatal() never returns, but this makes lint happy */
	return(0);
}


/*
 *	Convert an OOP from Xerox to BS format.
 *	Assumes that bytes have been swapped.
 *	Moves SmallInteger bit from low bit to high bit.
 *	Not changed if oop is not a SmallInt.
 *	Returns the converted oop.  
 */

oop_t
convertOop(oop)
register XeroxWord_t oop;
{
	register oop_t	res;

	if (sizeof(oop) != sizeof(short))
		fatal("convertOop only works for halfwords", (oop_t)oop);

	res = oop >> 1  &  0x7fff;	/* remove xerox smallInt flag */
	if (!(oop & 1))			/* is real pointer */
		;
	else if (oop & 0x8000)		/* is neg small int */
		res |= 0xffff8000;
	else
		res |= INTEGER_BITS;
	return (res);
}


/*
 *	Portions of Xerox file are padded out to 512-byte boundary.
 *	This routine skips past such padding.
 *
 *	Can simply use:
 * 	(void)xlseek(Xfd, roundUp(xlseek(Xfd, 0L, 1), 512), 0);
 */

void
skip_padding()
{
	long	where;

	where = xlseek(Xfd, 0L, 1);
	where = (where + 511)  &  ~511;
	(void)xlseek(Xfd, where, 0);
}


/*
 *	asserts that we must be at end of xerox file
 */

must_be_at_end()
{
	long	where;

	where = xlseek(Xfd, 0L, 1);
	if (where != xlseek(Xfd, 0L, 2))
		fatal("not at end", NullOop);
}


/*
 *	Returns size of a Xerox object
 */

u_int
xeroxSize(xoop)
int xoop;
{
	register struct	XeroxOTE *otep;
	XeroxWord_t	xlen, *p;

	otep = &XeroxOT[xoop];
	p = &XeroxObjs[xoffset(otep) + XeroxLengthOffset];
#	ifdef SM_SPACE
		(void)xlseek(Xfd, (long)p, 0);
		xread(Xfd, (char *)&xlen, (u_int)sizeof(xlen));
#	else
		xlen = *p;
#	endif
	return ((xlen - XeroxFieldsOffset) * sizeof(XeroxWord_t) 
	    -  (xo(otep) ? 1 : 0));
}


/*
 *	Returns class of Xerox object
 */

oop_t
xeroxClass(xoop)
register int xoop;
{
	register struct XeroxOTE *otep;
	XeroxWord_t	xcls, *p;

	otep = &XeroxOT[xoop];
	p = &XeroxObjs[xoffset(otep) + XeroxClassOffset];
#	ifdef SM_SPACE
		(void)xlseek(Xfd, (long)p, 0);
		xread(Xfd, (char *)&xcls, (u_int)sizeof(xcls));
#	else
		xcls = *p;
#	endif
	return (convertOop(xcls));
}


/*
 *	returns instance spec of Xerox object
 */

oop_t
xeroxSpec(xcls)
oop_t xcls;
{
	register struct XeroxOTE *otep;
	XeroxWord_t	xspec, *xoop;

	otep = &XeroxOT[xcls];
	xoop = &xFieldAddr(otep)[instanceSpecIndexCl];
# 	ifdef SM_SPACE
		(void)xlseek(Xfd, (long)xoop, 0);
		xread(Xfd, (char *)&xspec, (u_int)sizeof(xspec));
#	else
		xspec = *xoop;
#	endif
	return (convertOop(xspec));
}


/*
 *	Collapse LargePositiveInteger to SmallInteger if possible.  
 *	Oop is the LPI and p is the place to clobber.
 */

void
zapLPI(lpi, p)
register rp_t lpi;
register oop_t *p;
{
	register u_int	value = 0, size;

	size = lengthRp(lpi);
	if (size > 4)			/* larger than largePosInt */
		return;

	if (size >= 1)
		value = (u_int)byteFieldRp(lpi)[0];
	if (size >= 2)
		value += (u_int)byteFieldRp(lpi)[1] << 8;
	if (size >= 3)
		value += (u_int)byteFieldRp(lpi)[2] << 16;
	if (size == 4)
		value += (u_int)byteFieldRp(lpi)[3] << 24;

	if (isIntegerValue(value))
		*p = integerObjectOf(value);
}


/*
 *	not used
 */

void
printObjs()
{
	register oop_t	xoop;
	register rp_t	rp;
	char	buf[BUFSIZ];


	FOR_ALL_XOOPS(xoop) {
		rp = FullRegistry[xoop];
		if (rp == NullOop)
			continue;
		(void)asString(buf, (oop_t)rp);
		printf("oop 0x%d: %s\n", xoop, buf);
	}
}


/*
 *	The following stubs replace the display stuff that mem.c uses.
 */

u_int	pollCount;
/*ARGSUSED*/	checkSpaceLimits(acp) rp_t *acp; {}
/*ARGSUSED*/	ggcCursor(saved) BE_t *saved; {}
/*ARGSUSED*/	restoreCursor(saved) BE_t *saved; {}
/*ARGSUSED*/	swapCursor(saved) BE_t *saved; {}
/*ARGSUSED*/	initSnapshotCursor(b, s) BE_t *b, *s; {}
/*ARGSUSED*/	snapshotCursor(b) BE_t *b; {}
/*ARGSUSED*/	CloseWindow(force) boolean force; {}
synchronousSignal() {}
smashCache() {}
