static char rp_h[] = "$Header: rp.h,v 41.2 84/05/15 23:35:22 hilfingr Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 *
 *
 *	In other words, the object header and format.
 */
/* $Log:	rp.h,v $
 * Revision 41.2  84/05/15  23:35:22  hilfingr
 * Added definition of b2s_..._defined to aid in conditional compilations.
 *  */
#define b2s_rp_defined


/*
 *	Oop_t is either a pointer to an object (header), or a smallInteger.
 *	Int is the more general type.
 *	BE_t is a bitmap element, or 16-bit word for Smalltalk.
 *	byte_t is an unsigned byte for bytecodes and strings.
 *
 *	NullOop is the illegal Oop, used for failure reporting and 
 *	list termination.
 */

typedef int	oop_t;
typedef u_short	BE_t;
# define bitsPerBE		16
typedef u_char	byte_t;

# define NullOop		((oop_t)0)	/* for terminating lists */
# define NullRp			((rp_t)0)	/* for terminating lists */

/*
 *	Flags:  each object has a flag word.  All flags are initialized 
 *		to zero and cleared when object is freed.
 *
 *		REMEMBERED	This is an old object that points to a new one.
 *				Set when a new OOP is stored into an old object
 *				(ggcStore).  Cleared at beginning of GGC and GC.
 *				Recomputed by GGC and GC.
 *		MARKED		Used during garbage collection to avoid looping.
 *		SMC		unreferenced (reusable) small method context
 *		CONTEXT		is a context
 *		FORWARDED	object has been moved, see forwardingRp().
 *				This flag is only seen in the midst of
 *				memory ops like ggc, promotion, etc.
 */


/*
 *	Here is the header structure.  Rp is a "Real Pointer", because 
 *	it is not an integer pointer.  The structure here is private.  
 *	C source files should use access macro below.
 */

typedef struct rp {
	oop_t		rp_class;	/* points to this object's class */
	u_int		rp_length;	/* number of fields of whatever kind */
	byte_t		rp_flags;
#			define MARKED			'\1'
#			define REMEMBERED		'\2'
#			define SMC			'\4'
#			define CONTEXT			'\10'
#			define FORWARDED		'\20'
	byte_t		rp_age;		/* number of scavenges survived */
	byte_t		rp_format;
#			define OOPS		'\0'
#			define BYTES		'\1'
#			define BES		'\2'
#			define COMPILED_METHOD	'\3'
	byte_t		rp_fixedFields;	/* number of fixed fields in object */
	oop_t		rp_hash;	/* for dicts, esp. methodDicts */
	union {
		oop_t		rp_oop[1];
		BE_t		rp_BE[1];
		byte_t		rp_byte[1];
	} rp_data;
} *rp_t;


/*
 *	accessing macros - returns various fields of the structure
 *	pointed to by "rp".
 */

/*
 *	same thing.  Both returns the pointer to the class of this
 *	object.
 */

# define classRp(rp)			(*(rp_t *)&(rp)->rp_class)
# define forwardingRp(rp)		(*(rp_t *)&(rp)->rp_class)


/*
 *	macros for accessing other fields.
 */

# define ageRp(rp)			((rp)->rp_age)
# define fixedFieldsRp(rp)		((rp)->rp_fixedFields)
# define flagsRp(rp)			((rp)->rp_flags)
# define formatRp(rp)			((rp)->rp_format)
# define hashRp(rp)			((rp)->rp_hash)
# define lengthRp(rp)			((rp)->rp_length)


/*
 *	macros for accessing the union fields
 */
# define BEFieldRp(rp)			((rp)->rp_data.rp_BE)
# define byteFieldRp(rp)		((rp)->rp_data.rp_byte)
# define oopFieldRp(rp)			((rp)->rp_data.rp_oop)
# define rpFieldRp(rp)			((rp_t *)oopFieldRp(rp))


/*
 *	Gets end of an object.  Rp points to the object, and a pointer 
 *	to the end of the object is returned in the argument endOop.
 *	The returned pointer (endOop) will have the type endType.
 */

# define endRp(rp, endOop, endType)\
	if (TRUE) \
		switch (formatRp(rp)) {\
		case OOPS:\
			endOop = (endType)&oopFieldRp(rp)[lengthRp(rp)];\
			break;\
		case BES:\
			endOop = (endType)&BEFieldRp(rp)[roundUp(lengthRp(rp),\
			      sizeof(oop_t)/sizeof(BE_t))];\
			break;\
		case BYTES:\
		case COMPILED_METHOD:\
			endOop = (endType)&byteFieldRp(rp)[\
			    roundUp(lengthRp(rp),\
			    sizeof(oop_t)/sizeof(byte_t))];\
			break;\
		default:\
			fatal("endRp: bad format", NullOop);\
			break;\
		}\
	else


/*
 *	Returns the number of words between address "a" and "b".
 *	Faster, but "a" must be >= "b".
 */

# define wordPointerDifference(a, b)	((int)(a) - (int)(b)  >>  2)
