static char rcsId[] = "$Header: sends.c,v 41.1 84/03/14 20:04:53 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Implements general method send surbroutine.
 *	Called from executeMethod.[ch] and executeBlock.[ch].
 */

# include	<stdio.h>

# include	"utils.h"
# include	"mem.h"
# include	"prims.h"
# include	"execute.h"
# include	"mc.h"
# include	"primTable.h"


static void	nilStack();


/*
 *	methodSend performs the send, returning the new active context.
 */

rp_t
methodSend(ac, sel, nargs, super)
register rp_t ac;			/* old active context */
rp_t sel;
register u_int nargs;
boolean super;				/* true for super send */
{
	register rp_t	nc;		/* new context */
	register oop_t	*rp;
	register oop_t	rcvr;
	rp_t	class;
	struct methodCache	*cs;

	rp = fetchSP(ac) - nargs - 1;
	rcvr = *rp;

	/*
	 *	Get class of receiver.  
	 *	Receiver's class will be the class of its super class
	 *	if ``super'' is true.
	 */

	if (super) {
		register rp_t	method;

		method = isBlock(ac)  ?  methodMC(homeBC(ac))  :   methodMC(ac);
		class = superclassCl((rp_t)valueAs(
		    (rp_t)literalFrameCM(method)[
			literalCountCMH(headerCM(method)) - 1]));
		if (!isIntegerObject(rcvr)  &&  flagsRp((rp_t)rcvr) & SMC)
			nilStack((rp_t)rcvr);
	}
	else if (isIntegerObject(rcvr))
		class = classSmallIntegerOop;
	else {
		register rp_t	rcvrrp;

		rcvrrp = (rp_t)rcvr;
		class = classRp(rcvrrp);
		if (flagsRp(rcvrrp) & SMC)
			nilStack(rcvrrp);
	}


	/*
	 *	First do probe into method cache, call lookup to fill cache
	 *	slot on miss.  If lookup fails, send 
	 *	self  doesNotUnderstandSelector: Message  selector: selector 
	 *		arguments: (Array with: arg1 with: arg2 ...)
	 */

	cs = cacheSlot(class, sel);
	if (cs->mc_class != class  ||  cs->mc_sel != sel)
		if (!lookupMethodInClass(class, (oop_t)sel)) {

			/*
			 *	Method not found.
			 */

			rp_t	argumentArray, message;

			if (sel == doesNotUnderstandSelectorOop)
				fatal("recursive not understood error", rcvr);
			else if (isIntegerObject(sel))
				fatal("integer selector sent to", rcvr);
			argumentArray = instantiateClass(classArrayOop, nargs);
			copy_words(rp + 1,  oopFieldRp(argumentArray), nargs);

			message = instantiateClass(classMessageOop, 0);
			selectorMe(message) = sel;
			argumentsMe(message) = argumentArray;

			++rp;		/* point to first argument */
			*rp++ = (oop_t)message;
			storeSP(ac, rp);

			/*
			 *	This is a recursive call.  But 
			 *	the test above catches infinite recursion.
			 */

			return (methodSend(ac, 
			    doesNotUnderstandSelectorOop, 1, FALSE));
		}

	/*
	 *	Method is found/installed in method cache.  
	 *	cs now points to the cached method context.
	 */

	/*
	 *	Store SP anticipating short send or successful primitive
	 */

	storeSP(ac, rp + 1);
	switch (cs->mc_whatKind) {
	case RET_SELF:
		return (ac);

	case RET_INST:
		*rp = oopFieldRp((rp_t)rcvr)[cs->mc_index];
		return (ac);

	case PRIMITIVE:

		/*
		 *	Calls primitive methods using dispatch table
		 *	to do the mapping.
		 */

		nc = (*primitiveDispatchTable[cs->mc_index])(rp, ac, nargs);
		if (nc != NullOop)
			return (nc);
		break;

	case FULL_SEND:
		break;

	default:
		fatal("methodSend: bad kind", NullOop);
		/* NOTREACHED */
	}


	/*
	 *	full send.  First step: get a context.
	 */

	storeSP(ac, rp);		/* fix SP in old context */
	if (cs->mc_large)
		nc = instantiateClass(classMethodContextOop,
		    largeContextLength - fixedFieldsC);
	else if (isIntegerObject(FirstFreeContext))	/* no free context */
		nc = instantiateClass(classMethodContextOop,
		    smallContextLength - fixedFieldsC);
	else {
		register oop_t	*p, *endp;

		nc = FirstFreeContext;
		FirstFreeContext = nextSMC(nc);

		/*
		 *	if reusing one, must nil out temps
		 */

		for (p = &tempFrameC(nc)[nargs],  
			endp = &tempFrameC(nc)[cs->mc_tempCount];
		    p < endp;)
			*p++ = (oop_t)nilOop;
	}


	/*
	 *	Set up header of new context
	 */

	senderMC(nc) = ac;
	storePC(nc, cs->mc_method, cs->mc_pc);
	spC(nc) = integerObjectOf(cs->mc_tempCount);
	methodMC(nc) = cs->mc_method;
	receiverMC(nc) = rcvr;

	/*
	 *	Copy args, and return new context.
	 */

	copy_words(&rp[1], tempFrameC(nc), nargs);
	return (nc);
}


/*
 *	Nil above top of stack when rcvr is a context.
 *	Src is a nilOop in a register for speed.
 */

static void
nilStack(con)
register rp_t con;
{
	register oop_t	*dstp, *endp;
	register rp_t	src;

	src = nilOop;
	dstp = fetchSP(con);
	endp = &oopFieldRp(con)[lengthRp(con)];
	while (dstp < endp)
		*dstp++ = (oop_t)src;
}
