static char si_h[] = "$Header: si.h,v 41.1 84/03/14 20:04:59 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	SmallInteger macros
 *
 *	The representation for SmallInteger in BS is different from 
 *	that described in Goldberg and Robson.
 *	Checking for SmallIntegers is faster in this representaion.
 *
 *	If the high bit is set, the lower 31 bits represents a 31-bit
 *	integer in 2's complement.  In other words, the sign bit is 
 *	the 2nd bit from the left.
 *
 *	These macros are crufty, but fast.
 */


/*
 *	The leftmost bit indicates whether the pointer is an
 *	integer object, and the second bit from the left is the 
 *	sign bit if the pointer is and integer object.  
 *	These are the masks for the two bits.
 */

# define INTEGER_BITS	0x80000000	/* integer bit mask */
# define NEGATIVE_BIT	0x40000000	/* sign bit mask */


/*
 * 	Returns integer value of "oop".  If "oop" is negative, the
 *	high order bit (integer bit) need not be removed, since it
 *	serves as the extended sign bit for the 2's complement number.
 */

# define integerValueOf(oop) \
	((oop_t)(oop) & NEGATIVE_BIT \
		?  (oop_t)(oop) 	/* oop is negative */ \
		: nonNegIntegerValueOf((oop_t)(oop))) /* oop is non-negative */


/*
 *	Converts "oop" to a 32-bit 2's complement integer.  The integer
 *	bit (leftmost bit) will be removed only when the integer represented
 *	by the pointer is non-negative.  The integer bit will not be removed
 *	if the number represented is negative since the sign bit has to 
 *	be sign extended there anyway.
 */

# define convertToInteger(oop) \
	if (!((oop) & NEGATIVE_BIT)) \
		oop &= ~INTEGER_BITS;\
	else


# define convertToNonNegInteger(oop) \
	oop &= ~INTEGER_BITS


# define nonNegIntegerValueOf(oop)	((oop_t)(oop) - INTEGER_BITS)


# define integerObjectOf(value)		((value) | INTEGER_BITS)


/*
 *	Converts the integer value "value" to a pointer having the same
 *	integer value.
 */

# define convertToObject(value)		((value) |= INTEGER_BITS)


/*
 *	Tests whether "oop" is a SmallInteger by checking its sign bit.
 *	Returns true if "oop" is a SmallInteger.
 */

# define isIntegerObject(oop)		((int)(oop) < 0)


/*
 *	Tests whether "value" is a legal integer value.  Legal values
 *	are 31-bit 2's complement numbers.  In other words, 
 *	the two high order bits should be either both 0's or both 1's.
 */

# define isIntegerValue(value) \
	(!((value) & (INTEGER_BITS | NEGATIVE_BIT))\
	    ||  (u_int)(value)  >=  (u_int)(INTEGER_BITS | NEGATIVE_BIT))


/*
 *	Increment the SmallInteger.
 */

# define incrementSmallInteger(oop)	/* fast way */\
	++((oop_t)(oop))		/* TRICKY */


/*
 *	Add "delta" to "oop".  
 */

# define addToSmallInteger(oop, delta)\
	((oop_t)(oop) += (delta))	/* TRICKY */


# define immediate(oop) /* oop is immed ptr */\
	isIntegerObject(oop)		/* nil, true, false are real objects */
