static char rcsId[] = "$Header: snapshot.c,v 42.3 84/07/25 15:40:25 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Contains the heart of swoop; the memory system.
 */

# include	<stdio.h>
# include	<sys/file.h>
# include	<sys/vadvise.h>

# include	"utils.h"
# include	"mem.h"
# include	"prims.h"
# include	"execute.h"
# include	"cursor.h"


/*
 *	Fix pointers.  Used only in translate() to adjust pointer
 *	difference due to different base address.  See translate()
 *	for a more detailed description.
 */

# define fixup(from, to)\
	if ((u_int)(from) < (u_int)dividingLine)\
		(to) = (from) + sdelta;\
	else if (isIntegerObject(from)) \
		(to) = (from);\
	else if (TRUE)\
		(to) = (from) + odelta;\
	else


static	void	spaceTransput(), seekToNextBlock(), translate();
void		xread(), xwrite();

rp_t		Registry[lastRegXoop + 1];	/* this file needs size */


/*
 *	header of snapshot file
 */

struct snapHeader {
	struct space	sh_survivor, sh_old;
	u_int		sh_PastSurvivorIndex;
	u_int		sh_NextRemembered;
	rp_t		sh_Registry[lengthof(Registry)];
	rp_t		sh_Remembered[lengthof(Remembered)];
};


/*
 *	Prints the string "s" to the standard output and flushes it.
 */

# define	flushString(s)\
	if (TRUE) {\
		printf(s);\
		(void)fflush(stdout);\
	}\
	else



/*
 *	reading, writing, initing images
 */

/*
 *	save a snapshot on a file.  The caller must have saved the 
 *	Oop of the active context in the suspended context field 
 *	of the appropriate process object.  Uses `raw' system calls 
 *	since the stdio calls buffer their arguments, which blows 
 *	the Sun's memory limits.
 */

dumpToFile(fname, acp)
char fname[];
register oop_t *acp;			/* active context pointer */
{
	register int fd;
	struct snapHeader sh;

	fd =  open(fname, O_WRONLY | O_CREAT | O_TRUNC, 0640);
	if (fd < 0) {
		char buf[BUFSIZ];

		warning(sprintf(buf,
		    "dumpToFile: cannot open file <%s> for snapshot",
		    fname));
		return;
	}

	/*
	 * 	Garbage collect before taking snapshot.
	 */

	ggc((rp_t *)acp);

	/*
	 * 	Save spaces, Remembered and Registry into 
	 *	snapshot header.
	 */

	sh.sh_survivor = PastSurvivors;
	sh.sh_old = Old;
	sh.sh_NextRemembered =  wordPointerDifference(NextRemembered,
	    Remembered);
	wcopy((long *)Remembered, (long *)sh.sh_Remembered,
	    sh.sh_NextRemembered);
	wcopy((long *)Registry, (long *)sh.sh_Registry, lengthof(Registry));
	sh.sh_PastSurvivorIndex = PastSurvivorIndex;

	/*
	 *	Write snapshot Header to file then do snapshotting.
	 */

	xwrite(fd, (char *)&sh, sizeof(sh));
	vadvise(VA_SEQL);
	spaceTransput(xwrite, fd, &PastSurvivors, TRUE);
	spaceTransput(xwrite, fd, &Old, TRUE);
	vadvise(VA_NORM);

	(void)close(fd);
	sync();
}


/*
 *	Read a snapshot from a file.  The snapshot header is read first,
 *	and restored to the Registry, PastSurvivors and Old spaces.
 *	Then the PastSurvivors and Old space are read.
 *	
 */

loadFromFile(filename)
char filename[];
{
	struct snapHeader sh;
	int fd;

	initMem();
	flushString("reading image");
	if ((fd = open(filename, O_RDONLY, 0)) < 0) {
		char buf[BUFSIZ];

		(void)sprintf(buf, "loadFromFile cannot open %s", filename);
		fatal(buf, NullOop);
	}

	xread(fd, (char *)&sh, (u_int)sizeof(sh));
	NextRemembered = &Remembered[sh.sh_NextRemembered];
	wcopy((long *)sh.sh_Remembered,  (long *)Remembered, 
	    sh.sh_NextRemembered);
	wcopy((long *)sh.sh_Registry, (long *)Registry, lengthof(Registry));

	PastSurvivorIndex = sh.sh_PastSurvivorIndex;
	PastSurvivors.sp_next = sh.sh_survivor.sp_next -
	    sh.sh_survivor.sp_bottom + PastSurvivors.sp_bottom;
	if (PastSurvivors.sp_next > PastSurvivors.sp_top)
		fatal("loadFromFile: PastSurvivors space overflow", NullOop);

	Old.sp_next = sh.sh_old.sp_next - sh.sh_old.sp_bottom + Old.sp_bottom;
	if (Old.sp_next > Old.sp_top)
		fatal("loadFromFile: Old space overflow", NullOop);

	vadvise(VA_SEQL);
	spaceTransput(xread, fd, &PastSurvivors, FALSE);
	spaceTransput(xread, fd, &Old, FALSE);
	vadvise(VA_NORM);
	putchar('\n');

	(void)close(fd);

	translate(&sh);

	LowSpaceSemaphore = SignalAtTickSemaphore = InputSemaphore   
	    = DisplayForm = nilOop;
}


/*
 *	Reads/writes an area of virtual memory (pointed to by sp) 
 *	to/from file.  Copying always starts from block boundary.
 *	During snapshotting, the cursor will be changed to a bucket.
 *
 *	Transfers are done in blocks of transferSize characters.
 *	and one more bit will be added to the bucket cursor
 *	if updateCursor is TRUE.
 */

static void
spaceTransput(operation, fd, sp, updateCursor)
void (*operation)();			/* xread or xwrite */
int fd;
register struct space *sp;
boolean updateCursor;
{
	/*
	 *	Number of characters transfered for each xread or xwrite
	 */

# 	define	transferSize	(50 * BUFSIZ)

	/*
	 *	read/write "n" bytes from/to memory, then adds one more
	 *	dot to the bucket cursor or flushes a period "." to 
	 *	standard output depending on updateCursor.
	 */

#	define	transput(n)\
		if (TRUE) {\
			(*operation)(fd, (char *)sp->sp_next - size, n);\
			if (updateCursor)\
				snapshotCursor(bucketCursor);\
			else\
				flushString(" .");\
		}\
		else


	u_int	size;			/* total no. of chars to transfer */
	BE_t	bucketCursor[cursorBitmapLength];
	BE_t	savedCursor[cursorBitmapLength];

	if (updateCursor)
		initSnapshotCursor(bucketCursor, savedCursor);
	seekToNextBlock(fd);		/* find next block boundary */
	for (size = (char *)sp->sp_next - (char *)sp->sp_bottom;
	    size > transferSize; size -= transferSize)
		transput(transferSize);
	transput(size);
	if (updateCursor)
		restoreCursor(savedCursor);
}


# ifndef L_SET
#	define L_SET 0
# endif

# ifndef L_INCR
#	define L_INCR 1
# endif

/*
 *	Move the read/write pointer to the beginning of the next 
 *	block (where BUFSIZ is the block size).  roundUp() converts 
 *	the current location to a multiple of BUFSIZ.
 */
static void
seekToNextBlock(fd)
int fd;
{
	extern long	lseek();

	(void)lseek(fd, (long)roundUp(lseek(fd, 0L, L_INCR), BUFSIZ), L_SET);
}


/*
 *	Used when reading snapshot to compensate for different
 *	base addresses.  Traverse Registry, Remembered and spaces,
 *	to fix pointers.
 *
 *	For the Registry, if the pointer is a survivor, add sdelta
 *	to it.  If it is an integer object, do not chanage it.
 *	Otherwise, add odelta.
 *
 *	For the Remembered array, increment each pointer by odelta.
 *
 *	For spaces (PastSurvivors and Old), fix pointers which points
 *	to either PastSurvivors or Old.
 */

static void
translate(shp)
struct snapHeader *shp;			/* snapshot header */
{
	/*
	 *	computes differnece between two spaces.
	 */
# 	define spaceDelta(from, to)\
		((char *)(to)->sp_bottom - (char *)(from)->sp_bottom)

	oop_t	*p;
	register rp_t	rp;
	register u_int 	sdelta; 	/* difference between PastSurvivors */
	register u_int	odelta;		/* difference between Old spaces */
	register oop_t	dividingLine;

	sdelta = spaceDelta(&shp->sh_survivor, &PastSurvivors);
	odelta = spaceDelta(&shp->sh_old, &Old);
	dividingLine = (oop_t)shp->sh_old.sp_bottom;

	if (sdelta == 0 && odelta == 0) {
		flushString("no translation needed.\n");
		return;
	}

	flushString("translating .");
	for (p = (oop_t *)Registry; 
	     p < (oop_t *)&Registry[lastRegXoop + 1]; p++)
		fixup(*p, *p);

	flushString(" .");
	for (p = (oop_t *)Remembered; p < (oop_t *)NextRemembered;)
		*p++ += odelta;

	flushString(" .");
	vadvise(VA_SEQL);
	for (rp = (rp_t)PastSurvivors.sp_bottom; 
	    rp < (rp_t)PastSurvivors.sp_next; )
		traverseObject(TRUE, fixup);

	flushString(" .");
	for (rp = (rp_t)Old.sp_bottom; rp < (rp_t)Old.sp_next; )
		traverseObject(TRUE, fixup);
	vadvise(VA_NORM);

	flushString(" done\n");
}
