static char rcsId[] = "$Header: sunprim.c,v 44.4 84/07/26 01:17:00 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/* 
 *	Routines to manipulate Sun's display.  Consists of routines to
 *	initialize display, copy and write bitmaps to the display.
 */

# include	<sys/mman.h>
# include	<sys/ioctl.h>

# include	<fcntl.h>
# include	<sun/fbio.h>
# include	<sunwindow/window_hs.h>
# include	<pixrect/pixrect_hs.h>
# include	"framebuf.h"

# include	<ctype.h>
# include	<stdio.h>
# include	<signal.h>

# include	"utils.h"
# include	"io.h"
# include	"rp.h"
# include	"cursor.h"


int		GXBase;
long		GXConfig;

int		BSWindowFd;
struct pixwin	*BSPixwin;


# ifdef RELEASE1_0
#	define	DisplayDevice	"/dev/bw0"
# else
#	define	DisplayDevice	"/dev/bwone0"
# endif



/* 
 *	Allocate memory for bit map.
 *	Open the console, the link to the graphics device 
 *	Map the bit map into memory.
 *	to the bit map's base address.
 *
 *	This takes advantage of the fact that Sun 2 uses the same graphics
 *	board as Sun 1.5.  Should be changed to pixrects to accommodate 
 *	future hardware releases.
 */

void
SunInitialize()
{
	int		fd, v;
	char		*Base, msg[100];

	extern int	errno;

	Base = xvalloc((u_int)GXaddrRange);
	fd = xopen(DisplayDevice, 2);
	if ((v = mmap(Base,
		      GXaddrRange,
		      PROT_WRITE | PROT_READ, MAP_SHARED,
		      fd, 0)) != 0) {
		(void)sprintf(msg, "SunInitailize: mmap failed = %d, errno= %d",
			    v, errno);
		fatal(msg, NullOop);
	}

	GXBase = (int)Base;
	GXConfig = GXCPresent;
	GXcontrol = GXvideoEnable;

	BSWindowFd = nullFd;
	BSPixwin = (struct pixwin *)NULL;
}


/*
 *	Parameters for displayScreen will be read from argv.
 *	Default values will be used if none is provided.
 */

void
OpenWindow(argv)
char **argv;
{
	char			on = 1;
	static char		**prevArgv;
	struct inputmask	bsInputMask;
#	ifdef RELEASE1_0
		static struct screen    displayScreen
		    = { DisplayDevice, FBTYPE_SUN1BW,  TRUE };
#	else
		static struct screen	displayScreen;
#	endif
		
	sigset(SIGWINCH, SIG_IGN);
#       ifdef RELEASE1_0
		if ((BSWindowFd = win_getnewwindow()) == NULL)
			fatal("OpenWindow: cannot get new window.", NullOop);
		win_screennew(BSWindowFd, &displayScreen);
#	else
		/*
		 *	prevArgv remembers what OpenWindow() was called
		 *	with so that it know what to do when reopening it
		 *	after a suspend.
		 */

		if (argv != NULL)
			prevArgv = argv;
		else
			argv = prevArgv;
		if (win_initscreenfromargv(&displayScreen, argv) < 0)
			fatal("OpenWindow: argument error.", NullOop);
		if ((BSWindowFd = win_screennew(&displayScreen)) == NULL)
			fatal("OpenWindow: cannot open screen.", NullOop);
#	endif

	input_imnull(&bsInputMask);

	if (EncodedKeyboard)
		bsInputMask.im_flags  |=  IM_ASCII | IM_NEGEVENT;
	else
		bsInputMask.im_flags  |=  IM_UNENCODED | IM_NEGEVENT;

	win_setinputcodebit(&bsInputMask, MS_LEFT);
	win_setinputcodebit(&bsInputMask, MS_MIDDLE);
	win_setinputcodebit(&bsInputMask, MS_RIGHT);
	win_setinputcodebit(&bsInputMask, LOC_MOVE);

	win_setinputmask(BSWindowFd, &bsInputMask, &bsInputMask, WIN_NULLLINK);

	ioctl(BSWindowFd, (int)FIONBIO, (char *)&on);

	BSPixwin = pw_open(BSWindowFd);
	InitCursor();
}


/*
 *	Closes Sun Windows so that output will be directed back 
 *	to /dev/console.  The argument force is ignored.
 */

void
CloseWindow(force)
{
	if (BSPixwin != (struct pixwin *)NULL) {
		pw_close(BSPixwin);
		BSPixwin = (struct pixwin *)NULL;
	}
	if (BSWindowFd != nullFd) {
		close(BSWindowFd);
		BSWindowFd = nullFd;
	}
}



/*
 *	Copies the bitmap from "map" to coordinates (x,y) on sun's display.  
 *	"map" should be a 16x16 bitmap.  This is used for changing the cursor.
 */

void
WriteCursor(x, y, map, paintingNotRestoring)
int x, y;
short map[];
boolean paintingNotRestoring;
{
	register int	*intMap;
	register int	*yStart;

	GXfunction = paintingNotRestoring ? GXpaintInverted : GXcopy;
	GXwidth = bitsPerBE;

	/*
	 *	Sets graphics board's x-register to x.  Data value 1 is ignored.
	 */
	((short *)(GXBase + GXselectX))[x] = 1;


	/*
	 *	Unrolled loop for speed.
	 *	Bitmap is 16 words long and 16 bits wide.  Funny stuff
	 *	with 32 bit instructions.
	 */

	yStart = (int *)
	    &((short *)(GXBase + GXupdate + GXsource + GXselectY))[y];

	intMap = (int *)map;

	*yStart++ = *intMap++;  *yStart++ = *intMap++;
	*yStart++ = *intMap++;  *yStart++ = *intMap++;
	*yStart++ = *intMap++;  *yStart++ = *intMap++;
	*yStart++ = *intMap++;  *yStart = *intMap;
}


/*
 *	Reads the cursor at position (x, y) on the sun's display and 
 *	saves it in map.  Used in cursor.c to save the current cursor.
 */
 
void
ReadCursor(x, y, map)
int x, y;
register short *map;
{
	register int *yStart;
	register int *intMap;
	short dummy;

	GXwidth = bitsPerBE;
	*((short *)(GXBase + GXselectX) + x) = 1;

	yStart = (int *)((short *)(GXBase + GXsource + GXselectY) + y);

	/*
	 *	The first assignment statement doesn't really copy 
	 *	anything, it is used for initializing the hardware.
	 */

	intMap = (int *)map;

	dummy = *(short *)yStart;
	yStart = (int *)((short *)yStart + 1);

	*intMap++ = *yStart++;	*intMap++ = *yStart++;
	*intMap++ = *yStart++;	*intMap++ = *yStart++;
	*intMap++ = *yStart++;	*intMap++ = *yStart++;
	*intMap++ = *yStart++;

	((short *)intMap)[0] = *(short *)yStart;
	((short *)intMap)[1] = *(short *)yStart;
}
