static char rcsId[] = "$Header: sysPrims.c,v 44.3 84/07/25 20:09:40 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	System Primitives.  This file contains the primitives that are
 *	implemented in the interpreter for efficiency.  Some of the 
 *	arguments to the primitives will not be used, but have been 
 *	included for compatibility.
 */

# include	<stdio.h>
# include	<math.h>
# include	<sgtty.h>
# include	<signal.h>
# include	<ctype.h>

# include	<sys/types.h>
# include	<sys/vtimes.h>
# include	<sys/timeb.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"prims.h"
# include	"mc.h"


/*
 *	Smalltalk implementation of the same primitive.
 *
 *	String replaceFrom: start to: stop with: aString
 *		startingAt: repStart 
 *
 *		"This destructively replaces elements from start to stop in the
 *		receiver starting at index, repStart, in the byte array,
 *		aByteArray.  Answer the receiver."
 *
 *		| index repOff characterTable |
 *		<primitive: 105>
 *		repOff _ repStart - start.
 *		index _ start - 1.
 *		[(index _ index + 1) <= stop]
 *			whileTrue: 
 *				[self at: index
 *				 	put: (aString at: repOff + index)]
 */

rp_t
StringReplPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	/*
	 *	argument offsets on stack
	 */

#	define fromArg		1
#	define toArg		2
#	define withArg		3
#	define startingAtArg	4
#	define argCount		4

	register rp_t	rcvr, with;
	register oop_t	from, to, startingAt;


	/*
	 *	get arguments
	 */

	rcvr = (rp_t)*rp;

	getInt(rp[fromArg], from, NullRp);
	getInt(rp[toArg], to, NullRp);
	getInt(rp[startingAtArg], startingAt, NullRp);

	with = (rp_t)rp[withArg];
	if (isIntegerObject((oop_t)with)  ||  formatRp(with) != BYTES  
	    || fixedFieldsRp(with))
		return (NullOop);

	/*
	 *	convert to zero origin (for C arrays) and makes sure 
	 *	that bounds are positive.
	 */

	if (from-- < 1  ||  to-- < 1  ||  startingAt-- < 1)
		return (NullOop);

	/*
	 *	bounds check
	 */

	if (from > to)
		;
	else if (to  >=  lengthRp(rcvr)) {
		warning("stringReplace: starting pos. out of bounds");
		return (NullOop);
	}
	else if (startingAt + to - from  >=  lengthRp(with)) {
		warning("stringReplace: replacement too short");
		return (NullOop);
	}
	else
		bcopy((char *)&byteFieldRp(with)[startingAt], 
		    (char *)&byteFieldRp(rcvr)[from],
		    (u_int)(to - from + 1));

	return (ac);
}


/*
 *	SystemDictionary suspend
 *	Suspends BS.
 */

/*ARGSUSED*/
rp_t
SDSuspendPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	/*
	 *	Return failure if we cannot suspend bs for any reason.
	 */

	if (!canSuspend())
		return(NullOop);

	OnSusp();
	return(ac);
}


/*
 *	SystemDictionary quit
 *	Sets halt flag.
 */

/*ARGSUSED*/
rp_t
SDQuitPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	halt(QuitMsg);
	return (ac);
}


/*
 *	SystemDictionary exitToDebugger
 */

/*ARGSUSED*/
rp_t
SDExitPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	warning("exitToDebugger invoked\n");
	abort();
}


/*
 *	SystemDictionary reorganize
 */

/*ARGSUSED*/
rp_t
SDReorgPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	reorganize(&ac);
	return (ac);
}


/*
 *	Behavior flushCache
 *	Must flush method cache when adding a method.
 */

/*ARGSUSED*/
rp_t
BehavFlsCchPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	smashCache();
	return (ac);
}


/*
 *	Flushes method cache, i.e. setting all oops in method cache to null.
 */

void
smashCache()
{
	register struct methodCache	*p, *endp;

	for (p = MethodCache,  endp = &MethodCache[cacheSize];
	    p < endp; ++p)
		p->mc_class = NullOop;
}
