static char rcsId[] = "$Header: timing.c,v 41.2 84/03/29 21:10:58 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Code for BS timing commands.
 */

# include	<stdio.h>
# include	<sys/types.h>
# include	<sys/time.h>
# include	<sys/resource.h>

# include	"utils.h"
# include	"mem.h"
# include	"execute.h"


/*
 * 	For timing stats:  *Old are from last split, *New are current.
 */

# define 	USEC	((double) 1000000.0)

static	struct	rusage	ruOld, ruNew;
static	struct	timeval	tpOld, tpNew;
static	struct	timezone tzp;


/*
 *	timingSplitCycle remembers when the last timing reset was done
 */

u_int		cycle;
static	u_int	timingSplitCycle;


/*
 *	Find the amount of resource used initially.  Called at the
 *	beginning so that we can compute time used.
 */

void
initTimeData()
{
	(void)getrusage(RUSAGE_SELF, (struct rlimit *)&ruOld);
	(void)gettimeofday(&tpOld, &tzp);
	timingSplitCycle = cycle;
}


/*
 *	Find user time, system, real and idle time spend so far.
 *	Called to print timing statistics when BS terminates.
 */

void
reportTimeData()
{
	double	us, ss, rs, is;
	register u_int	cyclesSinceSplit;

	(void)getrusage(RUSAGE_SELF, (struct rlimit *)&ruNew);
	us = (double)(ruNew.ru_utime.tv_sec - ruOld.ru_utime.tv_sec) +
	    (double)(ruNew.ru_utime.tv_usec - ruOld.ru_utime.tv_usec)  /  USEC;
	ss = (double)(ruNew.ru_stime.tv_sec - ruOld.ru_stime.tv_sec) +
	    (double)(ruNew.ru_stime.tv_usec - ruOld.ru_stime.tv_usec)  /  USEC;

	(void)gettimeofday(&tpNew, &tzp);
	rs = (double)(tpNew.tv_sec - tpOld.tv_sec)   +
	    (double)(tpNew.tv_usec - tpOld.tv_usec) / USEC;

	is = rs - us - ss;		/* Idle time */

	cyclesSinceSplit = cycle - timingSplitCycle;
	printf("%d bytecodes, %.3f u-sec, %.3f s-sec, %.3f real sec, %.3f idle\r\n",
	    cyclesSinceSplit,  us,  ss,  rs,  is);

	if (cyclesSinceSplit == 0) {
		warning("no bytecodes counted since timing initialization");
		return;
	}
	printf("%.0f bc/u-s, %.0f bc/CPU-s, %.0f bc/real-sec\r\n",
		(double)cyclesSinceSplit / us,
		(double)cyclesSinceSplit / (us + ss),
		(double)cyclesSinceSplit / rs);
}
