static char rcsId[] = "$Header: toolBitblt.c,v 1.2 84/07/25 22:25:51 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	This file contains implementation of the class Bitblt described
 *	in the Book.  Most of the routines are direct translation of 
 *	those written in Smalltalk in the Book.  However, the more 
 *	important cases have been optimized so that it will run faster on
 *	the SUN.
 *
 *	The MSB bit of each Smalltalk 16-bit word is bit zero.
 *	Bit coordinates are zero-origin.  The top left is (0, 0).
 *
 *	Can be compiled with YAK for debugging.
 */

# include	<stdio.h>
# include	<pixrect/pixrect_hs.h>
# include	<sunwindow/window_hs.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"bitblt.h"
# include	"prims.h"
# include	"cursor.h"


static boolean	getArgs(), clipForX(), clipForY(), clipToDisplay();
void	BBSetup(), internalCopyBits();

/*
 *	For batched text display.
 */

static void	batchInit(), batchIt(), batchDoIt();


/*
 *	From Bitblt description in Book, p350.
 */

int		combinationRule;	/* 0 thru 15 */
int		destX, destY;
int		width, height;		/* bit size of xfer */
int		sourceX, sourceY;
int		clipX, clipY, clipWidth, clipHeight;


/*
 *	Replaces df, sf, hf.
 */

struct pixrect	*dstPr, *srcPr, *htPr;


/*
 *	Instance variables in class BitBltSimulation.
 *	A lot of them are not necessary because of the use of pixrects
 *	and Rop's.
 */

static short		sx, dx, w, sy, dy, h;	/* after clipping */


/*
 *	Macros from bitblt.h
 */

/*
 *	getForm() takes apart form arguments that are passed in.
 */

# define getForm(foop, fpr) \
	if (TRUE) {\
		register rp_t		rp;\
		register struct pixrect	*pr;\
		\
		pr = fpr;\
		getInt(widthFm(foop), pr->pr_width, FALSE);\
		getInt(heightFm(foop), pr->pr_height, FALSE);\
		\
		rp = bitmapFm(foop);\
		if (isIntegerObject((oop_t)rp)) {\
			warning("getForm: bad bitmap");\
			return (FALSE);\
		}\
		if (pr->pr_width < 0  ||  pr->pr_height < 0) {\
			warning("getForm: form has neg size");\
			pr->pr_width = pr->pr_height = 0;\
		}\
		if (foop != DisplayForm\
		    &&  pr->pr_width * pr->pr_height\
		      >  wordsToBits(lengthRp(rp))) {\
			warning("getForm: bitmap has bad size");\
			pr->pr_width = pr->pr_height = 0;\
			if (DisplayForm == nilOop) {\
				warning("getForm: forcing beDisplay");\
				DisplayForm = foop;\
			}\
		}\
		mpr_d(pr)->md_linebytes =\
		    mpr_linebytes(pr->pr_width, PIXEL_DEPTH);\
		mpr_d(pr)->md_image = (short *)BEFieldRp(rp);\
		mpr_d(pr)->md_offset.x = mpr_d(pr)->md_offset.y = 0;\
		mpr_d(pr)->md_primary = 0;\
		mpr_d(pr)->md_flags = 0;\
	}\
	else



/*
 *	Do whatever you want.  Call at the beginning of each BBSrc*Dst*()
 */

# define	traceBB(name)	/* nothing */


/*
 *	Index table for translating  Smalltalk's combination rule 
 *	to sunwindow's rasterops.  Cannot use PIX_DONTCLIP because 
 *	BS may be covered up by other suntool windows.
 */

static int	sunRopMap[] = {
	PIX_CLR,				/* 0 */
	PIX_SRC & PIX_DST,			/* 1 */
	PIX_SRC & PIX_NOT(PIX_DST),		/* 2 */
	PIX_SRC,				/* 3 */
	PIX_NOT(PIX_SRC) & PIX_DST,		/* 4 */
	PIX_DST,				/* 5 */
	PIX_SRC ^ PIX_DST,			/* 6 */
	PIX_SRC | PIX_DST,			/* 7 */
	PIX_NOT(PIX_SRC) & PIX_NOT(PIX_DST),	/* 8 */
	PIX_NOT(PIX_SRC) ^ PIX_DST,		/* 9 */
	PIX_NOT(PIX_DST),			/* 10 */
	PIX_SRC | PIX_NOT(PIX_DST),		/* 11 */
	PIX_NOT(PIX_SRC),			/* 12 */
	PIX_NOT(PIX_SRC) | PIX_DST,		/* 13 */
	PIX_NOT(PIX_SRC) | PIX_NOT(PIX_DST),	/* 14 */
	PIX_SET  				/* 15 */
};


/*
 *	The SUN hardware requires us to treat these cases differently.
 *	Used later by internalCopyBits() (modified
 *	version of copyLoop described in the Book) to decide which
 *	copy routine to call.
 */

static boolean		dstIsDisplay, srcIsDisplay, htIsNil, srcIsNil;


/*
 *	They will get the correct value when copybits is actually performed
 */

InitBitblt()
{
	/* dummy values */
	dstPr = mem_create(1, 1, PIXEL_DEPTH);
	srcPr = mem_create(1, 1, PIXEL_DEPTH);
	htPr  = mem_create(1, 1, PIXEL_DEPTH);
}


/*
 *	external interface to BitBlt
 */

boolean
copybits(bb)
rp_t bb;
{
	if (!getArgs(bb, TRUE, FALSE))
		return (FALSE);
	internalCopyBits(bb);
	return (TRUE);
}


/*
 *	Hook for scanCharacters and turtleDrawLoop.
 *	Follows book except for optimizations. 
 *	Dispatches to appropriate routine for special case.
 */

void

internalCopyBits(bb)
rp_t bb;
{
	if (!clipForX()  ||  !clipForY())
		return;

	if (dstIsDisplay  &&  !clipToDisplay())
		return;

	if (srcIsNil)
		if (htIsNil)
			if (dstIsDisplay)
				BBDstDisp();
			else
				BBDstMem();
		else
			if (dstIsDisplay)
				BBHtDstDisp();
			else
				BBHtDstMem();
	else
		if (htIsNil)
			if (dstIsDisplay)
				if (srcIsDisplay)
					BBSrcDispDstDisp();
				else
					BBSrcMemDstDisp();
			else
				if (srcIsDisplay)
					BBSrcDispDstMem();
				else
					BBSrcMemDstMem();
		else
			if (dstIsDisplay)
				if (srcIsDisplay)
					BBSrcDispHtDstDisp();
				else
					BBSrcMemHtDstDisp();
			else
				if (srcIsDisplay)
					BBSrcDispHtDstMem();
				else
					BBSrcMemHtDstMem();
}


/*
 *	Gets all the arguments.
 *	Extracts fields from BitBlt and Scanner as needed.
 *	Can do some optimizations by nothing special cases in 
 *	the halftone form.
 */

static boolean
getArgs(bb, willCopybits, willScan)
rp_t bb;
boolean willCopybits, willScan;
{
	register rp_t	rp;

	if (willCopybits) {
		if (destFormBB(bb) == nilOop) {
			warning("bitblt failing: no destForm");
			return (FALSE);
		}
		getInt(combinationRuleBB(bb), combinationRule, FALSE);
		combinationRule &= 15;
		getInt(destYBB(bb), destY, FALSE);
		getInt(heightBB(bb), height, FALSE);
		getInt(sourceYBB(bb), sourceY, FALSE);
		getInt(clipXBB(bb), clipX, FALSE);
		getInt(clipYBB(bb), clipY, FALSE);
		getInt(clipWidthBB(bb), clipWidth, FALSE);
		getInt(clipHeightBB(bb), clipHeight, FALSE);

		getForm(destFormBB(bb), dstPr);
		dstIsDisplay =  destFormBB(bb) == DisplayForm;

		if (!(srcIsNil  =  sourceFormBB(bb) == nilOop)) {
			getForm(sourceFormBB(bb), srcPr);
			srcIsDisplay =  sourceFormBB(bb) == DisplayForm;
		}
		
		switch (combinationRule) {
		case 0:
		case 5:
		case 10:
		case 15:
			srcIsNil = htIsNil = TRUE;
			break;
		default:
			if (!(htIsNil  =  halftoneFormBB(bb) == nilOop)) {
				register short	*p;

				getForm(halftoneFormBB(bb), htPr);
				p = mpr_d(htPr)->md_image;
				if (isOnesBitmap(p))
					htIsNil = TRUE;
				else if (isZerosBitmap(p)) {
					/*
					 *	Reduced rules when bitmap is 
					 *	all zeros.
					 */
					static short rm[] = {
						0, 0, 0, 0,
						5, 5, 5, 5,
						10, 10, 10, 10,
						15, 15, 15, 15
					};
				
					combinationRule = rm[combinationRule];
					htIsNil = srcIsNil = TRUE;
				}
			}
			break;
		}
	}
	getInt(destXBB(bb), destX, FALSE);

	/*
	 *	If called by ScanCharacters(), must also check for
	 *	the validity of the font table.
	 */

	if (willScan) {
		register rp_t	xt;

		xt = xTableCS(bb);
		if (isIntegerObject((oop_t)xt)  
		    ||  classRp(xt)  != classArrayOop) {
			warning("scanCharacters failing: XTable not array");
			return (FALSE);
		}
		rp = stopConditionsCS(bb);
		if (isIntegerObject((oop_t)rp)
		    ||  classRp(rp) != classArrayOop) {
			warning("scanCharacters failing: stopConditions not array");
			return (FALSE);
		}
	}
	else {
		getInt(widthBB(bb), width, FALSE);
		getInt(sourceXBB(bb), sourceX, FALSE);
	}
	return (TRUE);
}


/*
 *	Basically out of the Book (p356).  Appears as method
 *	clipRange in the Book, but it has been broken down 
 *	into two separate routines here since they are logically
 *	separate.
 *	This adjusts the width of the window in the destination.
 *	Must ensure that we don't clobber memory by clipping.
 *	Must also take min of source, dest, and clip rects.
 *	Returns FALSE if clipped window shrinkds to zero or
 *	negative size.
 */

static boolean
clipForX()
{
	register int	i;

	if (width <= 0)
		return (FALSE);

	/*
	 *	Clips dest to clip rectangle so that sx lies within
	 *	the destination rectangle.
	 */

	i = clipX - destX;
	if (i <= 0) {			/* set origin */
		sx = sourceX;
		dx = destX;
		w = width;
	}
	else {
		sx = i + sourceX;
		dx = clipX;
		w = width - i;
	}
	if (dx < 0) {
		sx += -dx;
		w -= -dx;
		dx = 0;
	}

	/*
	 *	Sets the width in case part of the clipped window
	 *	falls outside the destination window.
	 */

	i = dx + w  -  (clipX + clipWidth);
	if (i > 0)
		w -= i;
	i = dx + w - dstPr->pr_width;
	if (i > 0)
		w -= i;

	/*
	 *	Now clip to source rectangle.  Adjust sx and w in case 
	 *	the source is larger than the destination.
	 */

	if (!srcIsNil) {	/* now clip to source rect */
		if (sx < 0) {
			dx += -sx;
			w -= -sx;
			sx = 0;
		}
		i = sx + w - srcPr->pr_width;
		if (i > 0)
			w -= i;
	}
	if (w <= 0)
		return (FALSE);
	return (TRUE);
}


/*
 *	Now, do the same for the height (along vertical axis).
 */

static boolean
clipForY()
{
	register int i;

	if (height <= 0)
		return (FALSE);

	i = clipY - destY;
	if (i <= 0) {
		sy = sourceY;
		dy = destY;
		h = height;
	}
	else {
		sy = i + sourceY;
		dy = clipY;
		h = height - i;
	}
	if (dy < 0) {
		sy += -dy;
		h -= -dy;
		dy = 0;
	}

	i = dy + h  -  (clipY + clipHeight);
	if (i > 0)
		h -= i;
	i = dy + h - dstPr->pr_height;
	if (i > 0)
		h -= i;

	if (!srcIsNil) {	/* now clip to source rect */
		if (sy < 0) {
			dy += -sy;
			h -= -sy;
			sy = 0;
		}
		i = sy + h - srcPr->pr_height;
		if (i > 0)
			h -= i;
	}
	if (h <= 0)
		return (FALSE);
	return (TRUE);
}


/*
 *	Clip the already clipped sx, sy, dx, dy to the size of the
 *	window.  Should only be called when the destination is the
 *	display.
 */

static
boolean
clipToDisplay()
{
	register int	winWidth, winHeight;

	winWidth = win_getwidth(BSWindowFd);
	winHeight = win_getheight(BSWindowFd);

	/*
	 *	Entire destination is outside window.
	 */

	if (dx > winWidth)
		return(FALSE);
	if (dy > winHeight)
		return(FALSE);

	/*
	 *	Part of destination is outside window.
	 */
	
	if (dx + w  >  winWidth)
		w = winWidth - dx;
	if (dy + h  >  winHeight)
		h = winHeight - dy;
	return(TRUE);
}


/*
 *	Bitblt setup: nothing to do right now.
 *	There may be some if optimizations are put in.
 */

void
BBSetup(bb)
rp_t bb;
{
}


/*
 *	ScanCharacters implemented from the source code.  This 
 *	routine is translated from the Smalltalk source code for:
 *
 *	scanCharactersFrom: startIndex to: stopIndex in: sourceString
 *		rightX: rightX stopConditions: stops displaying: display 
 *		| lastIndex  nextDestX |
 *		lastIndex _ startIndex.
 *		[ lastIndex <= stopIndex ] whileTrue: [
 *			ascii _ (sourceString at: lastIndex) asciiValue.
 *			(stopConditions at: ascii + 1) ~~ nil  
 *			    ifTrue: [ ^stops at: ascii + 1 ].
 *			sourceX _ xTable at: ascii + 1.
 *			nextDestX _ destX + 
 *			    (width _ (xTable at: ascii + 2) - sourceX).
 *			nextDestX > rightX
 *			    ifTrue: [ ^stops at: CrossedX ].
 *			display ifTrue: [ self copyBits ].
 *			destX _ nextDestX.
 *			lastIndex _ lastIndex + 1
 *		].
 *		^stops at: EndOfRun
 *
 *
 *	This is the inner loop of text display.
 *	March through source string from startIndex to stopIndex.
 *	If any character is flagged with a non-nil entry in stops, 
 *	then return the corresponding value. Determine width of 
 *	each character from xTable. If dextX would exceed rightX, 
 *	then return stops at: 258. If displaying is true, then 
 *	display the character. Advance destX by the width of the 
 *	character. If stopIndex has been reached, then return
 *	stops at: 257.  Fail under the same conditions that the 
 *	Smalltalk code would cause an error. Optional.
 *
 *	This code is optimized for the SUN; as each character's position
 *	is computed, the character itself is displayed on the screen with
 *	unrolled loops.
 *
 *	The original Smalltalk code is repeated in the comments
 *	to help out.
 */

boolean
scanCharacters(rp)
oop_t *rp;
{
	register rp_t	rpreg;		/* scratch register */
	register rp_t	bb;		/* receiver bitblt */
	register oop_t	*xTablep;	/* font table */
	register rp_t	*stopConditionsp; /* argument stopCondition */
	register byte_t	*sourceStringp;	  /* argument sourceString */

	register u_int	xTableLength, stopConditionsLength;
	register oop_t	oop;

	oop_t	startIndex, stopIndex, rightX;	/* arguments */
	u_int	sourceStringSize;
	rp_t	*stopsp;
	u_int	stopsLength;
	boolean	willCopybits;		/* argument display */

	byte_t	*lastSourcep, *stopSourcep;

	u_int	ascii;
	int	nextDestX;

	u_int	ret;			/* return value from method */

	/*
	 *	Remembers which of them have been changed.  Have to 
	 *	update its value in aBitblt if true.
	 */

	boolean	changedSourceX = FALSE,  changedDestX = FALSE, 
	    changedWidth = FALSE;
	


	getInt(startIndexRp(rp), startIndex, FALSE);
	if (startIndex < 1)
		startIndex = 1;

	getInt(stopIndexRp(rp), stopIndex, FALSE);

	getInt(rightXRp(rp), rightX, FALSE);

	rpreg = sourceStringRp(rp);
	if (isIntegerObject((oop_t)rpreg)  ||  formatRp(rpreg) != BYTES) {
		warning("scanCharacters failing: not byte object");
		return (FALSE);
	}
	sourceStringp = byteFieldRp(rpreg);
	sourceStringSize = lengthRp(rpreg);

	rpreg = stopsRp(rp);
	stopsp = rpFieldRp(rpreg);
	stopsLength = lengthRp(rpreg);

	rpreg = displayRp(rp);
	if (rpreg == trueOop) {
		willCopybits = TRUE;
	}
	else if (rpreg == falseOop) {
		willCopybits = FALSE;
	}
	else {
		warning("scanCharacters failing: display not boolean");
		return (FALSE);
	}


	/*
	 *	Get bitblt object to display, and extract various fields.
	 */

	bb = (rp_t)*rp;
	if (!getArgs(bb, willCopybits, TRUE)){
		warning("scanCharacters failing: bad Bitblt");
		return (FALSE);
	}

	if (stopIndex > sourceStringSize)
		stopIndex = sourceStringSize;

	if (minStopsLength > stopsLength) {
		warning( "scanCharacters failing due to stops subscript");
		return (FALSE);
	}


	/*
	 *	Starts Smalltalk code
	 */

	/*
	 *	lastIndex <- startIndex.
	 *
	 *	lastIndex indexes the next character to be printed, and
	 *	lastSourcep is its address.  stopIndex indexes the last
	 *	character to be printed, and stopSourcep is its address.
	 */

	lastSourcep = &sourceStringp[startIndex - 1];
	stopSourcep = &sourceStringp[stopIndex - 1];

	xTablep = oopFieldRp(xTableCS(bb));
	xTableLength = lengthRp(xTableCS(bb));

	/*
	 *	stopConditionsp is the address of an array of oops.
	 *	A character constitutes a stop condition if the element
	 *	indexed by its ascii value is not nil.
	 */

	stopConditionsp = rpFieldRp(stopConditionsCS(bb));
	stopConditionsLength = lengthRp(stopConditionsCS(bb));

	batchInit();			/* get ready for batching chars */

	for (;;) {
		/*
		 *	[lastIndex <= stopIndex] whileTrue:  [
		 */

		if (lastSourcep > stopSourcep) {
# 			ifdef YAK
				warning("scanCharacters: stopIndex");
# 			endif
			lastSourcep = stopSourcep;
			ret = stopIndexIndex - 1;
			break;
		}

		/*
		 *	ascii <- (sourceString at: lastIndex) asciiValue.
		 */

		ascii = *lastSourcep;

		/*
		 *	(stopConditions at: ascii + 1) ~~ nil ifTrue:
		 *
		 *	The ASCII character set starts at zero, but Smalltalk
		 *	arrays starts at one.  That's why we add one to "ascii"
		 *	in the Smalltalk code.  This does not happen in our 
		 *	C code since C has zero-origin arrays.
		 */

		if (ascii >= stopConditionsLength) {
			warning("scanCharacters: stopConditions subscript");
			return (FALSE);
		}
		if (stopConditionsp[ascii] != nilOop) {
#			ifdef YAK
				warning("scanCharacters: stopCondition");
#			endif

		 	/*
			 *	[^stops at: ascii + 1].
			 */

			ret = ascii;
			break;
		}

		/*
		 *	sourceX <- xTable at: ascii + 1.
		 *
		 *	Get bit position of character to copy.
		 */

		if (ascii + 1  >=  xTableLength) {
			warning("scanCharacters failing: ascii >= xTblLen");
			return (FALSE);
		}
#		ifdef YAK
			putchar(ascii);
			fflush(stdout);
#		endif

		getInt(xTablep[ascii], oop, FALSE);
		sourceX = oop;
		changedSourceX = TRUE;

		/*
		 *	nextDestX <-
		 *	    destX + (width <- (xTable at: ascii + 2) - sourceX).
		 *
		 *	Find the width of the character to be printed, and
		 *	use that to compute where the carat will stop at.
		 */

		getInt(xTablep[ascii + 1], oop, FALSE);
		nextDestX = destX + (width = oop - sourceX);
		changedWidth = TRUE;

		/*
		 *	nextDestX > rightX ifTrue: [^stops at: CrossedX].
		 */

		if (nextDestX > rightX) {
# 			ifdef YAK
				warning("\r\nscanCharacters: nextDestX > rightX");
# 			endif
			ret = crossedXIndex - 1;
			break;
		}

		/*
		 *	display ifTrue: [self copyBits].
		 *
		 *	Uses batchrop.
		 */

		if (willCopybits)
			batchIt();


		/*
		 *	destX <- nextDestX.
		 */

		destX = nextDestX;
		changedDestX = TRUE;

		/*
		 *	lastIndex <- lastIndex + 1].
		 */

		++lastSourcep;
	}

	batchDoIt();

	/*
	 *	^stops at: EndOfRun
	 */

	rpreg = stopsp[ret];
	resultRp(rp) = rpreg;

	/*
	 *	End Smalltalk code
	 */


	/*
	 *	lastIndex <- stopIndex.
	 *
	 *	last - first is last, but with one-origin must increment.
	 */

	putInt(lastSourcep - sourceStringp + 1,  lastIndexCS(bb));

	/*
	 *	Write all changed values back to receiver Bitblt.
	 */

	if (changedDestX)
		putInt(destX, destXBB(bb));

	if (changedSourceX)
		putInt(sourceX, sourceXBB(bb));

	if (changedWidth)
		putInt(width, widthBB(bb));

#	ifdef YAK
		warning("scanCharacters succeded");
#	endif

	return (TRUE);
}


/*
 *	Straight from method  drawLoopX:Y:  in the book, p352.
 */

boolean
drawLoopXY(rp)
oop_t *rp;
{
	register oop_t	px, py;		/* xDelta, yDelta */
	register int	signX, signY, p, i;
	register rp_t	bb;		/* receiver bitblt */

	bb = (rp_t)*rp;
	if (!getArgs(bb, TRUE, FALSE))
		return (FALSE);
	getInt(rp[1], px, FALSE);
	getInt(rp[2], py, FALSE);

	if (px < 0) {
		px = -px;
		signX = -1;
	}
	else
		signX = 1;

	if (py < 0) {
		py = -py;
		signY = -1;
	}
	else
		signY = 1;

	internalCopyBits(bb);

	if (py > px) 			/* more horizontal */
		for (i = 1,  p = py >> 1;   i <= py;   i++) {
			destX += signX;
			if ((p -= px) < 0) {
				destY += signY;
				p += py;
			}
			internalCopyBits(bb);
		}
	else				/* more vertical */
		for (i = 1,  p = px >> 1;   i <= px;   i++) {
			destY += signY;
			if ((p -= py) < 0) {
				destX += signX;
				p += px;
			}
			internalCopyBits(bb);
		}

	return (TRUE);
}

/*
 *	The following are subroutines for internalCopyBits().
 *	Each one of them handles a special case of copyLoop.
 *
 *	Some of the code can be factored out.
 */

/*
 *	Temporary pixrect.  Used for performing rasterops with halftones.
 *	Can only be used when the clipped source is less than 800x1024.
 *	Using a portion of maskedPr reduces the need to call mem_create 
 *	everytime a bitblt has a halftone.
 *	The choice of the size of maskedPr is completely arbitrary.
 *	But making it too small will defeat its purpose.
 */


# define	maskedRectX		1024
# define	maskedRectY		800

static short	maskedBitmap[bitsToWords(maskedRectX) * maskedRectY];

static mpr_static(maskedPr, maskedRectX, maskedRectY, PIXEL_DEPTH,
    maskedBitmap);


/*
 *	MaskedRegion() "cuts" a portion out of maskedPr and return it
 *	as a pixrect in result.  If maskedPr is no large enough for the
 *	requested width and height, a brand new pixrect is created and 
 *	alloc is set to TRUE to remind the caller that result should
 *	be destroyed (via pr_destroy) when it is no longer needed.
 */

# define	maskedRegion(result, width, height, alloc)\
	if ((width) > maskedRectX  ||  (height) > maskedRectY) {\
		(result) = mem_create(width, height, PIXEL_DEPTH);\
		(alloc) = TRUE;\
	}\
	else if (TRUE) {\
		(result) = &maskedPr;\
		(result)->pr_width = width;\
		(result)->pr_height = height;\
		(alloc) = FALSE;\
	}\
	else


/*
 *	Bitblt with no source or halftone (i.e. halftone = ones) to memory.
 */

BBDstMem()
{
	register struct pixrect	*regDstPr;

	traceBB(DstMem);
	regDstPr = dstPr;
	switch (combinationRule) {
	case 1:				/* merging dst with nothing */
	case 5:
	case 9:
	case 13:
		break;
	default:
		pr_rop(regDstPr, dx, dy, w, h,
		    PIX_BLACK | sunRopMap[combinationRule], NULL, 0, 0);
		break;
	}
}



/*
 *	Bitblt routine: no src, with halftone, dest in memory.
 */

BBHtDstMem()
{
	register struct pixrect	*tmpPr, *regDstPr, *regHtPr;
	boolean			created;

	traceBB(HtDstMem);
	regDstPr = dstPr;
	regHtPr = htPr;

	/*
	 *	No need to mask source with halftone first since 
	 *	source is nil.
	 */

	pr_replrop(regDstPr, dx, dy, w, h, sunRopMap[combinationRule],
	    regHtPr, halftoneOffset(dx), halftoneOffset(dy));
}


/*
 *	Bitblt routine: src in memory, no halftone, dest in memory.
 */

BBSrcMemDstMem()
{
	register struct pixrect	*regDstPr, *regSrcPr;

	traceBB(SrcMemDstMem);
	regDstPr = dstPr;
	regSrcPr = srcPr;
	pr_rop(regDstPr, dx, dy, w, h, sunRopMap[combinationRule],
	    regSrcPr, sx, sy);
}


/*
 *	Bitblt routine: src in memory, with halftone, dest in memory.
 */

BBSrcMemHtDstMem()
{
	register struct pixrect	*tmpPr, *regDstPr, *regSrcPr, *regHtPr;
	boolean			created;

	traceBB(SrcMemHtDstMem);
	regDstPr = dstPr;
	regSrcPr = srcPr;
	regHtPr = htPr;
	maskedRegion(tmpPr, w, h, created);
	pr_rop(tmpPr, 0, 0, w, h, PIX_COPY, regSrcPr, sx, sy);
	pr_replrop(tmpPr, 0, 0, w, h, PIX_MASK, regHtPr, 0, 0);
	pr_rop(regDstPr, dx, dy, w, h, sunRopMap[combinationRule],
	    tmpPr, 0, 0);
	if (created)
		pr_destroy(tmpPr);
}


/*
 *	Bitblt routine: src in display, no halftone, dest in memory
 *	Cursor must be turned off while changing the display, otherwise
 *	it will be moved together with the background.
 */


BBSrcDispDstMem()
{
	register struct pixrect	*regDstPr, *regDispPr;

	traceBB(SrcDispDstMem);
	regDstPr = dstPr;
	regDispPr = BSPixwin->pw_pixrect;
	pw_read(regDstPr, dx, dy, w, h, sunRopMap[combinationRule],
	    BSPixwin, sx, sy);
}


/*
 *	Src is display, with halftone to memory: moves horizontally.
 */

BBSrcDispHtDstMem()
{
	register struct pixrect	*tmpPr, *regDstPr, *regDispPr, *regHtPr;
	boolean			created;

	traceBB(SrcDispHtDstMem);
	regDstPr = dstPr;
	regDispPr = BSPixwin->pw_pixrect;
	regHtPr = htPr;
	maskedRegion(tmpPr, w, h, created);
	pw_read(tmpPr, 0, 0, w, h, PIX_COPY, BSPixwin, sx, sy);
	pr_replrop(tmpPr, 0, 0, w, h, PIX_MASK, regHtPr, halftoneOffset(dx),
	    halftoneOffset(dy));
	pr_rop(regDstPr, dx, dy, w, h, sunRopMap[combinationRule],
	    tmpPr, 0, 0);
	if (created)
		pr_destroy(tmpPr);
}


 /*
  *	Bitblt to display, no src no ht.
  */

BBDstDisp()
{
	register struct pixrect	*regDispPr;

	traceBB(DstDisp);
	regDispPr = BSPixwin->pw_pixrect;
	pw_writebackground(BSPixwin, dx, dy, w, h,
	    PIX_BLACK | sunRopMap[combinationRule]);
}


/*
 *	Bitblt to display with halftone only.
 *	No need to mask source with halftone since the source is all
 *	ones.
 */

BBHtDstDisp()
{
	register struct pixrect	*tmpPr, *regDispPr, *regHtPr;
	boolean			created;

	traceBB(HtDstDisp);
	regDispPr = BSPixwin->pw_pixrect;
	regHtPr = htPr;
	pw_replrop(BSPixwin, dx, dy, w, h, sunRopMap[combinationRule],
	    regHtPr, halftoneOffset(dx), halftoneOffset(dy));
}


/*
 *	Bitblt from memory to display, no haftone.
 */

BBSrcMemDstDisp()
{
	register struct pixrect	*regDispPr, *regSrcPr;

	traceBB(SrcMemDstDisp);
	regDispPr = BSPixwin->pw_pixrect;
	regSrcPr = srcPr;
	pw_write(BSPixwin, dx, dy, w, h, sunRopMap[combinationRule],
	    regSrcPr, sx, sy);
}


/*
 *	Bitblt from memory to display, with halftone.
 */

BBSrcMemHtDstDisp()
{
	register struct pixrect	*tmpPr, *regSrcPr, *regHtPr, *regDispPr;
	boolean			created;

	traceBB(SrcMemHtDstDisp);
	regSrcPr = srcPr;
	regHtPr = htPr;
	regDispPr = BSPixwin->pw_pixrect;
	maskedRegion(tmpPr, w, h, created);
	pr_rop(tmpPr, 0, 0, w, h, PIX_COPY, regSrcPr, sx, sy);
	pr_replrop(tmpPr, 0, 0, w, h, PIX_MASK, regHtPr, halftoneOffset(dx),
	    halftoneOffset(dy));
	pw_write(BSPixwin, dx, dy, w, h, sunRopMap[combinationRule],
	    tmpPr, 0, 0);
	if (created)
		pr_destroy(tmpPr);
}


 /*
  *	Bitblt routine src and dst are display, no halftone.
  */

BBSrcDispDstDisp()
{
	register struct pixrect	*regDispPr;

	traceBB(SrcDispDstDisp);
	regDispPr = BSPixwin->pw_pixrect;
	pw_copy(BSPixwin, dx, dy, w, h, sunRopMap[combinationRule],
	    BSPixwin, sx, sy);
}


/*
 *	Bitblt routine src, dst is display with halftone.
 */

BBSrcDispHtDstDisp()
{
	register struct pixrect	*tmpPr, *regDispPr, *regHtPr;
	boolean			created;

	traceBB(SrcDispHtDstDisp);
	regDispPr = BSPixwin->pw_pixrect;
	regHtPr = htPr;
	maskedRegion(tmpPr, w, h, created);
	pw_read(tmpPr, 0, 0, w, h, PIX_COPY, BSPixwin, sx, sy);
	pr_replrop(tmpPr, 0, 0, w, h, PIX_MASK, regHtPr, halftoneOffset(dx),
	    halftoneOffset(dy));
	pw_write(BSPixwin, dx, dy, w, h, sunRopMap[combinationRule],
	    tmpPr, 0, 0);
	if (created)
		pr_destroy(tmpPr);
}

/*
 *	Batches the destinations together and execute a batched rasterop.
 *	Calls pr_batchrop() when the batch array filled up.
 *	Configured to be used by scanCharacters().
 *
 *	For some reason, batchitem was not defined in the pixrect header
 *	files.
 */

struct batchitem {
	struct pixrect	*bi_pr;
	struct pr_pos	bi_pos;
};

static struct batchitem	batch[BUFSIZ];		/* buffer for batched items */
static int		batchIndex,		/* index to batch[] */
			prevDstX, prevDstY, 	/* previous destination */
			startDstX, startDstY;	/* where to start */


static void
batchInit()
{
	batchIndex = 0;
	prevDstX = prevDstY = 0;
	startDstX = startDstY = 0;
}


static void
batchIt()
{
	register struct batchitem	*bip;
	register struct pixrect		*regHtPr, *regSrcPr;

	if (!clipForX()  ||  !clipForY())
		return;

	if (batchIndex == BUFSIZ)
		batchDoIt();

	/*
	 *	If there is a halftone,
	 *	must create a new pixrect for batching and apply halftone
	 *	to it now.
	 *	Creating a pixrect for every character batched is 
	 *	expensive.  I wish there is a better way to do this.
	 */

	if (htIsNil)
		batch[batchIndex].bi_pr = pr_region(srcPr, sx, sy, w, h);
	else {
		regHtPr = htPr;
		regSrcPr = mem_create(w, h, PIXEL_DEPTH);	/* YUK!! */
		if (srcIsNil)
			pr_replrop(regSrcPr, 0, 0, w, h, PIX_COPY, regHtPr,
			    halftoneOffset(dx), halftoneOffset(dy));
		else {
			pr_rop(regSrcPr, 0, 0, w, h, PIX_COPY, srcPr, sx, sy);
			pr_replrop(regSrcPr, 0, 0, w, h, PIX_MASK, regHtPr,
			    halftoneOffset(dx), halftoneOffset(dy));
			batch[batchIndex].bi_pr = regSrcPr;
		}
	}


	bip = &batch[batchIndex];
	if (batchIndex++) {		/* not first batch item */
		bip->bi_pos.x = dx - prevDstX;
		bip->bi_pos.y = dy - prevDstY;
	}
	else {
		startDstX = dx;
		startDstY = dy;
		bip->bi_pos.x = bip->bi_pos.y = 0;
	}
	prevDstX = dx;
	prevDstY = dy;
}


/*
 *	Called when batch buffer is full, or when all items have 
 *	been batched.
 */

static void
batchDoIt()
{
	register struct batchitem	*bip;
	register struct pixrect		*regDstPr, *regSrcPr;
	register int			width, height, i;

	if (batchIndex == 0)		/* nothing batched */
		return;

	if (dstIsDisplay)
		pw_batchrop(BSPixwin, startDstX, startDstY,
		    sunRopMap[combinationRule], batch, batchIndex);
	else {
		regDstPr = dstPr;
		if (batchIndex == 1) {	/* optimized for single char */
			regSrcPr = batch[0].bi_pr;
			width = regSrcPr->pr_width;
			height = regSrcPr->pr_height;
			pr_rop(regDstPr, startDstX, startDstY, width, height,
			    sunRopMap[combinationRule], regSrcPr, 0, 0);
		}
		else
			pr_batchrop(regDstPr, startDstX, startDstY,
			    sunRopMap[combinationRule], batch, batchIndex);
	}

	/*
	 *	Release the temporary pixrects.
	 */

	for (i = 0;  i < batchIndex;  i++) {
		regSrcPr = batch[i].bi_pr;
		pr_destroy(regSrcPr);
	}
	
	batchInit();
}
