static char rcsId[] = "$Header: toolSunprim.c,v 1.2 84/07/26 00:28:01 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/* 
 *	Routines to manipulate Sun's display.  Consists of routines to
 *	initialize display, copy and write bitmaps to the display.
 */

# include	<sys/ioctl.h>

# include	<fcntl.h>
# include	<sun/fbio.h>
# include	<suntool/tool_hs.h>
# include	<sunwindow/window_hs.h>
# include	<pixrect/pixrect_hs.h>

# include	<ctype.h>
# include	<stdio.h>
# include	<signal.h>
# include	<errno.h>

# include	"utils.h"
# include	"io.h"
# include	"rp.h"
# include	"cursor.h"
# include	"bitblt.h"


/*
 *	suntools is true if bs is invoked from suntools.
 */

boolean			suntools;
int			BSWindowFd = nullFd;
struct pixwin		*BSPixwin;
struct rect		WindowRect;

static struct tool	*BSTool;
static int		pid;
void			sigwincher(), sigwincher1(), sigchlder();

# include		"bs.icon"
mpr_static(bsIconPr, TOOL_ICONWIDTH, TOOL_ICONHEIGHT, 1, icon_data);
static struct icon	bsIcon = {
	TOOL_ICONWIDTH, TOOL_ICONHEIGHT, (struct pixrect *)NULL,
	{0, 0, TOOL_ICONWIDTH, TOOL_ICONHEIGHT},
	&bsIconPr, {0, 0, 0, 0}, NULL, NULL, 0
};


# ifdef RELEASE1_0
#	define	DisplayDevice	"/dev/bw0"
# else
#	define	DisplayDevice	"/dev/bwone0"
# endif


/* 
 *	Not much.
 */

void
SunInitialize()
{
	BSWindowFd = nullFd;
	InitBitblt();
}



/*
 *	Define RELEASE1_0 if you are running Sun Unix release 1.0.
 *	The major difference is in the calling sequence for win_screennew()
 *	and the definition of struct screen.
 *
 *	See Sun Window System Programmer's Manual
 *	for the default setting in struct screen.
 *	Modify it accordingly if your configuration
 *	is different.
 */

void
OpenWindow(argv)
char **argv;
{
	char			on = 1;
	static char		**prevArgv;
	struct inputmask	bsInputMask;
	static struct toolsw	*BSToolsw;
	static char 		myName[] = " Berkeley Smalltalk Version 2.0";
	static struct rect	screenRect = { 50, 50, 800, 600 };

#       ifdef RELEASE1_0
		static struct screen    displayScreen
		    = { DisplayDevice, FBTYPE_SUN1BW,  TRUE };
#	else
		static struct screen	displayScreen;
#	endif


	if (windowOpened(BSWindowFd))
		return;

#	ifdef RELEASE1_0
		sigset(SIGWINCH, SIG_IGN);
		sigset(SIGCHLD, SIG_IGN);
		suntools = FALSE;
		if ((BSWindowFd = win_getnewwindow()) == NULL)
			fatal("OpenWindow: cannot get new window.", NullOop);
		win_screennew(BSWindowFd, &displayScreen);
#	else
		/*
		 *	Remembers previous argument so that it knows
		 *	what to init displayScreen to when it is re-opened
		 *	after a suspend.
		 */
		if (argv != NULL)
			prevArgv = argv;
		else
			argv = prevArgv;
		if (win_initscreenfromargv(&displayScreen, argv) < 0)
			fatal("OpenWindow: argument error.", NullOop);
		if (!getenv("WINDOW_PARENT")) {	/* not from suntools */
			suntools = FALSE;
			sigset(SIGWINCH, SIG_IGN);
			sigset(SIGCHLD, SIG_IGN);
			if ((BSWindowFd = win_screennew(&displayScreen))
			    == NULL)
				fatal("OpenWindow: cannot open screen.",
				    NullOop);
		}
		else {
			suntools = TRUE;
			if ((BSTool = tool_create(myName, TOOL_NAMESTRIPE,
			    &screenRect, &bsIcon)) == (struct tool *)0)
				fatal("OpenWindow: cannot create tool.",
				    NullOop);
			if ((BSToolsw = tool_createsubwindow(BSTool, myName,
			    TOOL_SWEXTENDTOEDGE, TOOL_SWEXTENDTOEDGE)) 
			      == (struct toolsw *)0)
				fatal("OpenWindow: cannot create subwindow.",
				    NullOop);
			sigset(SIGWINCH, sigwincher);
			sigset(SIGCHLD, sigchlder);
			tool_install(BSTool);

			/*
			 *	Must fork to execute a tool_select() in order
			 *	to get a boundary manager.
			 */

			if (pid = fork()) {
				if (pid == -1) {
					perror("bs");
					fatal("cannot fork.", NullOop);
				}
				sigset(SIGWINCH, sigwincher1);
				tool_select(BSTool, 1);	/* wait for child */
				tool_destroy(BSTool);
				/*
				 *	Kill child one more time in case 
				 *	it is still around.
				 */
				kill(pid, SIGKILL);
				exit(0);
			}
			else	/* child */
				;
			BSWindowFd = BSToolsw->ts_windowfd;
		}
#	endif

	BSPixwin = pw_open(BSWindowFd);

	/*
	 *	Set up input mask.  The input mask selects what input
	 *	events should be reported.
	 */

	input_imnull(&bsInputMask);

	if (EncodedKeyboard)
		bsInputMask.im_flags  |=  IM_ASCII | IM_NEGEVENT;
	else
		bsInputMask.im_flags  |=  IM_UNENCODED | IM_NEGEVENT;

	win_setinputcodebit(&bsInputMask, MS_LEFT);
	win_setinputcodebit(&bsInputMask, MS_MIDDLE);
	win_setinputcodebit(&bsInputMask, MS_RIGHT);
	win_setinputcodebit(&bsInputMask, LOC_MOVE);

	win_setinputmask(BSWindowFd, &bsInputMask, &bsInputMask,
	    WIN_NULLLINK);

	ioctl(BSWindowFd, (int)FIONBIO, (char *)&on);

	/*
	 *	cannot suspend bs when invoked from suntools.
	 *	There is no need to do so.
	 */

	setSuspend(!suntools);

	/*
	 *	Initializes the cursor when window is opened
	 */

	InitCursor();
	if (suntools)
		ClearWindow();

}


/*
 *	Closes Sun Windows so that output will be directed back 
 *	to /dev/console.  But don't do that if bs is invoked from
 *	suntools.  Closing the window will cause it to go away!
 *
 *	If force is TRUE, window will always be closed.
 */

void
CloseWindow(force)
boolean force;
{
	if (!force  &&  suntools)
		return;
	pw_close(BSPixwin);
	close(BSWindowFd);
	BSWindowFd = nullFd;
}


ClearWindow()
{
	pw_writebackground(BSPixwin, 0, 0,
	    win_getwidth(BSWindowFd), win_getheight(BSWindowFd),
	    PIX_WHITE | PIX_SRC | PIX_DONTCLIP);
}


/*
 *	Signal handlers for SIGWINCH and SIGCHLD.
 */

/*
 *	Child's SIGWINCH handler.
 */

void
sigwincher()
{
	tool_sigwinch(BSTool);
}


/*
 *	Parent's SIGWINCH handler.  Notify child.
 */

void
sigwincher1()
{
	tool_sigwinch(BSTool);
	kill(pid, SIGWINCH);
}


/*
 *	SIGCHLD handler.  Notify tool_select().
 */

void
sigchlder()
{
	tool_sigchld(BSTool);
}
