static char rcsId[] = "$Header: unixPrims.c,v 41.8 84/07/17 22:02:30 dmu Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Smalltalk primitives that interface with the Unix file system.
 *	These primitives function like the Unix system calls, 
 *	except the arguments and return values are Smalltalk objects.
 */

# include	<stdio.h>
# include	<math.h>
# include	<sgtty.h>
# include	<signal.h>
# include	<ctype.h>

# include	<sys/types.h>
# include	<sys/vtimes.h>
# include	<sys/timeb.h>
# include	<sys/time.h>
# include	<sys/resource.h>
# include	<sys/dir.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"prims.h"

static boolean	primitiveOpenCreat(), primitiveReadWrite(), primitiveClose();


/*
 *	UnixDirectory class primitiveOpen: filename mode: mode.
 *	Shares code with creat.
 */

rp_t
UDOpenPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	extern int	open();

	if (!primitiveOpenCreat(rp, open))
		return (NullOop);
	return (ac);
}


/*
 *	UnixFile class primitiveSeekFd: fd  offset: offset  ptr: ptr
 *	``res'' is the resulting byte offset.
 */

rp_t
UFCSeekPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register oop_t	fd, offset, ptr, res;

	getInt(rp[1], fd, NullRp);
	getInt(rp[2], offset, NullRp);
	getInt(rp[3], ptr, NullRp);

	res = lseek(fd, (long)offset, ptr);

	rp[0] = integerObjectOf(res);

	return (ac);
}


/*
 *	UnixFile class primitiveReadFd: fd  buffer: buf  length: n.
 *	Shares code with write.
 */

rp_t
UFCReadPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	extern int	read();

	if (!primitiveReadWrite(rp, read))
		return (NullOop);
	return (ac);
}


/*
 *	UnixFile class primitiveWriteFd: fd  buffer: buf  length: n.
 */

rp_t
UFCWritePrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	extern int	write();
	extern int	write();

	if (!primitiveReadWrite(rp, write))
		return (NullOop);
	return (ac);
}


/*
 *	UnixFile class primitiveClose: fd
 */

rp_t
UFCClosePrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	if (!primitiveClose(rp))
		return (NullOop);
	return (ac);
}


/*
 *	UnixDirectory primitiveCreat: filename  mode: mode
 */

rp_t
UDCreatPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	extern int	creat();

	if (!primitiveOpenCreat(rp, creat))
		return (NullOop);
	return (ac);
}


/*
 *	UnixFile class primtiveBoot
 *
 *	How do you read in the file system code when you haven't
 *	got any in the image?
 *	Copy the file into boot, the run this primitive.
 *	It returns a string containing the contents of ``boot''.
 */

rp_t
UFCBootPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t	res;
	register int	fd;
	register long	size;
	register long	count;
	register char	*cp;

	fd = open("boot", 0);
	if (fd < 0)
		return (NullOop);

	size = lseek(fd, 0L, 2);
	if ((int)size < 0)
		size = 0;
	(void)xlseek(fd, 0L, 0);

	res = instantiateClass(classStringOop, (u_int)size);
	(void)read(fd, (char *)byteFieldRp(res), (int)size);
	(void)close(fd);
	
	/*
	 *	Converts all newlines to carriage returns.
	 *	Smalltalk likes returns but Unix likes newlines.
	 */

	for (cp = (char *)byteFieldRp(res), count = 0L;
	    count < size;  cp++, count++)
		if (*cp == '\n')
			*cp = '\r';

	*rp = (oop_t)res;
	return (ac);
}


/*
 *	Does either an open or creat, returns true for success.
 *	Must null terminate string and convert return to smallInteger.
 */

static boolean
primitiveOpenCreat(rp, openOrCreat)
register oop_t *rp;
int (*openOrCreat)();
{
	register oop_t	mode;
	char	namebuf[BUFSIZ];

	getString(rp[1], namebuf, FALSE);

	getInt(rp[2], mode, FALSE);
	{
		register int	fd;

		fd = openOrCreat(namebuf, mode);
		*rp = integerObjectOf(fd);
	}
	return (TRUE);
}


/*
 *	Routine to either read or write a buffer.
 *	The buffer is passed as an argument to the Smalltalk primitive.
 *	Returns the number of bytes read/written as a SmallInteger.
 */

static boolean
primitiveReadWrite(rp, readOrWrite)
register oop_t *rp;
int (*readOrWrite)();
{
	register oop_t	fd, size;
	register rp_t	string;
	u_int	bufSize;

	getInt(rp[1], fd, FALSE);

	string = (rp_t)rp[2];
	if (isIntegerObject((oop_t)string)  ||  formatRp(string) != BYTES)
		return (FALSE);
	bufSize = lengthRp(string);

	getInt(rp[3], size, FALSE);
	if (bufSize < size)
		size = bufSize;

	size =  (*readOrWrite)(fd, byteFieldRp(string), size);

	*rp = integerObjectOf(size);
	return (TRUE);
}


/*
 *	Closes a file.  Returns the return code (0 for success, -1
 *	for error) as a SmallInteger.
 */

static boolean
primitiveClose(rp)
oop_t *rp;
{
	register oop_t	fd;

	getInt(rp[1], fd, FALSE);

	*rp = integerObjectOf(close(fd));
	return (TRUE);
}


/*
 *	aCompiledMethod initialPC
 *	Answer the program counter for the receiver's first bytecode.
 *	Is a primitive cause the Smalltalk code is wordsize dependant:
 *	^self numLiterals * 2 + 3
 *
 *	OBSOLETE with v2 of the ST image
 */

rp_t
CMInitialPC(rp, ac)
oop_t *rp;
rp_t ac;
{
	*rp = integerObjectOf(
	    initialInstructionPointerCMH(headerCM((rp_t)*rp)));
	return (ac);
}


/*
 *	SystemDictionary profile: aBoolean.  
 *	En/disable gprofiling.
 */

rp_t
SDProfPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	register rp_t	aBool;

	aBool = (rp_t)(rp[1]);
	if (aBool == trueOop)
		moncontrol(1);
	else if (aBool == falseOop)
		moncontrol(0);
	else
		return (NullOop);
	return (ac);
}


/*
 *	Implements named snapshots.  Sets SnapName with the desired name
 *	and checks for accessability by opening filename for writing.
 */

rp_t
UFCSnapPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	int	fd;

	getString(rp[1], SnapName, NullOop);

	if ((fd = creat(SnapName, 0660)) < 0) {
		SnapName[0] = '\0';
		return (NullOop);
	}
	(void)close(fd);
	return (ac);
}


/*
 *	UnixFile class ftruncate: fd  to: size
 *
 *	Truncates a file to size ``size''.
 */

rp_t
UFCFtruncPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	oop_t	fd, len;

	getInt(rp[1], fd, NullRp);
	getInt(rp[2], len, NullRp);
	if (ftruncate(fd, len) < 0)
		return (NullOop);
	return (ac);
}


/*
 *	UnixFile class unlink: fd
 */

rp_t
UFCUnlinkPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	char	name[BUFSIZ];

	getString(rp[1], name, NullOop);
	if (unlink(name) < 0)
		return (NullOop);
	return (ac);
}


/*
 *	UnixFile class link: oldPath  to: newPath
 */

rp_t
UFCLinkPrm(rp, ac)
register oop_t *rp;
rp_t ac;
{
	char	oldpath[BUFSIZ], newpath[BUFSIZ];

	getString(rp[1], oldpath, NullOop);
	getString(rp[2], newpath, NullOop);
	if (link(oldpath, newpath) < 0)
		return (NullOop);
	return (ac);
}


/*
 *	UnixResourceUsage getResourceUsagePrimitive
 */

rp_t
URUGetRUsagePrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	register rp_t rcvr, p;
	struct rusage ru;
	register u_int i;

	getrusage(RUSAGE_SELF, &ru);
	rcvr = (rp_t)*rp;
	i = 0;

	floatObjectOf(
	    (float)ru.ru_utime.tv_sec + (float)ru.ru_utime.tv_usec / 1000000.0,
	    p);
	rpFieldRp(rcvr)[i++] = p;
	floatObjectOf(
	    (float)ru.ru_stime.tv_sec + (float)ru.ru_stime.tv_usec / 1000000.0,
	    p);
	rpFieldRp(rcvr)[i++] = p;

#	define get(component)\
		oopFieldRp(rcvr)[i++] = integerObjectOf(ru.component)

	get(ru_maxrss), get(ru_ixrss), get(ru_idrss), get(ru_isrss);
	get(ru_minflt), get(ru_majflt);
	get(ru_nswap);
	get(ru_inblock), get(ru_oublock);
	get(ru_msgsnd), get(ru_msgrcv);
	get(ru_nsignals);
	get(ru_nvcsw), get(ru_nivcsw);

	if (i > fixedFieldsRp(rcvr))
		fatal("URUGetRUsagePrm: bad UnixResourceUsage definition",
		    rcvr);
	
	return (ac);
}


static DIR *dirp = NULL;

/*
 *	UDOpenDirPrm() open directory for scanning
 *	only works on current directory.
 */

rp_t
UDOpenDirPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	char name[BUFSIZ];

	if (dirp != (DIR *)NULL) {
		closedir(dirp);
		dirp = (DIR *)NULL;
	}
	getString(rp[1], name, NullOop);
	dirp = opendir(name);
	return (ac);
}


/*
 *	UDReadDirPrm() read directory for scanning
 *	only works on current directory.
 */

rp_t
UDReadDirPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	struct direct *direct;

	if (dirp == (DIR *)NULL)
		return (NullOop);

	direct = readdir(dirp);
	if (direct == (struct direct *)NULL) {
		closedir(dirp);
		*rp = (oop_t)nilOop;
	}
	else {
		*rp = (oop_t)instantiateClass(classStringOop,
		    (u_int)direct->d_namlen);
		bcopy(direct->d_name, byteFieldRp(*(rp_t *)rp), 
		    (u_int)direct->d_namlen);
	}
	return (ac);
}
