static char rcsId[] = "$Header: utils.c,v 44.7 84/07/25 20:07:58 pkl Exp $ (Berkeley)";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 *
 *	The ones that start with x are encapsulations of
 *	library routines that cannot fail.  In case of 
 *	failures, "fatal" will be called, and causes the
 *	program to be aborted.  User may then select to 
 *	interrupt or quit (causing a core dump).
 */

# include	<stdio.h>
# include	<signal.h>

# include	"utils.h"
# include	"mem.h"

extern char	*sbrk();
extern long	lseek();


xfread(p, s, n, fp)
char *p;
u_int s, n;
FILE *fp;
{
	if (fread(p, (int)s, (int)n, fp) != n)
		fatal("read error/eof", NullOop);
}


/*
 *	If a fatal error occurs, print the message, and let the user
 *	decide whether or not to take a dump.
 */

void
fatal(s, oop)
char s[];
oop_t oop;
{
	char buf[BUFSIZ];

	CloseWindow(TRUE);

	/* 
	 *	convert second argument to 
	 *	a string.
	 */

	if (NullOop == oop)
		buf[0] = '\0';
	else
		(void)asString(buf, oop);

	fprintf(stderr,
	    "fatal error: ``%s <%s>''.  Interrupt or kill me.\r\n",
	    s, buf);
	sigset(SIGINT, (void(*)())SIG_DFL); /* be consistent with IOPrims */
	fflush(stderr);
	/* 
	 * wait for signal
	 */
	for (;;)
		pause();
}


int
xopen(s, f)
char s[];
int f;
{
	int i;

	if ((i = open(s, f)) >= 0)
		return (i);
	else {
		fprintf(stderr, "could not open ``%s''\n", s);
		fatal("xopen error", NullOop);
		/*NOTREACHED*/
	}
}


void
xread(fd, s, n)
int fd;
char s[];
u_int n;
{
	if (read(fd, s, (int)n) != n)	/* abort on read error */
		fatal("xread error/eof", NullOop);
}


char *
xvalloc(n)
u_int n;
{
	char *p;			/* points to block returned */

	if ((p = valloc((int)n)) == NULL) {	/* abort if no more space */
		fprintf(stderr, "cannot xvalloc %u bytes\n", n);
		fprintf(stderr, "already have %u bytes\n", sbrk(0));
		fatal("xvalloc error", NullOop);
		/* NORETURN */
	}
	return (p);
}


char *
xmalloc(n)
u_int n;
{
	char *p;			/* points to block returned */

	fprintf(stderr, "xmalloc: %u\n", n);
	if ((p = malloc(n)) == NULL) {	/* abort when no more space */
		fprintf(stderr, "cannot malloc %u bytes\n", n);
		fprintf(stderr, "already have %u bytes\n", sbrk(0));
		fatal("xmalloc error", NullOop);
		/* NORETURN */
	}
	return (p);
}


long
xlseek(fd, offset, whence)
int fd;
long offset;
int whence;
{
	long ret;			/* resulting pointer location */

	if ((ret = lseek(fd, offset, whence)) < 0)
		fatal("xlseek error", NullOop);
	return (ret);
}
	

/*
 *	Print warning message "s".
 */

warning(s)
char s[];
{
	fprintf(stderr, "%s\r\n", s);
	(void)fflush(stderr);
}


FILE *
xfopen(s, mode)
char s[];
char mode[];
{
	FILE *ret;			/* file pointer returned */

	if ((ret = fopen(s, mode)) == NULL) { /* cannot access file */
		fprintf(stderr, "cannot fopen ``%s''\n", s);
		fatal("xfopen error", NullOop);
	}
	return (ret);
}


/*
 *	Will ask for instruction if there are any errors.  Abort only 
 *	when EOF or 'n' is typed.
 */

xfwrite(buf, size, n, fp)
char buf[];
u_int size, n;
FILE *fp;
{
	char c;

	if (fwrite(buf, (int)size, (int)n, fp) != n) {
		warning("xfwrite error.  Do you wish to proceed?");
		do {			/* ask for instructions */
			c = getchar();
			if (c == 'y'  ||  c == 'n')
				break;
			warning("type y or n");
		} while (!feof(stdin));
		if (feof(stdin)  ||  c == 'n')
			fatal("xfwrite error", NullOop);
	}
}


/*
 *	Will ask for instruction if there is any error.  Typing 'y' 
 *	will continue, while 'n' or EOF will abort program.
 */

void
xwrite(fd, buf, size)
int fd;
char buf[];
int size;
{
	char c;				/* character in typed response */

	if (write(fd, buf, size) != size) {
		warning("xwrite error.  Do you wish to proceed?");
		do {			/* get response */
			c = getchar();
			if (c == 'y'  ||  c == 'n')
				break;
			warning("type y or n");
		} while (!feof(stdin));
		if (feof(stdin)  ||  c == 'n')
			fatal("xwrite error", NullOop);
	}
}


long
xfseek(fp, pos, ptr)
FILE *fp;
long pos;
int ptr;
{
	if ((pos = fseek(fp, pos, ptr)) < 0)	/* improper seeks */
		fatal("xfseek error", NullOop);
	return (pos);
}


/*
 *	An easy breakpoint.  
 */

void
_adb()
{
}


/*
 *	Long integer copy - Copy "count" words of type 
 *	long from the address "from" to the address "to."
 */

wcopy(from, to, count)
register long *from, *to;
u_int count;
{
	register long *end;		/* address after last address to copy */

	for (end = &from[count];  from < end;  *to++ = *from++)
		;
}


/*
 *	Short copy - Copies "count" integers from the 
 *	address "from" to address "to."
 */

void
scopy(from, to, count)
register short *from, *to;
u_int count;
{
	register short *end;		/* address after last address to copy */

	if (((int)from ^ (int)(to))  &  sizeof(short))	/* differ in twos bit */
		for (end = &from[count];  from < end;  *to++ = *from++)
			;
	else {				/* same twos bit */
		if ((int)from & sizeof(short)  &&  count)
			*to++ = *from++,  --count;
		wcopy((long *)from,  (long *)to,  count >> 1);
		if (count & 1)		/* if odd */
			to[count - 1] = from[count - 1];
	}
}


/*
 *	Byte copy - copy "count" bytes from address "from" to 
 *	address "to."
 */

bcopy(from, to, count)
register char *from, *to;
register u_int count;
{
	register char *end;		/* address after last address to copy */

	if (((int)from ^ (int)(to))  &  sizeof(char))	/* differ in low bit */
		for (end = &from[count];  from < end;  *to++ = *from++)
			;
	else {					/* same low bit */
		if ((int)from & sizeof(char)  &&  count)
			*to++ = *from++,  --count;
		scopy((short *)from,  (short *)to,  count >> 1);
		if (count & 1)
			to[count - 1] = from[count - 1];
	}
}


/*
 * 	Word fill - Fill up "count" words with the value "fill" 
 *	starting from address "to."
 */

wfill(fill, to, count)
register int fill, *to;
u_int count;
{
	register int *end;	/* address after last address to be filled */

	for (end = &to[count];  to < end;  *to++ = fill)
		;
}


/*
 *	Short (integer) fill - fill up "count" words with "fill" 
 *	starting from address "to."
 */

sfill(fill, to, count)
register short fill, *to;
u_int count;
{
	if ((int)to & 2  &&  count) {
		*to++ = fill;
		--count;
	}
	wfill((int)fill << 16  |  (int)fill & 0xffff,  (int *)to,  count >> 1);
	if (count & 1)
		to[count - 1] = fill;
}



# ifdef	NEEDMONCONTROL
	/*
	 *	stub for now
	 *	Have to define a dummy one when not compiled with -g or -pg.
	 */
/* PNH: I don't know what this is for.  It interferes with the compilation
 * of b2s and newb2s.
 */

	/*ARGSUSED*/ 
	moncontrol(sel)
	int sel;
	{
	}
# endif

